/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsp.backgroundjob.service;

import dk.nsp.backgroundjob.stack.StackOperation;
import dk.nsp.backgroundjob.util.DateTimeProvider;
import org.apache.log4j.Logger;

import java.time.Duration;
import java.time.ZonedDateTime;
import java.util.Collections;
import java.util.List;
import java.util.Stack;
import java.util.stream.Collectors;

public class StackedOperationsService {
    private final Logger log = Logger.getLogger(StackedOperationsService.class);

    private final StackOperation defaultStackOperation;
    private final Duration desiredExecutionDuration;
    private final DateTimeProvider dateTimeProvider;
    private final Stack<StackOperation> stack;
    private String serviceName = "Operations service";
    private boolean latestRunSuccessful = true;

    public StackedOperationsService(StackOperation defaultStackOperation, Duration desiredExecutionDuration, DateTimeProvider dateTimeProvider) {
        this.defaultStackOperation = defaultStackOperation;
        this.desiredExecutionDuration = desiredExecutionDuration;
        this.dateTimeProvider = dateTimeProvider;
        stack = new Stack<>();
    }

    public void run() {
        ZonedDateTime startTime = dateTimeProvider.now();
        try {
            synchronizedRun(startTime);
            latestRunSuccessful = true;
        } catch (Exception e) {
            latestRunSuccessful = false;
            throw e;
        }
    }

    public boolean latestRunSuccessful() {
        return latestRunSuccessful;
    }

    private synchronized void synchronizedRun(ZonedDateTime startTime) {
        log.trace(String.format("%s: Execution started.", serviceName));
        if (!desiredExecutionDurationReached(startTime)) {
            fillStackIfEmpty();
        }

        try {
            while(!desiredExecutionDurationReached(startTime) && !stack.isEmpty()) {
                runOperationAndPopulateStack(stack.pop());
            }
        } finally {
            logRemaining();
            log.info(String.format("%s: Execution finished. Time spent: %d seconds", serviceName, Duration.between(startTime, dateTimeProvider.now()).getSeconds()));
        }
    }

    private void fillStackIfEmpty() {
        if (stack.isEmpty()) {
            log.trace(serviceName + ": Stack is empty");
            runOperationAndPopulateStack(defaultStackOperation);
        }
    }

    private void runOperationAndPopulateStack(StackOperation stackOperation) {
        List<StackOperation> newStackOperations = stackOperation.run();
        Collections.shuffle(newStackOperations);
        newStackOperations.forEach(stack::push);
        log.info(String.format("%s: %d new operations added to stack from %s", serviceName, newStackOperations.size(), stackOperation.getClass().getSimpleName()));
    }

    private boolean desiredExecutionDurationReached(ZonedDateTime startTime) {
        return Duration.between(startTime, dateTimeProvider.now()).compareTo(desiredExecutionDuration) > 0;
    }

    private void logRemaining() {
        log.info(String.format("%s: remaining=%d on stack", serviceName, stack.size()));
        stack.stream().collect(Collectors.groupingBy(stackOperation -> stackOperation.getClass().getSimpleName(), Collectors.counting()))
                .forEach((operationName, count) ->
                        log.info(String.format("%s: remaining=%d of type %s on stack", serviceName, count, operationName)));
    }
}
