/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsp.backgroundjob.service;

import dk.nsp.backgroundjob.stack.StackOperation;
import dk.nsp.backgroundjob.util.DateTimeProvider;
import org.junit.Before;
import org.junit.Test;
import org.mockito.Mockito;

import java.time.Duration;
import java.time.ZonedDateTime;
import java.util.Arrays;
import java.util.Collections;
import java.util.concurrent.atomic.AtomicBoolean;

import static org.junit.Assert.*;

public class StackedOperationsServiceTest {

    private StackOperation defaultStackOperation;
    private StackedOperationsService stackedOperationsService;
    private DateTimeProvider dateTimeProvider;
    private Duration desiredExecutionDuration;

    @Before
    public void setUp() throws Exception {
        defaultStackOperation = Mockito.mock(StackOperation.class);
        dateTimeProvider = Mockito.mock(DateTimeProvider.class);
        Mockito.when(dateTimeProvider.now()).thenReturn(ZonedDateTime.now());
        desiredExecutionDuration = Duration.ofHours(1);
        stackedOperationsService = new StackedOperationsService(defaultStackOperation, desiredExecutionDuration, dateTimeProvider);
    }

    @Test
    public void run_WithOneOperationFromDefaultOperation_RunsOperation() {
        // Arrange
        StackOperation operation = Mockito.mock(StackOperation.class);
        Mockito.when(defaultStackOperation.run()).thenReturn(Collections.singletonList(operation));

        // Act
        stackedOperationsService.run();

        // Assert
        Mockito.verify(operation, Mockito.times(1)).run();
    }

    @Test
    public void run_With0OperationsFromDefaultOperation_RunsDefaultOperationOnce() {
        // Arrange
        Mockito.when(defaultStackOperation.run()).thenReturn(Collections.emptyList());

        // Act
        stackedOperationsService.run();

        // Assert
        Mockito.verify(defaultStackOperation, Mockito.times(1)).run();
    }

    @Test
    public void run_TwiceWith0OperationsFromDefaultOperation_RunsDefaultOperationTwice() {
        // Arrange
        Mockito.when(defaultStackOperation.run()).thenReturn(Collections.emptyList());

        // Act
        stackedOperationsService.run();
        stackedOperationsService.run();

        // Assert
        Mockito.verify(defaultStackOperation, Mockito.times(2)).run();
    }

    @Test
    public void run_WithMultipleOperationsFromDefaultOperation_RunsOperations() {
        // Arrange
        StackOperation operation1 = Mockito.mock(StackOperation.class);
        StackOperation operation2 = Mockito.mock(StackOperation.class);
        Mockito.when(defaultStackOperation.run()).thenReturn(Arrays.asList(operation1, operation2));

        // Act
        stackedOperationsService.run();

        // Assert
        Mockito.verify(operation1, Mockito.times(1)).run();
        Mockito.verify(operation2, Mockito.times(1)).run();
    }

    @Test
    public void run_WithOperationsFromOperationFromOperation_RunsOperations() {
        // Arrange
        StackOperation operation1 = Mockito.mock(StackOperation.class);
        StackOperation operation2 = Mockito.mock(StackOperation.class);
        StackOperation operation3 = Mockito.mock(StackOperation.class);
        StackOperation operation4 = Mockito.mock(StackOperation.class);
        Mockito.when(defaultStackOperation.run()).thenReturn(Collections.singletonList(operation1));
        Mockito.when(operation1.run()).thenReturn(Collections.singletonList(operation2));
        Mockito.when(operation2.run()).thenReturn(Arrays.asList(operation3, operation4));

        // Act
        stackedOperationsService.run();

        // Assert
        Mockito.verify(operation1, Mockito.times(1)).run();
        Mockito.verify(operation2, Mockito.times(1)).run();
        Mockito.verify(operation3, Mockito.times(1)).run();
        Mockito.verify(operation4, Mockito.times(1)).run();
    }

    @Test
    public void run_WhenDesiredExecutionDurationIsReachedAfterFirstOperation_RunsOnlyFirstOperation() {
        // Arrange
        StackOperation operation1 = Mockito.mock(StackOperation.class);
        StackOperation operation2 = Mockito.mock(StackOperation.class);
        Mockito.when(defaultStackOperation.run()).thenReturn(Arrays.asList(operation1, operation2));
        ZonedDateTime startTime = ZonedDateTime.now();
        Mockito.when(dateTimeProvider.now())
                .thenReturn(startTime) // start time
                .thenReturn(startTime) // default operation run
                .thenReturn(startTime) // first operation run
                .thenReturn(startTime.plus(desiredExecutionDuration).plus(Duration.ofSeconds(1))); // second operation run

        AtomicBoolean operation1Called = new AtomicBoolean(false);
        AtomicBoolean operation2Called = new AtomicBoolean(false);
        Mockito.doAnswer(a -> {
            operation1Called.set(true);
            return Collections.emptyList();
        }).when(operation1).run();
        Mockito.doAnswer(a -> {
            operation2Called.set(true);
            return Collections.emptyList();
        }).when(operation2).run();

        // Act
        stackedOperationsService.run();

        // Assert
        assertTrue(operation1Called.get() ^ operation2Called.get());
    }

    @Test
    public void run_WhenDesiredExecutionDurationIsReachedBeforeFillingStack_RunsNothing() {
        // Arrange
        ZonedDateTime startTime = ZonedDateTime.now();
        Mockito.when(dateTimeProvider.now())
                .thenReturn(startTime) // start time
                .thenReturn(startTime.plus(desiredExecutionDuration).plus(Duration.ofSeconds(1))); // default operation run

        // Act
        stackedOperationsService.run();

        // Assert
        Mockito.verifyZeroInteractions(defaultStackOperation);
    }

    @Test
    public void run_WithTwoOperationsThatThrowsExceptions_RunsOneOperationAndThrowsException() {
        // Arrange
        StackOperation operation1 = Mockito.mock(StackOperation.class);
        StackOperation operation2 = Mockito.mock(StackOperation.class);
        Mockito.when(defaultStackOperation.run()).thenReturn(Arrays.asList(operation1, operation2));
        AtomicBoolean operation1Called = new AtomicBoolean(false);
        AtomicBoolean operation2Called = new AtomicBoolean(false);
        Mockito.doAnswer(a -> {
            operation1Called.set(true);
            throw new Exception();
        }).when(operation1).run();
        Mockito.doAnswer(a -> {
            operation2Called.set(true);
            throw new Exception();
        }).when(operation2).run();

        // Act
        assertThrows(Exception.class, () -> stackedOperationsService.run());

        // Assert
        assertTrue(operation1Called.get() ^ operation2Called.get());
    }

    @Test
    public void run_TwiceWithTwoOperationsThatThrowsExceptions_RunsBothOperationsAndThrowsExceptions() {
        // Arrange
        StackOperation operation1 = Mockito.mock(StackOperation.class);
        StackOperation operation2 = Mockito.mock(StackOperation.class);
        Exception operation1Exception = new RuntimeException();
        Exception operation2Exception = new RuntimeException();
        Mockito.doThrow(operation1Exception).when(operation1).run();
        Mockito.doThrow(operation2Exception).when(operation2).run();
        Mockito.when(defaultStackOperation.run()).thenReturn(Arrays.asList(operation1, operation2));

        // Act
        Exception e1 = assertThrows(RuntimeException.class, () -> stackedOperationsService.run());
        Exception e2 = assertThrows(RuntimeException.class, () -> stackedOperationsService.run());

        // Assert
        assertNotEquals(e1, e2);
        assertTrue(e1 == operation1Exception || e1 == operation2Exception);
        assertTrue(e2 == operation1Exception || e2 == operation2Exception);
    }

    @Test
    public void latestRunSuccessful_WhenRunHasNotBeenExecuted_ReturnsTrue() {
        boolean latestRunSuccessful = stackedOperationsService.latestRunSuccessful();

        assertTrue(latestRunSuccessful);
    }

    @Test
    public void latestRunSuccessful_WhenLatestRunExecutedSuccessfully_ReturnsTrue() {
        stackedOperationsService.run();

        boolean latestRunSuccessful = stackedOperationsService.latestRunSuccessful();

        assertTrue(latestRunSuccessful);
    }

    @Test
    public void latestRunSuccessful_WhenLatestRunThrewException_ReturnsFalse() {
        Mockito.doThrow(RuntimeException.class).when(defaultStackOperation).run();
        assertThrows(RuntimeException.class, () -> stackedOperationsService.run());

        boolean latestRunSuccessful = stackedOperationsService.latestRunSuccessful();

        assertFalse(latestRunSuccessful);
    }

    @Test
    public void latestRunSuccessful_WhenRunHasFailedAndThenExecutedSuccessfully_ReturnsTrue() {
        Mockito.when(defaultStackOperation.run())
                .thenThrow(RuntimeException.class)
                .thenReturn(Collections.emptyList());
        assertThrows(RuntimeException.class, () -> stackedOperationsService.run());
        stackedOperationsService.run();

        boolean latestRunSuccessful = stackedOperationsService.latestRunSuccessful();

        assertTrue(latestRunSuccessful);
    }

    private static <T extends Throwable> T assertThrows(Class<T> expectedThrowable, ThrowingRunnable runnable) {
        try {
            runnable.run();
        } catch (Throwable t) {
            assertEquals(expectedThrowable, t.getClass());
            return (T) t;
        }
        fail("No exception thrown");
        throw new RuntimeException(); // Never happens
    }

    private interface ThrowingRunnable {
        void run() throws Throwable;
    }
}