/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.conf;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.annotation.PropertySource;
import org.springframework.stereotype.Component;

import javax.annotation.PostConstruct;
import java.time.Duration;
import java.time.Period;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.time.format.DateTimeParseException;
import java.util.Arrays;
import java.util.List;
import java.util.stream.Collectors;

@Component
@PropertySource("classpath:application.properties")
public class Properties {

    private static final Logger logger = LogManager.getLogger(Properties.class);

    @Value("${datasource.ltr.jndi-name:}")
    private String dataSourceLtrJNDIName;

    @Value("${datasource.btr.jndi-name:}")
    private String dataSourceBtrJNDIName;

    @Value("${dcc.endpoint.ltr:}")
    private String dccLTREndpoint;

    @Value("${dcc.endpoint.btr:}")
    private String dccBTREndpoint;

    @Value("${minlog.read-activity-text.ltr:}")
    private String minLogReadActivityTextLTR;

    @Value("${minlog.read-activity-text.btr:}")
    private String minLogReadActivityTextBTR;
    
    @Value("${minlog.create-activity-text.btr:}")
    private String minLogCreateActivityTextBTR;

    @Value("${minlog.delete-activity-text.ltr:}")
    private String minLogDeleteActivityTextLTR;

    @Value("${minlog.delete-activity-text.btr:}")
    private String minLogDeleteActivityTextBTR;

    @Value("${minlog.update-activity-text.ltr:}")
    private String minLogUpdateActivityTextLTR;

    @Value("${minlog.update-activity-text.btr:}")
    private String minLogUpdateActivityTextBTR;

    @Value("${minlog.citizen.user.title:}")
    private String minlogCitizenUserTitle;

    @Value("${minlog.healthcareprofessional.unknown.user.title}")
    private String minlogHealthcareProfessionalUnknownUserTitle;

    @Value("${btr.open-date:}")
    private String btrOpenDateString;
    private ZonedDateTime btrOpenDate;

    @Value("${personInformation.url:#{null}}")
    String PersonInformationUrl;

    @Value("${personInformation.minage:18}")
    private int personInformationMinAge;

    @Value("${personInformation.validationlevel:}")
    private String PersonInformationValidationLevel;

    @Value("#{'${personInformation.inactive.status:30,50,60}'.split(',')}")
    private List<Integer> inactiveStatusList;

    @Value("${personInformation.maxTotalConnections:200}")
    private int maxTotalConnections;

    @Value("${personInformation.defaultMaxConnectionsPerRoute:20}")
    private int defaultMaxConnectionsPerRoute;

    @Value("${whitelisted.level3.cvrs:}")
    private String whitelistedLevel3Cvrs;

    @Value("${allowed.idws.audience:https://fsk}")
    private String allowedIdwsAudience;

    @Value("${app.endpoint:}")
    private String appEndpoint;

    @Value("${nas.app.name}")
    private String appName;

    @Value("${nas.app.shortName}")
    private String appShortName;

    @Value("${nas.fail.threshold}")
    private String nasFailThreshold;

    @Value("${nas.connect.timeout.millis}")
    private String nasConnectTimeoutMillis;

    @Value("${nas.read.timeout.millis}")
    private String nasReadTimeoutMillis;

    @Value("${nas.max.total.connections}")
    private String nasMaxTotalConnections;

    @Value("${nas.default.max.connections.per.route}")
    private String nasDefaultMaxConnectionsPerRoute;

    @Value("${nas.service.url}")
    private String nasServiceUrl;

    @Value("${nas.idcard.subject.id.type}")
    private String idcardSubjectIdType;

    @Value("${nas.idcard.subject.id}")
    private String idcardSubjectId;

    @Value("${nas.idcard.subject.name}")
    private String idcardSubjectName;

    @Value("${nas.idcard.system.name}")
    private String idcardSystemName;

    @Value("${nas.sts.endpoint}")
    private String stsEndpoint;

    @Value("${NSP_TEST_FEDERATION:false}")
    private boolean stsTestMode;

    @Value("${nas.sts.keystore}")
    private String stsKeystore;

    @Value("${nas.sts.keystore.password}")
    private String stsKeystorePassword;

    @Value("${minlog.read-activity-text-with-only-forced-treatment.btr:}")
    private String minLogReadActivityTextWithOnlyForcedTreatmentBTR;

    @Value("${deletion.save.deceased}")
    private String deletionSaveDeceased;

    @Value("${deletion.jobs.batchSize}")
    private int deletionJobsBatchSize;

    @Value("${personinformation.error.tolerance:50}")
    private int personInformationErrorTolerance;

    @Value("${desired.execution.duration}")
    private String desiredExecutionDuration;

    @PostConstruct
    private void init() {
        try {
            btrOpenDate = ZonedDateTime.parse(btrOpenDateString);
            logger.info("BTR open date set to: " + getBtrOpenDate().format(DateTimeFormatter.ISO_DATE_TIME));
        } catch (DateTimeParseException e) {
            logger.warn("The date in the property 'btr.open-date' could not be parsed. BTR registrations can be created.");
        }

        logger.info(toString());
    }

    @Override
    public String toString() {
        return "Configuration properties: {" +
                "dataSourceLtrJNDIName=" + getDataSourceLtrJNDIName() +
                ", dataSourceBtrJNDIName=" + getDataSourceBtrJNDIName() +
                ", dccLTREndpoint=" + getDccLTREndpoint() +
                ", dccBTREndpoint=" + getDccBTREndpoint() +
                ", minLogReadActivityTextLTR=" + getMinLogReadActivityTextLTR() +
                ", minLogReadActivityTextBTR=" + getMinLogReadActivityTextBTR() +
                ", minLogCreateActivityTextBTR=" + getMinLogCreateActivityTextBTR() +
                ", minLogDeleteActivityTextLTR=" + getMinLogDeleteActivityTextLTR() +
                ", minLogDeleteActivityTextBTR=" + getMinLogDeleteActivityTextBTR() +
                ", minLogUpdateActivityTextLTR=" + getMinLogUpdateActivityTextLTR() +
                ", minLogUpdateActivityTextBTR=" + getMinLogUpdateActivityTextBTR() +
                ", btrOpenDate=" + getBtrOpenDate() +
                ", whitelistedLevel3Cvrs=" + getWhitelistedLevel3Cvrs() +
                ", allowedIdwsAudience=" + getAllowedIdwsAudience() +
                ", deceasedPeriod=" + getDeceasedPeriod() +
                ", personInformationUrl=" + getPersonInformationUrl() +
                ", personInformationMinAge=" + getPersonInformationMinAge() +
                ", personInformationValidationLevel=" + getPersonInformationValidationLevel() +
                ", inactiveStatusList=" + getInactiveStatusList() +
                ", maxTotalConnections=" + getMaxTotalConnections() +
                ", defaultMaxConnectionsPerRoute=" + getDefaultMaxConnectionsPerRoute() +
                ", appName=" + getAppName() +
                ", appShortName=" + getAppShortName() +
                ", nasFailThreshold=" + getNasFailThreshold() +
                ", nasConnectTimeoutMillis=" + getNasConnectTimeoutMillis() +
                ", nasReadTimeoutMillis=" + getNasReadTimeoutMillis() +
                ", nasMaxTotalConnections=" + getNasMaxTotalConnections() +
                ", nasDefaultMaxConnectionsPerRoute=" + getNasDefaultMaxConnectionsPerRoute() +
                ", nasServiceUrl=" + getNasServiceUrl() +
                ", idcardSubjectIdType=" + getIdcardSubjectIdType() +
                ", idcardSubjectId=" + getIdcardSubjectId() +
                ", idcardSubjectName=" + getIdcardSubjectName() +
                ", idcardSystemName=" + getIdcardSystemName() +
                ", stsEndpoint=" + getStsEndpoint() +
                ", stsTestMode=" + getStsTestMode() +
                ", stsKeystore=" + getStsKeystore() +
                ", stsKeystorePassword=" + getStsKeystorePassword() +
                ", minLogReadActivityTextWithOnlyForcedTreatmentBTR=" + getMinLogReadActivityTextWithOnlyForcedTreatmentBTR() +
                ", DeletionJobsBatchSize=" + getDeletionJobsBatchSize() +
                ", personInformationErrorTolerance=" + getPersonInformationErrorTolerance() +
                ", desiredExecutionDuration=" + desiredExecutionDuration +
                '}';
    }

    public String getDataSourceLtrJNDIName() {
        return dataSourceLtrJNDIName;
    }

    public String getDataSourceBtrJNDIName() {
        return dataSourceBtrJNDIName;
    }

    public String getDccLTREndpoint() {
        return dccLTREndpoint;
    }

    public String getDccBTREndpoint() {
        return dccBTREndpoint;
    }

    public String getMinLogReadActivityTextLTR() {
        return minLogReadActivityTextLTR;
    }

    public String getMinLogReadActivityTextBTR() {
        return minLogReadActivityTextBTR;
    }
    
    public String getMinLogCreateActivityTextBTR() {
        return minLogCreateActivityTextBTR;
    }
    
    public String getMinLogDeleteActivityTextLTR() {
        return minLogDeleteActivityTextLTR;
    }

    public String getMinLogDeleteActivityTextBTR() {
        return minLogDeleteActivityTextBTR;
    }
    
    public String getMinLogUpdateActivityTextLTR() {
        return minLogUpdateActivityTextLTR;
    }

    public String getMinLogUpdateActivityTextBTR() {
        return minLogUpdateActivityTextBTR;
    }

    public String getMinlogCitizenUserTitle() {
        return minlogCitizenUserTitle;
    }

    public String getMinlogHealthcareProfessionalUnknownUserTitle() {
        return minlogHealthcareProfessionalUnknownUserTitle;
    }

    public ZonedDateTime getBtrOpenDate() {
        return btrOpenDate;
    }

    public String getPersonInformationUrl() {
        return PersonInformationUrl;
    }

    public int getPersonInformationMinAge() { return personInformationMinAge; }

    public String getPersonInformationValidationLevel() {
        return PersonInformationValidationLevel;
    }

    public List<Integer> getInactiveStatusList() {
        return inactiveStatusList;
    }

    public int getMaxTotalConnections() {
        return maxTotalConnections;
    }

    public int getDefaultMaxConnectionsPerRoute() {
        return defaultMaxConnectionsPerRoute;
    }

    public List<String> getWhitelistedLevel3Cvrs() {
        return Arrays.stream(whitelistedLevel3Cvrs.split(",")).map(String::trim).filter(cvr -> !cvr.isEmpty()).collect(Collectors.toList());
    }

    public String getAllowedIdwsAudience() {
        return allowedIdwsAudience;
    }

    public String getAppName() {
        return appName;
    }

    public String getAppShortName() {
        return appShortName;
    }

    public String getNasFailThreshold() {
        return nasFailThreshold;
    }

    public String getNasConnectTimeoutMillis() {
        return nasConnectTimeoutMillis;
    }

    public String getNasReadTimeoutMillis() {
        return nasReadTimeoutMillis;
    }

    public String getNasMaxTotalConnections() {
        return nasMaxTotalConnections;
    }

    public String getNasDefaultMaxConnectionsPerRoute() {
        return nasDefaultMaxConnectionsPerRoute;
    }

    public String getNasServiceUrl() {
        return nasServiceUrl;
    }

    public String getIdcardSubjectIdType() {
        return idcardSubjectIdType;
    }

    public String getIdcardSubjectId() {
        return idcardSubjectId;
    }

    public String getIdcardSubjectName() {
        return idcardSubjectName;
    }

    public String getIdcardSystemName() {
        return idcardSystemName;
    }

    public String getStsEndpoint() {
        return stsEndpoint;
    }

    public boolean getStsTestMode() {
        return stsTestMode;
    }

    public String getStsKeystore() {
        return stsKeystore;
    }

    public String getStsKeystorePassword() {
        return stsKeystorePassword;
    }

    public String getMinLogReadActivityTextWithOnlyForcedTreatmentBTR() {return minLogReadActivityTextWithOnlyForcedTreatmentBTR; }

    public Period getDeceasedPeriod() {
        return Period.parse(deletionSaveDeceased);
    }

    public int getDeletionJobsBatchSize() { return deletionJobsBatchSize; }

    public int getPersonInformationErrorTolerance() {
        return personInformationErrorTolerance;
    }

    public Duration desiredExecutionDuration() {
        return Duration.parse(desiredExecutionDuration);
    }
}
