/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.context;

import java.util.List;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;

import dk.sds.nsp.security.SecurityContext;
import dk.sds.nsp.security.SecurityContext.Organisation.OrganisationIdentifierFormat;
import dk.sds.nsp.security.SecurityContext.User.UserType;
import dk.sundhedsdatastyrelsen.livstestamente.util.FaultCode;
import dk.sundhedsdatastyrelsen.livstestamente.util.FaultUtil;
import dk.sundhedsdatastyrelsen.livstestamente.conf.Properties;

public class ServiceActorProvider {

	private static final Logger logger = LogManager.getLogger(ServiceActorProvider.class);

	private String allowedAudience;
	private List<String> whiteListedLevel3Cvrs;
	private UserTitleProvider userTitleProvider;
	
	public ServiceActorProvider(@Autowired Properties properties, @Autowired UserTitleProvider userTitleProvider) {
		this(properties.getAllowedIdwsAudience(), properties.getWhitelistedLevel3Cvrs(), userTitleProvider);
	}
	
	public ServiceActorProvider(String allowedAudience, List<String> whiteListedLevel3Cvrs, UserTitleProvider userTitleProvider) {
		this.allowedAudience = allowedAudience;
		this.whiteListedLevel3Cvrs = whiteListedLevel3Cvrs;
		this.userTitleProvider = userTitleProvider;
	}

	public ServiceActor getServiceActor(SecurityContext securityContext, boolean wantsAdmin) {
		validateTicket(securityContext);
		return mapSecurityContextToValidatedServiceActor(securityContext, wantsAdmin);
	}

	private ServiceActor mapSecurityContextToValidatedServiceActor(SecurityContext securityContext, boolean wantsAdmin) {


		// Gennemløb de understøttede aktørtyper for denne service og valider og map
		
		// Tjekke for sundhedsfaglig og admin brugere
		if (securityContext.getActingUser().filter(u -> u.getUserType().equals(UserType.HealthcareProfessional)).isPresent()) {
			// Ingen delegering
			if (securityContext.getPrincipalUser().isPresent()) {
				FaultUtil.throwFault(FaultCode.AUTHORIZATION_ERROR_INGEN_DELEGERING);
			}
			
			ServiceActor result = mapToServiceActor(securityContext, wantsAdmin ? dk.sundhedsdatastyrelsen.livstestamente.domain.UserType.Admin : dk.sundhedsdatastyrelsen.livstestamente.domain.UserType.HealthCareProfessional, true);
			return result;
		}
		// Borgerkald
		if (securityContext.getActingUser().filter(u -> u.getUserType().equals(UserType.Citizen)).isPresent()) {
			if (wantsAdmin) {
				FaultUtil.throwFault(FaultCode.AUTHORIZATION_ERROR_USERTYPE_ADMIN_NOT_ALLOWED);
			}
			// Ingen fuldmagt for borgere
			if (securityContext.getPrincipalUser().isPresent()) {
				FaultUtil.throwFault(FaultCode.AUTHORIZATION_ERROR_INGEN_DELEGERING);
			}
			// Der skal være audience tilstede
			if (!securityContext.getTicket().get().getAudience().isPresent()) {
				FaultUtil.throwFault(FaultCode.AUTHORIZATION_ERROR_INVALID_HEADER);
			}

	        // Validering af audience
	        if (!securityContext.getTicket().get().getAudience().get().equals(allowedAudience)) {
				FaultUtil.throwFault(FaultCode.AUTHORIZATION_ERROR_WRONG_AUDIENCE);
	        }
			
			ServiceActor result = mapToServiceActor(securityContext, dk.sundhedsdatastyrelsen.livstestamente.domain.UserType.Citizen, false);
			return result;
		}
		// Systembruger
		if (!securityContext.getActingUser().isPresent() && securityContext.getOrganisation().isPresent()) {
			
			if (wantsAdmin) {
				FaultUtil.throwFault(FaultCode.AUTHORIZATION_ERROR_USERTYPE_ADMIN_NOT_ALLOWED);
			}

			ServiceActor result = mapToServiceActor(securityContext, dk.sundhedsdatastyrelsen.livstestamente.domain.UserType.SystemUser, true);
			
			// Er whitelisting ok?
			if (!whiteListedLevel3Cvrs.contains(result.getOrganizationId())) {
				FaultUtil.throwFault(FaultCode.AUTHORIZATION_ERROR_CVR_NOT_WHITELISTED);
			}
			return result;
		}

		FaultUtil.throwFault(FaultCode.AUTHORIZATION_ERROR_INGEN_DELEGERING);
		return null;
	}


	private ServiceActor mapToServiceActor(SecurityContext securityContext, dk.sundhedsdatastyrelsen.livstestamente.domain.UserType userType, boolean requireCvr) {
		ServiceActor result = new ServiceActor(userType);
		
	/*	if (securityContext.getClient().isPresent()) {
			mapDataFromClient(result, securityContext.getClient().get());
		}*/
		if (securityContext.getOrganisation().isPresent()) {
			mapDataFromOrganisation(result, securityContext.getOrganisation().get());
		} else if (requireCvr) {
			FaultUtil.throwFault(FaultCode.AUTHORIZATION_ERROR_INGEN_CVR);
		}
		if (securityContext.getActingUser().isPresent()) {
			mapAndValidateDataFromUser(result, securityContext.getActingUser().get());
		}

		switch (userType) {
			case Admin:
			case HealthCareProfessional:
				result.setUserTitle(userTitleProvider.getSundhedsfagligUserTitle(securityContext.getActingUser().get()));
				break;
			case Citizen:
				result.setUserTitle(userTitleProvider.getBorgerUserTitle());
				break;
		}

		return result;
	}


	private void mapDataFromOrganisation(ServiceActor serviceActor, SecurityContext.Organisation organisation) {
		if (organisation.getIdentifierFormat().equals(OrganisationIdentifierFormat.CVR)) {
			serviceActor.setOrganizationId(organisation.getIdentifier());			
			serviceActor.setOrganizationIdSource(OrganisationIdentifierFormat.CVR);
			serviceActor.setOrganizationName(organisation.getName().orElse(null));			
		} else {
			FaultUtil.throwFault(FaultCode.AUTHORIZATION_ERROR_INGEN_CVR);
		}
	}
/*
	private void mapDataFromClient(ServiceActor ServiceActor, SecurityContext.Client client) {
		ServiceActor.setItSystemName(client.getName().orElse(null));
	}
*/
	private void mapAndValidateDataFromUser(ServiceActor aerviceActor, SecurityContext.User user) {
		if (user.getIdentifierFormat().equals(SecurityContext.User.PersonIdentifierFormat.CPR)) {
			aerviceActor.setPersonIdentifier(user.getIdentifier());
		} else {
			FaultUtil.throwFault(FaultCode.AUTHORIZATION_ERROR_USER_NOT_CPR);
		}
		aerviceActor.setFirstName(user.getGivenName().orElse(null));
		aerviceActor.setLastName(user.getSurname().orElse(null));
		if (user.getCredentials().isPresent()) {
			SecurityContext.User.Credentials credentials = user.getCredentials().get();
			aerviceActor.setAuthorisationNumber(credentials.getAuthorizationCode().orElse(null));
			aerviceActor.setEducationCode(credentials.getEducationCode().orElse(null));
		}
	}

	private void validateTicket(SecurityContext securityContext) {
		if (securityContext == null || !securityContext.getTicket().isPresent() || !securityContext.getTicket().get().isValid()) {
			FaultUtil.throwFault(FaultCode.AUTHORIZATION_ERROR_INVALID_HEADER);
		}
	}
}
