/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.dao.impl;

import dk.sundhedsdatastyrelsen.livstestamente.dao.TreatmentWillDAO;
import dk.sundhedsdatastyrelsen.livstestamente.dao.impl.abstr.AbstractDAOImpl;
import dk.sundhedsdatastyrelsen.livstestamente.domain.AcceptanceNeeded;
import dk.sundhedsdatastyrelsen.livstestamente.domain.TreatmentWillVO;
import dk.sundhedsdatastyrelsen.livstestamente.exceptions.DAOException;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.dao.DataAccessException;
import org.springframework.dao.support.DataAccessUtils;
import org.springframework.jdbc.core.namedparam.MapSqlParameterSource;
import org.springframework.jdbc.core.namedparam.NamedParameterJdbcTemplate;
import org.springframework.jdbc.support.GeneratedKeyHolder;
import org.springframework.stereotype.Repository;

import javax.sql.DataSource;
import java.sql.Types;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.LinkedList;
import java.util.List;
import java.util.Objects;

@Repository
public class TreatmentWillDAOImpl extends AbstractDAOImpl implements TreatmentWillDAO {
    private final DataSource dataSource;

    @Autowired
    public TreatmentWillDAOImpl(DataSource btrDataSource) {
        super(btrDataSource);
        this.dataSource = btrDataSource;
    }

    @Override
    public TreatmentWillVO getCurrentTreatmentWill(String personIdentifier, Instant validAt) throws DAOException {
        try {
            final String sql = "SELECT t.* FROM TreatmentWill t " +
                    "WHERE t.PersonIdentifier = :PersonIdentifier " +
                    "AND t.ValidFrom <= :ValidAt AND :ValidAt < t.ValidTo " +
                    "ORDER BY t.ValidTo DESC LIMIT 1";
            final MapSqlParameterSource parameterSource = new MapSqlParameterSource()
                    .addValue("PersonIdentifier", personIdentifier, Types.VARCHAR)
                    .addValue("ValidAt", java.sql.Timestamp.from(validAt.truncatedTo(ChronoUnit.SECONDS)), Types.TIMESTAMP);
            return DataAccessUtils.singleResult(namedParamJdbcTemplate.query(sql, parameterSource, new TreatmentWillRowMapper()));
        } catch (DataAccessException e) {
            throw new DAOException("An error occurred reading TreatmentWill", e);
        }
    }

    @Override
    public boolean existsTreatmentWillRecords(String personIdentifier) throws DAOException {
        try {
            final String sql = "SELECT COUNT(*) FROM TreatmentWill t " +
                    "WHERE t.PersonIdentifier = :PersonIdentifier";
            final MapSqlParameterSource parameterSource = new MapSqlParameterSource()
                    .addValue("PersonIdentifier", personIdentifier, Types.VARCHAR);
            final long count = Objects.requireNonNull(namedParamJdbcTemplate.queryForObject(sql, parameterSource, Long.class));
            return count > 0;
        } catch (DataAccessException | NullPointerException e) {
            throw new DAOException("An error occurred counting TreatmentWill", e);
        }
    }

    @Override
    public long createTreatmentWill(TreatmentWillVO treatmentWillVO) throws DAOException {

        final String sql = "INSERT INTO TreatmentWill " +
                "(Version, ValidFrom, ValidTo, ModifiedDateTime, PersonIdentifier, " +
                "NoLifeProlongingIfDying, NoLifeProlongingIfDyingAcceptanceNeeded, " +
                "NoLifeProlongingIfSeverelyDegraded, NoLifeProlongingIfSeverelyDegradedAcceptanceNeeded, " +
                "NoLifeProlongingIfSeverePain, NoLifeProlongingIfSeverePainAcceptanceNeeded, " +
                "NoForcedTreatmentIfIncapable, NoForcedTreatmentIfIncapableAcceptanceNeeded) " +
                "VALUES (:Version, :ValidFrom, :ValidTo, :ModifiedDateTime, :PersonIdentifier, " +
                ":NoLifeProlongingIfDying, :NoLifeProlongingIfDyingAcceptanceNeeded, " +
                ":NoLifeProlongingIfSeverelyDegraded, :NoLifeProlongingIfSeverelyDegradedAcceptanceNeeded, " +
                ":NoLifeProlongingIfSeverePain, :NoLifeProlongingIfSeverePainAcceptanceNeeded, " +
                ":NoForcedTreatmentIfIncapable, :NoForcedTreatmentIfIncapableAcceptanceNeeded)";
        try {
            final MapSqlParameterSource parameterSource = new MapSqlParameterSource()
                    .addValue("Version", treatmentWillVO.getVersion(), Types.INTEGER)
                    .addValue("ValidFrom", java.sql.Timestamp.from(treatmentWillVO.getValidFrom().truncatedTo(ChronoUnit.SECONDS)), Types.TIMESTAMP)
                    .addValue("ValidTo", java.sql.Timestamp.from(treatmentWillVO.getValidTo().truncatedTo(ChronoUnit.SECONDS)), Types.TIMESTAMP)
                    .addValue("ModifiedDateTime", java.sql.Timestamp.from(treatmentWillVO.getModifiedDateTime().truncatedTo(ChronoUnit.SECONDS)), Types.TIMESTAMP)
                    .addValue("PersonIdentifier", treatmentWillVO.getPersonIdentifier(), Types.VARCHAR)
                    .addValue("NoLifeProlongingIfDying", treatmentWillVO.isNoLifeProlongingIfDying(), Types.TINYINT)
                    .addValue("NoLifeProlongingIfDyingAcceptanceNeeded", getAcceptanceNeededDatabaseValue(treatmentWillVO.getNoLifeProlongingIfDyingAcceptanceNeeded()), Types.VARCHAR)
                    .addValue("NoLifeProlongingIfSeverelyDegraded", treatmentWillVO.isNoLifeProlongingIfSeverelyDegraded(), Types.TINYINT)
                    .addValue("NoLifeProlongingIfSeverelyDegradedAcceptanceNeeded", getAcceptanceNeededDatabaseValue(treatmentWillVO.getNoLifeProlongingIfSeverelyDegradedAcceptanceNeeded()), Types.VARCHAR)
                    .addValue("NoLifeProlongingIfSeverePain", treatmentWillVO.isNoLifeProlongingIfSeverePain(), Types.TINYINT)
                    .addValue("NoLifeProlongingIfSeverePainAcceptanceNeeded", getAcceptanceNeededDatabaseValue(treatmentWillVO.getNoLifeProlongingIfSeverePainAcceptanceNeeded()), Types.VARCHAR)
                    .addValue("NoForcedTreatmentIfIncapable", treatmentWillVO.isNoForcedTreatmentIfIncapable(), Types.TINYINT)
                    .addValue("NoForcedTreatmentIfIncapableAcceptanceNeeded", getAcceptanceNeededDatabaseValue(treatmentWillVO.getNoForcedTreatmentIfIncapableAcceptanceNeeded()), Types.VARCHAR);
            final GeneratedKeyHolder generatedKeyHolder = new GeneratedKeyHolder();
            namedParamJdbcTemplate.update(sql, parameterSource, generatedKeyHolder);
            final long pid = generatedKeyHolder.getKey().longValue();
            treatmentWillVO.setTreatmentWillPID(pid);
            return pid;
        } catch (DataAccessException e) {
            throw new DAOException("An error occurred inserting TreatmentWill", e);
        }
    }

    @Override
    public long createTreatmentWillSkipNoLifeProlongingIfDyingAcceptanceNeeded(TreatmentWillVO treatmentWillVO) throws DAOException {

        final String sql = "INSERT INTO TreatmentWill " +
                "(Version, ValidFrom, ValidTo, ModifiedDateTime, PersonIdentifier, " +
                "NoLifeProlongingIfDying, " +
                "NoLifeProlongingIfSeverelyDegraded, NoLifeProlongingIfSeverelyDegradedAcceptanceNeeded, " +
                "NoLifeProlongingIfSeverePain, NoLifeProlongingIfSeverePainAcceptanceNeeded, " +
                "NoForcedTreatmentIfIncapable, NoForcedTreatmentIfIncapableAcceptanceNeeded) " +
                "VALUES (:Version, :ValidFrom, :ValidTo, :ModifiedDateTime, :PersonIdentifier, " +
                ":NoLifeProlongingIfDying, " +
                ":NoLifeProlongingIfSeverelyDegraded, :NoLifeProlongingIfSeverelyDegradedAcceptanceNeeded, " +
                ":NoLifeProlongingIfSeverePain, :NoLifeProlongingIfSeverePainAcceptanceNeeded, " +
                ":NoForcedTreatmentIfIncapable, :NoForcedTreatmentIfIncapableAcceptanceNeeded)";
        try {
            final MapSqlParameterSource parameterSource = new MapSqlParameterSource()
                    .addValue("Version", treatmentWillVO.getVersion(), Types.INTEGER)
                    .addValue("ValidFrom", java.sql.Timestamp.from(treatmentWillVO.getValidFrom().truncatedTo(ChronoUnit.SECONDS)), Types.TIMESTAMP)
                    .addValue("ValidTo", java.sql.Timestamp.from(treatmentWillVO.getValidTo().truncatedTo(ChronoUnit.SECONDS)), Types.TIMESTAMP)
                    .addValue("ModifiedDateTime", java.sql.Timestamp.from(treatmentWillVO.getModifiedDateTime().truncatedTo(ChronoUnit.SECONDS)), Types.TIMESTAMP)
                    .addValue("PersonIdentifier", treatmentWillVO.getPersonIdentifier(), Types.VARCHAR)
                    .addValue("NoLifeProlongingIfDying", treatmentWillVO.isNoLifeProlongingIfDying(), Types.TINYINT)
                    .addValue("NoLifeProlongingIfSeverelyDegraded", treatmentWillVO.isNoLifeProlongingIfSeverelyDegraded(), Types.TINYINT)
                    .addValue("NoLifeProlongingIfSeverelyDegradedAcceptanceNeeded", getAcceptanceNeededDatabaseValue(treatmentWillVO.getNoLifeProlongingIfSeverelyDegradedAcceptanceNeeded()), Types.VARCHAR)
                    .addValue("NoLifeProlongingIfSeverePain", treatmentWillVO.isNoLifeProlongingIfSeverePain(), Types.TINYINT)
                    .addValue("NoLifeProlongingIfSeverePainAcceptanceNeeded", getAcceptanceNeededDatabaseValue(treatmentWillVO.getNoLifeProlongingIfSeverePainAcceptanceNeeded()), Types.VARCHAR)
                    .addValue("NoForcedTreatmentIfIncapable", treatmentWillVO.isNoForcedTreatmentIfIncapable(), Types.TINYINT)
                    .addValue("NoForcedTreatmentIfIncapableAcceptanceNeeded", getAcceptanceNeededDatabaseValue(treatmentWillVO.getNoForcedTreatmentIfIncapableAcceptanceNeeded()), Types.VARCHAR);
            final GeneratedKeyHolder generatedKeyHolder = new GeneratedKeyHolder();
            namedParamJdbcTemplate.update(sql, parameterSource, generatedKeyHolder);
            final long pid = generatedKeyHolder.getKey().longValue();
            treatmentWillVO.setTreatmentWillPID(pid);
            return pid;
        } catch (DataAccessException e) {
            throw new DAOException("An error occurred inserting TreatmentWill", e);
        }
    }


    @Override
    public void updateValidToTimestamp(long pid, Instant newValidTo) throws DAOException {
        try {
            final String sql = "UPDATE TreatmentWill SET ValidTo = :ValidTo " +
                    "WHERE TreatmentWillPID = :TreatmentWillPID";
            final MapSqlParameterSource parameters = new MapSqlParameterSource()
                    .addValue("ValidTo", java.sql.Timestamp.from(newValidTo.truncatedTo(ChronoUnit.SECONDS)), Types.TIMESTAMP)
                    .addValue("TreatmentWillPID", pid, Types.INTEGER);
            namedParamJdbcTemplate.update(sql, parameters);
        } catch (DataAccessException e) {
            throw new DAOException("An error updating valid to timestamp", e);
        }
    }

    @Override
    public void deleteTreatmentWillRecords(List<String> personIdentifiers) throws DAOException {
        try {
            String sql = "DELETE FROM TreatmentWill WHERE PersonIdentifier IN (:PersonIdentifiers)";
            final MapSqlParameterSource parameters = new MapSqlParameterSource()
                    .addValue("PersonIdentifiers", personIdentifiers, Types.VARCHAR);
            namedParamJdbcTemplate.update(sql, parameters);
        } catch (DataAccessException e) {
            throw new DAOException("An error occurred deleting from TreatmentWill", e);
        }
    }

    @Override
    public LinkedList<String> getAllPersonsFromBTR() throws DAOException {
        try {
            final String sql = "SELECT DISTINCT PersonIdentifier FROM TreatmentWill";
            return new LinkedList<>(namedParamJdbcTemplate.query(sql, (resultSet, i) -> resultSet.getNString("PersonIdentifier")));
        } catch (DataAccessException | NullPointerException e) {
            throw new DAOException("An error occurred fetching all rows from TreatmentWill", e);
        }
    }

    @Override
    public LinkedList<String> getAllPersonsFromBTR(int limit) throws DAOException {
        try {
            final String sql = "SELECT DISTINCT PersonIdentifier FROM TreatmentWill limit :limit";

            MapSqlParameterSource parameters = new MapSqlParameterSource();
            parameters.addValue("limit", limit);

            NamedParameterJdbcTemplate template = new NamedParameterJdbcTemplate(dataSource);
            return new LinkedList<>(template.query(sql, parameters, (resultSet, i) -> resultSet.getNString("PersonIdentifier")));
        } catch (DataAccessException | NullPointerException e) {
            throw new DAOException("An error occurred fetching all rows from TreatmentWill", e);
        }
    }

    @Override
    public LinkedList<String> getAllPersonsFromLTRWithPrefix(String prefix) throws DAOException {
        try {
            final String sql = "SELECT DISTINCT PersonIdentifier FROM TreatmentWill " +
                    "where PersonIdentifier like :prefix";;

            MapSqlParameterSource parameters = new MapSqlParameterSource();
            parameters.addValue("prefix", prefix + "%");

            NamedParameterJdbcTemplate template = new NamedParameterJdbcTemplate(dataSource);
            List<String> resultList = template.queryForList(sql, parameters, String.class);
            return new LinkedList<>(resultList);
        } catch (DataAccessException | NullPointerException e) {
            throw new DAOException("An error occurred fetching all rows from TreatmentWill", e);
        }
    }

    private String getAcceptanceNeededDatabaseValue(AcceptanceNeeded noLifeProlongingIfDyingAcceptanceNeeded) {
        return noLifeProlongingIfDyingAcceptanceNeeded == null ? null : noLifeProlongingIfDyingAcceptanceNeeded.getDatabaseValue();
    }

}
