/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.integrations.personInformation.impl;


import dk.sundhedsdatastyrelsen.livstestamente.auditlog.AuditLogger;
import dk.sundhedsdatastyrelsen.livstestamente.exceptions.ValidationException;
import dk.sundhedsdatastyrelsen.livstestamente.exceptions.ValidationServiceException;
import dk.sundhedsdatastyrelsen.livstestamente.integrations.personInformation.PersonInformation;
import org.apache.log4j.Logger;

import java.time.Instant;
import java.util.LinkedList;

public class PersonInformationImpl implements PersonInformation {

	private static final Logger LOGGER = Logger.getLogger(PersonInformationImpl.class);

	private PersonInformationServiceClient personinformationServiceClient;

    private final PersonInformation.ValidationLevel validationLevel;
	private final int cprMinValidationAge;
	private final AuditLogger auditLogger;

	public PersonInformationImpl(ValidationLevel validationLevel,
								 PersonInformationServiceClientFactory personInformationServiceClientFactory,
								 int cprMinValidationAge,
								 AuditLogger auditLogger) {
		this.validationLevel = validationLevel;
		this.personinformationServiceClient = personInformationServiceClientFactory.createClient();
        this.cprMinValidationAge = cprMinValidationAge;
		this.auditLogger = auditLogger;
	}

	 // For test purposes. Exists because autowired PersonInformation fields need another PersonInformationServiceClient
	 // than the one created in WebServiceConfig.
	@Override
	public void setServiceClient(PersonInformationServiceClient sc) { personinformationServiceClient = sc; }

	@Override
    public void validateAge(String patientId) throws ValidationException {
        int age = personinformationServiceClient.getAge(patientId);
        if (age < cprMinValidationAge) {
            throw new ValidationException(
                    "Patienten afvises, da patienten er " + age + " år gammel, men minimum skal være " + cprMinValidationAge + " år gammel.");
        }
    }

	@Override
	public boolean validateCpr(String patientId) throws ValidationException {
		
		if (!ValidationLevel.OFF.equals(validationLevel)) {
			try {
				PersonInformationCprStatus personInformationCprStatus = personinformationServiceClient.validateCpr(patientId);
				boolean cprOk = (personInformationCprStatus != null && !personInformationCprStatus.isInactive());
				if (!cprOk && ValidationLevel.WARNING.equals(validationLevel)) {
					if (personInformationCprStatus == null) {
						auditLogger.logCprDoesNotExist(patientId);
					} else {
						auditLogger.logCprInactive(patientId);
					}
					
					return false;
				} else if (!cprOk && ValidationLevel.REJECT.equals(validationLevel)) {
					if (personInformationCprStatus == null) {
						throw new ValidationException("Patientid eller status eksisterer ikke");	
					} else {
						throw new ValidationException("Patientens status er inaktiv");
					}
				}
			} catch (ValidationServiceException e) {
				LOGGER.error("Kaldet til PersonInformation fejlede", e);
				throw new ValidationException(e.getMessage());
			}
		}	
		return true;
	}

	@Override
	public LinkedList<String> getPersonsDeceased(LinkedList<String> listOfCpr, Instant latestDeceasedTime) throws ValidationException {
		return personinformationServiceClient.getDeceased(listOfCpr, latestDeceasedTime);
	}
}
