/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.integrations.personInformation.impl;

import dk.sundhedsdatastyrelsen.livstestamente.exceptions.ValidationServiceException;

import java.time.Instant;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

import dk.sundhedsdatastyrelsen.livstestamente.util.HttpClientFactory;
import org.apache.commons.httpclient.HttpStatus;
import org.apache.http.client.HttpClient;
import org.apache.http.client.methods.HttpUriRequest;
import org.junit.*;
import org.mockito.Mockito;
import org.mockserver.client.MockServerClient;
import org.mockserver.integration.ClientAndServer;
import org.mockserver.model.HttpRequest;
import org.mockserver.model.HttpResponse;
import org.mockserver.model.MediaType;

import static org.junit.Assert.*;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

public class PersonInformationClientFactoryTest {

	private static final String CPR_EXISTING_AND_ACTIVE = "1111";
	private static final String CPR_NONEXISTING = "2222";
	private static final String CPR_EXISTING_AND_INACTIVE = "3333";
	private static final String CPR_WRONG_ANSWER = "WRONG";
	private static final String CPR_ERROR_404_LB = "404";
	
	private static final String STATUS_ACTIVE = "1";
	private static final String STATUS_INACTIVE = "30";
	
	private List<Integer> inactiveStatusList = Arrays.asList(30, 50, 60);

	private PersonInformationServiceClientFactory subject;

	private static ClientAndServer mockServer;

	@BeforeClass
	public static void startServer() {
		mockServer = ClientAndServer.startClientAndServer(1080);
	}

	@AfterClass 
	public static void stopServer() { 
		mockServer.stop();
	}
	
	@Before
	public void setup() {
		MockServerClient msc = new MockServerClient("localhost", 1080);

		msc.when(
			HttpRequest.request()
				.withMethod("GET")
				.withPath("/v1/person/"+CPR_EXISTING_AND_ACTIVE+"/status")
		).respond(
			HttpResponse.response()
				.withContentType(MediaType.APPLICATION_JSON)
				.withBody("{ \"status\": "+STATUS_ACTIVE+"}")
				.withStatusCode(HttpStatus.SC_OK)
		);

		msc.when(
				HttpRequest.request()
				.withMethod("GET")
				.withPath("/v1/person/"+CPR_NONEXISTING+"/status")
				).respond(
						HttpResponse.response()
						.withContentType(MediaType.APPLICATION_JSON)
						.withBody("{}")
						.withStatusCode(HttpStatus.SC_NOT_FOUND)
						);

		msc.when(
				HttpRequest.request()
				.withMethod("GET")
				.withPath("/v1/person/"+CPR_EXISTING_AND_INACTIVE+"/status")
				).respond(
						HttpResponse.response()
						.withContentType(MediaType.APPLICATION_JSON)
						.withBody("{ \"status\": "+STATUS_INACTIVE+"}")
						.withStatusCode(HttpStatus.SC_OK)
						);

		msc.when(
				HttpRequest.request()
				.withMethod("GET")
				.withPath("/v1/person/"+CPR_WRONG_ANSWER+"/status")
				).respond(
						HttpResponse.response()
						.withContentType(MediaType.ANY_AUDIO_TYPE)
						.withBody("blabla")
						.withStatusCode(HttpStatus.SC_PAYMENT_REQUIRED)
						);

		msc.when(
				HttpRequest.request()
						.withMethod("GET")
						.withPath("/v1/person/"+CPR_ERROR_404_LB+"/status")
		).respond(
				HttpResponse.response()
						.withStatusCode(HttpStatus.SC_NOT_FOUND)
						.withContentType(MediaType.TEXT_PLAIN)
						.withBody("/v1/status/"+CPR_ERROR_404_LB)
		);


		HttpClientFactory httpClientFactory = new HttpClientFactory();
		subject = new PersonInformationServiceClientFactory("http://localhost:1080/v1", httpClientFactory, 1, HttpStatus.SC_NON_AUTHORITATIVE_INFORMATION, inactiveStatusList);
	}

	@Test
	public void testPersonInformationServiceClientThrowsCorrectExceptionIfAgeDoesNotExistsForCpr() {
		HttpClientFactory httpCF = mock(HttpClientFactory.class);
		when(httpCF.getClient()).thenReturn(new Return404HttpClient());
		PersonInformationServiceClient sc = new PersonInformationServiceClient(httpCF,
				"dummy",
				new PersonInformationServiceClientFactory.ServiceErrorCallback(),
				new LinkedList<>());

		final String nonExistingCpr = "9998885553";
		String exMsg = assertThrows(ValidationServiceException.class, () -> sc.getAge(nonExistingCpr)).getMessage();
		assertEquals(
				"CPR-nummeret eller alder kunne ikke valideres, fordi kaldet til PersonInformation-servicen returnerede en " +
						"uventet statuskode: 404",
				exMsg);
	}

	@Test
	public void testExistingCprWithActiveStatus() {

		// Given

		// When
		PersonInformationServiceClient client = subject.createClient();
		PersonInformationCprStatus personInformationCprStatus = client.validateCpr(CPR_EXISTING_AND_ACTIVE);

		// Then
		Assert.assertFalse(personInformationCprStatus.isInactive());
		Assert.assertEquals(HttpStatus.SC_OK, subject.getStatusCode());
	}
	
	@Test
	public void testExistingCprWithInactiveStatus() {

		// Given

		// When
		PersonInformationServiceClient client = subject.createClient();
		PersonInformationCprStatus personInformationCprStatus = client.validateCpr(CPR_EXISTING_AND_INACTIVE);

		// Then
		Assert.assertTrue(personInformationCprStatus.isInactive());
		Assert.assertEquals(HttpStatus.SC_OK, subject.getStatusCode());
	}

	@Ignore
	@Test(expected = ValidationServiceException.class)
	public void testErrorServiceConfig() {

		// Given

		// When
		PersonInformationServiceClient client = subject.createClient();
		try {
			client.validateCpr(CPR_ERROR_404_LB);
		} catch (ValidationServiceException e) {
			// Then
			Assert.assertEquals(HttpStatus.SC_NON_AUTHORITATIVE_INFORMATION, subject.getStatusCode());
			Assert.assertEquals("Fejl i retursvar fra PersonInformation", e.getMessage());
			throw e;
		}
	}
	
	@Test(expected = ValidationServiceException.class)
	public void testUnexpectedResponseFromService() {

		// Given

		// When
		PersonInformationServiceClient client = subject.createClient();
		try {
			client.validateCpr(CPR_WRONG_ANSWER);
		} catch (ValidationServiceException e) {
			// Then
			Assert.assertEquals(HttpStatus.SC_NON_AUTHORITATIVE_INFORMATION, subject.getStatusCode());
			Assert.assertEquals("CPR-nummeret eller alder kunne ikke valideres, fordi kaldet til PersonInformation-servicen returnerede en uventet statuskode: 402", e.getMessage());
			throw e;
		}
	}

	@Test
	public void testGetDeceasedReturnsCprsFromRequestUri() throws Exception {
		LinkedList<String> cprs = new LinkedList<>();
		IntStream.range(0, 50000).forEach(i -> cprs.add(String.format("%010d", i)));
		Instant now = Instant.now();
		testBuildDeceasedRequestsHelper(cprs, now);
	}

	@Test
	public void testDeceasedRequestsWithEmptyCprList() throws Exception {
		LinkedList<String> cprs = new LinkedList<>();
		Instant now = Instant.now();
		testBuildDeceasedRequestsHelper(cprs, now);
	}

	private void testBuildDeceasedRequestsHelper(LinkedList<String> cprs, Instant now) throws Exception {
		HttpClientFactory httpCF = Mockito.mock(HttpClientFactory.class);
		HttpClient client = Mockito.mock(HttpClient.class);
		Mockito.when(httpCF.getClient()).thenReturn(client);

		PersonInformationServiceClient personInformationClient = new PersonInformationServiceClient(httpCF,
				"dummy",
				new PersonInformationServiceClientFactory.ServiceErrorCallback(),
				new LinkedList<>());

		HttpUriRequest request = personInformationClient.buildDeceasedRequests(cprs, now);

		String query = request.getURI().getQuery();
		List<String> cprsFromRequest = Arrays.stream(query.split("&"))
				.filter(param -> param.startsWith("cpr="))
				.map(param -> param.substring(4))
				.collect(Collectors.toList());
		Assert.assertEquals("Cpr request og cpr i uri burde være ens", cprs, cprsFromRequest);
	}
}
