/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.integrations.personInformation.impl;

import dk.sundhedsdatastyrelsen.livstestamente.auditlog.AuditLogger;
import dk.sundhedsdatastyrelsen.livstestamente.exceptions.ValidationException;
import dk.sundhedsdatastyrelsen.livstestamente.integrations.personInformation.PersonInformation;
import dk.sundhedsdatastyrelsen.livstestamente.util.HttpClientFactory;
import org.apache.http.Header;
import org.apache.http.HttpResponse;
import org.apache.http.HttpStatus;
import org.apache.http.ProtocolVersion;
import org.apache.http.client.HttpClient;
import org.apache.http.client.methods.HttpUriRequest;
import org.apache.http.entity.ContentType;
import org.apache.http.entity.StringEntity;
import org.apache.http.message.BasicStatusLine;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;
import org.mockito.Mockito;

import java.nio.charset.UnsupportedCharsetException;
import java.time.Instant;
import java.time.Period;
import java.time.ZonedDateTime;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import static org.mockito.Mockito.*;

public class PersonInformationImplTest {
	private PersonInformationServiceClientFactory personInformationServiceClientFactory;
	private PersonInformationServiceClient personinformationServiceClient;

	private static final String VALID_ACTIVE_CPR = "valid";
	private static final String VALID_INACTIVE_CPR = "inactive";
	private static final String INVALID_CPR = "invalid";
	private static final int CPR_EXISTS_MIN_AGE = 18;
	private static final int inactiveStatus = 999;
	private AuditLogger auditLogger;
	private final Instant latestDeceasedTime = ZonedDateTime.now().minus(Period.parse("P12M")).toInstant();
	private HttpResponse mockHttpResponse;

	@Before
	public void setup() throws Exception {
		HttpClientFactory httpClientFactory = mock(HttpClientFactory.class);
		HttpClient mockHttpClient = Mockito.mock(HttpClient.class);
		Mockito.when(httpClientFactory.getClient()).thenReturn(mockHttpClient);

		PersonInformationServiceClientFactory.ServiceErrorCallback errorCallback = Mockito.mock(PersonInformationServiceClientFactory.ServiceErrorCallback.class);
		List<Integer> inactiveStatusList = new ArrayList<>();
		inactiveStatusList.add(inactiveStatus);
		personinformationServiceClient = Mockito.spy(new PersonInformationServiceClient(httpClientFactory, "test", errorCallback, inactiveStatusList));

		// Mock the HttpResponse
		mockHttpResponse = Mockito.mock(HttpResponse.class);
		Mockito.when(mockHttpResponse.getStatusLine()).thenReturn(new BasicStatusLine(new ProtocolVersion("HTTP", 1, 1), HttpStatus.SC_OK, "OK"));
		Header mockHeader = Mockito.mock(Header.class);
		Mockito.when(mockHeader.getValue()).thenReturn("application/json");
		Mockito.when(mockHttpResponse.getFirstHeader(anyString())).thenReturn(mockHeader);

		// Mock the invokeService method to call handleResponse
		Mockito.doAnswer(invocation -> {
			HttpUriRequest request = invocation.getArgument(0);
			PersonInformationServiceClient.requestTypes requestType = invocation.getArgument(1);
			return personinformationServiceClient.handleResponse(mockHttpResponse, requestType);
		}).when(personinformationServiceClient).invokeService(Mockito.any(HttpUriRequest.class), Mockito.any(PersonInformationServiceClient.requestTypes.class));

		personInformationServiceClientFactory = Mockito.mock(PersonInformationServiceClientFactory.class);
		Mockito.when(personInformationServiceClientFactory.createClient()).thenReturn(personinformationServiceClient);

		auditLogger = Mockito.mock(AuditLogger.class);
	}

	@Test
	public void testValidationOffAndInvalidCpr() throws ValidationException {
		
		// Given
		PersonInformationImpl subject = new PersonInformationImpl(PersonInformation.ValidationLevel.OFF, personInformationServiceClientFactory,
				CPR_EXISTS_MIN_AGE, auditLogger);

		// When
		String cprWasInvalidResponse = "{}";
		mockResponse(cprWasInvalidResponse);
		boolean respons = subject.validateCpr(INVALID_CPR);
		
		// Then
		Assert.assertTrue(respons);
	}
	
	@Test
	public void testValidationOffAndInactiveCpr() throws ValidationException {
		
		// Given
		PersonInformationImpl subject = new PersonInformationImpl(PersonInformation.ValidationLevel.OFF, personInformationServiceClientFactory,
				CPR_EXISTS_MIN_AGE, auditLogger);

		// When
		String cprWasValidResponse = "{\"status\": 0}";
		mockResponse(cprWasValidResponse);
		boolean respons = subject.validateCpr(VALID_ACTIVE_CPR);
		
		// Then
		Assert.assertTrue(respons);
	}
	
	@Test
	public void testValidationWarningGeneratesAuditEntryOnInvalidCpr() throws ValidationException {
		
		// Given
		PersonInformationImpl subject = new PersonInformationImpl(PersonInformation.ValidationLevel.WARNING, personInformationServiceClientFactory,
				CPR_EXISTS_MIN_AGE, auditLogger);

		// When
		String cprWasValidResponse = "{\"status\": 1}";
		mockResponse(cprWasValidResponse);
		boolean respons = subject.validateCpr(VALID_ACTIVE_CPR);
		
		// Then
		Assert.assertTrue(respons);
	}
	
	@Test
	public void testValidationWarningGeneratesAuditEntryOnInactiveCpr() throws ValidationException {
		
		// Given
		PersonInformationImpl subject = new PersonInformationImpl(PersonInformation.ValidationLevel.WARNING, personInformationServiceClientFactory,
				CPR_EXISTS_MIN_AGE, auditLogger);

		// When
		String cprWasValidResponse = "{\"status\": " + inactiveStatus + "}";
		mockResponse(cprWasValidResponse);
		boolean respons = subject.validateCpr(VALID_INACTIVE_CPR);
		
		// Then
		Assert.assertFalse(respons);
		Mockito.verify(auditLogger, Mockito.times(1)).logCprInactive(VALID_INACTIVE_CPR);
	}

	@Test
	public void testValidationWarningGeneratesNoAuditEntryOnValidActiveCpr() throws ValidationException {
		
		// Given
		PersonInformationImpl subject = new PersonInformationImpl(PersonInformation.ValidationLevel.WARNING, personInformationServiceClientFactory,
				CPR_EXISTS_MIN_AGE, auditLogger);

		// When
		String cprWasValidResponse = "{\"status\": 0}";
		mockResponse(cprWasValidResponse);
		boolean respons = subject.validateCpr(VALID_ACTIVE_CPR);
		
		// Then
		Assert.assertTrue(respons);
		Mockito.verifyZeroInteractions(auditLogger);
	}

	
	@Test
	public void testValidationRejectOnValidActiveCpr() throws ValidationException {
		
		// Given
		PersonInformationImpl subject = new PersonInformationImpl(PersonInformation.ValidationLevel.REJECT, personInformationServiceClientFactory,
				CPR_EXISTS_MIN_AGE, auditLogger);

		// When
		mockResponse("{\"status\": 0}");
		boolean respons = subject.validateCpr(VALID_ACTIVE_CPR);
		
		// Then
		Assert.assertTrue(respons);
		Mockito.verify(personinformationServiceClient, Mockito.times(1)).validateCpr(Mockito.anyString());
	}

	@Test(expected = ValidationException.class)
	public void testValidationRejectOnInValidCpr() throws ValidationException {
		
		// Given
		PersonInformationImpl subject = new PersonInformationImpl(PersonInformation.ValidationLevel.REJECT, personInformationServiceClientFactory,
				CPR_EXISTS_MIN_AGE, auditLogger);
		String invalidPatientCpr = INVALID_CPR;

		// When
		try {
			mockResponse("{}");
			subject.validateCpr(invalidPatientCpr);
		} catch (ValidationException e) {
			// Then
			Mockito.verify(personinformationServiceClient, Mockito.times(1)).validateCpr(Mockito.anyString());
			throw e;
		}
	}

	@Test
	public void validationAgeYoungerThanMinAge() {
		int patientAge = CPR_EXISTS_MIN_AGE - 1;
		Assert.assertThrows(ValidationException.class, () -> validateAge(patientAge, PersonInformation.ValidationLevel.OFF));
		Assert.assertThrows(ValidationException.class, () -> validateAge(patientAge, PersonInformation.ValidationLevel.WARNING));
		Assert.assertThrows(ValidationException.class, () -> validateAge(patientAge, PersonInformation.ValidationLevel.REJECT));
	}

	@Test
	public void validationAgeAtMinAge() throws ValidationException {
		int patientAge = CPR_EXISTS_MIN_AGE;
		validateAge(patientAge, PersonInformation.ValidationLevel.OFF);
		validateAge(patientAge, PersonInformation.ValidationLevel.WARNING);
		validateAge(patientAge, PersonInformation.ValidationLevel.REJECT);
	}

	@Test
	public void testPersonsDeceasedOneYearAgoOffAndInvalidCpr() throws ValidationException {

		// Given
		PersonInformationImpl subject = new PersonInformationImpl(PersonInformation.ValidationLevel.OFF, personInformationServiceClientFactory,
				CPR_EXISTS_MIN_AGE, auditLogger);
		LinkedList<String> cprs = new LinkedList<>();
		cprs.add("1234567890");
		cprs.add("0987654321");

		// When
		mockResponse("[{\"cpr\": \"1234567890\"}, {\"cpr\": \"0987654321\"}]");
		LinkedList<String> respons = subject.getPersonsDeceased(cprs, latestDeceasedTime);

		// Then
		Assert.assertNotNull(respons);
		Assert.assertEquals(2, respons.size());
		Assert.assertEquals("1234567890", respons.get(0));
		Assert.assertEquals("0987654321", respons.get(1));
	}

	@Test
	public void testPersonsDeceasedOneYearAgoOffAndInvalidCpr2() throws ValidationException {
		// Given
		PersonInformationImpl subject = new PersonInformationImpl(PersonInformation.ValidationLevel.OFF, personInformationServiceClientFactory,
				CPR_EXISTS_MIN_AGE, auditLogger);
		LinkedList<String> cprs = new LinkedList<>();
		cprs.add("1234567890");
		cprs.add("0987654321");
		cprs.add("invalid");

		// When
		mockResponse("[{\"cpr\": \"1234567890\"}, {\"cpr\": \"0987654321\"}]");
		LinkedList<String> respons = subject.getPersonsDeceased(cprs , latestDeceasedTime);

		// Then
		Assert.assertNotNull(respons);
		Assert.assertEquals(2, respons.size());
		Assert.assertEquals("1234567890", respons.get(0));
		Assert.assertEquals("0987654321", respons.get(1));
	}

	private void validateAge(int validationAge, PersonInformation.ValidationLevel vl) throws ValidationException {
		mockResponse("{\"age\": " + validationAge + "}");
		new PersonInformationImpl(vl, personInformationServiceClientFactory, CPR_EXISTS_MIN_AGE, auditLogger).validateAge(VALID_ACTIVE_CPR);
	}

	private void mockResponse(String response) {
		try {
			Mockito.when(mockHttpResponse.getEntity()).thenReturn(new StringEntity(response, ContentType.APPLICATION_JSON));
		} catch (UnsupportedCharsetException e) {
			throw new RuntimeException(e);
		}
	}
}
