/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsp.joboperations.setup;

import dk.nsp.backgroundjob.service.StackedOperationsService;
import dk.nsp.backgroundjob.util.DateTimeProvider;
import dk.nsp.joboperations.factory.CleanupOperationsFactory;
import dk.nsp.joboperations.stack.PersonInformationService;
import dk.nsp.joboperations.stack.livingwilldeceasedcleanup.LivingWillDeceasedCleanupSupplier;
import dk.nsp.joboperations.stack.treatmentWilldeceasedcleanup.TreatmentWillDeceasedCleanupSupplier;
import dk.sds.nsp.audit.Audit;
import dk.sundhedsdatastyrelsen.livstestamente.dao.LivingWillDAO;
import dk.sundhedsdatastyrelsen.livstestamente.dao.TreatmentWillDAO;
import dk.sundhedsdatastyrelsen.livstestamente.auditlog.AuditLogger;
import dk.sundhedsdatastyrelsen.livstestamente.integrations.personInformation.PersonInformation;
import dk.sundhedsdatastyrelsen.livstestamente.integrations.personInformation.impl.PersonInformationImpl;
import dk.sundhedsdatastyrelsen.livstestamente.integrations.personInformation.impl.PersonInformationServiceClientFactory;
import dk.sundhedsdatastyrelsen.livstestamente.util.HttpClientFactory;
import dk.sundhedsdatastyrelsen.livstestamente.conf.Properties;
import org.apache.http.HttpStatus;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.*;

@Configuration
@Import({DBSetup.class})
@ComponentScan("dk.sundhedsdatastyrelsen.livstestamente")
public class CleanupJobConfiguration {

    private final Properties properties;

    @Autowired
    public CleanupJobConfiguration(Properties properties) {
        this.properties = properties;
    }

    @Bean
    public CleanupOperationsFactory cleanupOperationsFactory(LivingWillDAO livingWillDAO,
                                                             TreatmentWillDAO treatmentWillDAO,
                                                             DateTimeProvider dateTimeProvider,
                                                             PersonInformationService personInformationService) {
        return new CleanupOperationsFactory(
                livingWillDAO,
                treatmentWillDAO,
                dateTimeProvider,
                properties.getDeletionJobsBatchSize(),
                properties.getDeceasedPeriod(),
                personInformationService);
    }

    @Bean
    public DateTimeProvider dateTimeProvider() {
        return new DateTimeProvider();
    }

    @Bean
    PersonInformationService personInformationService(PersonInformation personInformation, PersonInformationServiceClientFactory personInformationServiceClientFactory) {
        return new PersonInformationService(personInformation, personInformationServiceClientFactory);
    }

    @Bean
    public PersonInformation personInformation(PersonInformation.ValidationLevel validationLevel,
                                               PersonInformationServiceClientFactory personInformationServiceClientFactory,
                                               AuditLogger auditLogger) {
        return new PersonInformationImpl(
                validationLevel,
                personInformationServiceClientFactory,
                properties.getPersonInformationMinAge(),
                auditLogger);
    }

    @Bean
    public PersonInformation.ValidationLevel getValidationLevel() {
        try {
            return PersonInformation.ValidationLevel.valueOf(properties.getPersonInformationValidationLevel());
        } catch (IllegalArgumentException e) {
            throw new IllegalArgumentException(
                    "Værdien af personInformation.validationlevel skal være OFF, REJECT eller WARNING", e);
        }
    }

    @Bean
    public PersonInformationServiceClientFactory personInformationServiceClientFactory(HttpClientFactory httpClientFactory) {
        if (properties.getPersonInformationUrl() == null) {
            throw new IllegalArgumentException("personInformation.url skal være sat, da denne skal bruges til " +
                    "påkrævet validering af alder.");
        }
        return new PersonInformationServiceClientFactory(
                properties.getPersonInformationUrl(),
                httpClientFactory,
                properties.getPersonInformationErrorTolerance(),
                (PersonInformation.ValidationLevel.REJECT.getValidationLevel()
                        .equals(properties.getPersonInformationValidationLevel()) ?
                        HttpStatus.SC_INTERNAL_SERVER_ERROR : HttpStatus.SC_NON_AUTHORITATIVE_INFORMATION),
                properties.getInactiveStatusList());
    }

    @Bean
    public HttpClientFactory getHttpClientFactory() {
        return new HttpClientFactory(properties.getMaxTotalConnections(),
                properties.getDefaultMaxConnectionsPerRoute());
    }

    @Bean
    public AuditLogger auditLogger() {
        return new AuditLogger(Audit::createAuditBuilder);
    }

    @Bean
    public StackedOperationsService livingWillDeceasedCleanupService(CleanupOperationsFactory cleanupOperationsFactory, DateTimeProvider dateTimeProvider) {
        LivingWillDeceasedCleanupSupplier defaultStackOperation = new LivingWillDeceasedCleanupSupplier(cleanupOperationsFactory);
        return new StackedOperationsService(defaultStackOperation, properties.desiredExecutionDuration(), dateTimeProvider);

    }

    @Bean
    public StackedOperationsService treatmentWillDeceasedCleanupService(CleanupOperationsFactory cleanupOperationsFactory, DateTimeProvider dateTimeProvider) {
        TreatmentWillDeceasedCleanupSupplier defaultStackOperation = new TreatmentWillDeceasedCleanupSupplier(cleanupOperationsFactory);
        return new StackedOperationsService(defaultStackOperation, properties.desiredExecutionDuration(), dateTimeProvider);

    }
}
