/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsp.joboperations.stack.treatmentWilldeceasedcleanup;

import dk.nsp.backgroundjob.stack.StackOperation;
import dk.nsp.backgroundjob.util.BackgroundJobException;
import dk.nsp.backgroundjob.util.BackgroundJobUtils;
import dk.nsp.backgroundjob.util.DateTimeProvider;
import dk.nsp.joboperations.factory.CleanupOperationsFactory;
import dk.nsp.joboperations.stack.PersonInformationService;

import java.time.Instant;
import java.time.Period;
import java.util.List;
import java.util.stream.Collectors;

public class TreatmentWillDeceasedPatientIdBatchCleanupSupplier implements StackOperation {
    private final List<String> cprs;
    private final DateTimeProvider dateTimeProvider;
    private final CleanupOperationsFactory cleanupOperationsFactory;
    private final int batchSize;
    private final Period deceasedPeriod;
    private final PersonInformationService personInformationService;

    public TreatmentWillDeceasedPatientIdBatchCleanupSupplier(List<String> cprs,
                                                              DateTimeProvider dateTimeProvider,
                                                              CleanupOperationsFactory cleanupOperationsFactory,
                                                              int batchSize,
                                                              Period deceasedPeriod,
                                                              PersonInformationService personInformationService) {
        this.cprs = cprs;
        this.dateTimeProvider = dateTimeProvider;
        this.cleanupOperationsFactory = cleanupOperationsFactory;
        this.batchSize = batchSize;
        this.deceasedPeriod = deceasedPeriod;
        this.personInformationService = personInformationService;
    }

    @Override
    public List<StackOperation> run() {
        Instant latestDeceasedTime = dateTimeProvider.now().minus(deceasedPeriod).toInstant();
        List<String> deceasedCitizens = getDeceasedCitizens(cprs, latestDeceasedTime);

        return BackgroundJobUtils.createShuffledBatches(deceasedCitizens, batchSize).stream()
                .map(batch -> cleanupOperationsFactory.createTreatmentWillBatchCleanupOperation(batch).withOperationName("TreatmentWill Deceased Cleanup"))
                .collect(Collectors.toList());
    }

    private List<String> getDeceasedCitizens(List<String> personsFromBTR, Instant latestDeceasedTime) {
        try {
            return personInformationService.filterToDeceased(personsFromBTR, latestDeceasedTime);
        } catch (Exception e) {
            throw new BackgroundJobException("Failed to retrieve deceased citizens with latestStatusTime " + latestDeceasedTime, e);
        }
    }
}
