/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsp.joboperations.stack;

import dk.nsp.backgroundjob.health.HealthIndication;
import dk.nsp.backgroundjob.util.BackgroundJobException;
import dk.nsp.backgroundjob.util.DateTimeProvider;
import dk.sundhedsdatastyrelsen.livstestamente.integrations.personInformation.impl.PersonInformationServiceClientFactory;
import dk.sundhedsdatastyrelsen.livstestamente.exceptions.ValidationException;
import dk.sundhedsdatastyrelsen.livstestamente.integrations.personInformation.PersonInformation;
import org.junit.Before;
import org.junit.Test;
import org.mockito.Mockito;

import java.time.Instant;
import java.util.*;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.doAnswer;

public class PersonInformationServiceTest {

    private PersonInformationService subject;
    private PersonInformation personInformationClient;
    private PersonInformationServiceClientFactory personInformationServiceClientFactory;
    private Instant latestDeceasedTime;

    @Before
    public void setUp() {
        personInformationClient = Mockito.mock(PersonInformation.class);
        personInformationServiceClientFactory = Mockito.mock(PersonInformationServiceClientFactory.class);
        subject = new PersonInformationService(personInformationClient, personInformationServiceClientFactory);
        DateTimeProvider dateTimeProvider = new DateTimeProvider();
        latestDeceasedTime = dateTimeProvider.now().toInstant();
    }

    @Test
    public void getDeceasedCitizens_WithNonExistingCitizen_ReturnsEmptyList() throws ValidationException {

        // Arrange
        String nonExistingCitizen = "0000000000";
        LinkedList<String> responseCprs = new LinkedList<>();
        doAnswer(arg -> responseCprs).when(personInformationClient).getPersonsDeceased(any(), any());

        // Act
        List<String> result = subject.filterToDeceased(Collections.singletonList(nonExistingCitizen), latestDeceasedTime);

        // Assert
        assertTrue(result.isEmpty());
    }

    @Test
    public void getDeceasedCitizens_WithEmptyCitizenList_ReturnsEmptyList() throws ValidationException {

        // Arrange
        LinkedList<String> responseCprs = new LinkedList<>();
        doAnswer(arg -> responseCprs).when(personInformationClient).getPersonsDeceased(any(), any());

        // Act
        List<String> result = subject.filterToDeceased(Collections.emptyList(), latestDeceasedTime);

        // Assert
        assertTrue(result.isEmpty());

    }

    @Test
    public void getDeceasedCitizens_WithOneOutOfThreeDeceasedCitizens_ReturnsDeceasedCitizen() throws ValidationException {

        // Arrange
        String nonExistingCitizen = "0000000000";
        String deceasedCitizen = "0000000090";
        String futureDeceasedCitizen = "0000000091";
        LinkedList<String> responseCprs = new LinkedList<>();
        responseCprs.add(deceasedCitizen);
        doAnswer(arg -> responseCprs).when(personInformationClient).getPersonsDeceased(any(), any(Instant.class));

        // Act
        List<String> result = subject.filterToDeceased(Arrays.asList(nonExistingCitizen, deceasedCitizen, futureDeceasedCitizen), latestDeceasedTime);

        // Assert
        assertEquals(1, result.size());
    }

    @Test
    public void getDeceasedCitizens_WithThreeOutOfThreeDeceasedCitizens_ReturnsDeceasedCitizen() throws ValidationException {

        // Arrange
        String deceasedCitizen1 = "0000000091";
        String deceasedCitizen2 = "0000000092";
        String deceasedCitizen3 = "0000000093";
        LinkedList<String> responseCprs = new LinkedList<>();
        responseCprs.add(deceasedCitizen1);
        responseCprs.add(deceasedCitizen2);
        responseCprs.add(deceasedCitizen3);
        doAnswer(arg -> responseCprs).when(personInformationClient).getPersonsDeceased(any(), any());

        // Act
        List<String> result = subject.filterToDeceased(Arrays.asList(deceasedCitizen1, deceasedCitizen2, deceasedCitizen3), latestDeceasedTime);

        // Assert
        assertEquals(3, result.size());
    }

    @Test(expected = BackgroundJobException.class)
    public void filterToDeceased_WhenClientThrowsException_ThrowsBackgroundJobException() throws ValidationException {
        // Arrange
        Mockito.doThrow(new ValidationException("Test Exception")).when(personInformationClient).getPersonsDeceased(any(), any());

        // Act
        subject.filterToDeceased(Collections.singletonList("0000000000"), latestDeceasedTime);
    }

    @Test
    public void getHealthInfo_WhenServiceIsHealthy_ReturnsOkStatus() {
        // Arrange
        Mockito.when(personInformationServiceClientFactory.isHealthy()).thenReturn(true);

        // Act
        HealthIndication healthIndication = subject.getHealthInfo();

        // Assert
        assertEquals(200, healthIndication.getStatus());
        assertEquals("\"PersonInformation\": \"OK\"", healthIndication.getText());
    }

    @Test
    public void getHealthInfo_WhenServiceIsUnhealthy_ReturnsErrorStatus() {
        // Arrange
        Mockito.when(personInformationServiceClientFactory.isHealthy()).thenReturn(false);

        // Act
        HealthIndication healthIndication = subject.getHealthInfo();

        // Assert
        assertEquals(500, healthIndication.getStatus());
        assertEquals("\"PersonInformation\": \"ERROR\"", healthIndication.getText());
    }
}
