/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsp.joboperations.stack.treatmentWilldeceasedcleanup;

import dk.nsp.backgroundjob.stack.StackOperation;
import dk.nsp.backgroundjob.util.BackgroundJobException;
import dk.nsp.backgroundjob.util.DateTimeProvider;
import dk.nsp.joboperations.factory.CleanupOperationsFactory;
import dk.nsp.joboperations.stack.PersonInformationService;
import dk.sundhedsdatastyrelsen.livstestamente.exceptions.DAOException;
import org.junit.Before;
import org.junit.Test;
import org.mockito.Mockito;

import java.time.Period;
import java.time.ZonedDateTime;
import java.util.Arrays;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import static dk.nsp.joboperations.stack.TestUtils.assertThrows;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

public class TreatmentWillDeceasedPatientIdBatchCleanupSupplierTest {
    private DateTimeProvider dateTimeProvider;
    private CleanupOperationsFactory cleanupOperationsFactory;
    private PersonInformationService personInformationService;

    @Before
    public void setUp() throws Exception {
        dateTimeProvider = Mockito.mock(DateTimeProvider.class);
        cleanupOperationsFactory = Mockito.mock(CleanupOperationsFactory.class);
        personInformationService = Mockito.mock(PersonInformationService.class);
    }

    @Test
    public void run_WithNoCprs_ReturnsEmptyList() throws DAOException {
        // Arrange
        LinkedList<String> citizenCprs = new LinkedList<>(Collections.singletonList("1234567890"));
        TreatmentWillDeceasedPatientIdBatchCleanupSupplier treatmentWillDeceasedPatientIdBatchCleanupSupplier =
                setupMocksForSuccessfulOperation(citizenCprs, new LinkedList<>());

        // Act
        List<StackOperation> stackOperations = treatmentWillDeceasedPatientIdBatchCleanupSupplier.run();

        // Assert
        assertEquals(0, stackOperations.size());
    }

    @Test
    public void run_With3CprsForCitizensBatchSize1_Returns2StackOperationWithCprEntries() throws DAOException {
        // Arrange
        LinkedList<String> citizenCprs = new LinkedList<>(Arrays.asList("0000000000", "1111111111", "2222222222"));
        LinkedList<String> deceasedCprs = new LinkedList<>(Arrays.asList("1111111111", "2222222222"));
        TreatmentWillDeceasedPatientIdBatchCleanupSupplier treatmentWillDeceasedPatientIdBatchCleanupSupplier =
                setupMocksForSuccessfulOperation(citizenCprs, deceasedCprs);

        TreatmentWillBatchCleanupOperation batchOperation1 = Mockito.mock(TreatmentWillBatchCleanupOperation.class);
        Mockito.when(batchOperation1.withOperationName(Mockito.anyString())).thenReturn(batchOperation1);
        Mockito.when(cleanupOperationsFactory.createTreatmentWillBatchCleanupOperation(Collections.singletonList(deceasedCprs.get(0))))
                .thenReturn(batchOperation1);

        TreatmentWillBatchCleanupOperation batchOperation2 = Mockito.mock(TreatmentWillBatchCleanupOperation.class);
        Mockito.when(batchOperation2.withOperationName(Mockito.anyString())).thenReturn(batchOperation2);
        Mockito.when(cleanupOperationsFactory.createTreatmentWillBatchCleanupOperation(Collections.singletonList(deceasedCprs.get(1))))
                .thenReturn(batchOperation2);

        // Act
        List<StackOperation> stackOperations = treatmentWillDeceasedPatientIdBatchCleanupSupplier.run();

        // Assert
        assertEquals(2, stackOperations.size());
        assertTrue(stackOperations.contains(batchOperation1));
        assertTrue(stackOperations.contains(batchOperation2));
    }

    @Test
    public void run_WhenRetrievingDeceasedCitizensFails_ThrowsCleanupException() throws DAOException {
        // Arrange
        TreatmentWillDeceasedPatientIdBatchCleanupSupplier treatmentWillDeceasedPatientIdBatchCleanupSupplier = prepareUnsuccessfulOperationMocks();
        Mockito.when(personInformationService.filterToDeceased(Mockito.anyList(), Mockito.any())).thenThrow(RuntimeException.class);

        // Act
        BackgroundJobException e = assertThrows(BackgroundJobException.class, () -> treatmentWillDeceasedPatientIdBatchCleanupSupplier.run());

        // Assert
        assertTrue(e.getMessage(), e.getMessage().contains("Failed to retrieve deceased citizens with latestStatusTime "));
    }

    private TreatmentWillDeceasedPatientIdBatchCleanupSupplier setupMocksForSuccessfulOperation(LinkedList<String> citizenCprs, LinkedList<String> deceasedCprs) throws DAOException {
        TreatmentWillDeceasedPatientIdBatchCleanupSupplier treatmentWillDeceasedPatientIdBatchCleanupSupplier = new TreatmentWillDeceasedPatientIdBatchCleanupSupplier(
                citizenCprs, dateTimeProvider, cleanupOperationsFactory, 1, Period.ofMonths(1), personInformationService);
        Mockito.when(dateTimeProvider.now()).thenReturn(ZonedDateTime.now());
        Mockito.when(personInformationService.filterToDeceased(Mockito.anyList(), Mockito.any())).thenReturn(deceasedCprs);
        return treatmentWillDeceasedPatientIdBatchCleanupSupplier;
    }

    private TreatmentWillDeceasedPatientIdBatchCleanupSupplier prepareUnsuccessfulOperationMocks() throws DAOException {
        Mockito.when(dateTimeProvider.now()).thenReturn(ZonedDateTime.now());
        TreatmentWillDeceasedPatientIdBatchCleanupSupplier treatmentWillDeceasedPatientIdBatchCleanupSupplier = new TreatmentWillDeceasedPatientIdBatchCleanupSupplier(
                Collections.emptyList(), dateTimeProvider, cleanupOperationsFactory, 1, Period.ofMonths(1), personInformationService);
        return treatmentWillDeceasedPatientIdBatchCleanupSupplier;
    }
}