/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsp.joboperations.stack.treatmentWilldeceasedcleanup;

import dk.nsp.backgroundjob.stack.StackOperation;
import dk.nsp.backgroundjob.util.BackgroundJobException;
import dk.nsp.joboperations.factory.CleanupOperationsFactory;
import dk.sundhedsdatastyrelsen.livstestamente.dao.TreatmentWillDAO;
import dk.sundhedsdatastyrelsen.livstestamente.exceptions.DAOException;
import org.junit.Before;
import org.junit.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.ArgumentMatcher;
import org.mockito.Mockito;

import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;
import java.util.stream.Collectors;

import static dk.nsp.joboperations.stack.TestUtils.assertThrows;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

public class TreatmentWillDeceasedPatientIdPrefixCleanupSupplierTest {
    private TreatmentWillDAO treatmentWillDAO;
    private CleanupOperationsFactory cleanupOperationsFactory;

    @Before
    public void setUp() {
        treatmentWillDAO = Mockito.mock(TreatmentWillDAO.class);
        cleanupOperationsFactory = Mockito.mock(CleanupOperationsFactory.class);
    }

    @Test
    public void run_WithNoPatientIdsForPrefix_ReturnsEmptyList() throws DAOException {
        // Arrange
        TreatmentWillDeceasedPatientIdPrefixCleanupSupplier treatmentWillDeceasedPatientIdPrefixCleanupSupplier =
                new TreatmentWillDeceasedPatientIdPrefixCleanupSupplier(treatmentWillDAO, cleanupOperationsFactory, 5, "01");
        Mockito.when(treatmentWillDAO.getAllPersonsFromLTRWithPrefix(Mockito.anyString()))
                .thenReturn(new LinkedList<>());

        // Act
        List<StackOperation> stackOperations = treatmentWillDeceasedPatientIdPrefixCleanupSupplier.run();

        // Assert
        assertEquals(0, stackOperations.size());
    }

    @Test
    public void run_With3PatientIdsBatchSize3_ReturnsOneSupplier() throws DAOException {
        // Arrange
        String prefix = "01";
        TreatmentWillDeceasedPatientIdPrefixCleanupSupplier treatmentWillDeceasedPatientIdPrefixCleanupSupplier =
                new TreatmentWillDeceasedPatientIdPrefixCleanupSupplier(treatmentWillDAO, cleanupOperationsFactory, 3, prefix);
        LinkedList<String> patientIds = new LinkedList<>(Arrays.asList("1", "2", "3"));
        Mockito.when(treatmentWillDAO.getAllPersonsFromLTRWithPrefix(Mockito.anyString()))
                .thenReturn(patientIds);
        TreatmentWillDeceasedPatientIdBatchCleanupSupplier cleanupSupplier = Mockito.mock(TreatmentWillDeceasedPatientIdBatchCleanupSupplier.class);
        Mockito.when(cleanupOperationsFactory.createTreatmentWillDeceasedPatientIdBatchCleanupSupplier(Mockito.argThat(new ArgumentMatcher<List<String>>() {
            @Override
            public boolean matches(List<String> argument) {
                return argument.size() == patientIds.size() && argument.containsAll(patientIds);
            }
        }))).thenReturn(cleanupSupplier);

        // Act
        List<StackOperation> stackOperations = treatmentWillDeceasedPatientIdPrefixCleanupSupplier.run();

        // Assert
        assertEquals(1, stackOperations.size());
        assertTrue(stackOperations.contains(cleanupSupplier));
    }

    @Test
    public void run_With3PatientIdsBatchSize2_ReturnsTwoSuppliers() throws DAOException {
        // Arrange
        String prefix = "01";
        TreatmentWillDeceasedPatientIdPrefixCleanupSupplier treatmentWillDeceasedPatientIdPrefixCleanupSupplier =
                new TreatmentWillDeceasedPatientIdPrefixCleanupSupplier(treatmentWillDAO, cleanupOperationsFactory, 2, prefix);
        LinkedList<String> patientIds =  new LinkedList<>(Arrays.asList("1", "2", "3"));
        Mockito.when(treatmentWillDAO.getAllPersonsFromLTRWithPrefix(Mockito.anyString()))
                .thenReturn(patientIds);
        TreatmentWillDeceasedPatientIdBatchCleanupSupplier cleanupSupplier1 = Mockito.mock(TreatmentWillDeceasedPatientIdBatchCleanupSupplier.class);
        TreatmentWillDeceasedPatientIdBatchCleanupSupplier cleanupSupplier2 = Mockito.mock(TreatmentWillDeceasedPatientIdBatchCleanupSupplier.class);
        Mockito.when(cleanupOperationsFactory.createTreatmentWillDeceasedPatientIdBatchCleanupSupplier(Mockito.anyList()))
                .thenReturn(cleanupSupplier1).thenReturn(cleanupSupplier2);

        // Act
        List<StackOperation> stackOperations = treatmentWillDeceasedPatientIdPrefixCleanupSupplier.run();

        // Assert
        assertEquals(2, stackOperations.size());
        assertTrue(stackOperations.contains(cleanupSupplier1));
        assertTrue(stackOperations.contains(cleanupSupplier2));
        ArgumentCaptor<List> patientIdsCaptor = ArgumentCaptor.forClass(List.class);
        Mockito.verify(cleanupOperationsFactory, Mockito.times(2)).createTreatmentWillDeceasedPatientIdBatchCleanupSupplier(patientIdsCaptor.capture());
        List<String> patientIdArguments = (List<String>) patientIdsCaptor.getAllValues().stream().flatMap(list -> list.stream()).collect(Collectors.toList());
        assertEquals(patientIds.size(), patientIdArguments.size());
        assertTrue(patientIdArguments.containsAll(patientIds));
    }

    @Test
    public void run_WhenRetrievingDocumentEntriesFails_ThrowsCleanupException() throws DAOException {
        Mockito.when(treatmentWillDAO.getAllPersonsFromLTRWithPrefix(Mockito.anyString())).thenThrow(RuntimeException.class);
        TreatmentWillDeceasedPatientIdPrefixCleanupSupplier treatmentWillDeceasedPatientIdPrefixCleanupSupplier =
                new TreatmentWillDeceasedPatientIdPrefixCleanupSupplier(treatmentWillDAO, cleanupOperationsFactory, 2, "01");

        BackgroundJobException e = assertThrows(BackgroundJobException.class, () -> treatmentWillDeceasedPatientIdPrefixCleanupSupplier.run());

        assertEquals("Failed to retrieve CPR entry using prefix 01", e.getMessage());
    }
}