/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsp.joboperations.util;

import dk.idptest.SOAPOperation20180501;
import dk.idptest.ServiceCaller;
import dk.nsp.test.idp.builder.Builders;
import dk.nsp.test.idp.model.CitizenIdentity;
import dk.nsp.test.idp.vault.RemoteCredentialVault;
import dk.sundhedsdatastyrelsen.behandlingstestamente._2018._05._01.*;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Unmarshaller;
import java.io.StringReader;
import java.net.URI;

import static org.junit.Assert.assertNotNull;

public class BtrFacade {

    private final ServiceCaller serviceCaller;
    private final Unmarshaller unmarshallerTreatmentWill;
    private final String uriTreatmentWill;

    public BtrFacade(String uri) {
        try {
            JAXBContext jaxbContextTreatmentWill = JAXBContext.newInstance(dk.sundhedsdatastyrelsen.behandlingstestamente._2018._05._01.ObjectFactory.class);
            unmarshallerTreatmentWill = jaxbContextTreatmentWill.createUnmarshaller();
            serviceCaller = new ServiceCaller();
            this.uriTreatmentWill = uri + "/btr";
        } catch (JAXBException e) {
            throw new RuntimeException(e);
        }
    }

    public void createTreatmentWill(String cpr) throws Exception {
        CreateTreatmentWillResponse response = callService(
                getCreateTreatmentWillRequestBody(cpr),
                CreateTreatmentWillRequest.class,
                CreateTreatmentWillResponse.class,
                SOAPOperation20180501.CreateTreatmentWill.getSoapAction(),
                uriTreatmentWill,
                cpr
        );
        assertNotNull(response);
    }

    public boolean hasTreatmentWill(String cpr) throws Exception {
        return callService(
                getHasTreatmentWillRequestBody(cpr),
                HasTreatmentWillRequest.class,
                HasTreatmentWillResponse.class,
                SOAPOperation20180501.HasTreatmentWill.getSoapAction(),
                uriTreatmentWill,
                cpr
        ).isWillExists();
    }


    private CreateTreatmentWillRequest getCreateTreatmentWillRequestBody(String cpr) {
        // Dette er 2018.05.01 versionen af CreateTreatmentWillRequest
        String requestStrin = String.format(
                "<ns3:CreateTreatmentWillRequest xmlns:ns2=\"urn:hl7-org:btr\" xmlns:ns4=\"urn:hl7-org:btr\" xmlns:ns3=\"http://sundhedsdatastyrelsen.dk/behandlingstestamente/2018/05/01/\">\n" +
                "<id root=\"1.2.208.176.1.2\" extension=\"%s\" assigningAuthorityName=\"CPR\"/>\n" +
                "<treatmentWill>\n" +
                "    <ns2:noLifeProlongingIfDying>true</ns2:noLifeProlongingIfDying>\n" +
                "    <ns2:noLifeProlongingIfSeverelyDegraded acceptanceNeeded=\"relativeAcceptanceRequired\">true</ns2:noLifeProlongingIfSeverelyDegraded>\n" +
                "    <ns2:noLifeProlongingIfSeverePain acceptanceNeeded=\"relativeAcceptanceRequired\">true</ns2:noLifeProlongingIfSeverePain>\n" +
                "    <ns2:noForcedTreatmentIfIncapable acceptanceNeeded=\"relativeAcceptanceRequired\">true</ns2:noForcedTreatmentIfIncapable>\n" +
                "</treatmentWill>\n" +
                "</ns3:CreateTreatmentWillRequest>", cpr);
        return unmarshalString(requestStrin, CreateTreatmentWillRequest.class, unmarshallerTreatmentWill);
    }

    private HasTreatmentWillRequest getHasTreatmentWillRequestBody(String cpr) {
        // Dette er 2018.05.01 versionen af HasTreatmentWillRequest
        String requestStrin = String.format(
                "<ns:HasTreatmentWillRequest xmlns:ns=\"http://sundhedsdatastyrelsen.dk/behandlingstestamente/2018/05/01/\">\n" +
                "    <id assigningAuthorityName=\"CPR\" extension=\"%s\" root=\"1.2.208.176.1.2\"/>\n" +
                "</ns:HasTreatmentWillRequest>\n", cpr);
        return unmarshalString(requestStrin, HasTreatmentWillRequest.class, unmarshallerTreatmentWill);
    }

    private <T, R> R callService(Object request, Class<T> requestClass, Class<R> responseClass, String soapAction, String uri, String cpr) throws Exception {
        JAXBContext jaxbContext = JAXBContext.newInstance(requestClass, responseClass);
        return serviceCaller.call(
                getActor(cpr),
                request,
                jaxbContext,
                soapAction,
                new URI(uri),
                responseClass
        );
    }

    private CitizenIdentity getActor(String cpr) {
        RemoteCredentialVault serviceConsumerVault = new RemoteCredentialVault(
                URI.create("https://www.nspop.dk/download/attachments/190481050/NSP_Test_Service_Consumer_sds.p12"),
                "Test1234".toCharArray(),
                true
        );

        RemoteCredentialVault identityProviderVault = new RemoteCredentialVault(
                URI.create("https://www.nspop.dk/download/attachments/190481050/NSP_Test_Identity_Provider_sds.p12"),
                "Test1234".toCharArray(),
                true
        );

        return Builders.citizenIdentity()
                .representing(
                        Builders.citizen().identifiedBy(cpr)
                )
                .usedBy(
                        Builders.serviceConsumer()
                                .named("Service Consumer Test")
                                .identifiedBy(serviceConsumerVault)
                )
                .verifiedBy(
                        Builders.identityProvider()
                                .named("https://idp.test.nspop.dk")
                                .withAudience("https://bootstrap.sts.nspop.dk/")
                                .identifiedBy(identityProviderVault)
                )
                .using(
                        Builders.securityTokenService()
                                .at(URI.create("http://test1.ekstern-test.nspop.dk:8080/sts/services/Bst2Idws"))
                )
                .consumedBy(
                        Builders.serviceProvider()
                                .withAudience("https://fsk")
                )
                .build();
    }

    private <T> T unmarshalString(String xmlString, Class<T> clazz, Unmarshaller unmarshaller) {
        try {
            StringReader reader = new StringReader(xmlString);
            return clazz.cast(unmarshaller.unmarshal(reader));
        } catch (JAXBException e) {
            throw new RuntimeException(e);
        }
    }
}
