/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.wildfly;

import dk.nsp.test.idp.EmployeeIdentities;
import dk.nsp.test.idp.model.EmployeeIdentity;
import dk.sundhedsdatastyrelsen.behandlingstestamente._2020._03._16.*;
import dk.sundhedsdatastyrelsen.livstestamente.service.util.TestHelperV2;
import dk.idptest.SOAPOperation20200316;
import org.hl7.btrv2.TreatmentWillWithOnlyForcedTreatment;
import org.hl7.v3.ANY;
import org.hl7.v3.II;
import org.hl7.v3.POCDMT000040ClinicalDocument;
import org.hl7.v3.POCDMT000040Entry;
import org.junit.Before;
import org.junit.FixMethodOrder;
import org.junit.Test;
import org.junit.runners.MethodSorters;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Unmarshaller;
import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.util.List;

import static org.junit.Assert.*;

@FixMethodOrder(MethodSorters.NAME_ASCENDING)
public class TreatmentWillV2DGWSITCase extends AbstractITCase {

    private static final String REQUEST_FILE_PATH = "/dgws/testfiles/2020.03.16/";
    private static final String personIdentifier = "1806889160";
    private String uri;
    private Unmarshaller unmarshaller;
    private EmployeeIdentity actor = EmployeeIdentities.lægeCharlesBabbage();

    @Before
    public void before() throws JAXBException, IOException {

        uri = super.getItAddress() + "/btrAdmin";
        if(uri.startsWith("null")) {
            uri = "http://localhost:8080/btr" + "/btrAdmin";
        }
        setUpMarshalling();
    }

    private void setUpMarshalling() throws JAXBException {
        unmarshaller = JAXBContext.newInstance(ObjectFactory.class).createUnmarshaller();
    }

    private List<POCDMT000040Entry> getAllEntries(POCDMT000040ClinicalDocument d) {
        return d.getComponent().getStructuredBody().getComponents().get(0).getSection().getEntries();
    }

    @Test
    public void step0_DgwsRequest_CreateTreatmentWill() throws Exception {
        JAXBContext jaxbContext = JAXBContext.newInstance(CreateTreatmentWillRequest.class, CreateTreatmentWillResponse.class);

        File requestFile = getRequestFile(REQUEST_FILE_PATH + "CreateTreatmentWill_request_body.xml");
        CreateTreatmentWillRequest request = (CreateTreatmentWillRequest) unmarshaller.unmarshal(requestFile);
        CreateTreatmentWillResponse response = serviceCaller.call(
                actor,
                request,
                jaxbContext,
                SOAPOperation20200316.CreateTreatmentWill.getSoapAction(),
                new URI(uri),
                CreateTreatmentWillResponse.class);

        assertNotNull(response);
    }

    @Test
    public void step1_DgwsRequest_GetTreatmentWillWithOnlyForcedTreatment() throws Exception {
        JAXBContext jaxbContext = JAXBContext.newInstance(GetTreatmentWillWithOnlyForcedTreatmentRequest.class, GetTreatmentWillWithOnlyForcedTreatmentResponse.class, ANY.class, TreatmentWillWithOnlyForcedTreatment.class);

        GetTreatmentWillWithOnlyForcedTreatmentRequest request = new GetTreatmentWillWithOnlyForcedTreatmentRequest();

        II value = new II();
        value.setRoot("1.2.208.176.1.2");
        value.setExtension(personIdentifier);
        value.setAssigningAuthorityName("CPR");
        request.setId(value);

        GetTreatmentWillWithOnlyForcedTreatmentResponse response = serviceCaller.call(
                actor,
                request,
                jaxbContext,
                SOAPOperation20200316.GetTreatmentWillWithOnlyForcedTreatment.getSoapAction(),
                new URI(uri),
                GetTreatmentWillWithOnlyForcedTreatmentResponse.class);

        // Assert
        // Test PersonIdentifier and acceptances
        assertNotNull(response);
        assertEquals(personIdentifier, response.getClinicalDocument().getRecordTargets().get(0).getPatientRole().getIds().get(0).getExtension());
        List<POCDMT000040Entry> allEntries = getAllEntries(response.getClinicalDocument());
        assertEquals(1, allEntries.size());
        assertEquals(1, allEntries.get(0).getObservation().getValues().size());

        TreatmentWillWithOnlyForcedTreatment treatmentWillElement = (TreatmentWillWithOnlyForcedTreatment) allEntries.get(0).getObservation().getValues().get(0);
        assertNotNull(treatmentWillElement);

        assertEquals(TestHelperV2.NO_FORCED_TREATMENT_IF_INCAPABLE, treatmentWillElement.getNoForcedTreatmentIfIncapable().isValue());
        assertEquals("relativeAcceptanceRequired", treatmentWillElement.getNoForcedTreatmentIfIncapable().getAcceptanceNeeded());
    }

    @Test
    public void step2_IDWSRequest_UpgradeToTreatmentWill() throws Exception {
        JAXBContext jaxbContext = JAXBContext.newInstance(UpgradeToTreatmentWillRequest.class, UpgradeToTreatmentWillResponse.class);

        // check if there is living will to upgrade
        LivingWillV2DGWSITCase livingWillTestClass = new LivingWillV2DGWSITCase();
        livingWillTestClass.before();
        if (livingWillTestClass.DGWSRequest_HasLivingWill("HasLivingWill_request_body.xml")){
            File requestFile = getRequestFile(REQUEST_FILE_PATH + "UpgradeToTreatmentWill_request_body.xml");
            UpgradeToTreatmentWillRequest requestT = (UpgradeToTreatmentWillRequest) unmarshaller.unmarshal(requestFile);
            UpgradeToTreatmentWillResponse responseT = serviceCaller.call(
                    actor,
                    requestT,
                    jaxbContext,
                    SOAPOperation20200316.UpgradeToTreatmentWill.getSoapAction(),
                    new URI(uri),
                    UpgradeToTreatmentWillResponse.class);

            // The response is empty, but should not be null
            assertNotNull(responseT);
        }
    }

    @Test
    public void step3_IDWSRequest_UpgradeToTreatmentWill_WithInvalidXml() throws Exception {
        JAXBContext jaxbContext = JAXBContext.newInstance(UpgradeToTreatmentWillRequest.class, UpgradeToTreatmentWillResponse.class);

        File requestFile = getRequestFile(REQUEST_FILE_PATH + "UpgradeToTreatmentWill_invalid_request_body.xml");
        UpgradeToTreatmentWillRequest request = (UpgradeToTreatmentWillRequest) unmarshaller.unmarshal(requestFile);

        try {
            UpgradeToTreatmentWillResponse response = serviceCaller.call(
                    actor,
                    request,
                    jaxbContext,
                    SOAPOperation20200316.UpgradeToTreatmentWill.getSoapAction(),
                    new URI(uri),
                    UpgradeToTreatmentWillResponse.class);

            fail("Should fail");
        } catch (Exception e) {
            String faultString = e.getMessage();
            assertNotNull("Schemavalidation should fail", faultString);
            assertTrue(faultString, faultString.matches(
                    ".*Invalid content was found starting with element.*noLifeProlongingIfSeverelyDegraded.*" +
                            "One of .*noLifeProlongingIfDying.* is expected.*"));

        }
    }


    @Test
    public void step4_DgwsRequest_DeleteTreatmentWill() throws Exception {
        JAXBContext jaxbContext = JAXBContext.newInstance(DeleteTreatmentWillRequest.class, DeleteTreatmentWillResponse.class);

        File requestFile = getRequestFile(REQUEST_FILE_PATH + "DeleteTreatmentWill_request_body.xml");
        DeleteTreatmentWillRequest request = (DeleteTreatmentWillRequest) unmarshaller.unmarshal(requestFile);
        DeleteTreatmentWillResponse response = serviceCaller.call(
                actor,
                request,
                jaxbContext,
                SOAPOperation20200316.DeleteTreatmentWill.getSoapAction(),
                new URI(uri),
                DeleteTreatmentWillResponse.class);

        assertNotNull(response);
    }

    public boolean DGWSRequest_HasTreatmentWill(String requestFileName) throws Exception {
        JAXBContext jaxbContext = JAXBContext.newInstance(HasTreatmentWillRequest.class, HasTreatmentWillResponse.class);

        File requestFile = getRequestFile(REQUEST_FILE_PATH + requestFileName);
        HasTreatmentWillRequest request = (HasTreatmentWillRequest) unmarshaller.unmarshal(requestFile);
        HasTreatmentWillResponse response = serviceCaller.call(
                actor,
                request,
                jaxbContext,
                SOAPOperation20200316.HasTreatmentWill.getSoapAction(),
                new URI(uri),
                HasTreatmentWillResponse.class);

        // The response is empty, but should not be null
        assertNotNull(response);
        return response.isWillExists();
    }

}
