/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.wildfly;

import dk.nsp.test.idp.builder.Builders;
import dk.nsp.test.idp.model.CitizenIdentity;
import dk.nsp.test.idp.vault.RemoteCredentialVault;
import dk.sundhedsdatastyrelsen.behandlingstestamente._2020._03._16.*;
import dk.sundhedsdatastyrelsen.livstestamente.service.util.TestHelperV2;
import dk.idptest.SOAPOperation20200316;
import org.hl7.btrv2.TreatmentWillV2;
import org.hl7.v3.POCDMT000040Entry;
import org.junit.Assert;
import org.junit.Before;
import org.junit.FixMethodOrder;
import org.junit.Test;
import org.junit.runners.MethodSorters;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Unmarshaller;
import java.io.File;
import java.net.URI;
import java.util.List;

import static org.junit.Assert.*;

/**
 * IDWS integration tests
 */
@FixMethodOrder(MethodSorters.NAME_ASCENDING)
public class TreatmentWillV2IDWSITCase extends AbstractITCase {
    private static final String RESPONSE_AUDIENCE_URI = "https://fsk";
    private static final String REQUEST_FILE_PATH = "/idws/testfiles/2020.03.16/";
    private static final String personIdentifier = "0501792275";
    private String uri;
    private Unmarshaller unmarshaller;
    private CitizenIdentity actor = Builders.citizenIdentity().representing(Builders.citizen().identifiedBy("0501792275")).usedBy(Builders.serviceConsumer().named("Service Consumer Test").identifiedBy(new RemoteCredentialVault(URI.create("https://www.nspop.dk/download/attachments/190481050/NSP_Test_Service_Consumer_sds.p12"), "Test1234".toCharArray(), true))).verifiedBy(Builders.identityProvider().named("https://idp.test.nspop.dk").withAudience("https://bootstrap.sts.nspop.dk/").identifiedBy(new RemoteCredentialVault(URI.create("https://www.nspop.dk/download/attachments/190481050/NSP_Test_Identity_Provider_sds.p12"), "Test1234".toCharArray(), true))).using(Builders.securityTokenService().at(URI.create("http://test1.ekstern-test.nspop.dk:8080/sts/services/Bst2Idws"))).consumedBy(Builders.serviceProvider().withAudience(RESPONSE_AUDIENCE_URI)).build();

    @Before
    public void before() throws JAXBException {
        uri = super.getItAddress() + "/btr";
        if(uri.startsWith("null")) {
            uri = "http://localhost:8080/btr" + "/btr";
        }
        setUpMarshalling();
    }

    private void setUpMarshalling() throws JAXBException {
        JAXBContext jaxbContext = JAXBContext.newInstance(ObjectFactory.class);
        unmarshaller = jaxbContext.createUnmarshaller();
    }

    @Test
    public void step0_IDWSRequest_HasTreatmentWill() throws Exception {
        JAXBContext jaxbContext = JAXBContext.newInstance(HasTreatmentWillRequest.class, HasTreatmentWillResponse.class);

        File requestFile = getRequestFile(REQUEST_FILE_PATH + "HasTreatmentWill_request_body.xml");
        HasTreatmentWillRequest request = (HasTreatmentWillRequest) unmarshaller.unmarshal(requestFile);
        HasTreatmentWillResponse response = serviceCaller.call(
                actor,
                request,
                jaxbContext,
                SOAPOperation20200316.HasTreatmentWill.getSoapAction(),
                new URI(uri),
                HasTreatmentWillResponse.class);

        assertFalse(response.isWillExists());
    }

    @Test
    public void step1_IDWSRequest_CreateTreatmentWill() throws Exception {
        JAXBContext jaxbContext = JAXBContext.newInstance(CreateTreatmentWillRequest.class, CreateTreatmentWillResponse.class);

        File requestFile = getRequestFile(REQUEST_FILE_PATH + "CreateTreatmentWill_request_body.xml");
        CreateTreatmentWillRequest request = (CreateTreatmentWillRequest) unmarshaller.unmarshal(requestFile);
        CreateTreatmentWillResponse response = serviceCaller.call(
                actor,
                request,
                jaxbContext,
                SOAPOperation20200316.CreateTreatmentWill.getSoapAction(),
                new URI(uri),
                CreateTreatmentWillResponse.class);

        // The response is empty, but should not be null
        assertNotNull(response);
    }

    @Test
    public void step2_IDWSRequest_GetTreatmentWill() throws Exception {
        JAXBContext jaxbContext = JAXBContext.newInstance(GetTreatmentWillRequest.class, GetTreatmentWillResponse.class);

        File requestFile = getRequestFile(REQUEST_FILE_PATH + "GetTreatmentWill_request_body.xml");
        GetTreatmentWillRequest request = (GetTreatmentWillRequest) unmarshaller.unmarshal(requestFile);
        GetTreatmentWillResponse response = serviceCaller.call(
                actor,
                request,
                jaxbContext,
                SOAPOperation20200316.GetTreatmentWill.getSoapAction(),
                new URI(uri),
                GetTreatmentWillResponse.class);

        // Assert
        // Test PersonIdentifier and acceptances
        assertNotNull(response);
        assertEquals(personIdentifier, response.getClinicalDocument().getRecordTargets().get(0).getPatientRole().getIds().get(0).getExtension());
        List<POCDMT000040Entry> allEntries = getAllEntries(response);
        assertEquals(1, allEntries.size());
        assertEquals(1, allEntries.get(0).getObservation().getValues().size());

        TreatmentWillV2 treatmentWillElement = (TreatmentWillV2) allEntries.get(0).getObservation().getValues().get(0);
        assertNotNull(treatmentWillElement);

        assertEquals(TestHelperV2.NO_LIFE_PROLONGING_IF_DYING, treatmentWillElement.getNoLifeProlongingIfDying().isValue());

        assertEquals(TestHelperV2.NO_LIFE_PROLONGING_IF_SEVERELY_DEGRADED, treatmentWillElement.getNoLifeProlongingIfSeverelyDegraded().isValue());
        assertEquals("relativeAcceptanceRequired", treatmentWillElement.getNoLifeProlongingIfSeverelyDegraded().getAcceptanceNeeded());

        assertEquals(TestHelperV2.NO_LIFE_PROLONGING_IF_SEVERE_PAIN, treatmentWillElement.getNoLifeProlongingIfSeverelyDegraded().isValue());
        assertEquals("relativeAcceptanceRequired", treatmentWillElement.getNoLifeProlongingIfSeverelyDegraded().getAcceptanceNeeded());

        assertEquals(TestHelperV2.NO_FORCED_TREATMENT_IF_INCAPABLE, treatmentWillElement.getNoForcedTreatmentIfIncapable().isValue());
        assertEquals("relativeAcceptanceRequired", treatmentWillElement.getNoForcedTreatmentIfIncapable().getAcceptanceNeeded());
    }

    private List<POCDMT000040Entry> getAllEntries(GetTreatmentWillResponse response) {
        return response.getClinicalDocument().getComponent().getStructuredBody().getComponents().get(0).getSection().getEntries();
    }

    @Test
    public void step3_IDWSRequest_UpdateTreatmentWill() throws Exception {
        JAXBContext jaxbContext = JAXBContext.newInstance(UpdateTreatmentWillRequest.class, UpdateTreatmentWillResponse.class);

        File requestFile = getRequestFile(REQUEST_FILE_PATH + "UpdateTreatmentWill_request_body.xml");
        UpdateTreatmentWillRequest request = (UpdateTreatmentWillRequest) unmarshaller.unmarshal(requestFile);
        UpdateTreatmentWillResponse response = serviceCaller.call(
                actor,
                request,
                jaxbContext,
                SOAPOperation20200316.UpdateTreatmentWill.getSoapAction(),
                new URI(uri),
                UpdateTreatmentWillResponse.class);

        // The response is empty, but should not be null
        assertNotNull(response);
    }

    @Test
    public void step4_IDWSRequest_GetTreatmentWill() throws Exception {
        JAXBContext jaxbContext = JAXBContext.newInstance(GetTreatmentWillRequest.class, GetTreatmentWillResponse.class);

        File requestFile = getRequestFile(REQUEST_FILE_PATH + "GetTreatmentWill_request_body_2.xml");
        GetTreatmentWillRequest request = (GetTreatmentWillRequest) unmarshaller.unmarshal(requestFile);
        GetTreatmentWillResponse response = serviceCaller.call(
                actor,
                request,
                jaxbContext,
                SOAPOperation20200316.GetTreatmentWill.getSoapAction(),
                new URI(uri),
                GetTreatmentWillResponse.class);

        // Test PersonIdentifier
        assertEquals(personIdentifier, response.getClinicalDocument().getRecordTargets().get(0).getPatientRole().getIds().get(0).getExtension());
    }

    @Test
    public void step5_IDWSRequest_HasTreatmentWill() throws Exception {
        JAXBContext jaxbContext = JAXBContext.newInstance(HasTreatmentWillRequest.class, HasTreatmentWillResponse.class);

        File requestFile = getRequestFile(REQUEST_FILE_PATH + "HasTreatmentWill_request_body_2.xml");
        HasTreatmentWillRequest request = (HasTreatmentWillRequest) unmarshaller.unmarshal(requestFile);
        HasTreatmentWillResponse response = serviceCaller.call(
                actor,
                request,
                jaxbContext,
                SOAPOperation20200316.HasTreatmentWill.getSoapAction(),
                new URI(uri),
                HasTreatmentWillResponse.class);

        // Test PersonIdentifier
        assertTrue(response.isWillExists());
    }

    @Test
    public void step6_IDWSRequest_DeleteTreatmentWill() throws Exception {
        JAXBContext jaxbContext = JAXBContext.newInstance(DeleteTreatmentWillRequest.class, DeleteTreatmentWillResponse.class);

        File requestFile = getRequestFile(REQUEST_FILE_PATH + "DeleteTreatmentWill_request_body.xml");
        DeleteTreatmentWillRequest request = (DeleteTreatmentWillRequest) unmarshaller.unmarshal(requestFile);
        DeleteTreatmentWillResponse response = serviceCaller.call(
                actor,
                request,
                jaxbContext,
                SOAPOperation20200316.DeleteTreatmentWill.getSoapAction(),
                new URI(uri),
                DeleteTreatmentWillResponse.class);

        assertNotNull(response);
    }

    @Test
    public void step7_IDWSRequest_GetTreatmentWillTooYoungShouldFail() throws Exception {
        JAXBContext jaxbContext = JAXBContext.newInstance(GetTreatmentWillRequest.class, GetTreatmentWillResponse.class);

        File requestFile = getRequestFile(REQUEST_FILE_PATH + "GetTreatmentWill_request_body_too_young_cpr.xml");
        GetTreatmentWillRequest request = (GetTreatmentWillRequest) unmarshaller.unmarshal(requestFile);
        Assert.assertThrows(RuntimeException.class, () ->  serviceCaller.call(
                actor,
                request,
                jaxbContext,
                SOAPOperation20200316.GetTreatmentWill.getSoapAction(),
                new URI(uri),
                GetTreatmentWillResponse.class));
    }

    @Test
    public void IDWSRequest_CreateTreatmentWill_WithInvalidXml() throws Exception {
        JAXBContext jaxbContext = JAXBContext.newInstance(CreateTreatmentWillRequest.class, CreateTreatmentWillResponse.class);

        File requestFile = getRequestFile(REQUEST_FILE_PATH + "CreateTreatmentWill_invalid_request_body.xml");
        CreateTreatmentWillRequest request = (CreateTreatmentWillRequest) unmarshaller.unmarshal(requestFile);

        try {
            CreateTreatmentWillResponse response = serviceCaller.call(
                    actor,
                    request,
                    jaxbContext,
                    SOAPOperation20200316.CreateTreatmentWill.getSoapAction(),
                    new URI(uri),
                    CreateTreatmentWillResponse.class);
                    fail("Should fail");
        } catch (Exception e) {
            String faultString = e.getMessage();
            assertNotNull("Schemavalidation should fail", faultString);
            assertTrue(faultString, faultString.matches(
                    ".*Invalid content was found starting with element.*noLifeProlongingIfSeverelyDegraded.*" +
                            "One of .*noLifeProlongingIfDying.* is expected.*"));

        }
    }
}
