/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service;

import dk.sundhedsdatastyrelsen.livstestamente.conf.Properties;
import org.apache.http.HttpHeaders;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;

import javax.servlet.ServletException;
import javax.servlet.annotation.WebServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URL;
import java.time.Instant;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;

/**
 * DCC konfigurationsservice (DKS)
 */
@WebServlet(name="DKSServlet", urlPatterns={"/dksconfig/ltr", "/dksconfig/btr", "/dksconfig/btrV2"})
public class DKSServlet extends AutowiredHttpServlet {

    @SuppressWarnings("unused") private static final Logger logger = LogManager.getLogger(DKSServlet.class);

    private static final String RESOURCE_NAME_LTR = "/dksconfig/ltr.xml";
    private static final String RESOURCE_NAME_BTR = "/dksconfig/btr.xml";
    private static final String RESOURCE_NAME_BTRv2 = "/dksconfig/btrV2.xml";
    private static final String DATETIME_FORMAT = "yyyy-MM-dd'T'HH:mm:ss'Z'";

    @Autowired
    private Properties properties;

    @Override
    public void doGet(HttpServletRequest request, HttpServletResponse response) throws IOException, ServletException {
        String responseBody;
        switch (request.getServletPath()) {
            case "/dksconfig/ltr":
                responseBody = getDksConfig(RESOURCE_NAME_LTR, properties.getDccLTREndpoint());
                break;
            case "/dksconfig/btr":
                responseBody = getDksConfig(RESOURCE_NAME_BTR, properties.getDccBTREndpoint());
                break;
            case "/dksconfig/btrV2":
                responseBody = getDksConfig(RESOURCE_NAME_BTRv2, properties.getDccBTREndpoint());
                break;
            default:
                super.doGet(request, response); // Sends error response
                return;
        }
        response.setStatus(200);
        response.setContentType("application/xml");
        response.addHeader(HttpHeaders.CACHE_CONTROL, "no-cache");
        response.getWriter().print(responseBody);
    }

    private String getDksConfig(String resourceName, String dccEndpoint) throws IOException {
        URL url = getClass().getResource(resourceName);
        
        ZonedDateTime zdt = ZonedDateTime.ofInstant(Instant.ofEpochMilli(url.openConnection().getLastModified()), ZoneId.systemDefault());
        DateTimeFormatter dateTimeFormatter = DateTimeFormatter.ofPattern(DATETIME_FORMAT);
        String fileTimestamp = dateTimeFormatter.format(zdt);

        InputStream in = getClass().getResourceAsStream(resourceName);
        BufferedReader reader = new BufferedReader(new InputStreamReader(in));
        StringBuilder out = new StringBuilder();
        String line;
        while ((line = reader.readLine()) != null) {
            out.append(line);
        }

        return getStringWithResolvedVariables(out, fileTimestamp, dccEndpoint);
    }

    private String getStringWithResolvedVariables(StringBuilder template, String filetimestamp, String dccEndpoint) {
        int timestampIndex = template.indexOf("TIMESTAMP");
        template = template.replace(timestampIndex, timestampIndex + 9, filetimestamp);

        return template.toString().replace("ENDPOINT_PLACEHOLDER", dccEndpoint);
    }
}
