/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service;

import dk.sundhedsdatastyrelsen.livstestamente.service.webservice.LtrBtrContextListener;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.dao.DataAccessException;
import org.springframework.jdbc.core.JdbcTemplate;

import javax.servlet.ServletContext;
import javax.servlet.annotation.WebServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.charset.StandardCharsets;
import java.time.Instant;
import java.util.Properties;

@WebServlet(name="IsAliveServlet", urlPatterns={"/status"})
public class IsAliveServlet extends AutowiredHttpServlet {

    @SuppressWarnings("unused") private static final Logger logger = LogManager.getLogger(IsAliveServlet.class);

    @Autowired
    private DataSource btrDataSource;

    private BuildData buildData;

    @Override
    public void doGet(HttpServletRequest request, HttpServletResponse response) throws IOException {
        int status;
        String statusMessage;

        if (datasourceResponding()) {
            status = 200;
            statusMessage = " OK";
        } else {
            status = 500;
            statusMessage = "Internal Server Error: Database check failed";
        }

        response.setStatus(status);
        response.setContentType("text/plain");

        response.getWriter().println(String.format("%s %s", status, statusMessage));
        response.getWriter().println();

        buildData = getBuildData(request.getServletContext());
        response.getWriter().println("Title: " + buildData.title);
        response.getWriter().println("Deployed: " + LtrBtrContextListener.DEPLOY_TIME);
        response.getWriter().println("Build-Date: " + buildData.date);
        response.getWriter().println("Build-Version: " + buildData.version);
        response.getWriter().println("Builder: " + buildData.builder);
        response.getWriter().println("Display time: " + Instant.now());
    }

    private boolean datasourceResponding() {
        try {
            new JdbcTemplate(btrDataSource).execute("SELECT 1");
        } catch (DataAccessException e) {
            return false;
        }
        return true;
    }

    private BuildData getBuildData(ServletContext servletContext) throws IOException {
        if (buildData == null) {
            InputStream stream = servletContext.getResourceAsStream("/META-INF/MANIFEST.MF");
            if (stream == null) {
                // Local
                stream = this.getClass().getResourceAsStream("/META-INF/MANIFEST.MF");
            }
            if (stream != null) {
                Properties prop = new Properties();
                prop.load(new InputStreamReader(stream, StandardCharsets.UTF_8));
                stream.close();
                buildData = new BuildData(prop.getProperty("Implementation-Title"), prop.getProperty("Implementation-Version"), prop.getProperty("Built-Date"), prop.getProperty("Built-By"));
            }
        }
        return buildData;
    }

    private static class BuildData {
        String title;
        String version;
        String date;
        String builder;

        BuildData(String title, String version, String date, String builder) {
            this.title = title;
            this.version = version;
            this.date = date;
            this.builder = builder;
        }
    }
}
