/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service;

import dk.sundhedsdatastyrelsen.livstestamente.service.util.WSDLHelper;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.springframework.util.FileCopyUtils;

import javax.servlet.ServletException;
import javax.servlet.annotation.WebServlet;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.nio.charset.StandardCharsets;
import java.util.HashMap;
import java.util.Map;

@WebServlet(name="WSDLServlet", urlPatterns={"/wsdl", "/wsdl/ltr/dgws", "/wsdl/ltr/idws", "/wsdl/btr/dgws", "/wsdl/btr/idws", "/wsdl/btrV2/dgws", "/wsdl/btrV2/idws", "/schema/*"})
public class WSDLServlet extends HttpServlet {

    private static final Logger logger = LogManager.getLogger(WSDLServlet.class);
    private static final Map<String, String> path2ResourceMap;

    static {
        path2ResourceMap = new HashMap<>();
        path2ResourceMap.put("/wsdl", "/static/wsdl_overview.html");
        path2ResourceMap.put("/wsdl/ltr/dgws", "/wsdl/ltr/livstestamente_dgws_service_2018_05_01.wsdl");
        path2ResourceMap.put("/wsdl/ltr/idws", "/wsdl/ltr/livstestamente_idws_service_2018_05_01.wsdl");
        path2ResourceMap.put("/wsdl/btr/dgws", "/wsdl/btr/behandlingstestamente_dgws_service_2018_05_01.wsdl");
        path2ResourceMap.put("/wsdl/btr/idws", "/wsdl/btr/behandlingstestamente_idws_service_2018_05_01.wsdl");
        path2ResourceMap.put("/wsdl/btrV2/dgws", "/wsdl/btr/behandlingstestamente_dgws_service_2020_03_16.wsdl");
        path2ResourceMap.put("/wsdl/btrV2/idws", "/wsdl/btr/behandlingstestamente_idws_service_2020_03_16.wsdl");
    }

    @Override
    public void doGet(HttpServletRequest request, HttpServletResponse response) throws IOException, ServletException {
        String path = request.getServletPath();
        logger.debug("WSDL request: " + path);
        String responseString;
        if (path2ResourceMap.containsKey(path)) {
            String requestUrl = path.equals("/wsdl") ? null : request.getRequestURL().toString();
            responseString = getResourceString(path2ResourceMap.get(path), requestUrl);
        } else {
            String pathInfo = request.getPathInfo() == null ? "" : request.getPathInfo(); // Whatever comes after "/schema"
            responseString = getResourceString(path + pathInfo, request.getRequestURL().toString());
        }
        if (responseString != null) {
            response.getWriter().print(responseString);
            response.setStatus(200);
            response.setContentType(path.equals("/wsdl") ? "text/html" : "text/xml");
        } else {
            response.sendError(404);
        }
    }

    private String getResourceString(String resourcePath, String requestURL) throws IOException {
        InputStream in = getClass().getResourceAsStream(resourcePath);
        if (in == null) {
            return null;
        }
        Reader reader = new InputStreamReader(in, StandardCharsets.UTF_8);
        String text = FileCopyUtils.copyToString(reader);

        if (requestURL != null) {
            text = WSDLHelper.replaceSchemaLocations(text, requestURL);
        }

        return text;
    }

}
