/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.business;

import org.springframework.stereotype.Component;

import dk.sundhedsdatastyrelsen.livstestamente.domain.RequestContextVO;
import dk.sundhedsdatastyrelsen.livstestamente.domain.UserType;
import dk.sundhedsdatastyrelsen.livstestamente.util.FaultCode;
import dk.sundhedsdatastyrelsen.livstestamente.util.FaultUtil;
import dk.sundhedsdatastyrelsen.livstestamente.util.ServiceException;

@Component
public abstract class AbstractWillAccessControlImpl implements AccessControl {

	abstract boolean isOpen(RequestContextVO context);

	@Override
	public void checkGetAccess(RequestContextVO context) throws ServiceException {
		final boolean isAuthorized = 
				(context.getServiceActor().getUserType() == UserType.Citizen && context.getPersonIdentifier().equals(context.getServiceActor().getPersonIdentifier()))
				|| context.getServiceActor().getUserType() == UserType.HealthCareProfessional
				|| context.getServiceActor().getUserType() == UserType.Admin;

		if (!isAuthorized) {
			FaultUtil.throwFault(FaultCode.AUTHORIZATION_ERROR_PERSON_IDENTIFIER_ACCESS_DENIED);
		}
	}

	@Override
	public void checkHasAccess(RequestContextVO context) throws ServiceException {
		final boolean isAuthorized = 
				(context.getServiceActor().getUserType() == UserType.Citizen && context.getPersonIdentifier().equals(context.getServiceActor().getPersonIdentifier()))
				|| context.getServiceActor().getUserType() == UserType.HealthCareProfessional
				|| context.getServiceActor().getUserType() == UserType.SystemUser
				|| context.getServiceActor().getUserType() == UserType.Admin;

		if (!isAuthorized) {
			FaultUtil.throwFault(FaultCode.AUTHORIZATION_ERROR_PERSON_IDENTIFIER_ACCESS_DENIED);
		}
	}

	@Override
	public void checkCreateAccess(RequestContextVO context) throws ServiceException {
		if (!isOpen(context)) {
			FaultUtil.throwFault(FaultCode.CREATE_NOT_ALLOWED);
		}

		final boolean isAuthorized = 
				(context.getServiceActor().getUserType() == UserType.Citizen && context.getPersonIdentifier().equals(context.getServiceActor().getPersonIdentifier()))
				|| context.getServiceActor().getUserType() == UserType.Admin;

		if (!isAuthorized) {
			FaultUtil.throwFault(FaultCode.AUTHORIZATION_ERROR_PERSON_IDENTIFIER_ACCESS_DENIED);
		}
	}

	@Override
	public void checkUpdateAccess(RequestContextVO context) throws ServiceException {
		final boolean isAuthorized = 
				(context.getServiceActor().getUserType() == UserType.Citizen && context.getPersonIdentifier().equals(context.getServiceActor().getPersonIdentifier()))
				|| context.getServiceActor().getUserType() == UserType.Admin;

		if (!isAuthorized) {
			FaultUtil.throwFault(FaultCode.AUTHORIZATION_ERROR_PERSON_IDENTIFIER_ACCESS_DENIED);
		}
	}

	@Override
	public void checkDeleteAccess(RequestContextVO context) throws ServiceException {
		final boolean isAuthorized = 
				(context.getServiceActor().getUserType() == UserType.Citizen && context.getPersonIdentifier().equals(context.getServiceActor().getPersonIdentifier()))
				|| context.getServiceActor().getUserType() == UserType.Admin;

		if (!isAuthorized) {
			FaultUtil.throwFault(FaultCode.AUTHORIZATION_ERROR_PERSON_IDENTIFIER_ACCESS_DENIED);
		}
	}

	@Override
	public void checkUpgradeAccess(RequestContextVO context) throws ServiceException {
		final boolean isAuthorized =
				context.getServiceActor().getUserType() == UserType.Admin;

		if (!isAuthorized) {
			FaultUtil.throwFault(FaultCode.AUTHORIZATION_ERROR_PERSON_IDENTIFIER_ACCESS_DENIED);
		}
	}
}
