/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.business;

import dk.sundhedsdatastyrelsen.livstestamente.service.business.validator.LivingWillValidator;
import dk.sundhedsdatastyrelsen.livstestamente.service.business.validator.TreatmentWillValidator;
import dk.sundhedsdatastyrelsen.livstestamente.dao.LivingWillDAO;
import dk.sundhedsdatastyrelsen.livstestamente.dao.TreatmentWillDAO;
import dk.sundhedsdatastyrelsen.livstestamente.domain.LivingWillVO;
import dk.sundhedsdatastyrelsen.livstestamente.domain.RequestContextVO;
import dk.sundhedsdatastyrelsen.livstestamente.domain.UserType;
import dk.sundhedsdatastyrelsen.livstestamente.exceptions.DAOException;
import dk.sundhedsdatastyrelsen.livstestamente.service.integrations.minlog.MinLogDataWrapper;
import dk.sundhedsdatastyrelsen.livstestamente.service.integrations.minlog.MinLogException;
import dk.sundhedsdatastyrelsen.livstestamente.service.integrations.nas.treatmentwill.NasException;
import dk.sundhedsdatastyrelsen.livstestamente.service.integrations.nas.treatmentwill.NotificationSender;
import dk.sundhedsdatastyrelsen.livstestamente.util.FaultCode;
import dk.sundhedsdatastyrelsen.livstestamente.util.FaultUtil;
import dk.sundhedsdatastyrelsen.livstestamente.conf.Properties;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.time.Instant;
import java.time.Year;
import java.time.ZoneOffset;

@Service
public class LivingWillServiceImpl implements LivingWillService {

    private static final Logger logger = LogManager.getLogger(LivingWillServiceImpl.class);

    static final String SERVICE_IDENTIFIER = "Livstestamente";

    private final Properties properties;
    private final TreatmentWillDAO treatmentWillDAO;
    private final LivingWillDAO livingWillDAO;
    private final MinLogDataWrapper minLogDataWrapper;
    private final AccessControl accessControl;
    private final TreatmentWillValidator treatmentWillValidator;
    private final LivingWillValidator livingWillValidator;

    static final Instant VALID_TO_DEFAULT = Year.of(8099).atDay(1).atStartOfDay(ZoneOffset.UTC).toInstant();
    static final int INITIAL_VERSION_DEFAULT = 1;

    @Autowired
    private NotificationSender notificationSender;

    @Autowired
    public LivingWillServiceImpl(
            Properties properties,
            TreatmentWillDAO treatmentWillDAO,
            LivingWillDAO livingWillDAO,
            MinLogDataWrapper minLogDataWrapper,
            @Qualifier("livingWillAccessControlImpl") AccessControl accessControl,
            TreatmentWillValidator treatmentWillValidator,
            LivingWillValidator livingWillValidator,
            NotificationSender notificationSender) {
        this.properties = properties;
        this.treatmentWillDAO = treatmentWillDAO;
        this.livingWillDAO = livingWillDAO;
        this.minLogDataWrapper = minLogDataWrapper;
        this.accessControl = accessControl;
        this.treatmentWillValidator = treatmentWillValidator;
        this.livingWillValidator = livingWillValidator;
        this.notificationSender = notificationSender;
    }

    @Override
    public LivingWillVO getLivingWill(RequestContextVO context) {
        LivingWillVO result = null;

        accessControl.checkGetAccess(context);

        try {
            // Get any existing TreatmentWill
            final String personIdentifier = context.getPersonIdentifier();
            result = livingWillDAO.getCurrentLivingWill(personIdentifier, context.getTime());

            // Validate existing TreatmentWill
            livingWillValidator.validateLivingWillExists(result);

            // Log to MinLog. If MinLog fails, it throws an exception
            minLogIfNotSystemUser(context, properties.getMinLogReadActivityTextLTR());
        } catch (DAOException | MinLogException e) {
            FaultUtil.throwFault(FaultCode.GET_LIVING_WILL_INTERNAL_SERVER_ERROR, e);
        }

        return result;
    }

    @Override
    public boolean hasLivingWill(RequestContextVO context) {
        boolean result = false;

        accessControl.checkHasAccess(context);

        try {
            final String personIdentifier = context.getPersonIdentifier();
            final LivingWillVO currentLivingWill = livingWillDAO.getCurrentLivingWill(personIdentifier, context.getTime());
            result = currentLivingWill != null;

            // No minLog needed
        } catch (DAOException e) {
            FaultUtil.throwFault(FaultCode.HAS_LIVING_WILL_INTERNAL_SERVER_ERROR, e);
        }
        return result;
    }

    @Override
    @Transactional("btrDataSourceTxManager")
    public void updateLivingWill(LivingWillVO livingWillVO, RequestContextVO context) {
        accessControl.checkUpdateAccess(context);

        try {
            // Get any existing LivingWill
            final String personIdentifier = context.getPersonIdentifier();
            final LivingWillVO currentVersionOfLivingWill = livingWillDAO.getCurrentLivingWill(personIdentifier, context.getTime());

            // Validate existing LivingWill
            livingWillValidator.validateLivingWillExists(currentVersionOfLivingWill);

            final long currentPid = currentVersionOfLivingWill.getLivingWillPID();
            final int currentVersion = currentVersionOfLivingWill.getVersion();
            int nextVersion = currentVersion + 1;

            Instant nowTimestamp = context.getTime();

            livingWillVO.setVersion(nextVersion);
            livingWillVO.setValidFrom(context.getTime());
            livingWillVO.setValidTo(VALID_TO_DEFAULT);
            livingWillVO.setModifiedDateTime(context.getTime());

            livingWillDAO.updateValidToTimestamp(currentPid, context.getTime());
            livingWillDAO.createLivingWill(livingWillVO);
            minLogIfNotSystemUser(context, properties.getMinLogUpdateActivityTextLTR());
            // Send a notification to NAS, indicating that the Living Will has changed.
            sendUpdateNotificationToNas(context, nowTimestamp);

        } catch (DAOException | MinLogException | NasException e) {
            FaultUtil.throwFault(FaultCode.UPDATE_LIVING_WILL_INTERNAL_SERVER_ERROR, e);
        }
    }

    @Override
    @Transactional("btrDataSourceTxManager")
    public void deleteLivingWill(RequestContextVO context, Boolean requireAdmin) {
        if(requireAdmin){
            accessControl.checkUpgradeAccess(context);
        }
        else {
            accessControl.checkDeleteAccess(context);
        }

        try {
            // Get any existing LivingWill
            final String personIdentifier = context.getPersonIdentifier();
            final LivingWillVO currentVersionOfLivingWill = livingWillDAO.getCurrentLivingWill(personIdentifier, context.getTime());

            Instant nowTimestamp = context.getTime();
            // Validate existing LivingWill
            livingWillValidator.validateLivingWillExists(currentVersionOfLivingWill);

            final long currentPid = currentVersionOfLivingWill.getLivingWillPID();
            livingWillDAO.updateValidToTimestamp(currentPid, context.getTime());
            minLogIfNotSystemUser(context, properties.getMinLogDeleteActivityTextLTR());

            // Send a notification to NAS, indicating that the Living Will has changed.
            sendUpdateNotificationToNas(context, nowTimestamp);
        } catch (DAOException | MinLogException | NasException e) {
            FaultUtil.throwFault(FaultCode.DELETE_LIVING_WILL_INTERNAL_SERVER_ERROR, e);
        }
    }
    
    private void minLogIfNotSystemUser(RequestContextVO context, String minLogText) throws MinLogException {
        if (context.getServiceActor().getUserType() != UserType.SystemUser) {
            logActivityToMinLog(context, minLogText);
        }
    }

    private void logActivityToMinLog(RequestContextVO context, String minLogText) throws MinLogException {
        minLogDataWrapper.logActivity(
                context,
                minLogText,
                context.getPersonIdentifier(),
                context.getServiceActor().getOrganizationId(),
                context.getServiceActor().getOrganizationName(),
                context.getServiceActor().getOrganizationIdSource(),
                context.getServiceActor().getPersonIdentifier(),
                null,
                SERVICE_IDENTIFIER,
                context.getServiceActor().getUserTitle());
    }

    private void sendUpdateNotificationToNas(RequestContextVO context, Instant currentTime) throws NasException {
        notificationSender.sendLivingWillNotification(context.getPersonIdentifier(), currentTime);
    }
}
