/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.conf;

import dk.sds.nsp.audit.Audit;
import dk.sundhedsdatastyrelsen.livstestamente.auditlog.AuditLogger;
import dk.sundhedsdatastyrelsen.livstestamente.context.UserTitleProvider;
import dk.sundhedsdatastyrelsen.livstestamente.service.integrations.minlog.MinLogClient;
import dk.sundhedsdatastyrelsen.livstestamente.service.integrations.minlog.MinLogException;
import dk.sundhedsdatastyrelsen.livstestamente.util.HttpClientFactory;
import dk.sundhedsdatastyrelsen.livstestamente.integrations.personInformation.PersonInformation;
import dk.sundhedsdatastyrelsen.livstestamente.integrations.personInformation.impl.PersonInformationImpl;
import dk.sundhedsdatastyrelsen.livstestamente.integrations.personInformation.impl.PersonInformationServiceClientFactory;
import dk.sundhedsdatastyrelsen.livstestamente.context.ServiceActorProvider;
import dk.sundhedsdatastyrelsen.livstestamente.conf.Properties;
import org.apache.http.HttpStatus;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.ComponentScan;
import org.springframework.context.annotation.Configuration;

@Configuration
@ComponentScan("dk.sundhedsdatastyrelsen.livstestamente")
public class WebServiceConfig {

    private static final String MINLOG_CONFIG_FILE_NAME = "/minlogclient.properties";

    private static final Logger logger = LogManager.getLogger(WebServiceConfig.class);

    private final Properties properties;

    @Autowired
    public WebServiceConfig(Properties properties) {
        this.properties = properties;
    }

    @Bean
    public HttpClientFactory getHttpClientFactory() {
        return new HttpClientFactory(properties.getMaxTotalConnections(),
                properties.getDefaultMaxConnectionsPerRoute());
    }

    @Bean
    public PersonInformation personInformation(@Autowired(required = false) PersonInformationServiceClientFactory personInformationServiceClientFactory,
                                                 @Autowired PersonInformation.ValidationLevel validationLevel,
                                                 @Autowired AuditLogger auditLogger) {
        return new PersonInformationImpl(validationLevel, personInformationServiceClientFactory, properties.getPersonInformationMinAge(), auditLogger);
    }

    @Bean
    public PersonInformationServiceClientFactory personInformationServiceClientFactory(@Autowired HttpClientFactory httpClientFactory) {
        if (properties.getPersonInformationUrl() == null) {
            throw new IllegalArgumentException("personInformation.url skal være sat, da denne skal bruges til " +
                    "påkrævet validering af alder.");
        }
        return new PersonInformationServiceClientFactory(
                properties.getPersonInformationUrl(),
                httpClientFactory,
                properties.getPersonInformationErrorTolerance(),
                (PersonInformation.ValidationLevel.REJECT.getValidationLevel()
                                .equals(properties.getPersonInformationValidationLevel()) ?
                                HttpStatus.SC_INTERNAL_SERVER_ERROR : HttpStatus.SC_NON_AUTHORITATIVE_INFORMATION),
                properties.getInactiveStatusList());
    }

    @Bean
    public PersonInformation.ValidationLevel getValidationLevel() {
        try {
            return PersonInformation.ValidationLevel.valueOf(properties.getPersonInformationValidationLevel());
        } catch (IllegalArgumentException e) {
            throw new IllegalArgumentException(
                    "Værdien af personInformation.validationlevel skal være OFF, REJECT eller WARNING", e);
        }
    }

    @Bean
    public ServiceActorProvider serviceActorProvider(@Autowired UserTitleProvider userTitleProvider) { return new ServiceActorProvider(properties, userTitleProvider); }

    @Bean
    public MinLogClient minLogClient() throws MinLogException {
        return new MinLogClient(MINLOG_CONFIG_FILE_NAME);
    }

    @Bean
    public AuditLogger auditLogger() {
        return new AuditLogger(Audit::createAuditBuilder);
    }

    @Bean
    public UserTitleProvider userTitleProvider() {
        return new UserTitleProvider(properties.getMinlogHealthcareProfessionalUnknownUserTitle(), properties.getMinlogCitizenUserTitle());
    }

}
