/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.integrations.nas.facade;

import dk.sdsd.nsp.slalog.api.SLALogConfig;
import dk.sundhedsdatastyrelsen.livstestamente.service.integrations.nas.facade.common.IsAlive;
import dk.sundhedsdatastyrelsen.livstestamente.service.integrations.nas.facade.common.SLALogFacade;
import dk.sundhedsdatastyrelsen.livstestamente.service.integrations.nas.facade.common.SLALogFacadeImpl;
import dk.sundhedsdatastyrelsen.livstestamente.service.integrations.nas.facade.dgws.*;
import dk.sundhedsdatastyrelsen.livstestamente.service.integrations.nas.facade.dto.NotifyInput;
import org.apache.log4j.Logger;

public class NasFacadeImpl implements NasFacade {
    private static Logger logger = Logger.getLogger(NasFacadeImpl.class);

    private DgwsHeaderFactory dgwsHeaderFactory;
    private SLALogConfig slaLogConfig;
    private NasProxy nasProxy;
    private IsAlive isAlive;

    public NasFacadeImpl(DgwsHeaderFactory dgwsHeaderFactory, SLALogConfig slaLogConfig, int failThreshold, NasProxy nasProxy) {
        this.dgwsHeaderFactory = dgwsHeaderFactory;
        this.slaLogConfig = slaLogConfig;
        this.nasProxy = nasProxy;
        this.isAlive = new IsAlive(failThreshold);
    }

    @Override
    public void notify(NotifyInput input) throws DGWSFault {
        logger.debug("Calling Notify. Input: " + input);

        SLALogFacade slaLogFacade = new SLALogFacadeImpl(slaLogConfig);
        try {
            slaLogFacade.startSLALog("callNotify", input.getSlaLogInput());

            invokeService(input);
            slaLogFacade.completeSLALog();
            isAlive.addStatus(true);
            logger.info("Successfully called notify.");
        }
        catch(Exception e) {
            isAlive.addStatus(false);
            slaLogFacade.completeSLALogFailure(e);
            logger.error("Error during call to notify.", e);
            throw e;
        }
    }

    private void invokeService(NotifyInput input) throws DGWSFault {
        NasRequestContext requestContext = new NasRequestContext();
        requestContext.setFlowId(input.getFlowId());

        try {
            DgwsHeaders dgwsHeaders = dgwsHeaderFactory.getDgwsHeaders(input.getFlowId());
            requestContext.setMedcomHeader(dgwsHeaders.getHeader());
            requestContext.setSecurityHeader(dgwsHeaders.getSecurity());
        }
        catch(StsException e) {
            logger.error("Kald til STS fejlede", e);
            throw new DGWSFault("Fejl i kommunikationen med STS", e.getMessage(), e);
        }

        NasDataMapper nasDataMapper = new NasDataMapper();
        nasProxy.notify(nasDataMapper.toNotify(input), requestContext);
    }

    @Override
    public boolean isAlive() {
        return isAlive.isAlive();
    }
}
