/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.integrations.nas.facade.dgws;

import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.*;
import org.oasis_open.docs.wss._2004._01.oasis_200401_wss_wssecurity_secext_1_0.Security;
import org.w3c.dom.Document;

import java.security.cert.X509Certificate;
import java.util.Calendar;
import java.util.Date;

public class NspServiceDgwsHeaderFactoryImpl implements DgwsHeaderFactory {

	private SOSIFactory sosiFactory;

	private StsClient stsClient;

	private String itSystemName;

	private String alternativeIdentifier;

	private CareProvider careProvider;

	private X509Certificate certificate;

	private IDCard cachedIdCard;
	
	public NspServiceDgwsHeaderFactoryImpl(SOSIFactory sosiFactory, StsClient stsClient, String itSystemName, String alternativeIdentifier, String careProviderIdType, String careProviderId, String careProviderName) {

		this.sosiFactory = sosiFactory;
		this.stsClient = stsClient;
		this.itSystemName = itSystemName;
		this.alternativeIdentifier = alternativeIdentifier;
		this.careProvider = new CareProvider(careProviderIdType, careProviderId, careProviderName);	
	}
	
	public synchronized IDCard getIdCard() throws StsException {
		
		Calendar cal = Calendar.getInstance();
		cal.add(Calendar.MINUTE, 10);
		Date tenMinutesFromNow = cal.getTime();
		if (cachedIdCard == null || cachedIdCard.getExpiryDate().before(tenMinutesFromNow)) {
			cachedIdCard = getNewIdCard();
		}
		
		return cachedIdCard;
	}
	
	public IDCard getNewIdCard() throws StsException {
		
		SystemIDCard systemIdCard = sosiFactory.createNewSystemIDCard(itSystemName, careProvider, AuthenticationLevel.VOCES_TRUSTED_SYSTEM, null, null, certificate, alternativeIdentifier);
		SecurityTokenRequest securityTokenRequest = sosiFactory.createNewSecurityTokenRequest();
		securityTokenRequest.setIDCard(systemIdCard);
		SecurityTokenResponse response = stsClient.requestToken(securityTokenRequest);
		if (response.isFault()) {
			throw new StsException(response.getFaultCode()+" "+response.getFaultString());
		}
		return response.getIDCard();
	}
	

	@Override
	public DgwsHeaders getDgwsHeaders(String flowId) throws StsException {
		Request request = sosiFactory.createNewRequest(false /* do not request non-repudiation */, flowId);
		request.setIDCard(getIdCard());
		request.setDGWSVersion("1.0.1");
		Document doc = request.serialize2DOMDocument();
		
		Security security = HeaderConverter.SECURITY.documentToJaxb(Security.class, doc);
		dk.medcom.dgws._2006._04.dgws_1_0.Header header = HeaderConverter.MEDCOM.documentToJaxb(dk.medcom.dgws._2006._04.dgws_1_0.Header.class, doc);
		return new DgwsHeaders(security, header);
	}
}
