/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.webservice;

import dk.sundhedsdatastyrelsen.livstestamente._2018._05._01.*;
import dk.sundhedsdatastyrelsen.livstestamente.auditlog.AuditLogger;
import dk.sundhedsdatastyrelsen.livstestamente.service.business.LivingWillService;
import dk.sundhedsdatastyrelsen.livstestamente.domain.LivingWillVO;
import dk.sundhedsdatastyrelsen.livstestamente.domain.RequestContextVO;
import dk.sundhedsdatastyrelsen.livstestamente.util.FaultCode;
import dk.sundhedsdatastyrelsen.livstestamente.service.webservice.hl7cda.LivingWillHL7CDAMapper;
import dk.sundhedsdatastyrelsen.livstestamente.service.webservice.hl7cda.LivingWillHL7CDAValidator;
import dk.sundhedsdatastyrelsen.livstestamente.integrations.personInformation.PersonInformation;
import org.apache.cxf.annotations.SchemaValidation;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.hl7.v3.POCDMT000040ClinicalDocument;
import org.springframework.beans.factory.annotation.Autowired;

import javax.jws.WebService;
import javax.servlet.annotation.WebServlet;


@WebServlet(name="LivingWillService", urlPatterns={"/ltr"})
@WebService(serviceName = "LivingWillService", wsdlLocation = "wsdl/ltr/livstestamente_service_2018_05_01.wsdl", targetNamespace = "http://sundhedsdatastyrelsen.dk/livstestamente/2018/05/01/", portName = "LivingWillPort")
@SchemaValidation
public class LivingWillWebService extends AbstractWebService implements LivingWillPortType {

    private static final Logger logger = LogManager.getLogger(LivingWillWebService.class);

    @Autowired
    private LivingWillService businessService;

    @Autowired
    private LivingWillHL7CDAMapper livingWillHL7CDAMapper;

    @Autowired
    private PersonInformation personInformation;

    @Autowired
    private AuditLogger auditLogger;

    public LivingWillWebService() { }

    @Override
    public UpdateLivingWillResponse updateLivingWill20180501(UpdateLivingWillRequest request) {
        final UpdateLivingWillResponse response = new ObjectFactory().createUpdateLivingWillResponse();

        try {
            RequestContextVO requestContext = getRequestContextVO(request.getId().getExtension());
            new LivingWillHL7CDAValidator(personInformation).validate(request);
            final LivingWillVO livingWillVO = livingWillHL7CDAMapper.createLivingWillFromRequest(
                    requestContext.getPersonIdentifier(),
                    request.getLivingWill()
            );
            businessService.updateLivingWill(livingWillVO, requestContext);
            auditLogger.logUpdateLivingWill(livingWillVO, requestContext.getServiceActor());
        } catch (Exception e) {
            commonExceptionHandler(e, FaultCode.UPDATE_LIVING_WILL_VALIDATION_ERROR, FaultCode.UPDATE_LIVING_WILL_INTERNAL_SERVER_ERROR);
        }

        return response;
    }

    @Override
    public DeleteLivingWillResponse deleteLivingWill20180501(DeleteLivingWillRequest request) {
        final DeleteLivingWillResponse response = new ObjectFactory().createDeleteLivingWillResponse();

        try {
            String personIdentifier = request.getId().getExtension();
            RequestContextVO requestContext = getRequestContextVO(personIdentifier);
            new LivingWillHL7CDAValidator(personInformation).validate(request);
            businessService.deleteLivingWill(requestContext, false);
            auditLogger.logDeleteLivingWill(personIdentifier, requestContext.getServiceActor());
        } catch (Exception e) {
            commonExceptionHandler(e, FaultCode.DELETE_LIVING_WILL_VALIDATION_ERROR, FaultCode.DELETE_LIVING_WILL_INTERNAL_SERVER_ERROR);
        }

        return response;
    }

    @Override
    public GetLivingWillResponse getLivingWill20180501(GetLivingWillRequest request) {
        final GetLivingWillResponse response = new GetLivingWillResponse();

        try {
            RequestContextVO requestContext = getRequestContextVO(request.getId().getExtension());
            new LivingWillHL7CDAValidator(personInformation).validate(request);

            final LivingWillVO livingWillVO = businessService.getLivingWill(requestContext);

            final POCDMT000040ClinicalDocument clinicalDocument = livingWillHL7CDAMapper.toClinicalDocument(livingWillVO, requestContext);
            response.setClinicalDocument(clinicalDocument);
            auditLogger.logGetLivingWill(livingWillVO, requestContext.getServiceActor());
        } catch (Exception e) {
            commonExceptionHandler(e, FaultCode.GET_LIVING_WILL_VALIDATION_ERROR, FaultCode.GET_LIVING_WILL_INTERNAL_SERVER_ERROR);
        }

        return response;
    }

    @Override
    public HasLivingWillResponse hasLivingWill20180501(HasLivingWillRequest request) {
        final HasLivingWillResponse response = new HasLivingWillResponse();
        final LivingWillHL7CDAValidator validator = new LivingWillHL7CDAValidator(personInformation);
        try {
            RequestContextVO requestContext = getRequestContextVO(request.getId().getExtension());
            validator.validate(request);
            final boolean hasLivingWill = businessService.hasLivingWill(requestContext);

            response.setWillExists(hasLivingWill);
            auditLogger.logHasLivingWill(requestContext.getPersonIdentifier(), requestContext.getServiceActor());
        } catch (Exception e) {
            boolean isCprValidationError = validator.getValidationErrors().stream()
                    .map(ve -> ve.cprValidation)
                    .reduce((val1, val2) -> val1 && val2)
                    .orElse(false);
            if (isCprValidationError) { response.setWillExists(false); return response; }
            commonExceptionHandler(e, FaultCode.HAS_LIVING_WILL_VALIDATION_ERROR, FaultCode.HAS_LIVING_WILL_INTERNAL_SERVER_ERROR);
        }

        return response;
    }
}
