/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.webservice;

import dk.sundhedsdatastyrelsen.behandlingstestamente._2018._05._01.*;
import dk.sundhedsdatastyrelsen.livstestamente.auditlog.AuditLogger;
import dk.sundhedsdatastyrelsen.livstestamente.service.business.TreatmentWillService;
import dk.sundhedsdatastyrelsen.livstestamente.domain.RequestContextVO;
import dk.sundhedsdatastyrelsen.livstestamente.domain.TreatmentWillVO;
import dk.sundhedsdatastyrelsen.livstestamente.util.FaultCode;
import dk.sundhedsdatastyrelsen.livstestamente.conf.Properties;
import dk.sundhedsdatastyrelsen.livstestamente.service.webservice.hl7cda.TreatmentWillHL7CDAMapper;
import dk.sundhedsdatastyrelsen.livstestamente.service.webservice.hl7cda.TreatmentWillHL7CDAValidator;
import dk.sundhedsdatastyrelsen.livstestamente.integrations.personInformation.PersonInformation;
import org.apache.cxf.annotations.SchemaValidation;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.hl7.v3.POCDMT000040ClinicalDocument;
import org.springframework.beans.factory.annotation.Autowired;

import javax.jws.WebService;
import javax.servlet.annotation.WebServlet;

@WebServlet(name="TreatmentWillService", urlPatterns={"/btr/20180501"})
@WebService(serviceName = "TreatmentWillService", wsdlLocation = "wsdl/btr/behandlingstestamente_service_2018_05_01.wsdl", targetNamespace = "http://sundhedsdatastyrelsen.dk/behandlingstestamente/2018/05/01/", portName = "TreatmentWillPort")
@SchemaValidation
public class TreatmentWillWebService extends AbstractWebService implements TreatmentWillPortType {
    private static final Logger logger = LogManager.getLogger(TreatmentWillWebService.class);

    @Autowired
    private TreatmentWillService businessService;

    @Autowired
    private TreatmentWillHL7CDAMapper treatmentWillHL7CDAMapper;

    @Autowired
    private PersonInformation personInformation;

    @Autowired
    private Properties properties;

    @Autowired
    private AuditLogger auditLogger;

    @Override
    public GetTreatmentWillResponse getTreatmentWill20180501(GetTreatmentWillRequest request) {
        final GetTreatmentWillResponse response = new GetTreatmentWillResponse();
        try {
            RequestContextVO requestContext = getRequestContextVO(request.getId().getExtension());
            new TreatmentWillHL7CDAValidator(personInformation).validate(request);

            final TreatmentWillVO treatmentWillVO = businessService.getTreatmentWill(requestContext, properties.getMinLogReadActivityTextBTR());

            final POCDMT000040ClinicalDocument clinicalDocument = treatmentWillHL7CDAMapper.toClinicalDocument(treatmentWillVO, requestContext);
            response.setClinicalDocument(clinicalDocument);
            auditLogger.logGetTreatmentWill(treatmentWillVO, requestContext.getServiceActor());
        } catch (Exception e) {
            commonExceptionHandler(e, FaultCode.GET_TREATMENT_WILL_VALIDATION_ERROR, FaultCode.GET_TREATMENT_WILL_INTERNAL_SERVER_ERROR);
        }

        return response;
    }

    @Override
    public HasTreatmentWillResponse hasTreatmentWill20180501(HasTreatmentWillRequest request) {
        final HasTreatmentWillResponse response = new HasTreatmentWillResponse();
        TreatmentWillHL7CDAValidator validator = new TreatmentWillHL7CDAValidator(personInformation);
        try {
            RequestContextVO requestContext = getRequestContextVO(request.getId().getExtension());
            validator.validate(request);
            final boolean hasTreatmentWill = businessService.hasTreatmentWill(requestContext);
            response.setWillExists(hasTreatmentWill);
            auditLogger.logHasTreatmentWill(requestContext.getPersonIdentifier(), requestContext.getServiceActor());
        } catch (Exception e) {
            boolean isCprValidationError = validator.getValidationErrors().stream()
                    .map(ve -> ve.cprValidation)
                    .reduce((val1, val2) -> val1 && val2)
                    .orElse(false);
            if (isCprValidationError) { response.setWillExists(false); return response; }
            commonExceptionHandler(e, FaultCode.HAS_TREATMENT_WILL_VALIDATION_ERROR, FaultCode.HAS_TREATMENT_WILL_INTERNAL_SERVER_ERROR);
        }

        return response;
    }

    @Override
    public CreateTreatmentWillResponse createTreatmentWill20180501(CreateTreatmentWillRequest request) {
        final CreateTreatmentWillResponse response = new CreateTreatmentWillResponse();

        try {
            RequestContextVO requestContext = getRequestContextVO(request.getId().getExtension());
            new TreatmentWillHL7CDAValidator(personInformation).validate(request);
            final TreatmentWillVO treatmentWillVO = treatmentWillHL7CDAMapper.fromTreatmentWill(
                    requestContext.getPersonIdentifier(),
                    request.getTreatmentWill()
            );
            businessService.createTreatmentWill(treatmentWillVO, requestContext, false, false);
            auditLogger.logCreateTreatmentWill(treatmentWillVO, requestContext.getServiceActor());
        } catch (Exception e) {
            commonExceptionHandler(e, FaultCode.CREATE_TREATMENT_WILL_VALIDATION_ERROR, FaultCode.CREATE_TREATMENT_WILL_INTERNAL_SERVER_ERROR);
        }

        return response;
    }

    @Override
    public UpdateTreatmentWillResponse updateTreatmentWill20180501(UpdateTreatmentWillRequest request) {
        final UpdateTreatmentWillResponse response = new UpdateTreatmentWillResponse();

        try {
            RequestContextVO requestContext = getRequestContextVO(request.getId().getExtension());
            new TreatmentWillHL7CDAValidator(personInformation).validate(request);
            final TreatmentWillVO treatmentWillVO = treatmentWillHL7CDAMapper.fromTreatmentWill(
                    requestContext.getPersonIdentifier(),
                    request.getTreatmentWill()
            );
            businessService.updateTreatmentWill(treatmentWillVO, requestContext, false);
            auditLogger.logUpdateTreatmentWill(treatmentWillVO, requestContext.getServiceActor());
        } catch (Exception e) {
            commonExceptionHandler(e, FaultCode.UPDATE_TREATMENT_WILL_VALIDATION_ERROR, FaultCode.UPDATE_TREATMENT_WILL_INTERNAL_SERVER_ERROR);
        }

        return response;
    }

    @Override
    public DeleteTreatmentWillResponse deleteTreatmentWill20180501(DeleteTreatmentWillRequest request) {
        final DeleteTreatmentWillResponse response = new DeleteTreatmentWillResponse();

        try {
            RequestContextVO requestContext = getRequestContextVO(request.getId().getExtension());
            new TreatmentWillHL7CDAValidator(personInformation).validate(request);
            businessService.deleteTreatmentWill(requestContext);
            auditLogger.logDeleteTreatmentWill(requestContext.getPersonIdentifier(), requestContext.getServiceActor());
        } catch (Exception e) {
            commonExceptionHandler(e, FaultCode.DELETE_TREATMENT_WILL_VALIDATION_ERROR, FaultCode.DELETE_TREATMENT_WILL_INTERNAL_SERVER_ERROR);
        }

        return response;
    }
}
