/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.webservice;

import dk.sundhedsdatastyrelsen.behandlingstestamente._2020._03._16.*;
import dk.sundhedsdatastyrelsen.livstestamente.auditlog.AuditLogger;
import dk.sundhedsdatastyrelsen.livstestamente.service.business.LivingWillService;
import dk.sundhedsdatastyrelsen.livstestamente.service.business.TreatmentWillService;
import dk.sundhedsdatastyrelsen.livstestamente.domain.RequestContextVO;
import dk.sundhedsdatastyrelsen.livstestamente.domain.TreatmentWillVO;
import dk.sundhedsdatastyrelsen.livstestamente.util.FaultCode;
import dk.sundhedsdatastyrelsen.livstestamente.conf.Properties;
import dk.sundhedsdatastyrelsen.livstestamente.service.webservice.hl7cda.TreatmentWillHL7CDAMapperV2;
import dk.sundhedsdatastyrelsen.livstestamente.service.webservice.hl7cda.TreatmentWillHL7CDAValidatorV2;
import dk.sundhedsdatastyrelsen.livstestamente.integrations.personInformation.PersonInformation;
import org.apache.cxf.annotations.SchemaValidation;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.hl7.v3.POCDMT000040ClinicalDocument;
import org.springframework.beans.factory.annotation.Autowired;

import javax.jws.WebService;
import javax.servlet.annotation.WebServlet;


@WebServlet(name="TreatmentWillServiceV2", urlPatterns={"/btr/20200316"})
@WebService(serviceName = "TreatmentWillService", wsdlLocation = "wsdl/btr/behandlingstestamente_service_2020_03_16.wsdl", targetNamespace = "http://sundhedsdatastyrelsen.dk/behandlingstestamente/2020/03/16/", portName = "TreatmentWillPort")
@SchemaValidation
public class TreatmentWillWebServiceV2 extends AbstractWebService implements TreatmentWillPortType {
    private static final Logger logger = LogManager.getLogger(TreatmentWillWebServiceV2.class);

    @Autowired
    private TreatmentWillService businessService;

    @Autowired
    private LivingWillService livingWillService;

    @Autowired
    private TreatmentWillHL7CDAMapperV2 treatmentWillHL7CDAMapper;

    @Autowired
    private PersonInformation personInformation;

    @Autowired
    private Properties properties;

    @Autowired
    private AuditLogger auditLogger;

    @Override
    public CreateTreatmentWillResponse createTreatmentWill20200316(CreateTreatmentWillRequest request) {
        final CreateTreatmentWillResponse response = new CreateTreatmentWillResponse();

        try {
            RequestContextVO requestContext = getRequestContextVO(request.getId().getExtension());
            new TreatmentWillHL7CDAValidatorV2(personInformation).validate(request);
            final TreatmentWillVO treatmentWillVO = treatmentWillHL7CDAMapper.fromTreatmentWill(
            		requestContext.getPersonIdentifier(),
                    request.getTreatmentWillV2()
            );
            businessService.createTreatmentWill(treatmentWillVO, requestContext, true, false);
            auditLogger.logCreateTreatmentWill(treatmentWillVO, requestContext.getServiceActor());
        } catch (Exception e) {
            commonExceptionHandler(e, FaultCode.CREATE_TREATMENT_WILL_VALIDATION_ERROR, FaultCode.CREATE_TREATMENT_WILL_INTERNAL_SERVER_ERROR);
        }

        return response;
    }

    @Override
    public UpgradeToTreatmentWillResponse upgradeToTreatmentWill20200316(UpgradeToTreatmentWillRequest request) {
        final UpgradeToTreatmentWillResponse response = new UpgradeToTreatmentWillResponse();

        try {
            RequestContextVO requestContext = getRequestContextVO(request.getId().getExtension());
            new TreatmentWillHL7CDAValidatorV2(personInformation).validate(request);
            final TreatmentWillVO treatmentWillVO = treatmentWillHL7CDAMapper.fromTreatmentWill(
                    requestContext.getPersonIdentifier(),
                    request.getTreatmentWillV2()
            );

            businessService.createTreatmentWill(treatmentWillVO, requestContext, true, true); // Also deletes living will
            auditLogger.logUpgradeToTreatmentWill(treatmentWillVO, requestContext.getServiceActor());
        } catch (Exception e) {
            commonExceptionHandler(e, FaultCode.CREATE_TREATMENT_WILL_VALIDATION_ERROR, FaultCode.CREATE_TREATMENT_WILL_INTERNAL_SERVER_ERROR);
        }

        return response;
    }

    @Override
    public UpdateTreatmentWillResponse updateTreatmentWill20200316(UpdateTreatmentWillRequest request) {
        final UpdateTreatmentWillResponse response = new UpdateTreatmentWillResponse();

        try {
            RequestContextVO requestContext = getRequestContextVO(request.getId().getExtension());
            new TreatmentWillHL7CDAValidatorV2(personInformation).validate(request);
            final TreatmentWillVO treatmentWillVO = treatmentWillHL7CDAMapper.fromTreatmentWill(
            		requestContext.getPersonIdentifier(),
                    request.getTreatmentWill()
            );
            businessService.updateTreatmentWill(treatmentWillVO, requestContext, true);
            auditLogger.logUpdateTreatmentWill(treatmentWillVO, requestContext.getServiceActor());
        } catch (Exception e) {
            commonExceptionHandler(e, FaultCode.UPDATE_TREATMENT_WILL_VALIDATION_ERROR, FaultCode.UPDATE_TREATMENT_WILL_INTERNAL_SERVER_ERROR);
        }

        return response;
    }

    @Override
    public DeleteTreatmentWillResponse deleteTreatmentWill20200316(DeleteTreatmentWillRequest request) {
        final DeleteTreatmentWillResponse response = new DeleteTreatmentWillResponse();

        try {
            RequestContextVO requestContext = getRequestContextVO(request.getId().getExtension());
            new TreatmentWillHL7CDAValidatorV2(personInformation).validate(request);
            businessService.deleteTreatmentWill(requestContext);
            auditLogger.logDeleteTreatmentWill(requestContext.getPersonIdentifier(), requestContext.getServiceActor());
        } catch (Exception e) {
            commonExceptionHandler(e, FaultCode.DELETE_TREATMENT_WILL_VALIDATION_ERROR, FaultCode.DELETE_TREATMENT_WILL_INTERNAL_SERVER_ERROR);
        }

        return response;
    }

    @Override
    public GetTreatmentWillResponse getTreatmentWill20200316(GetTreatmentWillRequest request) {
        final GetTreatmentWillResponse response = new GetTreatmentWillResponse();
        try {
            RequestContextVO requestContext = getRequestContextVO(request.getId().getExtension());
            new TreatmentWillHL7CDAValidatorV2(personInformation).validate(request.getId());

            final TreatmentWillVO treatmentWillVO = businessService.getTreatmentWill(requestContext, properties.getMinLogReadActivityTextBTR());

            final POCDMT000040ClinicalDocument clinicalDocument = treatmentWillHL7CDAMapper.toClinicalDocument(treatmentWillVO, requestContext);
            response.setClinicalDocument(clinicalDocument);
            auditLogger.logGetTreatmentWill(treatmentWillVO, requestContext.getServiceActor());
        } catch (Exception e) {
            commonExceptionHandler(e, FaultCode.GET_TREATMENT_WILL_VALIDATION_ERROR, FaultCode.GET_TREATMENT_WILL_INTERNAL_SERVER_ERROR);
        }

        return response;
    }

    @Override
    public GetTreatmentWillWithOnlyForcedTreatmentResponse getTreatmentWillWithOnlyForcedTreatment20200316(GetTreatmentWillWithOnlyForcedTreatmentRequest request) {
        final GetTreatmentWillWithOnlyForcedTreatmentResponse response = new GetTreatmentWillWithOnlyForcedTreatmentResponse();
        try {
            RequestContextVO requestContext = getRequestContextVO(request.getId().getExtension());
            new TreatmentWillHL7CDAValidatorV2(personInformation).validate(request.getId());

            final TreatmentWillVO treatmentWillVO = businessService.getTreatmentWill(requestContext, properties.getMinLogReadActivityTextWithOnlyForcedTreatmentBTR());

            final POCDMT000040ClinicalDocument clinicalDocument = treatmentWillHL7CDAMapper.toClinicalDocumentWithOnlyForcedTreatment(treatmentWillVO, requestContext);
            response.setClinicalDocument(clinicalDocument);
            auditLogger.logGetTreatmentWillWithOnlyForcedTreatment(treatmentWillVO, requestContext.getServiceActor());
        } catch (Exception e) {
            commonExceptionHandler(e, FaultCode.GET_TREATMENT_WILL_VALIDATION_ERROR, FaultCode.GET_TREATMENT_WILL_INTERNAL_SERVER_ERROR);
        }

        return response;
    }

    @Override
    public HasTreatmentWillResponse hasTreatmentWill20200316(HasTreatmentWillRequest request) {
        final HasTreatmentWillResponse response = new HasTreatmentWillResponse();
        TreatmentWillHL7CDAValidatorV2 validator = new TreatmentWillHL7CDAValidatorV2(personInformation);
        try {
            RequestContextVO requestContext = getRequestContextVO(request.getId().getExtension());
            validator.validate(request);
            final boolean hasTreatmentWill = businessService.hasTreatmentWill(requestContext);
            response.setWillExists(hasTreatmentWill);
            auditLogger.logHasTreatmentWill(requestContext.getPersonIdentifier(), requestContext.getServiceActor());
        } catch (Exception e) {
            boolean isCprValidationError = validator.getValidationErrors().stream()
                    .map(rve -> rve.cprValidation)
                    .reduce((val1, val2) -> val1 && val2)
                    .orElse(false);
            if (isCprValidationError) { response.setWillExists(false); return response; }
            commonExceptionHandler(e, FaultCode.HAS_TREATMENT_WILL_VALIDATION_ERROR, FaultCode.HAS_TREATMENT_WILL_INTERNAL_SERVER_ERROR);
        }

        return response;
    }
}
