/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.webservice.hl7cda;

import dk.sundhedsdatastyrelsen.livstestamente.AssigningAuthorities;
import dk.sundhedsdatastyrelsen.livstestamente.OID;
import org.hl7.v3.*;

import java.math.BigInteger;
import java.time.Instant;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

public abstract class AbstractHL7CDAMapper {

    static ObjectFactory o = new ObjectFactory();

    static final String CONFIDENTIALITY_CODE = "N";
    static final String LANGUAGE_CODE_DK = "da-DK";

    static final String TS_FORMAT = "yyyyMMddHHmmssZ"; // Output sample: 20171019144200+0200
    static final String DATE_ONLY_FORMAT = "yyyyMMdd";
    public static final String CPR = "CPR";
    static final String REALM_CODE_DK = "DK";
    static final String RECORD_TARGET_TYPE_CODE = "RCT";
    static final String RECORD_TARGET_CTX_CTRL_CODE = "OP";
    static final String PATIENT_CLASS_CODE = "PAT";
    static final String DETERMINER_CODE = "INSTANCE";
    static final String COMP_TYPE_CODE = "COMP";
    static final List<String> genderCodes = Collections.unmodifiableList(Arrays.asList("F", "M", "UN"));
    static final String PERSON_CLASS_CODE = "PSN";
    static final String MOOD_CODE_EVENT = "EVN";
    static final String CLASS_CODE_DOCCLIN = "DOCCLIN";
    static final String OBSERVATION_CLASSCODE = "OBS";

    POCDMT000040ClinicalDocument createDocument(String code, String displayName, String codeSystem) {
        // Create ClinicalDocument with classCode and MoodCode attributes
        final POCDMT000040ClinicalDocument doc = o.createPOCDMT000040ClinicalDocument();
        doc.setClassCode(ActClinicalDocument.DOCCLIN);
        doc.getMoodCodes().add(MOOD_CODE_EVENT);

        // Elements
        //

        // realmCode
        final CS dk = new CS();
        dk.setCode(REALM_CODE_DK);
        doc.getRealmCodes().add(dk);

        // typeId
        doc.setTypeId(createTypeId("POCD_HD000040"));

        // templateId
        final II templateId = new II();
        templateId.setRoot("1.2.208.184.15.1");
        doc.getTemplateIds().add(templateId);

        // id
        doc.setId(createId(OID.ISO_MEMBERBODY_DK_MEDCOM, "NA", "MedCom"));

        // code
        doc.setCode(createCode(code, codeSystem, displayName));

        // title
        doc.setTitle(createText(displayName));

        // confidentialityCode
        doc.setConfidentialityCode(createCode("N", OID.CONFIDENTIALITYCODE));

        // languageCode
        doc.setLanguageCode(createCode("da-DK"));

        return doc;
    }

    POCDMT000040Component2 createComponent(String displayName, String code, String codeSystem, ANY observationValue) {
        final POCDMT000040Component2 component = o.createPOCDMT000040Component2();
        final POCDMT000040StructuredBody structuredBody = new POCDMT000040StructuredBody();
        final POCDMT000040Component3 bodyComponent = new POCDMT000040Component3();
        final POCDMT000040Section section = new POCDMT000040Section();
        final StrucDocText text = new StrucDocText();
        text.getContent().add(displayName);
        section.setText(text);

        POCDMT000040Observation observation = o.createPOCDMT000040Observation();
        observation.getClassCodes().add(AbstractHL7CDAMapper.OBSERVATION_CLASSCODE);
        observation.setMoodCode(XActMoodDocumentObservation.EVN);

        observation.setCode(createCode(code, codeSystem, displayName));
        observation.getValues().add(observationValue);

        POCDMT000040Entry entry = o.createPOCDMT000040Entry();
        entry.setObservation(observation);
        section.getEntries().add(entry);

        bodyComponent.setSection(section);
        structuredBody.getComponents().add(bodyComponent);
        component.setStructuredBody(structuredBody);

        return component;
    }

    POCDMT000040Author createAuthor(String personIdentifier, Instant timestamp) {
        final POCDMT000040Author author = o.createPOCDMT000040Author();

        final POCDMT000040AssignedAuthor assignedAuthor = o.createPOCDMT000040AssignedAuthor();
        assignedAuthor.getIds().add(createId(OID.CPR, personIdentifier, AssigningAuthorities.CPR));
        author.setAssignedAuthor(assignedAuthor);
        author.setTime(createTimestamp(timestamp));
        return author;
    }

    POCDMT000040RecordTarget createRecordTarget(String personIdentifier) {
        final POCDMT000040RecordTarget recordTarget = o.createPOCDMT000040RecordTarget();
        recordTarget.getTypeCodes().add("RCT");
        recordTarget.setContextControlCode("OP");

        final POCDMT000040PatientRole patientRole = o.createPOCDMT000040PatientRole();
        patientRole.getClassCodes().add("PAT");

        // Person Identifier
        if (personIdentifier != null) {
            patientRole.getIds().add(createId(OID.CPR, personIdentifier, AssigningAuthorities.CPR));
        }
        recordTarget.setPatientRole(patientRole);

        return recordTarget;
    }

    CE createCode(String code, String codeSystem) {
        return createCode(code, codeSystem, null);
    }

    CE createCode(String code, String codeSystem, String displayName) {
        final CE c = o.createCE();
        c.setCode(code);
        c.setCodeSystem(codeSystem);
        c.setDisplayName(displayName);
        return c;
    }

    POCDMT000040Custodian createCustodian() {
        final POCDMT000040Custodian custodian = o.createPOCDMT000040Custodian();
        final POCDMT000040AssignedCustodian assignedCustodian = o.createPOCDMT000040AssignedCustodian();
        final POCDMT000040CustodianOrganization custodianOrganisation = o.createPOCDMT000040CustodianOrganization();
        final II id = o.createII();
        id.setRoot("NA");
        custodianOrganisation.getIds().add(id);
        assignedCustodian.setRepresentedCustodianOrganization(custodianOrganisation);
        custodian.setAssignedCustodian(assignedCustodian);

        return custodian;
    }

    public II createId(String root, String extension, String assigningAuth) {
        final II id = o.createII();
        if (root != null) {
            id.setRoot(root);
        }
        if (extension != null) {
            id.setExtension(extension);
        }
        if (assigningAuth != null) {
            id.setAssigningAuthorityName(assigningAuth);
        }
        return id;
    }

    INT createVersionNumber(int version) {
        final INT versionNumber = o.createINT();
        versionNumber.setValue(BigInteger.valueOf(version));
        return versionNumber;
    }

    POCDMT000040InfrastructureRootTypeId createTypeId(String extension) {
        final POCDMT000040InfrastructureRootTypeId typeId = new POCDMT000040InfrastructureRootTypeId();
        typeId.setRoot(OID.HL7_REGISTEREDE_RMIMS);
        typeId.setExtension(extension);
        return typeId;
    }

    TS createTimestamp(Instant instant) {
        final TS ts = o.createTS();
        final DateTimeFormatter formatter = DateTimeFormatter.ofPattern(TS_FORMAT).withZone(ZoneId.systemDefault());
        String format = formatter.format(instant);
        ts.setValue(format);
        return ts;
    }

    ST createText(String text) {
        final ST st = o.createST();
        st.getContent().add(text);
        return st;
    }

    CS createCode(String code) {
        final CS c = o.createCS();
        c.setCode(code);
        return c;
    }
}
