/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.webservice.hl7cda;

import dk.sundhedsdatastyrelsen.livstestamente.AssigningAuthorities;
import dk.sundhedsdatastyrelsen.livstestamente.OID;
import dk.sundhedsdatastyrelsen.livstestamente.exceptions.ValidationException;
import dk.sundhedsdatastyrelsen.livstestamente.integrations.personInformation.PersonInformation;
import org.hl7.v3.*;

import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.time.format.DateTimeParseException;
import java.util.Arrays;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.stream.Collectors;

import static dk.sundhedsdatastyrelsen.livstestamente.service.webservice.hl7cda.AbstractHL7CDAMapper.CPR;
import static dk.sundhedsdatastyrelsen.livstestamente.service.webservice.hl7cda.AbstractHL7CDAMapper.TS_FORMAT;

/*
 * Note: For single-threaded use only.
 * Also be aware that validation (and access to Getters) of the generated classes has side effects.
 * Access to list fields instantiates a new list, which has a different serialization output.
 * Consequently, the service responses cannot be validated without breaking response schema validation.
 */
public abstract class AbstractHL7CDAValidator {

    private PersonInformation personInformation;

    List<ValidatorError> validationErrors;

    public static class ValidatorError {
        public final String description;
        public final Boolean cprValidation;

        public ValidatorError(String description, Boolean cprValidation) {
            this.description = description;
            this.cprValidation = cprValidation;
        }
    }

    public List<ValidatorError> getValidationErrors() {
        return validationErrors;
    }

    public AbstractHL7CDAValidator(PersonInformation personInformation) {
        this.personInformation = personInformation;
        clearValidationErrors();
    }

    private void clearValidationErrors() {
        validationErrors = new LinkedList<>();
    }

    public String getCombinedErrorText() {
        return String.join(", ", validationErrors.stream()
                .map(ve -> ve.description)
                .collect(Collectors.toList())
        );
    }

    void validateClinicalDocument(POCDMT000040ClinicalDocument doc, String elementName, String expectedCodeSystem) {

        validateRecordTargets(doc.getRecordTargets(), elementName + ".recordTargets");
        validateClinicalDocumentComponent(doc.getComponent(), elementName + ".component");
        validateAuthors(doc.getAuthors(), elementName + ".authors");
        validateCD(doc.getCode(), expectedCodeSystem, elementName + ".code");
        validateCE(doc.getConfidentialityCode(), AbstractHL7CDAMapper.CONFIDENTIALITY_CODE, OID.CONFIDENTIALITYCODE, elementName + ".confidentialityCode");
        validateCustodian(doc.getCustodian(), elementName + ".custodian");
        validateTime(doc.getEffectiveTime(), elementName + ".effectiveTime");
        checkThatExists(doc.getId(), elementName + ".id");
        validateCS(doc.getLanguageCode(), AbstractHL7CDAMapper.LANGUAGE_CODE_DK, elementName + ".languageCode");
        validateCSListWithSingleElement(doc.getRealmCodes(), AbstractHL7CDAMapper.REALM_CODE_DK, elementName + ".realmCode");
        validateTemplateIds(doc.getTemplateIds(), elementName + ".templateIds");
        checkThatExists(doc.getTitle(), elementName + ".title");
        validateId(doc.getTypeId(), true, false, false, elementName + ".typeId");
        checkThatExists(doc.getVersionNumber(), elementName + ".versionNumber");
        checkThatListContainsSingleElementWithSpecificValue(doc.getMoodCodes(), AbstractHL7CDAMapper.MOOD_CODE_EVENT, elementName + ".moodCodes");

        final String classCodeElementName = elementName + ".classCode";
        if (checkThatExists(doc.getClassCode(), classCodeElementName)) {
            checkThatExistsAndContainsSupportedValue(doc.getClassCode().value(), AbstractHL7CDAMapper.CLASS_CODE_DOCCLIN, classCodeElementName);
        }

        checkThatListDoesNotExistOrContainElements(doc.getAuthenticators(), elementName + ".authenticators");
        checkThatListDoesNotExistOrContainElements(doc.getAuthorizations(), elementName + ".authorizations");
        checkThatDoesNotExist(doc.getComponentOf(), elementName + ".componentOf");
        checkThatDoesNotExist(doc.getCopyTime(), elementName + "copyTime");
        checkThatDoesNotExist(doc.getDataEnterer(), elementName + ".dataEnterer");
        checkThatListDoesNotExistOrContainElements(doc.getDocumentationOves(), elementName + ".documentationOves");
        checkThatListDoesNotExistOrContainElements(doc.getInformants(), elementName + ".informants");
        checkThatListDoesNotExistOrContainElements(doc.getInformationRecipients(), elementName + ".informationRecipients");
        checkThatListDoesNotExistOrContainElements(doc.getInFulfillmentOves(), elementName + ".inFulfillmentOves");
        checkThatDoesNotExist(doc.getLegalAuthenticator(), elementName + ".legalAuthenticator");
        checkThatListDoesNotExistOrContainElements(doc.getNullFlavors(), elementName + ".nullFlavors");
        checkThatListDoesNotExistOrContainElements(doc.getParticipants(), elementName + ".participants");
        checkThatListDoesNotExistOrContainElements(doc.getRelatedDocuments(), elementName + ".relatedDocuments");

    }

    private void validateTemplateIds(List<II> templateIds, String elementName) {
        if (checkThatListContainsSingleElement(templateIds, elementName + ".templateId")) {
            validateId(templateIds.get(0), false, false, false, elementName + ".templateId");
        }
    }

    private void validateRecordTargets(List<POCDMT000040RecordTarget> recordTargets, String elementName) {
        if (checkThatListContainsSingleElement(recordTargets, elementName)) {
            validateRecordTarget(recordTargets.get(0), elementName);
        }
    }

    private void validateRecordTarget(POCDMT000040RecordTarget recordTarget, String elementName) {
        checkThatExistsAndContainsSupportedValue(recordTarget.getContextControlCode(), AbstractHL7CDAMapper.RECORD_TARGET_CTX_CTRL_CODE,
                elementName + ".contextControlCode");
        checkThatListContainsSingleElementWithSpecificValue(recordTarget.getTypeCodes(), AbstractHL7CDAMapper.RECORD_TARGET_TYPE_CODE,
                elementName + ".typeCode");
        validatePatientRole(recordTarget.getPatientRole(), elementName + ".patientRole");

        checkThatListDoesNotExistOrContainElements(recordTarget.getNullFlavors(), elementName + ".nullFlavors");
        checkThatDoesNotExist(recordTarget.getTypeId(), elementName + ".typeId");
        checkThatListDoesNotExistOrContainElements(recordTarget.getRealmCodes(), elementName + ".realmCodes");
        checkThatListDoesNotExistOrContainElements(recordTarget.getTemplateIds(), elementName + ".templateIds");
    }

    private void validatePatientRole(POCDMT000040PatientRole patientRole, String elementName) {
        if (checkThatListContainsSingleElement(patientRole.getIds(), elementName + ".id")) {
            validateCPR(patientRole.getIds().get(0), elementName + ".id");
        }
        checkThatListContainsSingleElementWithSpecificValue(patientRole.getClassCodes(), AbstractHL7CDAMapper.PATIENT_CLASS_CODE,
                elementName + ".classCode");

        checkThatListDoesNotExistOrContainElements(patientRole.getTelecoms(), elementName + ".telecom");
        checkThatDoesNotExist(patientRole.getTypeId(), elementName + ".typeId");
        //checkThatListDoesNotExistOrContainElements(patientRole.getNullFlavors(), elementName + ".nullFlavors"); // NulFlavors is permitted here but not required (because some of the test code uses it and some doesn't)
        checkThatListDoesNotExistOrContainElements(patientRole.getRealmCodes(), elementName + ".realmCodes");
        checkThatListDoesNotExistOrContainElements(patientRole.getTemplateIds(), elementName + ".templateIds");
        checkThatDoesNotExist(patientRole.getProviderOrganization(), elementName + ".providerOrganization");

    }

    private void validateDeterminerCode(String determinerCode, String elementName) {
        checkThatExistsAndContainsSupportedValue(determinerCode, AbstractHL7CDAMapper.DETERMINER_CODE, elementName);
    }

    private void validateCSListWithSingleElement(List<CS> codes, String code, String elementName) {
        if (checkThatListContainsSingleElement(codes, elementName)) {
            validateCS(codes.get(0), code, elementName);
        }
    }

    private void validateCS(CS cs, String code, String elementName) {
        checkThatExistsAndContainsSupportedValue(cs.getCode(), code, elementName + ".code");

        checkThatListDoesNotExistOrContainElements(cs.getNullFlavors(), elementName + ".nullFlavors");
        checkThatDoesNotExist(cs.getCodeSystemName(), elementName + ".codeSystemName");
        checkThatDoesNotExist(cs.getCodeSystemVersion(), elementName + ".codeSystemVersion");
        checkThatDoesNotExist(cs.getOriginalText(), elementName + ".originalText");
        checkThatListDoesNotExistOrContainElements(cs.getQualifiers(), elementName + ".qualifiers");
        checkThatListDoesNotExistOrContainElements(cs.getTranslations(), elementName + ".translations");
        checkThatDoesNotExist(cs.getValueSet(), elementName + ".valueSet");
        checkThatDoesNotExist(cs.getValueSetVersion(), elementName + ".valueSetVersion");
    }


    private void validateCustodian(POCDMT000040Custodian custodian, String elementName) {
        validateAssignedCustodian(custodian.getAssignedCustodian(), elementName + ".assignedCustodian");

        checkThatDoesNotExist(custodian.getTypeId(), elementName + ".typeId");
        checkThatListDoesNotExistOrContainElements(custodian.getNullFlavors(), elementName + ".nullFlavors");
        checkThatListDoesNotExistOrContainElements(custodian.getRealmCodes(), elementName + ".realmCodes");
        checkThatListDoesNotExistOrContainElements(custodian.getTemplateIds(), elementName + ".templateIds");
    }

    private void validateAssignedCustodian(POCDMT000040AssignedCustodian assignedCustodian, String elementName) {
        validateRepresentedCustodianOrganization(assignedCustodian.getRepresentedCustodianOrganization(), elementName + ".representedCustodianOrganization");

        checkThatDoesNotExist(assignedCustodian.getTypeId(), elementName + ".typeId");
        checkThatListDoesNotExistOrContainElements(assignedCustodian.getNullFlavors(), elementName + ".nullFlavors");
        checkThatListDoesNotExistOrContainElements(assignedCustodian.getRealmCodes(), elementName + ".realmCodes");
        checkThatListDoesNotExistOrContainElements(assignedCustodian.getTemplateIds(), elementName + ".templateIds");
    }

    private void validateRepresentedCustodianOrganization(POCDMT000040CustodianOrganization reprCustodianOrg, String elementName) {
        checkThatListContainsSingleElement(reprCustodianOrg.getIds(), elementName + ".id");
        validateDeterminerCode(reprCustodianOrg.getDeterminerCode(), elementName + ".determinerCode");

        checkThatDoesNotExist(reprCustodianOrg.getAddr(), elementName + ".addr");
        checkThatDoesNotExist(reprCustodianOrg.getTelecom(), elementName + ".telecom");
        checkThatDoesNotExist(reprCustodianOrg.getTypeId(), elementName + ".typeId");
        checkThatListDoesNotExistOrContainElements(reprCustodianOrg.getNullFlavors(), elementName + ".nullFlavors");
        checkThatListDoesNotExistOrContainElements(reprCustodianOrg.getRealmCodes(), elementName + ".realmCodes");
        checkThatListDoesNotExistOrContainElements(reprCustodianOrg.getTemplateIds(), elementName + ".templateIds");
        checkThatDoesNotExist(reprCustodianOrg.getName(), elementName + ".name");
    }

    private void validateCE(CE ce, String expectedCode, String expectedCodeSystem, String elementName) {
        validateCE(ce, Collections.singletonList(expectedCode), expectedCodeSystem, elementName);
    }

    private void validateCE(CE ce, List<String> expectedCodes, String expectedCodeSystem, String elementName) {
        checkThatExistsAndContainsSupportedValue(ce.getCode(), expectedCodes, elementName + ".code");
        checkThatExistsAndContainsSupportedValue(ce.getCodeSystem(), expectedCodeSystem, elementName + ".codeSystem");

        checkThatListDoesNotExistOrContainElements(ce.getNullFlavors(), elementName + ".nullFlavors");
        checkThatDoesNotExist(ce.getCodeSystemName(), elementName + ".codeSystemName");
        checkThatDoesNotExist(ce.getCodeSystemVersion(), elementName + ".codeSystemVersion");
        checkThatDoesNotExist(ce.getOriginalText(), elementName + ".originalText");
        checkThatListDoesNotExistOrContainElements(ce.getQualifiers(), elementName + ".qualifiers");
        checkThatListDoesNotExistOrContainElements(ce.getTranslations(), elementName + ".translations");
        checkThatDoesNotExist(ce.getValueSet(), elementName + ".valueSet");
        checkThatDoesNotExist(ce.getValueSetVersion(), elementName + ".valueSetVersion");
    }

    private void validateClinicalDocumentComponent(POCDMT000040Component2 component, String elementName) {
        // If all we know about a patient is info from CPR, we have no components. That is valid.
        if (exists(component.getNonXMLBody())) {
            checkThatExists(component.getNonXMLBody().getText(), elementName + "nonXMLBody");
        } else {
            validateStructuredBody(component.getStructuredBody(), elementName + ".structuredBody");

            checkThatDoesNotExist(component.getTypeId(), elementName + ".typeId");
            checkThatListDoesNotExistOrContainElements(component.getNullFlavors(), elementName + ".nullFlavors");
            checkThatListDoesNotExistOrContainElements(component.getRealmCodes(), elementName + ".realmCodes");
            checkThatListDoesNotExistOrContainElements(component.getTemplateIds(), elementName + ".templateIds");
        }
    }

    private void validateStructuredBody(POCDMT000040StructuredBody structuredBody, String elementName) {
        if (checkThatListContainsSingleElement(structuredBody.getComponents(), elementName + ".component")) {
            validateStructuredBodyComponent(structuredBody.getComponents().get(0), elementName + ".component");
        }

        checkThatListDoesNotExistOrContainElements(structuredBody.getClassCodes(), elementName + ".classCodes");
        checkThatDoesNotExist(structuredBody.getConfidentialityCode(), elementName + ".confidentialityCode");
        checkThatDoesNotExist(structuredBody.getLanguageCode(), elementName + ".languageCode");
        checkThatListDoesNotExistOrContainElements(structuredBody.getMoodCodes(), elementName + ".moodCodes");
        checkThatListDoesNotExistOrContainElements(structuredBody.getNullFlavors(), elementName + ".nullFlavors");
        checkThatListDoesNotExistOrContainElements(structuredBody.getRealmCodes(), elementName + ".realmCodes");
        checkThatListDoesNotExistOrContainElements(structuredBody.getTemplateIds(), elementName + ".templateIds");
        checkThatDoesNotExist(structuredBody.getTypeId(), elementName + ".typeId");
    }

    private void validateStructuredBodyComponent(POCDMT000040Component3 component, String elementName) {
        validateComponentSection(component.getSection(), elementName + ".section");
        final String typeCodeElementName = elementName + ".typeCode";
        if (checkThatExists(component.getTypeCode(), typeCodeElementName)) {
            checkThatExistsAndContainsSupportedValue(component.getTypeCode().value(), AbstractHL7CDAMapper.COMP_TYPE_CODE, typeCodeElementName);
        }

        checkThatDoesNotExist(component.getTypeId(), elementName + ".typeId");
        checkThatListDoesNotExistOrContainElements(component.getNullFlavors(), elementName + ".nullFlavors");
        checkThatListDoesNotExistOrContainElements(component.getRealmCodes(), elementName + ".realmCodes");
        checkThatListDoesNotExistOrContainElements(component.getTemplateIds(), elementName + ".templateIds");
    }

    private void validateComponentSection(POCDMT000040Section section, String elementName) {
        checkThatExists(section.getText(), elementName + ".text");

        checkThatListDoesNotExistOrContainElements(section.getAuthors(), elementName + ".authors");
        checkThatListDoesNotExistOrContainElements(section.getClassCodes(), elementName + ".classCodes");
        checkThatDoesNotExist(section.getCode(), elementName + ".code");
        checkThatListDoesNotExistOrContainElements(section.getComponents(), elementName + ".components");
        checkThatDoesNotExist(section.getConfidentialityCode(), elementName + ".confidentialityCode");
        checkThatDoesNotExist(section.getId(), elementName + ".id");
        checkThatDoesNotExist(section.getId2(), elementName + ".id2");
        checkThatListDoesNotExistOrContainElements(section.getInformants(), elementName + ".informants");
        checkThatDoesNotExist(section.getLanguageCode(), elementName + ".languageCode");
        checkThatListDoesNotExistOrContainElements(section.getMoodCodes(), elementName + ".moodCodes");
        checkThatListDoesNotExistOrContainElements(section.getNullFlavors(), elementName + ".nullFlavors");
        checkThatListDoesNotExistOrContainElements(section.getRealmCodes(), elementName + ".realmCodes");
        checkThatDoesNotExist(section.getSubject(), elementName + ".subject");
        checkThatListDoesNotExistOrContainElements(section.getTemplateIds(), elementName + ".templateIds");
        checkThatDoesNotExist(section.getTitle(), elementName + ".title");
        checkThatDoesNotExist(section.getTypeId(), elementName + ".typeId");
    }

    private void validateAuthors(List<POCDMT000040Author> authors, String elementName) {
        if (checkThatListContainsSingleElement(authors, elementName)) {
            validateAuthor(authors.get(0), elementName);
        }
    }

    private void validateAuthor(POCDMT000040Author author, String elementName) {
        validateTime(author.getTime(), elementName + ".time");
        validateAssignedAuthor(author.getAssignedAuthor(), true, elementName + ".assignedAuthor");

        checkThatDoesNotExist(author.getFunctionCode(), elementName + ".functionCode");
        checkThatDoesNotExist(author.getTypeId(), elementName + ".typeId");
        checkThatListDoesNotExistOrContainElements(author.getTypeCodes(), elementName + ".typeCodes");
        checkThatListDoesNotExistOrContainElements(author.getNullFlavors(), elementName + ".nullFlavors");
        checkThatListDoesNotExistOrContainElements(author.getRealmCodes(), elementName + ".realmCodes");
        checkThatListDoesNotExistOrContainElements(author.getTemplateIds(), elementName + ".templateIds");
    }

    void throwExceptionIfValidationErrors() {
        if (validationErrors.size() > 0) { throw new IllegalArgumentException(getCombinedErrorText()); }
    }

    void validateIdWithPersonIdentifier(II id, String elementName) {
        validateId(id, true, true, true, elementName);
    }

    private void validateId(II id, boolean extensionRequired, boolean assigningAuthorityRequired, boolean isPersonIdentifier, String elementName) {
        if (!checkThatExists(id, elementName)) return;

        if (extensionRequired) {
            if (!checkThatExists(id.getExtension(), elementName + ".extension")) {
                return;
            }
            if (isPersonIdentifier)
                checkPersonIdLength(id.getExtension(), elementName + ".extension");
                if (CPR.equals(id.getAssigningAuthorityName())) {
                    validatePersonInformation(id.getExtension());
                }
        }
        if (assigningAuthorityRequired)
            if (checkThatExists(id.getAssigningAuthorityName(), elementName + ".assigningAuthorityName"))
                checkThatExistsAndContainsSupportedValue(id.getAssigningAuthorityName(), Collections.singletonList(AbstractHL7CDAMapper.CPR), elementName + ".assigningAuthorityName");
        checkThatExists(id.getRoot(), elementName + ".root");
        checkThatDoesNotExist(id.isDisplayable(), elementName + ".displayable");
    }

    private void checkPersonIdLength(String extension, String elementName) {
        if (extension == null || extension.length() != 10) {
            validationErrors.add(new ValidatorError("Person identifier not Valid. Requires 10 digits " + elementName + " [" + extension + "]", false));
        }
    }

    private void validatePersonInformation(String extension) {
        try {
            personInformation.validateCpr(extension);
            personInformation.validateAge(extension);
        } catch (ValidationException e) {
            validationErrors.add(new ValidatorError(e.getMessage(), true));
        }
    }

    private void validateAssignedAuthor(POCDMT000040AssignedAuthor assignedAuthor,
                                        boolean allowAuthorizationAndEducationCode, String elementName) {
        validateAssignedAuthorIds(assignedAuthor.getIds(), allowAuthorizationAndEducationCode, elementName);

        if (!checkThatExists(assignedAuthor, elementName)) return;

        checkThatExistsAndContainsSupportedValue(assignedAuthor.getClassCode(), Collections.singletonList("ASSIGNED"), elementName + ".classCode");

        checkThatDoesNotExist(assignedAuthor.getAssignedAuthoringDevice(), elementName + ".assignedAuthoringDevice");
        checkThatListDoesNotExistOrContainElements(assignedAuthor.getTelecoms(), elementName + ".telecom");
        checkThatDoesNotExist(assignedAuthor.getTypeId(), elementName + ".typeId");
        checkThatListDoesNotExistOrContainElements(assignedAuthor.getNullFlavors(), elementName + ".nullFlavors");
        checkThatDoesNotExist(assignedAuthor.getCode(), elementName + ".code");
        checkThatListDoesNotExistOrContainElements(assignedAuthor.getRealmCodes(), elementName + ".realmCodes");
        checkThatListDoesNotExistOrContainElements(assignedAuthor.getTemplateIds(), elementName + ".templateIds");
    }

    private void validateAssignedAuthorIds(List<II> ids, boolean allowAuthorizationAndEducationCode, String elementName) {
        final String idElement = ".id";
        if (allowAuthorizationAndEducationCode) {
            if (checkThatListContainsRightNumberOfElements(ids, 1, 3,
                    elementName + idElement)) {
                for (II id : ids) {
                    validateId(id,
                            Arrays.asList(OID.CPR, OID.AUTORISATIONSID, OID.UDDANNELSESKODE),
                            Arrays.asList(AssigningAuthorities.CPR, AssigningAuthorities.AUTORISATIONSID, AssigningAuthorities.UDDANNELSESKODE),
                            elementName + idElement);
                }
            }
        } else {
            if (checkThatListContainsSingleElement(ids, elementName + idElement)) {
                validateCPR(ids.get(0), elementName);
            }
        }
    }

    private void validateCPR(II id, String elementName) {
        validateId(id, Collections.singletonList(OID.CPR), Collections.singletonList(AssigningAuthorities.CPR), elementName + ".id");
    }

    private void validateId(II id, List<String> supportedRoots, List<String> supportedAssigningAuthorities, String elementName) {
        if (!checkThatExists(id, elementName)) return;
        checkThatExists(id.getExtension(), elementName + ".extension");
        checkThatExistsAndContainsSupportedValue(id.getAssigningAuthorityName(), supportedAssigningAuthorities,
                elementName + ".assigningAuthorityName");
        checkThatExistsAndContainsSupportedValue(id.getRoot(), supportedRoots, elementName + ".root");

        checkThatDoesNotExist(id.isDisplayable(), elementName + ".displayable");
    }

    private void validateTime(TS time, String elementName) {
        if (!checkThatExists(time, elementName)) return;
        if (checkThatExists(time.getValue(), elementName + ".value")) {
            checkDateTimeIsValid(time.getValue(), elementName + ".value");
        }

        checkThatListDoesNotExistOrContainElements(time.getNullFlavors(), elementName + ".nullFlavors");
    }

    private void validateCD(CD cd, String expectedCodeSystem, String elementName) {
        if (!checkThatExists(cd, elementName)) return;
        checkThatExists(cd.getCode(), elementName + ".code");
        checkThatExists(cd.getDisplayName(), elementName + ".displayName");
        checkThatExistsAndContainsSupportedValue(cd.getCodeSystem(), Collections.singletonList(expectedCodeSystem),
                elementName + ".codeSystem");

        checkThatListDoesNotExistOrContainElements(cd.getNullFlavors(), elementName + ".nullFlavors");
        checkThatDoesNotExist(cd.getCodeSystemName(), elementName + ".codeSystemName");
        checkThatDoesNotExist(cd.getCodeSystemVersion(), elementName + ".codeSystemVersion");
        checkThatDoesNotExist(cd.getOriginalText(), elementName + ".originalText");
        checkThatListDoesNotExistOrContainElements(cd.getQualifiers(), elementName + ".qualifiers");
        checkThatListDoesNotExistOrContainElements(cd.getTranslations(), elementName + ".translations");
        checkThatDoesNotExist(cd.getValueSet(), elementName + ".valueSet");
        checkThatDoesNotExist(cd.getValueSetVersion(), elementName + ".valueSetVersion");
    }

    // General purpose validation methods:

    private void checkThatExistsAndContainsSupportedValue(String s, String supportedValue, String elementName) {
        checkThatExistsAndContainsSupportedValue(s, Collections.singletonList(supportedValue), elementName);
    }

    private void checkThatExistsAndContainsSupportedValue(String s, List<String> supportedValues, String elementName) {
        if (!checkThatExists(s, elementName)) {
            return;
        }
        if (!supportedValues.contains(s)) {
            validationErrors.add(new ValidatorError("The " + elementName + " value " + s + " is not supported. Supported values are: "
                    + String.join(",", supportedValues), false));
        }
    }

    boolean checkThatExists(Object o, String elementName) {
        if (o == null || (o instanceof String && ((String) o).isEmpty())) {
            validationErrors.add(new ValidatorError("Required value was missing: " + elementName, false));
            return false;
        }
        return true;
    }

    private boolean exists(Object o) {
        return o != null && (!(o instanceof String) || !((String) o).isEmpty());
    }

    private void checkThatDoesNotExist(Object o, String elementName) {
        if (o != null) {
            validationErrors.add(new ValidatorError("Unsupported element found: " + elementName, false));
        }
    }

    void checkThatAttributeDoesNotExist(Object o, String attributeName) {
        if (o != null) {
            validationErrors.add(new ValidatorError("Illegal attribute found: " + attributeName, false));
        }
    }

    private <T> void checkThatListDoesNotExistOrContainElements(List<T> list, String elementName) {
        if (list != null && list.size() > 0) {
            validationErrors.add(new ValidatorError("Unsupported value was found: " + elementName, false));
        }
    }

    private void checkThatListContainsSingleElementWithSpecificValue(List<String> list, String expectedValue, String elementName) {
        if (checkThatListContainsSingleElement(list, elementName)) {
            checkThatExistsAndContainsSupportedValue(list.get(0), expectedValue, elementName);
        }
    }

    private <T> boolean checkThatListContainsSingleElement(List<T> list, String elementName) {
        return checkThatListContainsRightNumberOfElements(list, 1, 1, elementName);
    }

    private <T> boolean checkThatListContainsRightNumberOfElements(List<T> list, int minNumberOfElements,
                                                                   int maxNumberOfElements, String elementName) {
        if (list == null || list.size() == 0) {
            validationErrors.add(new ValidatorError("Required value was missing: " + elementName, false));
            return false;
        }
        if (list.size() < minNumberOfElements) {
            validationErrors.add(new ValidatorError(list.size() + " elements were found but at least " + minNumberOfElements + " elements are required: " + elementName, false));
            return false;
        }
        if (list.size() > maxNumberOfElements) {
            validationErrors.add(new ValidatorError(list.size() + " elements were found but at most " + maxNumberOfElements + " elements are supported: " + elementName, false));
            return false;
        }
        return true;
    }

    private void checkDateTimeIsValid(String dateTimeString, String elementName) {
        try {
            	DateTimeFormatter dateTimeFormatter = DateTimeFormatter.ofPattern(TS_FORMAT);
                LocalDateTime.parse(dateTimeString, dateTimeFormatter);
        } catch (DateTimeParseException  e) {
            	validationErrors.add(new ValidatorError("The timestamp " + dateTimeString + " for the element " + elementName + " does not comply with the required format: " + TS_FORMAT + ". " + e.getMessage(), false));
        }
    }

    String buildUnspecificErrorMessage(String requestName) {
        return buildUnspecificErrorMessage(requestName, "");
    }

    String buildUnspecificErrorMessage(String requestName, String errorMessage) {
        return "Unspecified validation error parsing " + requestName + " Request: " + errorMessage;
    }


}
