/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.webservice.hl7cda;

import dk.sundhedsdatastyrelsen.livstestamente.OID;
import dk.sundhedsdatastyrelsen.livstestamente._2018._05._01.*;
import dk.sundhedsdatastyrelsen.livstestamente.integrations.personInformation.PersonInformation;
import org.hl7.ltr.LivingWill;
import org.hl7.v3.II;
import org.springframework.stereotype.Component;

@Component
public class LivingWillHL7CDAValidator extends AbstractHL7CDAValidator {

    public LivingWillHL7CDAValidator(PersonInformation personInformation) {
        super(personInformation);
    }

    public void validateWithoutThrow(GetLivingWillResponse response) {
        try {
            validateClinicalDocument(response.getClinicalDocument(), "clinicalDocument", OID.LIVING_WILL_CODE_SYSTEM);
        } catch (Exception e) {
            validationErrors.add(new ValidatorError(buildUnspecificErrorMessage("Living Will Response", e.getMessage()), false));
        }
    }

    public void validate(GetLivingWillRequest request) {
        try {
            validateIdWithPersonIdentifier(request.getId(), "id");
        } catch (Exception e) {
            validationErrors.add(new ValidatorError(buildUnspecificErrorMessage("Get LivingWill", e.getMessage()), false));
        }
        throwExceptionIfValidationErrors();
    }

    public void validate(HasLivingWillRequest request) {
        try {
            validateIdWithPersonIdentifier(request.getId(), "id");
        } catch (Exception e) {
            validationErrors.add(new ValidatorError(buildUnspecificErrorMessage("Has LivingWill", e.getMessage()), false));
        }
        throwExceptionIfValidationErrors();
    }

    public void validate(UpdateLivingWillRequest request) {
        validateCreateOrUpdateLivingWill(request.getId(), request.getLivingWill(), buildUnspecificErrorMessage("Update LivingWill"));
    }

    public void validate(DeleteLivingWillRequest request) {
        try {
            validateIdWithPersonIdentifier(request.getId(), "id");
        } catch (Exception e) {
            validationErrors.add(new ValidatorError(buildUnspecificErrorMessage("Delete LivingWill", e.getMessage()), false));
        }
        throwExceptionIfValidationErrors();
    }

    private void validateCreateOrUpdateLivingWill(II id, LivingWill livingWill, String errorMessage) {
        try {
            validateIdWithPersonIdentifier(id, "id");
            validateLivingWill(livingWill, "livingWill");
        } catch (Exception e) {
            validationErrors.add(new ValidatorError(errorMessage + e.getMessage(), false));
        }
        throwExceptionIfValidationErrors();
    }

    private void validateLivingWill(LivingWill livingWill, String elementName) {
        checkThatExists(livingWill.isNoLifeProlongingIfDying(), elementName + ".noLifeProlongingIfDying");
        checkThatExists(livingWill.isNoLifeProlongingIfSeverelyDegraded(), elementName + ".noLifeProlongingIfSeverelyDegraded");
    }
}
