/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.webservice.hl7cda;

import dk.sundhedsdatastyrelsen.livstestamente.OID;
import dk.sundhedsdatastyrelsen.livstestamente.domain.AcceptanceNeeded;
import dk.sundhedsdatastyrelsen.livstestamente.domain.RequestContextVO;
import dk.sundhedsdatastyrelsen.livstestamente.domain.TreatmentWillVO;
import org.hl7.btr.TreatmentWill;
import org.hl7.btr.TreatmentWillValueType;
import org.hl7.v3.POCDMT000040ClinicalDocument;
import org.springframework.stereotype.Component;

import javax.xml.bind.ValidationException;

@Component
public class TreatmentWillHL7CDAMapper extends AbstractHL7CDAMapper {

    public static org.hl7.btr.ObjectFactory oBTR = new org.hl7.btr.ObjectFactory();

    private static final String TREATMENT_WILL_CODE = "TreatmentWill";
    private static final String TREATMENT_WILL_DISPLAY_NAME = "Behandlingstestamente";

    public POCDMT000040ClinicalDocument toClinicalDocument(TreatmentWillVO treatmentWillVO, RequestContextVO requestContextVO) {
        final POCDMT000040ClinicalDocument doc = createDocument(TREATMENT_WILL_CODE, TREATMENT_WILL_DISPLAY_NAME, OID.TREATMENT_WILL_CODE_SYSTEM);

        // effectiveTime
        doc.setEffectiveTime(createTimestamp(requestContextVO.getTime()));

        // versionNumber
        doc.setVersionNumber(createVersionNumber(treatmentWillVO.getVersion()));

        // recordTarget
        doc.getRecordTargets().add(createRecordTarget(treatmentWillVO.getPersonIdentifier()));

        // author
        doc.getAuthors().add(createAuthor(treatmentWillVO.getPersonIdentifier(), treatmentWillVO.getModifiedDateTime()));

        // custodian
        doc.setCustodian(createCustodian());

        // component
        TreatmentWill treatmentWill = toTreatmentWill(treatmentWillVO);
        doc.setComponent(createComponent(TREATMENT_WILL_DISPLAY_NAME, TREATMENT_WILL_CODE, OID.TREATMENT_WILL_CODE_SYSTEM, treatmentWill));

        return doc;
    }

    public TreatmentWillVO fromTreatmentWill(String personIdentifier, TreatmentWill treatmentWill) throws ValidationException {
        // Notice: It is assumed that the structure has been checked by the validator.

        final boolean noLifeProlongingIfDying = treatmentWill.getNoLifeProlongingIfDying().isValue();
        final AcceptanceNeeded noLifeProlongingIfDyingAcceptanceNeeded = fromAcceptanceNeededString(
                treatmentWill.getNoLifeProlongingIfDying().getAcceptanceNeeded());

        final boolean noLifeProlongingIfSeverelyDegraded = treatmentWill.getNoLifeProlongingIfSeverelyDegraded().isValue();
        final AcceptanceNeeded noLifeProlongingIfSeverelyDegradedAcceptanceNeeded = fromAcceptanceNeededString(
                treatmentWill.getNoLifeProlongingIfSeverelyDegraded().getAcceptanceNeeded());

        final boolean noLifeProlongingIfSeverePain = treatmentWill.getNoLifeProlongingIfSeverePain().isValue();
        final AcceptanceNeeded noLifeProlongingIfSeverePainAcceptanceNeeded = fromAcceptanceNeededString(
                treatmentWill.getNoLifeProlongingIfSeverePain().getAcceptanceNeeded());

        final boolean noForcedTreatmentIfIncapable = treatmentWill.getNoForcedTreatmentIfIncapable().isValue();
        final AcceptanceNeeded noForcedTreatmentIfIncapableAcceptanceNeeded = fromAcceptanceNeededString(
                treatmentWill.getNoForcedTreatmentIfIncapable().getAcceptanceNeeded());

        TreatmentWillVO treatmentWillVO = new TreatmentWillVO();
        treatmentWillVO.setPersonIdentifier(personIdentifier);
        treatmentWillVO.setNoLifeProlongingIfDying(noLifeProlongingIfDying);
        treatmentWillVO.setNoLifeProlongingIfDyingAcceptanceNeeded(noLifeProlongingIfDyingAcceptanceNeeded);
        treatmentWillVO.setNoLifeProlongingIfSeverelyDegraded(noLifeProlongingIfSeverelyDegraded);
        treatmentWillVO.setNoLifeProlongingIfSeverelyDegradedAcceptanceNeeded(noLifeProlongingIfSeverelyDegradedAcceptanceNeeded);
        treatmentWillVO.setNoLifeProlongingIfSeverePain(noLifeProlongingIfSeverePain);
        treatmentWillVO.setNoLifeProlongingIfSeverePainAcceptanceNeeded(noLifeProlongingIfSeverePainAcceptanceNeeded);
        treatmentWillVO.setNoForcedTreatmentIfIncapable(noForcedTreatmentIfIncapable);
        treatmentWillVO.setNoForcedTreatmentIfIncapableAcceptanceNeeded(noForcedTreatmentIfIncapableAcceptanceNeeded);
        return treatmentWillVO;
    }

    private TreatmentWill toTreatmentWill(TreatmentWillVO treatmentWillVO) {
        TreatmentWill treatmentWill = new TreatmentWill();

        final TreatmentWillValueType noLifeProlongingIfDying = toTreatmentWillValueType(
                treatmentWillVO.isNoLifeProlongingIfDying(),
                treatmentWillVO.getNoLifeProlongingIfDyingAcceptanceNeeded());
        treatmentWill.setNoLifeProlongingIfDying(noLifeProlongingIfDying);

        final TreatmentWillValueType noLifeProlongingIfSeverelyDegraded = toTreatmentWillValueType(
                treatmentWillVO.isNoLifeProlongingIfSeverelyDegraded(),
                treatmentWillVO.getNoLifeProlongingIfSeverelyDegradedAcceptanceNeeded());
        treatmentWill.setNoLifeProlongingIfSeverelyDegraded(noLifeProlongingIfSeverelyDegraded);

        final TreatmentWillValueType noLifeProlongingIfSeverePain = toTreatmentWillValueType(
                treatmentWillVO.isNoLifeProlongingIfSeverePain(),
                treatmentWillVO.getNoLifeProlongingIfSeverePainAcceptanceNeeded());
        treatmentWill.setNoLifeProlongingIfSeverePain(noLifeProlongingIfSeverePain);

        final TreatmentWillValueType noForcedTreatmentIfIncapable = toTreatmentWillValueType(
                treatmentWillVO.isNoForcedTreatmentIfIncapable(),
                treatmentWillVO.getNoForcedTreatmentIfIncapableAcceptanceNeeded());
        treatmentWill.setNoForcedTreatmentIfIncapable(noForcedTreatmentIfIncapable);

        return treatmentWill;
    }

    public TreatmentWillValueType toTreatmentWillValueType(boolean value, AcceptanceNeeded acceptanceNeeded) {
        TreatmentWillValueType treatmentWillValueType = new TreatmentWillValueType();
        treatmentWillValueType.setValue(value);
        treatmentWillValueType.setAcceptanceNeeded(toAcceptanceNeededString(acceptanceNeeded));

        return treatmentWillValueType;
    }

    private AcceptanceNeeded fromAcceptanceNeededString(String acceptanceNeeded) throws ValidationException {
        if (acceptanceNeeded == null) {
            return null;
        }

        switch (acceptanceNeeded) {
            case "relativeAcceptanceRequired":
                return AcceptanceNeeded.RELATIVE_ACCEPTANCE_REQUIRED;
            case "guardianAcceptanceRequired":
                return AcceptanceNeeded.GUARDIAN_ACCEPTANCE_REQUIRED;
            case "trustedAgentAcceptanceRequired":
                return AcceptanceNeeded.TRUSTED_AGENT_ACCEPTANCE_REQUIRED;
            default:
                throw new ValidationException("Value of acceptanceNeeded element not matched");
        }
    }

    private String toAcceptanceNeededString(AcceptanceNeeded acceptanceNeeded) {
        if (acceptanceNeeded == null) {
            return null;
        }

        switch (acceptanceNeeded) {
            case RELATIVE_ACCEPTANCE_REQUIRED:
                return "relativeAcceptanceRequired";
            case GUARDIAN_ACCEPTANCE_REQUIRED:
                return "guardianAcceptanceRequired";
            case TRUSTED_AGENT_ACCEPTANCE_REQUIRED:
                return "trustedAgentAcceptanceRequired";
            default:
                throw new IllegalArgumentException(acceptanceNeeded.name());
        }
    }
}
