/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.webservice.hl7cda;

import dk.sundhedsdatastyrelsen.behandlingstestamente._2018._05._01.*;
import dk.sundhedsdatastyrelsen.livstestamente.OID;
import dk.sundhedsdatastyrelsen.livstestamente.integrations.personInformation.PersonInformation;
import org.hl7.btr.TreatmentWill;
import org.hl7.btr.TreatmentWillValueType;
import org.hl7.v3.II;
import org.springframework.stereotype.Component;

@Component
public class TreatmentWillHL7CDAValidator extends AbstractHL7CDAValidator {

    public TreatmentWillHL7CDAValidator(PersonInformation personInformation) {
        super(personInformation);
    }

    public void validateWithoutThrow(GetTreatmentWillResponse response) {
        try {
            validateClinicalDocument(response.getClinicalDocument(), "clinicalDocument", OID.LIVING_WILL_CODE_SYSTEM);
        } catch (Exception e) {
            validationErrors.add(new ValidatorError(buildUnspecificErrorMessage("Living Will Response", e.getMessage()), false));
        }
    }

    public void validate(GetTreatmentWillRequest request) {
        try {
            validateIdWithPersonIdentifier(request.getId(), "id");
        } catch (Exception e) {
            validationErrors.add(new ValidatorError(buildUnspecificErrorMessage("Get TreatmentWill", e.getMessage()), false));
        }
        throwExceptionIfValidationErrors();
    }

    public void validate(HasTreatmentWillRequest request) {
        try {
            validateIdWithPersonIdentifier(request.getId(), "id");
        } catch (Exception e) {
            validationErrors.add(new ValidatorError(buildUnspecificErrorMessage("Has TreatmentWill", e.getMessage()), false));
        }
        throwExceptionIfValidationErrors();
    }

    public void validate(CreateTreatmentWillRequest request) {
        validateCreateOrUpdateTreatmentWill(request.getId(), request.getTreatmentWill(), buildUnspecificErrorMessage("Create TreatmentWill"));
    }

    public void validate(UpdateTreatmentWillRequest request) {
        validateCreateOrUpdateTreatmentWill(request.getId(), request.getTreatmentWill(), buildUnspecificErrorMessage("Update TreatmentWill"));
    }

    public void validate(DeleteTreatmentWillRequest request) {
        try {
            validateIdWithPersonIdentifier(request.getId(), "id");
        } catch (Exception e) {
            validationErrors.add(new ValidatorError(buildUnspecificErrorMessage("Delete TreatmentWill", e.getMessage()), false));
        }
        throwExceptionIfValidationErrors();
    }

    private void validateCreateOrUpdateTreatmentWill(II id, TreatmentWill treatmentWill, String errorMessage) {
        try {
            validateIdWithPersonIdentifier(id, "id");
            validateTreatmentWill(treatmentWill);
        } catch (Exception e) {
            validationErrors.add(new ValidatorError(errorMessage + e.getMessage(), false));
        }
        throwExceptionIfValidationErrors();
    }

    private void validateTreatmentWill(TreatmentWill treatmentWill) {
        validateTreatmentWillValueType(treatmentWill.getNoLifeProlongingIfDying(), "noLifeProlongingIfDying");
        validateTreatmentWillValueTypeAccpetanceNeededIsNull(treatmentWill.getNoLifeProlongingIfDying());
        
        validateTreatmentWillValueType(treatmentWill.getNoLifeProlongingIfSeverelyDegraded(), "noLifeProlongingIfSeverelyDegraded");
        validateTreatmentWillValueType(treatmentWill.getNoLifeProlongingIfSeverePain(), "noLifeProlongingIfSeverePain");
        validateTreatmentWillValueType(treatmentWill.getNoForcedTreatmentIfIncapable(), "noForcedTreatmentIfIncapable");
    }

    private void validateTreatmentWillValueType(TreatmentWillValueType element, String elementName) {
        if (!element.isValue()) {
            checkThatAttributeDoesNotExist(element.getAcceptanceNeeded(), elementName + "/@acceptanceNeeded");
        }
    }
    
    private void validateTreatmentWillValueTypeAccpetanceNeededIsNull(TreatmentWillValueType element) {
        if (element.getAcceptanceNeeded() != null) {
        	validationErrors.add(new ValidatorError("Unsupported value was found: " + element.getAcceptanceNeeded(), false));
        }
    }
    
}
