/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.auditlog;

import dk.sds.nsp.audit.AuditBuilder;
import dk.sundhedsdatastyrelsen.livstestamente.auditlog.AuditLogger;
import dk.sundhedsdatastyrelsen.livstestamente.domain.LivingWillVO;
import dk.sundhedsdatastyrelsen.livstestamente.domain.TreatmentWillVO;
import dk.sundhedsdatastyrelsen.livstestamente.service.util.TestHelper;
import dk.sundhedsdatastyrelsen.livstestamente.context.ServiceActor;
import org.junit.Before;
import org.junit.Test;
import org.mockito.Mockito;

public class AuditLoggerTest {

    private static final String COMPONENT = "BTR";

    private TestHelper testHelper;
    private AuditBuilder auditBuilder;
    private AuditLogger auditLogger;

    @Before
    public void setUp() throws Exception {
        testHelper = new TestHelper();
        auditBuilder = Mockito.mock(AuditBuilder.class);
        auditLogger = new AuditLogger(() -> auditBuilder);
    }

    @Test
    public void logGetTreatmentWill_LogsTreatmentWillAndServiceActor() {
        TreatmentWillVO treatmentWill = testHelper.getTreatmentWillExample();
        ServiceActor serviceActor = testHelper.getServiceActorExample();

        auditLogger.logGetTreatmentWill(treatmentWill, serviceActor);

        String context = "getTreatmentWill";
        verifyTreatmentWillAuditLogged(treatmentWill, context);
        verifyServiceActorAuditLogged(serviceActor, context);
    }

    @Test
    public void logGetTreatmentWillWithOnlyForcedTreatment_LogsTreatmentWillAndServiceActor() {
        TreatmentWillVO treatmentWill = testHelper.getTreatmentWillExample();
        ServiceActor serviceActor = testHelper.getServiceActorExample();

        auditLogger.logGetTreatmentWillWithOnlyForcedTreatment(treatmentWill, serviceActor);

        String context = "getTreatmentWillWithOnlyForcedTreatment";
        verifyTreatmentWillAuditLogged(treatmentWill, context);
        verifyServiceActorAuditLogged(serviceActor, context);
    }

    @Test
    public void logHasTreatmentWill_LogsCprAndServiceActor() {
        ServiceActor serviceActor = testHelper.getServiceActorExample();
        String cpr = "1111111111";

        auditLogger.logHasTreatmentWill(cpr, serviceActor);

        String context = "hasTreatmentWill";

        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.RegularPersonalInformation, "cpr", cpr);
        verifyServiceActorAuditLogged(serviceActor, context);
    }

    @Test
    public void logCreateTreatmentWill_LogsTreatmentWillAndServiceActor() {
        TreatmentWillVO treatmentWill = testHelper.getTreatmentWillExample();
        ServiceActor serviceActor = testHelper.getServiceActorExample();

        auditLogger.logCreateTreatmentWill(treatmentWill, serviceActor);

        String context = "createTreatmentWill";
        verifyTreatmentWillAuditLogged(treatmentWill, context);
        verifyServiceActorAuditLogged(serviceActor, context);
    }

    @Test
    public void logUpdateTreatmentWill_LogsTreatmentWillAndServiceActor() {
        TreatmentWillVO treatmentWill = testHelper.getTreatmentWillExample();
        ServiceActor serviceActor = testHelper.getServiceActorExample();

        auditLogger.logUpdateTreatmentWill(treatmentWill, serviceActor);

        String context = "updateTreatmentWill";
        verifyTreatmentWillAuditLogged(treatmentWill, context);
        verifyServiceActorAuditLogged(serviceActor, context);
    }

    @Test
    public void logDeleteTreatmentWill_LogsCprAndServiceActor() {
        ServiceActor serviceActor = testHelper.getServiceActorExample();
        String cpr = "1111111111";

        auditLogger.logDeleteTreatmentWill(cpr, serviceActor);

        String context = "deleteTreatmentWill";

        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.RegularPersonalInformation, "cpr", cpr);
        verifyServiceActorAuditLogged(serviceActor, context);
    }

    @Test
    public void logUpgradeToTreatmentWill_LogsTreatmentWillAndServiceActor() {
        TreatmentWillVO treatmentWill = testHelper.getTreatmentWillExample();
        ServiceActor serviceActor = testHelper.getServiceActorExample();

        auditLogger.logUpgradeToTreatmentWill(treatmentWill, serviceActor);

        String context = "upgradeToTreatmentWill";
        verifyTreatmentWillAuditLogged(treatmentWill, context);
        verifyServiceActorAuditLogged(serviceActor, context);
    }

    @Test
    public void logUpdateLivingWill_LogsLivingWillAndServiceActor() {
        LivingWillVO livingWill = testHelper.getLivingWillExample();
        ServiceActor serviceActor = testHelper.getServiceActorExample();

        auditLogger.logUpdateLivingWill(livingWill, serviceActor);

        String context = "updateLivingWill";
        verifyLivingWillAuditLogged(livingWill, context);
        verifyServiceActorAuditLogged(serviceActor, context);
    }

    @Test
    public void logDeleteLivingWill_LogsCprAndServiceActor() {
        ServiceActor serviceActor = testHelper.getServiceActorExample();
        String cpr = "1111111111";

        auditLogger.logDeleteLivingWill(cpr, serviceActor);

        String context = "deleteLivingWill";
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.RegularPersonalInformation, "cpr", cpr);
        verifyServiceActorAuditLogged(serviceActor, context);
    }

    @Test
    public void logGetLivingWill_LogsLivingWillAndServiceActor() {
        LivingWillVO livingWill = testHelper.getLivingWillExample();
        ServiceActor serviceActor = testHelper.getServiceActorExample();

        auditLogger.logGetLivingWill(livingWill, serviceActor);

        String context = "getLivingWill";
        verifyLivingWillAuditLogged(livingWill, context);
        verifyServiceActorAuditLogged(serviceActor, context);
    }

    @Test
    public void logHasLivingWill_LogsCprAndServiceActor() {
        ServiceActor serviceActor = testHelper.getServiceActorExample();
        String cpr = "1111111111";

        auditLogger.logHasLivingWill(cpr, serviceActor);

        String context = "hasLivingWill";
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.RegularPersonalInformation, "cpr", cpr);
        verifyServiceActorAuditLogged(serviceActor, context);
    }

    @Test
    public void logCprDoesNotExist_LogsCprDoesNotExist() {
        String cpr = "1111111111";

        auditLogger.logCprDoesNotExist(cpr);

        String context = "validateCpr";
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.RegularPersonalInformation, "cpr-does-not-exist", cpr);
    }

    @Test
    public void logCprInactive_LogsCprInactive() {
        String cpr = "1111111111";

        auditLogger.logCprInactive(cpr);

        String context = "validateCpr";
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.RegularPersonalInformation, "cpr-inactive", cpr);
    }

    private void verifyServiceActorAuditLogged(ServiceActor serviceActor, String context) {
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.RegularPersonalInformation, "actor-cpr", serviceActor.getPersonIdentifier());
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.RegularPersonalInformation, "actor-firstname", serviceActor.getFirstName());
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.RegularPersonalInformation, "actor-lastname", serviceActor.getLastName());
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.RegularPersonalInformation, "actor-usertype", serviceActor.getUserType());
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.RegularPersonalInformation, "actor-authorisation-number", serviceActor.getAuthorisationNumber());
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.RegularPersonalInformation, "actor-education-code", serviceActor.getEducationCode());
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.RegularPersonalInformation, "actor-organisation-id", serviceActor.getOrganizationId());
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.RegularPersonalInformation, "actor-organisation-id-source", serviceActor.getOrganizationIdSource());
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.RegularPersonalInformation, "actor-organisation-name", serviceActor.getOrganizationName());
    }

    private void verifyTreatmentWillAuditLogged(TreatmentWillVO treatmentWill, String context) {
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.NonPersonalInformation, "pid", treatmentWill.getTreatmentWillPID());
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.RegularPersonalInformation, "cpr", treatmentWill.getPersonIdentifier());
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.SensitivePersonalInformation, "no-life-prolonging-if-dying", treatmentWill.isNoLifeProlongingIfDying());
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.SensitivePersonalInformation, "no-life-prolonging-if-dying-acceptance-needed", treatmentWill.getNoLifeProlongingIfDyingAcceptanceNeeded());
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.SensitivePersonalInformation, "no-life-prolonging-if-severely-degraded", treatmentWill.isNoLifeProlongingIfSeverelyDegraded());
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.SensitivePersonalInformation, "no-life-prolonging-if-severely-degraded-acceptance-needed", treatmentWill.getNoLifeProlongingIfSeverelyDegradedAcceptanceNeeded());
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.SensitivePersonalInformation, "no-life-prolonging-if-severe-pain", treatmentWill.isNoLifeProlongingIfSeverePain());
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.SensitivePersonalInformation, "no-life-prolonging-if-severe-pain-acceptance-needed", treatmentWill.getNoLifeProlongingIfSeverePainAcceptanceNeeded());
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.SensitivePersonalInformation, "no-forced-treatment-if-incapable", treatmentWill.isNoForcedTreatmentIfIncapable());
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.SensitivePersonalInformation, "no-forced-treatment-if-incapable-acceptance-needed", treatmentWill.getNoForcedTreatmentIfIncapableAcceptanceNeeded());
    }

    private void verifyLivingWillAuditLogged(LivingWillVO livingWill, String context) {
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.NonPersonalInformation, "pid", livingWill.getLivingWillPID());
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.RegularPersonalInformation, "cpr", livingWill.getPersonIdentifier());
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.SensitivePersonalInformation, "no-life-prolonging-if-dying", livingWill.isNoLifeProlongingIfDying());
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.SensitivePersonalInformation, "no-life-prolonging-if-severely-degraded", livingWill.isNoLifeProlongingIfSeverelyDegraded());
    }
}