/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.business;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;
import static org.mockito.Mockito.*;

import java.time.Instant;

import dk.sundhedsdatastyrelsen.livstestamente.service.integrations.nas.treatmentwill.NasException;
import dk.sundhedsdatastyrelsen.livstestamente.service.integrations.nas.treatmentwill.NotificationSender;
import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Mock;
import org.mockito.junit.MockitoJUnit;
import org.mockito.junit.MockitoRule;
import org.mockito.quality.Strictness;

import dk.sds.nsp.security.SecurityContext.Organisation.OrganisationIdentifierFormat;
import dk.sundhedsdatastyrelsen.livstestamente.service.business.validator.LivingWillValidator;
import dk.sundhedsdatastyrelsen.livstestamente.service.business.validator.TreatmentWillValidator;
import dk.sundhedsdatastyrelsen.livstestamente.dao.LivingWillDAO;
import dk.sundhedsdatastyrelsen.livstestamente.dao.TreatmentWillDAO;
import dk.sundhedsdatastyrelsen.livstestamente.domain.LivingWillVO;
import dk.sundhedsdatastyrelsen.livstestamente.domain.RequestContextVO;
import dk.sundhedsdatastyrelsen.livstestamente.exceptions.DAOException;
import dk.sundhedsdatastyrelsen.livstestamente.service.integrations.minlog.MinLogDataWrapper;
import dk.sundhedsdatastyrelsen.livstestamente.service.integrations.minlog.MinLogException;
import dk.sundhedsdatastyrelsen.livstestamente.util.FaultCode;
import dk.sundhedsdatastyrelsen.livstestamente.conf.Properties;
import dk.sundhedsdatastyrelsen.livstestamente.util.ServiceException;
import dk.sundhedsdatastyrelsen.livstestamente.service.util.TestHelperV2;
import dk.sundhedsdatastyrelsen.livstestamente.context.ServiceActor;

public class LivingWillServiceTests extends TestHelperV2 {

    @Rule
    public MockitoRule rule = MockitoJUnit.rule().strictness(Strictness.STRICT_STUBS);

    @Mock
    private Properties properties;

    @Mock
    private TreatmentWillDAO treatmentWillDAO;

    @Mock
    private LivingWillDAO livingWillDAO;

    @Mock
    private MinLogDataWrapper minLogDataWrapper;

    @Mock
    private AccessControl accessControl;

    @Mock
    private TreatmentWillValidator treatmentWillValidator;

    @Mock
    private LivingWillValidator livingWillValidator;

    @Mock
    private NotificationSender notificationSender;

    private LivingWillServiceImpl classUnderTest;

    @Before
    public void setUp() {
        classUnderTest = new LivingWillServiceImpl(properties, treatmentWillDAO, livingWillDAO, minLogDataWrapper, accessControl, treatmentWillValidator, livingWillValidator, notificationSender);
    }

    @Test
    public void getLivingWill_all_DAO_fail() throws DAOException {
        // Note: These tests cover the thrown exceptions in the service,
        // but JaCoCo cannot detect lines that throw exceptions.

        // Arrange
        LivingWillVO livingWillVO = new LivingWillVO();
        ServiceActor serviceActor = getServiceActorExample();
        RequestContextVO requestContext = getRequestContextExample(serviceActor, MESSAGE_ID);
        ServiceActor serviceActorExample = getServiceActorExample();

        // Stub DAO method to throw an exception
        when(livingWillDAO.getCurrentLivingWill(anyString(), any(Instant.class)))
                .thenThrow(new DAOException("An error occurred reading LivingWill"));

        // Act
        try {
            classUnderTest.getLivingWill(requestContext);
            fail();
        } catch (ServiceException e) {
            // Assert
            assertEquals(FaultCode.GET_LIVING_WILL_INTERNAL_SERVER_ERROR, e.getFaultCode());
        }

        // Act
        try {
            classUnderTest.hasLivingWill(requestContext);
            fail();
        } catch (ServiceException e) {
            // Assert
            assertEquals(FaultCode.HAS_LIVING_WILL_INTERNAL_SERVER_ERROR, e.getFaultCode());
        }

        // Act
        try {
            classUnderTest.updateLivingWill(livingWillVO, requestContext);
            fail();
        } catch (ServiceException e) {
            // Assert
            assertEquals(FaultCode.UPDATE_LIVING_WILL_INTERNAL_SERVER_ERROR, e.getFaultCode());
        }

        // Act
        try {
            classUnderTest.deleteLivingWill(requestContext, false);
            fail();
        } catch (ServiceException e) {
            // Assert
            assertEquals(FaultCode.DELETE_LIVING_WILL_INTERNAL_SERVER_ERROR, e.getFaultCode());
        }
    }

    @Test
    public void getLivingWill_success() throws DAOException, MinLogException, NasException {
        // Arrange
        ServiceActor serviceActor = getServiceActorExample();
        RequestContextVO requestContext = getRequestContextExample(serviceActor, MESSAGE_ID);
        LivingWillVO livingWillVO = new LivingWillVO();

        // Stub DAO method to return the LivingWillVO
        when(livingWillDAO.getCurrentLivingWill(anyString(), any(Instant.class)))
                .thenReturn(livingWillVO);

        // Stub Properties to return a string
        String minLogReadActivityTextLTR = "test";
        when(properties.getMinLogReadActivityTextLTR()).thenReturn(minLogReadActivityTextLTR);

        // Act
        LivingWillVO livingWillReturned = classUnderTest.getLivingWill(requestContext);

        // Assert
        verify(accessControl).checkGetAccess(eq(requestContext));
        verify(livingWillDAO).getCurrentLivingWill(eq(PERSON_IDENTIFIER), eq(requestContext.getTime()));
        verify(livingWillValidator).validateLivingWillExists(eq(livingWillVO));
        verify(minLogDataWrapper, times(1)).logActivity(eq(requestContext), eq(minLogReadActivityTextLTR), anyString(), anyString(), anyString(), any(), eq(requestContext.getPersonIdentifier()), isNull(), eq(LivingWillServiceImpl.SERVICE_IDENTIFIER), eq(serviceActor.getUserTitle()));
        verify(notificationSender, times(0)).sendLivingWillNotification(any(), any());

        assertEquals(livingWillVO, livingWillReturned);
    }
    
    @Test
    public void citizenGetLivingWill_minLogCalled_success() throws DAOException, MinLogException, NasException {
    	getLivingWill_minLogCalled_success(getServiceActorExample(), 1);
    }
    @Test
    public void adminGetLivingWill_minLogCalled_success() throws DAOException, MinLogException, NasException {
    	getLivingWill_minLogCalled_success(getServiceActorAdminExample(), 1);
    }
    @Test
    public void systemGetLivingWill_minLogCalled_success() throws DAOException, MinLogException, NasException {
    	getLivingWill_minLogCalled_success(getServiceActorSystemUserExample(), 0);
    }
    @Test
    public void healthCareProfGetLivingWill_minLogCalled_success() throws DAOException, MinLogException, NasException {
    	getLivingWill_minLogCalled_success(getServiceActorHealthCareProfessionalExample(), 1);
    }
    
    private void getLivingWill_minLogCalled_success(ServiceActor serviceActor, int minlogExpectedCalls) throws DAOException, MinLogException, NasException {
        // Arrange
        RequestContextVO requestContext = getRequestContextExample(serviceActor, MESSAGE_ID);
        LivingWillVO livingWillVO = new LivingWillVO();

        // Stub DAO method to return the LivingWillVO
        when(livingWillDAO.getCurrentLivingWill(anyString(), any(Instant.class)))
                .thenReturn(livingWillVO);

        // Stub Properties to return a string
        String minLogReadActivityTextLTR = "test";
        when(properties.getMinLogReadActivityTextLTR()).thenReturn(minLogReadActivityTextLTR);

        // Act
        LivingWillVO livingWillReturned = classUnderTest.getLivingWill(requestContext);

        // Assert
        verify(minLogDataWrapper, times(minlogExpectedCalls)).logActivity(eq(requestContext), eq(minLogReadActivityTextLTR), anyString(), anyString(), anyString(), any(), eq(requestContext.getPersonIdentifier()), isNull(), eq(LivingWillServiceImpl.SERVICE_IDENTIFIER), eq(serviceActor.getUserTitle()));
        verify(notificationSender, times(0)).sendLivingWillNotification(any(), any());

    }

    @Test
    public void getLivingWill_failOnMinLogException() throws DAOException, MinLogException {
        // Arrange
        ServiceActor serviceActor = getServiceActorHealthCareProfessionalExample();
        RequestContextVO requestContext = getRequestContextExample(serviceActor, MESSAGE_ID);
        LivingWillVO livingWillVO = new LivingWillVO();

        // Stub DAO method to return the LivingWillVO
        when(livingWillDAO.getCurrentLivingWill(anyString(), any(Instant.class)))
                .thenReturn(livingWillVO);

        // Stub MinLog to throw an Exception
        doThrow(MinLogException.class).when(minLogDataWrapper)
                .logActivity(
                        any(RequestContextVO.class),
                        nullable(String.class),
                        nullable(String.class),
                        nullable(String.class),
                        nullable(String.class),
                        nullable(OrganisationIdentifierFormat.class),
                        nullable(String.class),
                        nullable(String.class),
                        nullable(String.class),
                        nullable(String.class)
                );

        // Stub Properties to return a string
        String minLogReadActivityTextLTR = "test";
        when(properties.getMinLogReadActivityTextLTR()).thenReturn(minLogReadActivityTextLTR);

        // Act
        try {
            classUnderTest.getLivingWill(requestContext);
            fail();
        } catch (ServiceException e) {
            // Assert
            verify(accessControl).checkGetAccess(eq(requestContext));
            verify(livingWillDAO).getCurrentLivingWill(eq(PERSON_IDENTIFIER), eq(requestContext.getTime()));
            verify(livingWillValidator).validateLivingWillExists(eq(livingWillVO));
            verify(minLogDataWrapper).logActivity(eq(requestContext), eq(minLogReadActivityTextLTR), anyString(), anyString(), anyString(), any(), eq(requestContext.getPersonIdentifier()), isNull(), eq(LivingWillServiceImpl.SERVICE_IDENTIFIER), eq(serviceActor.getUserTitle()));

            assertEquals(FaultCode.GET_LIVING_WILL_INTERNAL_SERVER_ERROR, e.getFaultCode());
        }
    }

    @Test
    public void hasLivingWill_true() throws DAOException {
        // Arrange
        ServiceActor serviceActor = getServiceActorExample();
        RequestContextVO requestContext = getRequestContextExample(serviceActor, MESSAGE_ID);
        LivingWillVO livingWillVO = new LivingWillVO();

        // Stub DAO method to return the LivingWillVO
        when(livingWillDAO.getCurrentLivingWill(anyString(), any(Instant.class)))
                .thenReturn(livingWillVO);

        // Act
        boolean hasLivingWill = classUnderTest.hasLivingWill(requestContext);

        // Assert
        verify(accessControl).checkHasAccess(eq(requestContext));
        verify(livingWillDAO).getCurrentLivingWill(eq(PERSON_IDENTIFIER), eq(requestContext.getTime()));

        assertTrue(hasLivingWill);
    }

    @Test
    public void hasLivingWill_false() throws DAOException {
        // Arrange
        ServiceActor serviceActor = getServiceActorExample();
        RequestContextVO requestContext = getRequestContextExample(serviceActor, MESSAGE_ID);

        // Stub DAO method to return null
        when(livingWillDAO.getCurrentLivingWill(anyString(), any(Instant.class)))
                .thenReturn(null);

        // Act
        boolean hasLivingWill = classUnderTest.hasLivingWill(requestContext);

        // Assert
        verify(accessControl).checkHasAccess(eq(requestContext));
        verify(livingWillDAO).getCurrentLivingWill(eq(PERSON_IDENTIFIER), eq(requestContext.getTime()));

        assertFalse(hasLivingWill);
    }

    @Test
    public void updateLivingWill_success() throws DAOException, NasException {
        // Arrange
        long pid = 1L;
        int version = 1;

        LivingWillVO livingWillVO = getLivingWillExample();
        ServiceActor serviceActorExample = getServiceActorExample();
        RequestContextVO requestContext = getRequestContextExample(serviceActorExample, MESSAGE_ID);

        // Stub DAO method to return an object with a PID and Version
        LivingWillVO livingWillVOFake = new LivingWillVO();
        livingWillVOFake.setLivingWillPID(pid);
        livingWillVOFake.setVersion(version);
        when(livingWillDAO.getCurrentLivingWill(anyString(), any(Instant.class)))
                .thenReturn(livingWillVOFake);

        // Act
        classUnderTest.updateLivingWill(livingWillVO, requestContext);

        // Assert
        verify(accessControl).checkUpdateAccess(eq(requestContext));
        verify(livingWillDAO).getCurrentLivingWill(eq(PERSON_IDENTIFIER), eq(requestContext.getTime()));
        verify(livingWillValidator).validateLivingWillExists(eq(livingWillVOFake));
        verify(livingWillDAO).updateValidToTimestamp(eq(pid), eq(requestContext.getTime()));
        verify(livingWillDAO).createLivingWill(eq(livingWillVO));
        verify(notificationSender).sendLivingWillNotification(serviceActorExample.getPersonIdentifier(), requestContext.getTime());

        assertEquals(livingWillVO.getPersonIdentifier(), livingWillVO.getPersonIdentifier());
        assertEquals(LivingWillServiceImpl.INITIAL_VERSION_DEFAULT + 1, livingWillVO.getVersion().intValue());
        assertEquals(requestContext.getTime(), livingWillVO.getValidFrom());
        assertEquals(LivingWillServiceImpl.VALID_TO_DEFAULT, livingWillVO.getValidTo());
        assertEquals(requestContext.getTime(), livingWillVO.getModifiedDateTime());
    }

    @Test
    public void citizenUpdateLivingWill__minLogCalled_success() throws DAOException, MinLogException, NasException {
    	updateLivingWill__minLogCalled_success(getServiceActorExample(), 1);
    }
    @Test
    public void adminUpdateLivingWill__minLogCalled_success() throws DAOException, MinLogException, NasException {
    	updateLivingWill__minLogCalled_success(getServiceActorAdminExample(), 1);
    }
    @Test
    public void systemUpdateLivingWill__minLogCalled_success() throws DAOException, MinLogException, NasException {
    	updateLivingWill__minLogCalled_success(getServiceActorSystemUserExample(), 0);
    }
    @Test
    public void healthCareProfUpdateLivingWill__minLogCalled_success() throws DAOException, MinLogException, NasException {
    	updateLivingWill__minLogCalled_success(getServiceActorHealthCareProfessionalExample(), 1);
    }
    
    private void updateLivingWill__minLogCalled_success(ServiceActor serviceActor, int minlogExpectedCalls) throws DAOException, MinLogException, NasException {
        // Arrange
        long pid = 1L;
        int version = 1;

        LivingWillVO livingWillVO = getLivingWillExample();
        RequestContextVO requestContext = getRequestContextExample(serviceActor, MESSAGE_ID);

        // Stub DAO method to return an object with a PID and Version
        LivingWillVO livingWillVOFake = new LivingWillVO();
        livingWillVOFake.setLivingWillPID(pid);
        livingWillVOFake.setVersion(version);
        when(livingWillDAO.getCurrentLivingWill(anyString(), any(Instant.class)))
                .thenReturn(livingWillVOFake);
        String minLogReadActivityTextLTR = "test";
        when(properties.getMinLogUpdateActivityTextLTR()).thenReturn(minLogReadActivityTextLTR);

        // Act
        classUnderTest.updateLivingWill(livingWillVO, requestContext);

        // Assert
        verify(minLogDataWrapper, times(minlogExpectedCalls)).logActivity(eq(requestContext), eq(minLogReadActivityTextLTR), anyString(), anyString(), anyString(), any(), eq(requestContext.getPersonIdentifier()), isNull(), eq(LivingWillServiceImpl.SERVICE_IDENTIFIER), eq(serviceActor.getUserTitle()));
        verify(notificationSender).sendLivingWillNotification(requestContext.getPersonIdentifier(), requestContext.getTime());
    }

    @Test
    public void deleteLivingWill_success() throws DAOException, NasException {
        // Arrange
        long pid = 1L;
        int version = 1;

        ServiceActor serviceActorExample = getServiceActorExample();
        RequestContextVO requestContext = getRequestContextExample(serviceActorExample, MESSAGE_ID);

        // Stub DAO method to return an object with a PID and Version
        LivingWillVO livingWillVOFake = new LivingWillVO();
        livingWillVOFake.setLivingWillPID(pid);
        livingWillVOFake.setVersion(version);
        when(livingWillDAO.getCurrentLivingWill(anyString(), any(Instant.class)))
                .thenReturn(livingWillVOFake);

        // Act
        classUnderTest.deleteLivingWill(requestContext, false);

        // Assert
        verify(accessControl).checkDeleteAccess(eq(requestContext));

        ArgumentCaptor<Long> updateValidToArgument1 = ArgumentCaptor.forClass(Long.class);
        ArgumentCaptor<Instant> updateValidToArgument2 = ArgumentCaptor.forClass(Instant.class);
        verify(livingWillDAO).updateValidToTimestamp(updateValidToArgument1.capture(), updateValidToArgument2.capture());
        verify(notificationSender).sendLivingWillNotification(serviceActorExample.getPersonIdentifier(), requestContext.getTime());

        assertEquals(pid, updateValidToArgument1.getValue().longValue());
        assertNotNull(updateValidToArgument2.getValue());
    }

    @Test
    public void citizenDeleteLivingWill_minLogCalled_success() throws DAOException, MinLogException, NasException {
    	adminDeleteLivingWill_minLogCalled_success(getServiceActorExample(), 1);
    }
    @Test
    public void adminDeleteLivingWill_minLogCalled_success() throws DAOException, MinLogException, NasException {
    	adminDeleteLivingWill_minLogCalled_success(getServiceActorAdminExample(), 1);
    }
    @Test
    public void systemDeleteLivingWill_minLogCalled_success() throws DAOException, MinLogException, NasException {
    	adminDeleteLivingWill_minLogCalled_success(getServiceActorSystemUserExample(), 0);
    }
    @Test
    public void healtCareProfDeleteLivingWill_minLogCalled_success() throws DAOException, MinLogException, NasException {
    	adminDeleteLivingWill_minLogCalled_success(getServiceActorHealthCareProfessionalExample(), 1);
    }
    
    private void adminDeleteLivingWill_minLogCalled_success(ServiceActor serviceActor, int minlogExpectedCalls) throws DAOException, MinLogException, NasException {
        // Arrange
        long pid = 1L;
        int version = 1;

        RequestContextVO requestContext = getRequestContextExample(serviceActor, MESSAGE_ID);

        // Stub DAO method to return an object with a PID and Version
        LivingWillVO livingWillVOFake = new LivingWillVO();
        livingWillVOFake.setLivingWillPID(pid);
        livingWillVOFake.setVersion(version);
        when(livingWillDAO.getCurrentLivingWill(anyString(), any(Instant.class)))
                .thenReturn(livingWillVOFake);
        String minLogReadActivityTextLTR = "test";
        when(properties.getMinLogDeleteActivityTextLTR()).thenReturn(minLogReadActivityTextLTR);


        // Act
        classUnderTest.deleteLivingWill(requestContext, false);

        // Assert
        verify(minLogDataWrapper, times(minlogExpectedCalls)).logActivity(eq(requestContext), eq(minLogReadActivityTextLTR), anyString(), anyString(), anyString(), any(), eq(requestContext.getPersonIdentifier()), isNull(), eq(LivingWillServiceImpl.SERVICE_IDENTIFIER), eq(serviceActor.getUserTitle()));
        verify(notificationSender).sendLivingWillNotification(requestContext.getPersonIdentifier(), requestContext.getTime());
    }
}
