/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.business;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;
import static org.mockito.Mockito.*;

import java.time.Instant;

import dk.sundhedsdatastyrelsen.livstestamente.dao.LivingWillDAO;
import dk.sundhedsdatastyrelsen.livstestamente.dao.TreatmentWillDAO;
import dk.sundhedsdatastyrelsen.livstestamente.domain.RequestContextVO;
import dk.sundhedsdatastyrelsen.livstestamente.domain.TreatmentWillVO;
import dk.sundhedsdatastyrelsen.livstestamente.exceptions.DAOException;
import dk.sundhedsdatastyrelsen.livstestamente.service.integrations.nas.treatmentwill.NasException;
import dk.sundhedsdatastyrelsen.livstestamente.service.integrations.nas.treatmentwill.NotificationSender;
import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Mock;
import org.mockito.junit.MockitoJUnit;
import org.mockito.junit.MockitoRule;
import org.mockito.quality.Strictness;

import dk.sds.nsp.security.SecurityContext.Organisation.OrganisationIdentifierFormat;
import dk.sundhedsdatastyrelsen.livstestamente.service.business.validator.LivingWillValidator;
import dk.sundhedsdatastyrelsen.livstestamente.service.business.validator.TreatmentWillValidator;
import dk.sundhedsdatastyrelsen.livstestamente.service.integrations.minlog.MinLogDataWrapper;
import dk.sundhedsdatastyrelsen.livstestamente.service.integrations.minlog.MinLogException;
import dk.sundhedsdatastyrelsen.livstestamente.util.FaultCode;
import dk.sundhedsdatastyrelsen.livstestamente.conf.Properties;
import dk.sundhedsdatastyrelsen.livstestamente.util.ServiceException;
import dk.sundhedsdatastyrelsen.livstestamente.service.util.TestHelperV2;
import dk.sundhedsdatastyrelsen.livstestamente.context.ServiceActor;

public class TreatmentWillServiceTests extends TestHelperV2 {

    @Rule
    public MockitoRule rule = MockitoJUnit.rule().strictness(Strictness.STRICT_STUBS);

    @Mock
    private Properties properties;

    @Mock
    private TreatmentWillDAO treatmentWillDAO;

    @Mock
    private LivingWillDAO livingWillDAO;

    @Mock
    private MinLogDataWrapper minLogDataWrapper;

    @Mock
    private AccessControl accessControl;

    @Mock
    private TreatmentWillValidator treatmentWillValidator;

    @Mock
    private LivingWillValidator livingWillValidator;

    @Mock
    private NotificationSender notificationSender;

    private TreatmentWillServiceImpl classUnderTest;

    @Before
    public void setUp() {
        classUnderTest = new TreatmentWillServiceImpl(properties, treatmentWillDAO, livingWillDAO, minLogDataWrapper, accessControl, treatmentWillValidator, livingWillValidator, notificationSender);
    }

    @Test
    public void getTreatmentWill_all_DAO_fail() throws DAOException {
        // Note: These tests cover the thrown exceptions in the service,
        // but JaCoCo cannot detect lines that throw exceptions.

        // Arrange
        TreatmentWillVO treatmentWillVO = new TreatmentWillVO();
        ServiceActor serviceActor = getServiceActorExample();
        RequestContextVO requestContext = getRequestContextExample(serviceActor, MESSAGE_ID);
        ServiceActor serviceActorExample = getServiceActorExample();

        // Stub DAO method to throw an exception
        when(treatmentWillDAO.getCurrentTreatmentWill(anyString(), any(Instant.class)))
                .thenThrow(new DAOException("An error occurred reading TreatmentWill"));

        // Act
        try {
            classUnderTest.getTreatmentWill(requestContext, "Logning (TEST)");
            fail();
        } catch (ServiceException e) {
            // Assert
            assertEquals(FaultCode.GET_TREATMENT_WILL_INTERNAL_SERVER_ERROR, e.getFaultCode());
        }

        // Act
        try {
            classUnderTest.hasTreatmentWill(requestContext);
            fail();
        } catch (ServiceException e) {
            // Assert
            assertEquals(FaultCode.HAS_TREATMENT_WILL_INTERNAL_SERVER_ERROR, e.getFaultCode());
        }

        // Act
        try {
            classUnderTest.createTreatmentWill(treatmentWillVO, requestContext, true, false);
            fail();
        } catch (ServiceException e) {
            // Assert
            assertEquals(FaultCode.CREATE_TREATMENT_WILL_INTERNAL_SERVER_ERROR, e.getFaultCode());
        }

        // Act
        try {
            classUnderTest.updateTreatmentWill(treatmentWillVO, requestContext, true);
            fail();
        } catch (ServiceException e) {
            // Assert
            assertEquals(FaultCode.UPDATE_TREATMENT_WILL_INTERNAL_SERVER_ERROR, e.getFaultCode());
        }

        // Act
        try {
            classUnderTest.deleteTreatmentWill(requestContext);
            fail();
        } catch (ServiceException e) {
            // Assert
            assertEquals(FaultCode.DELETE_TREATMENT_WILL_INTERNAL_SERVER_ERROR, e.getFaultCode());
        }
    }

    @Test
    public void citizenGetTreatmentWill_success() throws DAOException, MinLogException, NasException {
    	getTreatmentWill_success(getServiceActorExample(), 1);
    }
    @Test
    public void adminGetTreatmentWill_success() throws DAOException, MinLogException, NasException {
    	getTreatmentWill_success(getServiceActorAdminExample(), 1);
    }
    @Test
    public void systemGetTreatmentWill_success() throws DAOException, MinLogException, NasException {
    	getTreatmentWill_success(getServiceActorSystemUserExample(), 0);    	
    }
    @Test
    public void healthCareProfGetTreatmentWill_success() throws DAOException, MinLogException, NasException {
    	getTreatmentWill_success(getServiceActorHealthCareProfessionalExample(), 1);
    }
    
    public void getTreatmentWill_success(ServiceActor serviceActor,  int minlogExpectedCalls) throws DAOException, MinLogException, NasException {
        // Arrange
        RequestContextVO requestContext = getRequestContextExample(serviceActor, MESSAGE_ID);
        TreatmentWillVO treatmentWillVO = new TreatmentWillVO();

        // Stub DAO method to return the TreatmentWillVO
        when(treatmentWillDAO.getCurrentTreatmentWill(anyString(), any(Instant.class)))
                .thenReturn(treatmentWillVO);

        // Stub Properties to return a string
        String minLogReadActivityTextBTR = "test";

        // Act
        TreatmentWillVO treatmentWillReturned = classUnderTest.getTreatmentWill(requestContext, minLogReadActivityTextBTR);

        // Assert
        verify(accessControl).checkGetAccess(eq(requestContext));
        verify(treatmentWillDAO).getCurrentTreatmentWill(eq(PERSON_IDENTIFIER), eq(requestContext.getTime()));
        verify(treatmentWillValidator).validateTreatmentWillExists(eq(treatmentWillVO));
        verify(minLogDataWrapper, times(minlogExpectedCalls)).logActivity(eq(requestContext), eq(minLogReadActivityTextBTR), anyString(), anyString(), anyString(), any(), eq(requestContext.getPersonIdentifier()), isNull(), eq(TreatmentWillServiceImpl.SERVICE_IDENTIFIER), eq(serviceActor.getUserTitle()));
        verify(notificationSender, times(0)).sendTreatmentWillNotification(any(), any());

        assertEquals(treatmentWillVO, treatmentWillReturned);
    }

    @Test
    public void getTreatmentWill_failOnMinLogException() throws DAOException, MinLogException {
        // Arrange
        ServiceActor serviceActor = getServiceActorHealthCareProfessionalExample();
        RequestContextVO requestContext = getRequestContextExample(serviceActor, MESSAGE_ID);
        TreatmentWillVO treatmentWillVO = new TreatmentWillVO();

        // Stub DAO method to return the TreatmentWillVO
        when(treatmentWillDAO.getCurrentTreatmentWill(anyString(), any(Instant.class)))
                .thenReturn(treatmentWillVO);

        // Stub MinLog to throw an Exception
        doThrow(MinLogException.class).when(minLogDataWrapper)
                .logActivity(
                        any(RequestContextVO.class),
                        nullable(String.class),
                        nullable(String.class),
                        nullable(String.class),
                        nullable(String.class),
                        nullable(OrganisationIdentifierFormat.class),
                        nullable(String.class),
                        nullable(String.class),
                        nullable(String.class),
                        nullable(String.class)
                );

        // Stub Properties to return a string
        String minLogReadActivityTextBTR = "test";

        // Act
        try {
            classUnderTest.getTreatmentWill(requestContext, minLogReadActivityTextBTR);
            fail();
        } catch (ServiceException e) {
            verify(accessControl).checkGetAccess(eq(requestContext));
            verify(treatmentWillDAO).getCurrentTreatmentWill(eq(PERSON_IDENTIFIER), eq(requestContext.getTime()));
            verify(treatmentWillValidator).validateTreatmentWillExists(eq(treatmentWillVO));
            verify(minLogDataWrapper).logActivity(eq(requestContext), eq(minLogReadActivityTextBTR), anyString(), anyString(), anyString(), any(), eq(requestContext.getPersonIdentifier()), isNull(), eq(TreatmentWillServiceImpl.SERVICE_IDENTIFIER), eq(serviceActor.getUserTitle()));

            assertEquals(FaultCode.GET_TREATMENT_WILL_INTERNAL_SERVER_ERROR, e.getFaultCode());
        }
    }

    @Test
    public void hasTreatmentWill_true() throws DAOException {
        // Arrange
        ServiceActor serviceActor = getServiceActorExample();
        RequestContextVO requestContext = getRequestContextExample(serviceActor, MESSAGE_ID);
        TreatmentWillVO treatmentWillVO = new TreatmentWillVO();

        // Stub DAO method to return the TreatmentWillVO
        when(treatmentWillDAO.getCurrentTreatmentWill(anyString(), any(Instant.class)))
                .thenReturn(treatmentWillVO);

        // Act
        boolean hasTreatmentWill = classUnderTest.hasTreatmentWill(requestContext);

        // Assert
        verify(accessControl).checkHasAccess(eq(requestContext));
        verify(treatmentWillDAO).getCurrentTreatmentWill(eq(PERSON_IDENTIFIER), eq(requestContext.getTime()));

        assertTrue(hasTreatmentWill);
    }

    @Test
    public void hasTreatmentWill_false() throws DAOException {
        // Arrange
        ServiceActor serviceActor = getServiceActorExample();
        RequestContextVO requestContext = getRequestContextExample(serviceActor, MESSAGE_ID);

        // Stub DAO method to return null
        when(treatmentWillDAO.getCurrentTreatmentWill(anyString(), any(Instant.class)))
                .thenReturn(null);

        // Act
        boolean hasTreatmentWill = classUnderTest.hasTreatmentWill(requestContext);

        // Assert
        verify(accessControl).checkHasAccess(eq(requestContext));
        verify(treatmentWillDAO).getCurrentTreatmentWill(eq(PERSON_IDENTIFIER), eq(requestContext.getTime()));

        assertFalse(hasTreatmentWill);
    }

    @Test
    public void createTreatmentWill_success() throws DAOException, NasException {
        // Arrange
        TreatmentWillVO treatmentWillVO = getTreatmentWillExample();
        ServiceActor serviceActorExample = getServiceActorExample();
        RequestContextVO requestContext = getRequestContextExample(serviceActorExample, MESSAGE_ID);

        // Stub DAO method to return null (although the mock would do that anyway)
        when(livingWillDAO.getCurrentLivingWill(anyString(), any(Instant.class))).thenReturn(null);
        when(treatmentWillDAO.getCurrentTreatmentWill(anyString(), any(Instant.class))).thenReturn(null);

        // Act
        classUnderTest.createTreatmentWill(treatmentWillVO, requestContext, true, false);

        // Assert
        verify(accessControl).checkCreateAccess(eq(requestContext));
        verify(livingWillDAO).getCurrentLivingWill(eq(PERSON_IDENTIFIER), eq(requestContext.getTime()));
        verify(treatmentWillDAO).getCurrentTreatmentWill(eq(PERSON_IDENTIFIER), eq(requestContext.getTime()));
        verify(livingWillValidator).validateLivingWillNotExists(isNull());

        verify(treatmentWillValidator).validateTreatmentWillNotExists(isNull());
        verify(treatmentWillDAO).createTreatmentWillSkipNoLifeProlongingIfDyingAcceptanceNeeded(eq(treatmentWillVO));
        verify(notificationSender).sendTreatmentWillNotification(serviceActorExample.getPersonIdentifier(), requestContext.getTime());
        assertEquals(treatmentWillVO.getPersonIdentifier(), treatmentWillVO.getPersonIdentifier());
        assertEquals(TreatmentWillServiceImpl.INITIAL_VERSION_DEFAULT, treatmentWillVO.getVersion().intValue());
        assertEquals(requestContext.getTime(), treatmentWillVO.getValidFrom());
        assertEquals(TreatmentWillServiceImpl.VALID_TO_DEFAULT, treatmentWillVO.getValidTo());
        assertEquals(requestContext.getTime(), treatmentWillVO.getModifiedDateTime());
    }

    @Test
    public void adminCreateTreatmentWill_deleteLivingWill_success() throws DAOException, NasException {
        // Arrange
        TreatmentWillVO treatmentWillVO = getTreatmentWillExample();
        ServiceActor serviceActorExample = getServiceActorAdminExample();
        RequestContextVO requestContext = getRequestContextExample(serviceActorExample, MESSAGE_ID);

        // Stub DAO method to return null (although the mock would do that anyway)
        when(livingWillDAO.getCurrentLivingWill(anyString(), any(Instant.class))).thenReturn(null);
        when(treatmentWillDAO.getCurrentTreatmentWill(anyString(), any(Instant.class))).thenReturn(null);

        // Act
        classUnderTest.createTreatmentWill(treatmentWillVO, requestContext, true, true);

        // Assert
        verify(accessControl).checkUpgradeAccess(eq(requestContext));
        verify(livingWillDAO).getCurrentLivingWill(eq(PERSON_IDENTIFIER), eq(requestContext.getTime()));
        verify(treatmentWillDAO).getCurrentTreatmentWill(eq(PERSON_IDENTIFIER), eq(requestContext.getTime()));
        verify(livingWillValidator).validateLivingWillNotExists(isNull());
        verify(treatmentWillValidator).validateTreatmentWillNotExists(isNull());
        verify(treatmentWillDAO).createTreatmentWillSkipNoLifeProlongingIfDyingAcceptanceNeeded(eq(treatmentWillVO));
        verify(notificationSender).sendTreatmentWillNotification(serviceActorExample.getPersonIdentifier(), requestContext.getTime());
        assertEquals(treatmentWillVO.getPersonIdentifier(), treatmentWillVO.getPersonIdentifier());
        assertEquals(TreatmentWillServiceImpl.INITIAL_VERSION_DEFAULT, treatmentWillVO.getVersion().intValue());
        assertEquals(requestContext.getTime(), treatmentWillVO.getValidFrom());
        assertEquals(TreatmentWillServiceImpl.VALID_TO_DEFAULT, treatmentWillVO.getValidTo());
        assertEquals(requestContext.getTime(), treatmentWillVO.getModifiedDateTime());
    }

    @Test
    public void citizenCreateTreatmentWill_minLogCalled_success() throws DAOException, MinLogException, NasException {
    	createTreatmentWill_minLogCalled_success(getServiceActorExample(), 1);
    }
    @Test
    public void adminCreateTreatmentWill_minLogCalled_success() throws DAOException, MinLogException, NasException {
    	createTreatmentWill_minLogCalled_success(getServiceActorAdminExample(), 1);
    }
    @Test
    public void systemCreateTreatmentWill_minLogCalled_success() throws DAOException, MinLogException, NasException {
    	createTreatmentWill_minLogCalled_success(getServiceActorSystemUserExample(), 0);
    }
    @Test
    public void healthCareProfCreateTreatmentWill_minLogCalled_success() throws DAOException, MinLogException, NasException {
    	createTreatmentWill_minLogCalled_success(getServiceActorHealthCareProfessionalExample(), 1);
    }
    
    private void createTreatmentWill_minLogCalled_success(ServiceActor serviceActor,  int minlogExpectedCalls) throws DAOException, MinLogException, NasException {
        // Arrange
        TreatmentWillVO treatmentWillVO = getTreatmentWillExample();
        RequestContextVO requestContext = getRequestContextExample(serviceActor, MESSAGE_ID);

        // Stub DAO method to return null (although the mock would do that anyway)
        when(livingWillDAO.getCurrentLivingWill(anyString(), any(Instant.class))).thenReturn(null);
        when(treatmentWillDAO.getCurrentTreatmentWill(anyString(), any(Instant.class))).thenReturn(null);
        String minLogReadActivityTextBTR = "test";
        when(properties.getMinLogCreateActivityTextBTR()).thenReturn(minLogReadActivityTextBTR);

        // Act
        classUnderTest.createTreatmentWill(treatmentWillVO, requestContext, true, false);

        // Assert
        verify(minLogDataWrapper, times(minlogExpectedCalls)).logActivity(eq(requestContext), eq(minLogReadActivityTextBTR), anyString(), anyString(), anyString(), any(), eq(requestContext.getPersonIdentifier()), isNull(), eq(TreatmentWillServiceImpl.SERVICE_IDENTIFIER), eq(serviceActor.getUserTitle()));
        verify(notificationSender).sendTreatmentWillNotification(requestContext.getPersonIdentifier(), requestContext.getTime());
    }

    @Test
    public void updateTreatmentWill_success() throws DAOException, NasException {
        // Arrange
        long pid = 1L;
        int version = 1;

        TreatmentWillVO treatmentWillVO = getTreatmentWillExample();
        ServiceActor serviceActorExample = getServiceActorExample();
        RequestContextVO requestContext = getRequestContextExample(serviceActorExample, MESSAGE_ID);

        // Stub DAO method to return an object with a PID and Version
        TreatmentWillVO treatmentWillVOFake = new TreatmentWillVO();
        treatmentWillVOFake.setTreatmentWillPID(pid);
        treatmentWillVOFake.setVersion(version);
        when(treatmentWillDAO.getCurrentTreatmentWill(anyString(), any(Instant.class)))
                .thenReturn(treatmentWillVOFake);

        // Act
        classUnderTest.updateTreatmentWill(treatmentWillVO,requestContext, true);

        // Assert
        verify(accessControl).checkUpdateAccess(eq(requestContext));
        verify(treatmentWillDAO).getCurrentTreatmentWill(eq(PERSON_IDENTIFIER), eq(requestContext.getTime()));
        verify(treatmentWillValidator).validateTreatmentWillExists(eq(treatmentWillVOFake));
        verify(treatmentWillDAO).updateValidToTimestamp(eq(pid), eq(requestContext.getTime()));
        verify(treatmentWillDAO).createTreatmentWillSkipNoLifeProlongingIfDyingAcceptanceNeeded(eq(treatmentWillVO));
        verify(notificationSender).sendTreatmentWillNotification(serviceActorExample.getPersonIdentifier(), requestContext.getTime());

        assertEquals(treatmentWillVO.getPersonIdentifier(), treatmentWillVO.getPersonIdentifier());
        assertEquals(TreatmentWillServiceImpl.INITIAL_VERSION_DEFAULT + 1, treatmentWillVO.getVersion().intValue());
        assertEquals(requestContext.getTime(), treatmentWillVO.getValidFrom());
        assertEquals(TreatmentWillServiceImpl.VALID_TO_DEFAULT, treatmentWillVO.getValidTo());
        assertEquals(requestContext.getTime(), treatmentWillVO.getModifiedDateTime());
    }

    @Test
    public void citizenUpdateTreatmentWill_minLogCalled_success() throws DAOException, MinLogException, NasException {
    	updateTreatmentWill_minLogCalled_success(getServiceActorExample(), 1);
    }
    @Test
    public void adminUpdateTreatmentWill_minLogCalled_success() throws DAOException, MinLogException, NasException {
    	updateTreatmentWill_minLogCalled_success(getServiceActorAdminExample(), 1);
    }
    @Test
    public void systemUpdateTreatmentWill_minLogCalled_success() throws DAOException, MinLogException, NasException {
    	updateTreatmentWill_minLogCalled_success(getServiceActorSystemUserExample(), 0);
    }
    @Test
    public void healthCareProfUpdateTreatmentWill_minLogCalled_success() throws DAOException, MinLogException, NasException {
    	updateTreatmentWill_minLogCalled_success(getServiceActorHealthCareProfessionalExample(), 1);
    }
    
    private void updateTreatmentWill_minLogCalled_success(ServiceActor serviceActor,  int minlogExpectedCalls) throws DAOException, MinLogException, NasException {
        // Arrange
        long pid = 1L;
        int version = 1;

        TreatmentWillVO treatmentWillVO = getTreatmentWillExample();
        RequestContextVO requestContext = getRequestContextExample(serviceActor, MESSAGE_ID);

        // Stub DAO method to return an object with a PID and Version
        TreatmentWillVO treatmentWillVOFake = new TreatmentWillVO();
        treatmentWillVOFake.setTreatmentWillPID(pid);
        treatmentWillVOFake.setVersion(version);
        when(treatmentWillDAO.getCurrentTreatmentWill(anyString(), any(Instant.class)))
                .thenReturn(treatmentWillVOFake);
        String minLogReadActivityTextBTR = "test";
        when(properties.getMinLogUpdateActivityTextBTR()).thenReturn(minLogReadActivityTextBTR);

        // Act
        classUnderTest.updateTreatmentWill(treatmentWillVO,requestContext, true);

        // Assert
        verify(minLogDataWrapper, times(minlogExpectedCalls)).logActivity(eq(requestContext), eq(minLogReadActivityTextBTR), anyString(), anyString(), anyString(), any(), eq(requestContext.getPersonIdentifier()), isNull(), eq(TreatmentWillServiceImpl.SERVICE_IDENTIFIER), eq(serviceActor.getUserTitle()));
        verify(notificationSender).sendTreatmentWillNotification(requestContext.getPersonIdentifier(), requestContext.getTime());
    }

    @Test
    public void deleteTreatmentWill_success() throws DAOException, NasException {
        // Arrange
        long pid = 1L;
        int version = 1;

        ServiceActor serviceActorExample = getServiceActorExample();
        RequestContextVO requestContext = getRequestContextExample(serviceActorExample, MESSAGE_ID);

        // Stub DAO method to return an object with a PID and Version
        TreatmentWillVO treatmentWillVOFake = new TreatmentWillVO();
        treatmentWillVOFake.setTreatmentWillPID(pid);
        treatmentWillVOFake.setVersion(version);
        when(treatmentWillDAO.getCurrentTreatmentWill(anyString(), any(Instant.class)))
                .thenReturn(treatmentWillVOFake);

        // Act
        classUnderTest.deleteTreatmentWill(requestContext);

        // Assert
        verify(accessControl).checkDeleteAccess(eq(requestContext));

        ArgumentCaptor<Long> updateValidToArgument1 = ArgumentCaptor.forClass(Long.class);
        ArgumentCaptor<Instant> updateValidToArgument2 = ArgumentCaptor.forClass(Instant.class);
        verify(treatmentWillDAO).updateValidToTimestamp(updateValidToArgument1.capture(), updateValidToArgument2.capture());
        verify(notificationSender).sendTreatmentWillNotification(serviceActorExample.getPersonIdentifier(), requestContext.getTime());

        assertEquals(pid, updateValidToArgument1.getValue().longValue());
        assertNotNull(updateValidToArgument2.getValue());
    }

    @Test
    public void citizenDeleteTreatmentWill_minLogCalled_success() throws DAOException, MinLogException, NasException {
    	deleteTreatmentWill_minLogCalled_success(getServiceActorExample(), 1);
    }
    @Test
    public void adminDeleteTreatmentWill_minLogCalled_success() throws DAOException, MinLogException, NasException {
    	deleteTreatmentWill_minLogCalled_success(getServiceActorAdminExample(), 1);
    }
    @Test
    public void systemDeleteTreatmentWill_minLogCalled_success() throws DAOException, MinLogException, NasException {
    	deleteTreatmentWill_minLogCalled_success(getServiceActorSystemUserExample(), 0);
    }
    @Test
    public void healthCareProfDeleteTreatmentWill_minLogCalled_success() throws DAOException, MinLogException, NasException {
    	deleteTreatmentWill_minLogCalled_success(getServiceActorHealthCareProfessionalExample(), 1);
    }
    
    public void deleteTreatmentWill_minLogCalled_success(ServiceActor serviceActor,  int minlogExpectedCalls) throws DAOException, MinLogException, NasException {
        // Arrange
        long pid = 1L;
        int version = 1;

        RequestContextVO requestContext = getRequestContextExample(serviceActor, MESSAGE_ID);

        // Stub DAO method to return an object with a PID and Version
        TreatmentWillVO treatmentWillVOFake = new TreatmentWillVO();
        treatmentWillVOFake.setTreatmentWillPID(pid);
        treatmentWillVOFake.setVersion(version);
        when(treatmentWillDAO.getCurrentTreatmentWill(anyString(), any(Instant.class)))
                .thenReturn(treatmentWillVOFake);
        String minLogReadActivityTextBTR = "test";
        when(properties.getMinLogDeleteActivityTextBTR()).thenReturn(minLogReadActivityTextBTR);

        // Act
        classUnderTest.deleteTreatmentWill(requestContext);

        // Assert
        verify(minLogDataWrapper,  times(minlogExpectedCalls)).logActivity(eq(requestContext), eq(minLogReadActivityTextBTR), anyString(), anyString(), anyString(), any(), eq(requestContext.getPersonIdentifier()), isNull(), eq(TreatmentWillServiceImpl.SERVICE_IDENTIFIER), eq(serviceActor.getUserTitle()));
        verify(notificationSender).sendTreatmentWillNotification(requestContext.getPersonIdentifier(), requestContext.getTime());
    }
}
