/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.dao;


import dk.sundhedsdatastyrelsen.livstestamente.dao.LivingWillDAO;
import dk.sundhedsdatastyrelsen.livstestamente.domain.LivingWillVO;
import dk.sundhedsdatastyrelsen.livstestamente.service.AbstractSpringTest;
import org.junit.Assert;
import org.junit.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.transaction.annotation.Transactional;

import java.time.Instant;
import java.time.Year;
import java.time.ZoneOffset;
import java.time.temporal.ChronoUnit;
import java.util.LinkedList;

@Transactional("btrDataSourceTxManager")
public class LivingWillDAOImplTests extends AbstractSpringTest {

    private final Instant now = Instant.now();

    private final Integer version = 1;
    private final Instant validFrom = now;
    private final Instant validTo = Year.of(8099).atDay(1).atStartOfDay(ZoneOffset.UTC).toInstant();
    private final String personIdentifier = "12345678";
    private final Instant modifiedDateTime = now;
    private final boolean noLifeProlongingIfDying = true;
    private final boolean noLifeProlongingIfSeverelyDegraded = true;

    @Autowired
    private LivingWillDAO livingWillDAO;

    @Test
    public void createAndGetCurrentLivingWill() throws Exception {
        // Arrange
        final LivingWillVO livingWillVO = createTestLivingWillVO();

        // Act
        final long pid = livingWillDAO.createLivingWill(livingWillVO);

        // Assert
        final LivingWillVO actualLivingWillVO = livingWillDAO.getCurrentLivingWill(personIdentifier, livingWillVO.getValidFrom());
        Assert.assertNotNull(actualLivingWillVO);
        Assert.assertTrue(pid >= 0);
        Assert.assertEquals(personIdentifier, actualLivingWillVO.getPersonIdentifier());
        Assert.assertEquals(version, actualLivingWillVO.getVersion());
        Assert.assertEquals(validFrom.truncatedTo(ChronoUnit.SECONDS), actualLivingWillVO.getValidFrom());
        Assert.assertEquals(validTo.truncatedTo(ChronoUnit.SECONDS), actualLivingWillVO.getValidTo());
        Assert.assertEquals(modifiedDateTime.truncatedTo(ChronoUnit.SECONDS), actualLivingWillVO.getModifiedDateTime());
        Assert.assertEquals(noLifeProlongingIfDying, actualLivingWillVO.isNoLifeProlongingIfDying());
        Assert.assertEquals(noLifeProlongingIfSeverelyDegraded, actualLivingWillVO.isNoLifeProlongingIfSeverelyDegraded());
    }

    @Test
    public void updateValidToTimeStamp() throws Exception {
        // Arrange
        final LivingWillVO livingWillVO = createTestLivingWillVO();
        final long pid = livingWillDAO.createLivingWill(livingWillVO);
        final Instant newValidTo = livingWillVO.getValidFrom().plusSeconds(1); // Add 1 second

        // Act
        livingWillDAO.updateValidToTimestamp(pid, newValidTo);

        // Assert
        final LivingWillVO actualLivingWillVO = livingWillDAO.getCurrentLivingWill(personIdentifier, livingWillVO.getValidFrom());
        Assert.assertNotNull(actualLivingWillVO);
        Assert.assertEquals(newValidTo.truncatedTo(ChronoUnit.SECONDS), actualLivingWillVO.getValidTo());
    }

    @Test
    public void GetWithLimit() throws Exception {
        // Arrange
        final LivingWillVO livingWillVO1 = createTestLivingWillVO();
        livingWillVO1.setPersonIdentifier("0501792277");
        livingWillDAO.createLivingWill(livingWillVO1);
        final LivingWillVO livingWillVO2 = createTestLivingWillVO();
        livingWillVO2.setPersonIdentifier("0501792278");
        livingWillDAO.createLivingWill(livingWillVO2);
        final LivingWillVO livingWillVO3 = createTestLivingWillVO();
        livingWillVO3.setPersonIdentifier("0501792279");
        livingWillDAO.createLivingWill(livingWillVO3);
        final LivingWillVO livingWillVO4 = createTestLivingWillVO();
        livingWillVO4.setPersonIdentifier("0309691444");
        livingWillDAO.createLivingWill(livingWillVO4);

        // Act
        LinkedList<String> actualLivingWill1 = livingWillDAO.getAllPersonsFromLTR(1);
        LinkedList<String> actualLivingWill2 = livingWillDAO.getAllPersonsFromLTR(2);
        LinkedList<String> actualLivingWill3 = livingWillDAO.getAllPersonsFromLTR(3);
        LinkedList<String> actualLivingWillMoreThenThereIsInDb = livingWillDAO.getAllPersonsFromLTR(999);

        // Assert
        Assert.assertNotNull(actualLivingWill1);
        Assert.assertEquals(actualLivingWill1.size(), 1);

        Assert.assertNotNull(actualLivingWill2);
        Assert.assertEquals(actualLivingWill2.size(), 2);

        Assert.assertNotNull(actualLivingWill3);
        Assert.assertEquals(actualLivingWill3.size(), 3);

        Assert.assertNotNull(actualLivingWillMoreThenThereIsInDb);
        Assert.assertEquals(actualLivingWillMoreThenThereIsInDb.size(), 4);
    }

    @Test
    public void GetWithPrefix() throws Exception {
        // Arrange
        final LivingWillVO livingWillVO1 = createTestLivingWillVO();
        livingWillVO1.setPersonIdentifier("0501792277");
        livingWillDAO.createLivingWill(livingWillVO1);
        final LivingWillVO livingWillVO2 = createTestLivingWillVO();
        livingWillVO2.setPersonIdentifier("0501792278");
        livingWillDAO.createLivingWill(livingWillVO2);
        final LivingWillVO livingWillVO3 = createTestLivingWillVO();
        livingWillVO3.setPersonIdentifier("0501792279");
        livingWillDAO.createLivingWill(livingWillVO3);
        final LivingWillVO livingWillVO4 = createTestLivingWillVO();
        livingWillVO4.setPersonIdentifier("0309691444");
        livingWillDAO.createLivingWill(livingWillVO4);

        // Act
        LinkedList<String> actualLivingWill1 = livingWillDAO.getAllPersonsFromLTRWithPrefix("05");
        LinkedList<String> actualLivingWill2 = livingWillDAO.getAllPersonsFromLTRWithPrefix("03");
        LinkedList<String> actualLivingWill3 = livingWillDAO.getAllPersonsFromLTRWithPrefix("99");

        // Assert
        Assert.assertNotNull(actualLivingWill1);
        Assert.assertEquals(actualLivingWill1.size(), 3);
        Assert.assertTrue(actualLivingWill1.contains("0501792277"));
        Assert.assertTrue(actualLivingWill1.contains("0501792278"));
        Assert.assertTrue(actualLivingWill1.contains("0501792279"));

        Assert.assertNotNull(actualLivingWill2);
        Assert.assertEquals(actualLivingWill2.size(), 1);
        Assert.assertTrue(actualLivingWill2.contains("0309691444"));

        Assert.assertNotNull(actualLivingWill3);
        Assert.assertEquals(actualLivingWill3.size(), 0);
    }

    private LivingWillVO createTestLivingWillVO() {
        final LivingWillVO livingWillVO = new LivingWillVO();
        livingWillVO.setPersonIdentifier(personIdentifier);
        livingWillVO.setNoLifeProlongingIfDying(noLifeProlongingIfDying);
        livingWillVO.setNoLifeProlongingIfSeverelyDegraded(noLifeProlongingIfSeverelyDegraded);
        livingWillVO.setVersion(version);
        livingWillVO.setModifiedDateTime(modifiedDateTime);
        livingWillVO.setValidFrom(validFrom);
        livingWillVO.setValidTo(validTo);
        return livingWillVO;
    }
}
