/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.dao;

import dk.sundhedsdatastyrelsen.livstestamente.dao.TreatmentWillDAO;
import dk.sundhedsdatastyrelsen.livstestamente.domain.AcceptanceNeeded;
import dk.sundhedsdatastyrelsen.livstestamente.domain.TreatmentWillVO;
import dk.sundhedsdatastyrelsen.livstestamente.service.AbstractSpringTest;
import org.junit.Assert;
import org.junit.Ignore;
import org.junit.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.transaction.annotation.Transactional;

import java.time.Instant;
import java.time.Year;
import java.time.ZoneOffset;
import java.time.temporal.ChronoUnit;
import java.util.LinkedList;

@Transactional("btrDataSourceTxManager")
public class TreatmentWillDAOImplTests extends AbstractSpringTest {

    private final Instant now = Instant.now();

    private final Integer version = 1;
    private final Instant validFrom = now;
    private final Instant validTo = Year.of(8099).atDay(1).atStartOfDay(ZoneOffset.UTC).toInstant();
    private final String personIdentifier = "12345678";
    private final Instant modifiedDateTime = now;
    private final boolean noLifeProlongingIfDying = true;
    private final AcceptanceNeeded noLifeProlongingIfDyingAcceptanceNeeded = AcceptanceNeeded.RELATIVE_ACCEPTANCE_REQUIRED;
    private final boolean noLifeProlongingIfSeverelyDegraded = true;
    private final AcceptanceNeeded noLifeProlongingIfSeverelyDegradedAcceptanceNeeded = AcceptanceNeeded.RELATIVE_ACCEPTANCE_REQUIRED;
    private final boolean noLifeProlongingIfSeverePain = true;
    private final AcceptanceNeeded noLifeProlongingIfSeverePainAcceptanceNeeded = AcceptanceNeeded.RELATIVE_ACCEPTANCE_REQUIRED;
    private final boolean noForcedTreatmentIfIncapable = true;
    private final AcceptanceNeeded noForcedTreatmentIfIncapableAcceptanceNeeded = AcceptanceNeeded.RELATIVE_ACCEPTANCE_REQUIRED;

    @Autowired
    private TreatmentWillDAO treatmentWillDAO;

    @Test
    public void createAndGetCurrentTreatmentWill() throws Exception {
        // Arrange
        final TreatmentWillVO treatmentWillVO = createTestTreatmentWillVO();

        // Act
        final long pid = treatmentWillDAO.createTreatmentWill(treatmentWillVO);

        // Assert
        final TreatmentWillVO actualTreatmentWillVO = treatmentWillDAO.getCurrentTreatmentWill(personIdentifier, treatmentWillVO.getValidFrom());
        Assert.assertNotNull(actualTreatmentWillVO);
        Assert.assertTrue(pid >= 0);
        Assert.assertEquals(personIdentifier, actualTreatmentWillVO.getPersonIdentifier());
        Assert.assertEquals(version, actualTreatmentWillVO.getVersion());
        Assert.assertEquals(validFrom.truncatedTo(ChronoUnit.SECONDS), actualTreatmentWillVO.getValidFrom());
        Assert.assertEquals(validTo.truncatedTo(ChronoUnit.SECONDS), actualTreatmentWillVO.getValidTo());
        Assert.assertEquals(modifiedDateTime.truncatedTo(ChronoUnit.SECONDS), actualTreatmentWillVO.getModifiedDateTime());

        Assert.assertEquals(noLifeProlongingIfDying, actualTreatmentWillVO.isNoLifeProlongingIfDying());
        Assert.assertEquals(noLifeProlongingIfDyingAcceptanceNeeded, actualTreatmentWillVO.getNoLifeProlongingIfDyingAcceptanceNeeded());

        Assert.assertEquals(noLifeProlongingIfSeverelyDegraded, actualTreatmentWillVO.isNoLifeProlongingIfSeverelyDegraded());
        Assert.assertEquals(noLifeProlongingIfSeverelyDegradedAcceptanceNeeded, actualTreatmentWillVO.getNoLifeProlongingIfSeverelyDegradedAcceptanceNeeded());

        Assert.assertEquals(noLifeProlongingIfSeverePain, actualTreatmentWillVO.isNoLifeProlongingIfSeverePain());
        Assert.assertEquals(noLifeProlongingIfSeverePainAcceptanceNeeded, actualTreatmentWillVO.getNoLifeProlongingIfSeverePainAcceptanceNeeded());

        Assert.assertEquals(noForcedTreatmentIfIncapable, actualTreatmentWillVO.isNoForcedTreatmentIfIncapable());
        Assert.assertEquals(noForcedTreatmentIfIncapableAcceptanceNeeded, actualTreatmentWillVO.getNoForcedTreatmentIfIncapableAcceptanceNeeded());
    }

    @Test
    public void createAndGetCurrentTreatmentWill_null() throws Exception {
        // Arrange
        final TreatmentWillVO treatmentWillVO = createTestTreatmentWillVO();
        treatmentWillVO.setNoLifeProlongingIfDyingAcceptanceNeeded(null); // <--- is null
        treatmentWillVO.setNoLifeProlongingIfSeverelyDegradedAcceptanceNeeded(null); // <--- is null
        treatmentWillVO.setNoLifeProlongingIfSeverePainAcceptanceNeeded(null); // <--- is null
        treatmentWillVO.setNoForcedTreatmentIfIncapableAcceptanceNeeded(null); // <--- is null

        // Act
        final long pid = treatmentWillDAO.createTreatmentWill(treatmentWillVO);

        // Assert
        final TreatmentWillVO actualTreatmentWillVO = treatmentWillDAO.getCurrentTreatmentWill(personIdentifier, treatmentWillVO.getValidFrom());
        Assert.assertNotNull(actualTreatmentWillVO);
        Assert.assertTrue(pid >= 0);
        Assert.assertNull(actualTreatmentWillVO.getNoLifeProlongingIfDyingAcceptanceNeeded()); // <--- is null
        Assert.assertNull(actualTreatmentWillVO.getNoLifeProlongingIfSeverelyDegradedAcceptanceNeeded()); // <--- is null
        Assert.assertNull(actualTreatmentWillVO.getNoLifeProlongingIfSeverePainAcceptanceNeeded()); // <--- is null
        Assert.assertNull(actualTreatmentWillVO.getNoForcedTreatmentIfIncapableAcceptanceNeeded()); // <--- is null
    }

    @Ignore // disabled until a fix is implemented
    @Test
    public void createAndGetCurrentTreatmentWill_winterTime() throws Exception {
        // Arrange
        // Create a record a record with validFrom 10 minutes after the change to winter time (in Denmark)
        final Instant afterWinterTimeChange = Instant.parse("2018-10-28T00:10:00.000Z");

        final TreatmentWillVO treatmentWillVO = createTestTreatmentWillVO();
        treatmentWillVO.setValidFrom(afterWinterTimeChange);

        // Act
        treatmentWillDAO.createTreatmentWill(treatmentWillVO);

        // Assert
        // We should be able to find that record 55 minutes later
        final TreatmentWillVO actualTreatmentWillVO = treatmentWillDAO.getCurrentTreatmentWill(
                personIdentifier,
                afterWinterTimeChange.plus(55, ChronoUnit.MINUTES));
        Assert.assertNotNull(actualTreatmentWillVO);
        Assert.assertEquals(afterWinterTimeChange, actualTreatmentWillVO.getValidFrom());
    }

    @Test
    public void createAndGetCurrentTreatmentWill_summerTime() throws Exception {
        // Arrange
        // Create a record a record with validFrom 10 minutes after the change to summer time (in Denmark)
        final Instant afterSummerTimeChange = Instant.parse("2018-03-25T02:10:00.000Z");
        final TreatmentWillVO treatmentWillVO = createTestTreatmentWillVO();
        treatmentWillVO.setValidFrom(afterSummerTimeChange);

        // Act
        treatmentWillDAO.createTreatmentWill(treatmentWillVO);

        // Assert
        // We should be able to find that record 55 minutes later
        final TreatmentWillVO actualTreatmentWillVO = treatmentWillDAO.getCurrentTreatmentWill(
                personIdentifier,
                afterSummerTimeChange.plus(55, ChronoUnit.MINUTES));
        Assert.assertNotNull(actualTreatmentWillVO);
        Assert.assertEquals(afterSummerTimeChange, actualTreatmentWillVO.getValidFrom());
    }

    @Test
    public void
    updateValidToTimeStamp() throws Exception {
        // Arrange
        final TreatmentWillVO treatmentWillVO = createTestTreatmentWillVO();
        final long pid = treatmentWillDAO.createTreatmentWill(treatmentWillVO);
        final Instant newValidTo = treatmentWillVO.getValidFrom().plusSeconds(1); // Add 1 second

        // Act
        treatmentWillDAO.updateValidToTimestamp(pid, newValidTo);

        // Assert
        final TreatmentWillVO actualTreatmentWillVO = treatmentWillDAO.getCurrentTreatmentWill(personIdentifier, treatmentWillVO.getValidFrom());
        Assert.assertNotNull(actualTreatmentWillVO);
        Assert.assertEquals(newValidTo.truncatedTo(ChronoUnit.SECONDS), actualTreatmentWillVO.getValidTo());
    }

    @Test
    public void createAndGetCurrentTreatmentWillSkipNoLifeProlongingIfDyingAcceptanceNeeded() throws Exception {

        // Arrange
        final TreatmentWillVO treatmentWillVO = createTestTreatmentWillVO();

        // Act
        final long pid = treatmentWillDAO.createTreatmentWillSkipNoLifeProlongingIfDyingAcceptanceNeeded(treatmentWillVO);

        // Assert
        final TreatmentWillVO actualTreatmentWillVO = treatmentWillDAO.getCurrentTreatmentWill(personIdentifier, treatmentWillVO.getValidFrom());
        Assert.assertNotNull(actualTreatmentWillVO);
        Assert.assertTrue(pid >= 0);
        Assert.assertEquals(personIdentifier, actualTreatmentWillVO.getPersonIdentifier());
        Assert.assertEquals(version, actualTreatmentWillVO.getVersion());
        Assert.assertEquals(validFrom.truncatedTo(ChronoUnit.SECONDS), actualTreatmentWillVO.getValidFrom());
        Assert.assertEquals(validTo.truncatedTo(ChronoUnit.SECONDS), actualTreatmentWillVO.getValidTo());
        Assert.assertEquals(modifiedDateTime.truncatedTo(ChronoUnit.SECONDS), actualTreatmentWillVO.getModifiedDateTime());

        Assert.assertEquals(noLifeProlongingIfDying, actualTreatmentWillVO.isNoLifeProlongingIfDying());
        Assert.assertEquals(null, actualTreatmentWillVO.getNoLifeProlongingIfDyingAcceptanceNeeded());

        Assert.assertEquals(noLifeProlongingIfSeverelyDegraded, actualTreatmentWillVO.isNoLifeProlongingIfSeverelyDegraded());
        Assert.assertEquals(noLifeProlongingIfSeverelyDegradedAcceptanceNeeded, actualTreatmentWillVO.getNoLifeProlongingIfSeverelyDegradedAcceptanceNeeded());

        Assert.assertEquals(noLifeProlongingIfSeverePain, actualTreatmentWillVO.isNoLifeProlongingIfSeverePain());
        Assert.assertEquals(noLifeProlongingIfSeverePainAcceptanceNeeded, actualTreatmentWillVO.getNoLifeProlongingIfSeverePainAcceptanceNeeded());

        Assert.assertEquals(noForcedTreatmentIfIncapable, actualTreatmentWillVO.isNoForcedTreatmentIfIncapable());
        Assert.assertEquals(noForcedTreatmentIfIncapableAcceptanceNeeded, actualTreatmentWillVO.getNoForcedTreatmentIfIncapableAcceptanceNeeded());
    }

    @Test
    public void GetWithLimit() throws Exception {
        // Arrange
        final TreatmentWillVO treatmentWillVO1 = createTestTreatmentWillVO();
        treatmentWillVO1.setPersonIdentifier("0501792277");
        treatmentWillDAO.createTreatmentWill(treatmentWillVO1);
        final TreatmentWillVO treatmentWillVO2 = createTestTreatmentWillVO();
        treatmentWillVO2.setPersonIdentifier("0501792278");
        treatmentWillDAO.createTreatmentWill(treatmentWillVO2);
        final TreatmentWillVO treatmentWillVO3 = createTestTreatmentWillVO();
        treatmentWillVO3.setPersonIdentifier("0501792279");
        treatmentWillDAO.createTreatmentWill(treatmentWillVO3);
        final TreatmentWillVO treatmentWillVO4 = createTestTreatmentWillVO();
        treatmentWillVO4.setPersonIdentifier("0309691444");
        treatmentWillDAO.createTreatmentWill(treatmentWillVO4);

        // Act
        LinkedList<String> actualTreatmentWill1 = treatmentWillDAO.getAllPersonsFromBTR(1);
        LinkedList<String> actualTreatmentWill2 = treatmentWillDAO.getAllPersonsFromBTR(2);
        LinkedList<String> actualTreatmentWill3 = treatmentWillDAO.getAllPersonsFromBTR(3);
        LinkedList<String> actualTreatmentWillMoreThenThereIsInDb = treatmentWillDAO.getAllPersonsFromBTR(999);

        // Assert
        Assert.assertNotNull(actualTreatmentWill1);
        Assert.assertEquals(actualTreatmentWill1.size(), 1);

        Assert.assertNotNull(actualTreatmentWill2);
        Assert.assertEquals(actualTreatmentWill2.size(), 2);

        Assert.assertNotNull(actualTreatmentWill3);
        Assert.assertEquals(actualTreatmentWill3.size(), 3);

        Assert.assertNotNull(actualTreatmentWillMoreThenThereIsInDb);
        Assert.assertEquals(actualTreatmentWillMoreThenThereIsInDb.size(), 4);
    }

    @Test
    public void GetWithPrefix() throws Exception {
        // Arrange
        final TreatmentWillVO treatmentWillVO1 = createTestTreatmentWillVO();
        treatmentWillVO1.setPersonIdentifier("0501792277");
        treatmentWillDAO.createTreatmentWill(treatmentWillVO1);
        final TreatmentWillVO treatmentWillVO2 = createTestTreatmentWillVO();
        treatmentWillVO2.setPersonIdentifier("0501792278");
        treatmentWillDAO.createTreatmentWill(treatmentWillVO2);
        final TreatmentWillVO treatmentWillVO3 = createTestTreatmentWillVO();
        treatmentWillVO3.setPersonIdentifier("0501792279");
        treatmentWillDAO.createTreatmentWill(treatmentWillVO3);
        final TreatmentWillVO treatmentWillVO4 = createTestTreatmentWillVO();
        treatmentWillVO4.setPersonIdentifier("0309691444");
        treatmentWillDAO.createTreatmentWill(treatmentWillVO4);

        // Act
        LinkedList<String> actualTreatmentWill1 = treatmentWillDAO.getAllPersonsFromLTRWithPrefix("05");
        LinkedList<String> actualTreatmentWill2 = treatmentWillDAO.getAllPersonsFromLTRWithPrefix("03");
        LinkedList<String> actualTreatmentWill3 = treatmentWillDAO.getAllPersonsFromLTRWithPrefix("99");

        // Assert
        Assert.assertNotNull(actualTreatmentWill1);
        Assert.assertEquals(actualTreatmentWill1.size(), 3);
        Assert.assertTrue(actualTreatmentWill1.contains("0501792277"));
        Assert.assertTrue(actualTreatmentWill1.contains("0501792278"));
        Assert.assertTrue(actualTreatmentWill1.contains("0501792279"));

        Assert.assertNotNull(actualTreatmentWill2);
        Assert.assertEquals(actualTreatmentWill2.size(), 1);
        Assert.assertTrue(actualTreatmentWill2.contains("0309691444"));

        Assert.assertNotNull(actualTreatmentWill3);
        Assert.assertEquals(actualTreatmentWill3.size(), 0);
    }

    private TreatmentWillVO createTestTreatmentWillVO() {
        final TreatmentWillVO treatmentWillVO = new TreatmentWillVO();
        treatmentWillVO.setPersonIdentifier(personIdentifier);
        treatmentWillVO.setNoLifeProlongingIfDying(noLifeProlongingIfDying);
        treatmentWillVO.setNoLifeProlongingIfDyingAcceptanceNeeded(noLifeProlongingIfDyingAcceptanceNeeded);
        treatmentWillVO.setNoLifeProlongingIfSeverelyDegraded(noLifeProlongingIfSeverelyDegraded);
        treatmentWillVO.setNoLifeProlongingIfSeverelyDegradedAcceptanceNeeded(noLifeProlongingIfSeverelyDegradedAcceptanceNeeded);
        treatmentWillVO.setNoLifeProlongingIfSeverePain(noLifeProlongingIfSeverePain);
        treatmentWillVO.setNoLifeProlongingIfSeverePainAcceptanceNeeded(noLifeProlongingIfSeverePainAcceptanceNeeded);
        treatmentWillVO.setNoForcedTreatmentIfIncapable(noForcedTreatmentIfIncapable);
        treatmentWillVO.setNoForcedTreatmentIfIncapableAcceptanceNeeded(noForcedTreatmentIfIncapableAcceptanceNeeded);
        treatmentWillVO.setVersion(version);
        treatmentWillVO.setModifiedDateTime(modifiedDateTime);
        treatmentWillVO.setValidFrom(validFrom);
        treatmentWillVO.setValidTo(validTo);
        return treatmentWillVO;
    }

}
