/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.webservice;

import dk.sdsd.nsp.slalog.api.SLALogItem;
import dk.sundhedsdatastyrelsen.livstestamente._2018._05._01.*;
import dk.sundhedsdatastyrelsen.livstestamente.domain.LivingWillVO;
import dk.sundhedsdatastyrelsen.livstestamente.domain.RequestContextVO;
import dk.sundhedsdatastyrelsen.livstestamente.domain.UserType;
import dk.sundhedsdatastyrelsen.livstestamente.auditlog.AuditLogger;
import dk.sundhedsdatastyrelsen.livstestamente.service.business.LivingWillService;
import dk.sundhedsdatastyrelsen.livstestamente.util.ServiceException;
import dk.sundhedsdatastyrelsen.livstestamente.service.util.TestHelper;
import dk.sundhedsdatastyrelsen.livstestamente.exceptions.ValidationException;
import dk.sundhedsdatastyrelsen.livstestamente.service.webservice.hl7cda.LivingWillHL7CDAMapper;
import dk.sundhedsdatastyrelsen.livstestamente.integrations.personInformation.PersonInformation;
import dk.sundhedsdatastyrelsen.livstestamente.service.context.RequestContextProvider;
import dk.sundhedsdatastyrelsen.livstestamente.context.ServiceActor;
import org.hl7.ltr.LivingWill;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.junit.MockitoJUnitRunner;

import javax.servlet.ServletRequest;
import javax.xml.ws.WebServiceContext;
import javax.xml.ws.handler.MessageContext;

import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertThrows;
import static org.mockito.Mockito.*;

@RunWith(MockitoJUnitRunner.class)
public class LivingWillWebServiceMockedDependenciesTests {

    @Mock
    protected WebServiceContext context;

    @Mock
    private RequestContextProvider requestContextProvider;

    @Mock
    private LivingWillService businessService;

    @Mock
    private LivingWillHL7CDAMapper livingWillHL7CDAMapper;

    @Mock
    private PersonInformation personInformation;

    @Mock
    private AuditLogger auditLogger;

    @InjectMocks
    private LivingWillWebService webService;

    private TestHelper testHelper;
    private RequestContextVO requestContext;

    @Before
    public void setup() {
        MessageContext mc = mock(MessageContext.class);
        when(mc.get(MessageContext.SERVLET_REQUEST)).thenReturn(mock(ServletRequest.class));
        when(context.getMessageContext()).thenReturn(mc);
        testHelper = new TestHelper();

        ServiceActor serviceActor = new ServiceActor(UserType.Citizen);
        requestContext = new RequestContextVO(serviceActor, mock(SLALogItem.class));
        requestContext.setPersonIdentifier("1111111111");
        when(requestContextProvider.getRequestContext(any(), anyBoolean(), any()))
                .thenReturn(requestContext);
    }

    @Test
    public void hasLivingWill_cprDoesNotExist_falseIsReturned() throws ValidationException {
        doThrow(new ValidationException("Failing validateCpr")).when(personInformation).validateCpr(anyString());
        HasLivingWillRequest request = createHasLivingWillRequest(requestContext.getPersonIdentifier());

        HasLivingWillResponse response = webService.hasLivingWill20180501(request);

        assertFalse(response.isWillExists());
    }

    @Test
    public void hasLivingWill_ageTooLow_falseIsReturned() throws ValidationException {
        doThrow(new ValidationException("Failing validateAge")).when(personInformation).validateAge(anyString());
        HasLivingWillRequest request = createHasLivingWillRequest(requestContext.getPersonIdentifier());

        HasLivingWillResponse response = webService.hasLivingWill20180501(request);

        assertFalse(response.isWillExists());
    }

    @Test
    public void updateLivingWill_WhenNothingFails_AuditLogs() {
        UpdateLivingWillRequest request = createUpdateLivingWillRequest(requestContext.getPersonIdentifier());
        LivingWillVO livingWill = testHelper.getLivingWillExample();
        when(livingWillHL7CDAMapper.createLivingWillFromRequest(requestContext.getPersonIdentifier(), request.getLivingWill()))
                .thenReturn(livingWill);

        webService.updateLivingWill20180501(request);

        verify(auditLogger).logUpdateLivingWill(livingWill, requestContext.getServiceActor());
    }

    @Test
    public void updateLivingWill_WhenOperationFails_ThrowsServiceExceptionAndDoesNotAuditLog() {
        UpdateLivingWillRequest request = createUpdateLivingWillRequest(requestContext.getPersonIdentifier());
        doThrow(RuntimeException.class).when(businessService).updateLivingWill(any(), any());

        assertThrows(ServiceException.class, () -> webService.updateLivingWill20180501(request));

        verifyZeroInteractions(auditLogger);
    }

    @Test
    public void deleteLivingWill_WhenNothingFails_AuditLogs() {
        DeleteLivingWillRequest request = createDeleteLivingWillRequest(requestContext.getPersonIdentifier());

        webService.deleteLivingWill20180501(request);

        verify(auditLogger).logDeleteLivingWill(requestContext.getPersonIdentifier(), requestContext.getServiceActor());
    }

    @Test
    public void deleteLivingWill_WhenOperationFails_ThrowsServiceExceptionAndDoesNotAuditLog() {
        DeleteLivingWillRequest request = createDeleteLivingWillRequest(requestContext.getPersonIdentifier());
        doThrow(RuntimeException.class).when(businessService).deleteLivingWill(any(), any());

        assertThrows(ServiceException.class, () -> webService.deleteLivingWill20180501(request));

        verifyZeroInteractions(auditLogger);
    }

    @Test
    public void getLivingWill_WhenNothingFails_AuditLogs() {
        GetLivingWillRequest request = createGetLivingWillRequest(requestContext.getPersonIdentifier());
        LivingWillVO livingWill = testHelper.getLivingWillExample();
        when(businessService.getLivingWill(requestContext)).thenReturn(livingWill);

        webService.getLivingWill20180501(request);

        verify(auditLogger).logGetLivingWill(livingWill, requestContext.getServiceActor());
    }

    @Test
    public void getLivingWill_WhenOperationFails_ThrowsServiceExceptionAndDoesNotAuditLog() {
        GetLivingWillRequest request = createGetLivingWillRequest(requestContext.getPersonIdentifier());
        doThrow(RuntimeException.class).when(businessService).getLivingWill(any());

        assertThrows(ServiceException.class, () -> webService.getLivingWill20180501(request));

        verifyZeroInteractions(auditLogger);
    }

    @Test
    public void hasLivingWill_WhenNothingFails_AuditLogs() {
        HasLivingWillRequest request = createHasLivingWillRequest(requestContext.getPersonIdentifier());

        webService.hasLivingWill20180501(request);

        verify(auditLogger).logHasLivingWill(requestContext.getPersonIdentifier(), requestContext.getServiceActor());
    }

    @Test
    public void hasLivingWill_WhenOperationFails_ThrowsServiceExceptionAndDoesNotAuditLog() {
        HasLivingWillRequest request = createHasLivingWillRequest(requestContext.getPersonIdentifier());
        doThrow(RuntimeException.class).when(businessService).hasLivingWill(any());

        assertThrows(ServiceException.class, () -> webService.hasLivingWill20180501(request));

        verifyZeroInteractions(auditLogger);
    }

    private UpdateLivingWillRequest createUpdateLivingWillRequest(String cpr) {
        UpdateLivingWillRequest request = new UpdateLivingWillRequest();
        request.setId(testHelper.getId(cpr));
        request.setLivingWill(new LivingWill());
        return request;
    }

    private DeleteLivingWillRequest createDeleteLivingWillRequest(String cpr) {
        DeleteLivingWillRequest request = new DeleteLivingWillRequest();
        request.setId(testHelper.getId(cpr));
        return request;
    }

    private GetLivingWillRequest createGetLivingWillRequest(String cpr) {
        GetLivingWillRequest request = new GetLivingWillRequest();
        request.setId(testHelper.getId(cpr));
        return request;
    }

    private HasLivingWillRequest createHasLivingWillRequest(String cpr) {
        HasLivingWillRequest request = new HasLivingWillRequest();
        request.setId(testHelper.getId(cpr));
        return request;
    }
}
