/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.webservice;

import static dk.sundhedsdatastyrelsen.livstestamente.service.util.TestHelperV2.*;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.util.List;

import dk.sundhedsdatastyrelsen.livstestamente.domain.LivingWillVO;
import dk.sundhedsdatastyrelsen.livstestamente.domain.RequestContextVO;
import org.hl7.ltr.LivingWill;
import org.hl7.v3.ANY;
import org.hl7.v3.POCDMT000040Entry;
import org.junit.Test;
import org.mockito.Mockito;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.transaction.annotation.Transactional;

import dk.sundhedsdatastyrelsen.livstestamente.OID;
import dk.sundhedsdatastyrelsen.livstestamente._2018._05._01.DeleteLivingWillRequest;
import dk.sundhedsdatastyrelsen.livstestamente._2018._05._01.DeleteLivingWillResponse;
import dk.sundhedsdatastyrelsen.livstestamente._2018._05._01.GetLivingWillRequest;
import dk.sundhedsdatastyrelsen.livstestamente._2018._05._01.GetLivingWillResponse;
import dk.sundhedsdatastyrelsen.livstestamente._2018._05._01.HasLivingWillRequest;
import dk.sundhedsdatastyrelsen.livstestamente._2018._05._01.HasLivingWillResponse;
import dk.sundhedsdatastyrelsen.livstestamente._2018._05._01.ObjectFactory;
import dk.sundhedsdatastyrelsen.livstestamente._2018._05._01.UpdateLivingWillRequest;
import dk.sundhedsdatastyrelsen.livstestamente._2018._05._01.UpdateLivingWillResponse;
import dk.sundhedsdatastyrelsen.livstestamente.service.business.LivingWillServiceImpl;
import dk.sundhedsdatastyrelsen.livstestamente.util.FaultCode;
import dk.sundhedsdatastyrelsen.livstestamente.util.ServiceException;
import dk.sundhedsdatastyrelsen.livstestamente.service.util.TestHelperV2;
import dk.sundhedsdatastyrelsen.livstestamente.service.webservice.hl7cda.LivingWillHL7CDAMapper;
import dk.sundhedsdatastyrelsen.livstestamente.service.webservice.hl7cda.LivingWillHL7CDAValidator;
import dk.sundhedsdatastyrelsen.livstestamente.context.ServiceActor;
import dk.sundhedsdatastyrelsen.livstestamente.context.ServiceActorProvider;

@Transactional("btrDataSourceTxManager")
public class LivingWillWebServiceTests extends AbstractWebServiceTest {

	@Autowired
	@Qualifier("livingWillWebService")
	private LivingWillWebService serviceUnderTest;

	@Autowired
	@Qualifier("livingWillAdminWebService")
	private LivingWillWebService adminServiceUnderTest;

	@Autowired
	private LivingWillServiceImpl businessService;

	@Autowired
	private LivingWillHL7CDAMapper livingWillHL7CDAMapper;

	@Autowired
	private LivingWillHL7CDAValidator validator;

	@Autowired
    private ServiceActorProvider serviceActorProvider;

	private static final ObjectFactory factory = new ObjectFactory();

	@Test
	public void getLivingWill_testResponseStructure() {
		// Arrange
		ServiceActor serviceActorExample = testHelperV2.getServiceActorWithExistingLivingWillExample();

		Mockito.when(serviceActorProvider.getServiceActor(Mockito.any(), Mockito.anyBoolean())).thenReturn(serviceActorExample);

		// Act
		final GetLivingWillRequest request = createGetLivingWillRequest(PERSON_IDENTIFIER_2);
		GetLivingWillResponse response = serviceUnderTest.getLivingWill20180501(request);

		// Assert
		assertNotNull(response);
		final List<POCDMT000040Entry> allEntries = getAllEntries(response);

		// Should be 1 entry element
		assertEquals(1, allEntries.size());

		// Should be 1 value element
		List<ANY> values = allEntries.get(0).getObservation().getValues();
		assertEquals(1, values.size());

		// Cast the value element to a LivingWill-type
		LivingWill livingWillElement = (LivingWill) values.get(0);

		// Test values of LivingWill-element
		assertEquals(NO_LIFE_PROLONGING_IF_DYING, livingWillElement.isNoLifeProlongingIfDying());
		assertEquals(NO_LIFE_PROLONGING_IF_SEVERELY_DEGRADED, livingWillElement.isNoLifeProlongingIfSeverelyDegraded());
	}

	@Test
	public void adminGetLivingWill_testResponseStructure() {
		// Arrange
		ServiceActor serviceActorExample = testHelperV2.getServiceActorWithExistingLivingWillExample();

		Mockito.when(serviceActorProvider.getServiceActor(Mockito.any(), Mockito.anyBoolean())).thenReturn(serviceActorExample);

		// Act
		final GetLivingWillRequest request = createGetLivingWillRequest(PERSON_IDENTIFIER_2);
		GetLivingWillResponse response = serviceUnderTest.getLivingWill20180501(request);

		// Assert
		assertNotNull(response);
		final List<POCDMT000040Entry> allEntries = getAllEntries(response);

		// Should be 1 entry element
		assertEquals(1, allEntries.size());

		// Should be 1 value element
		List<ANY> values = allEntries.get(0).getObservation().getValues();
		assertEquals(1, values.size());

		// Cast the value element to a LivingWill-type
		LivingWill livingWillElement = (LivingWill) values.get(0);

		// Test values of LivingWill-element
		assertEquals(NO_LIFE_PROLONGING_IF_DYING, livingWillElement.isNoLifeProlongingIfDying());
		assertEquals(NO_LIFE_PROLONGING_IF_SEVERELY_DEGRADED, livingWillElement.isNoLifeProlongingIfSeverelyDegraded());
	}

	@Test
	public void getLivingWill_validateResponseStructure() {
		// Arrange
		ServiceActor serviceActorExample = testHelperV2.getServiceActorWithExistingLivingWillExample();

		Mockito.when(serviceActorProvider.getServiceActor(Mockito.any(), Mockito.anyBoolean())).thenReturn(serviceActorExample);

		// Act
		final GetLivingWillRequest request = createGetLivingWillRequest(PERSON_IDENTIFIER_2);
		GetLivingWillResponse response = serviceUnderTest.getLivingWill20180501(request);

		// Assert
		assertNotNull(response);

		validator.validateWithoutThrow(response);

		final int errorCount = validator.getValidationErrors().size();
		assertEquals(errorCount + " validation errors were found: " + validator.getCombinedErrorText(), 0, errorCount);
	}

	@Test
	public void hasLivingWill_testResponseStructure() {
		// Arrange
		ServiceActor serviceActorExample = testHelperV2.getServiceActorWithExistingLivingWillExample();

		Mockito.when(serviceActorProvider.getServiceActor(Mockito.any(), Mockito.anyBoolean())).thenReturn(serviceActorExample);

		// Act
		final HasLivingWillRequest request = createHasLivingWillRequest(PERSON_IDENTIFIER_2);
		HasLivingWillResponse response = serviceUnderTest.hasLivingWill20180501(request);

		// Assert
		assertNotNull(response);
		assertTrue(response.isWillExists());
	}

	@Test
	public void updateLivingWill_testRequestMapping() {
		// Arrange
		final boolean noLifeProlongingIfDyingUpdated = false;
		final boolean noLifeProlongingIfSeverelyDegradedUpdated = true;

		LivingWill livingWill = getLivingWillExample();
		livingWill.setNoLifeProlongingIfDying(noLifeProlongingIfDyingUpdated);
		livingWill.setNoLifeProlongingIfSeverelyDegraded(noLifeProlongingIfSeverelyDegradedUpdated);

		ServiceActor serviceActorExample = testHelperV2.getServiceActorWithExistingLivingWillExample();
		RequestContextVO requestContextExample = testHelperV2.getRequestContextExample(serviceActorExample, MESSAGE_ID);

		Mockito.when(serviceActorProvider.getServiceActor(Mockito.any(), Mockito.anyBoolean())).thenReturn(requestContextExample.getServiceActor());

		// Act
		final UpdateLivingWillRequest request = createUpdateLivingWillRequest(PERSON_IDENTIFIER_2, livingWill);
		UpdateLivingWillResponse response = serviceUnderTest.updateLivingWill20180501(request);

		// Assert using business layer
		assertNotNull(response);
		final LivingWillVO livingWillVO = testHelperV2.getLivingWill(businessService, PERSON_IDENTIFIER_2);
		assertNotNull(livingWillVO);
		assertNotNull(livingWillVO.getLivingWillPID());
		assertEquals(2, livingWillVO.getVersion().intValue());
		assertTrue(livingWillVO.getValidFrom().isBefore(livingWillVO.getValidTo()));
		assertEquals(noLifeProlongingIfDyingUpdated, livingWillVO.isNoLifeProlongingIfDying());
		assertEquals(noLifeProlongingIfSeverelyDegradedUpdated, livingWillVO.isNoLifeProlongingIfSeverelyDegraded());
	}

	@Test
	public void adminUpdateLivingWill_testRequestMapping() {
		// Arrange
		final boolean noLifeProlongingIfDyingUpdated = false;
		final boolean noLifeProlongingIfSeverelyDegradedUpdated = true;

		LivingWill livingWill = getLivingWillExample();
		livingWill.setNoLifeProlongingIfDying(noLifeProlongingIfDyingUpdated);
		livingWill.setNoLifeProlongingIfSeverelyDegraded(noLifeProlongingIfSeverelyDegradedUpdated);

		RequestContextVO requestContextExample = getSecurityDataForAdminUser(PERSON_IDENTIFIER_3, ADMIN_PERSON_IDENTIFIER);

		Mockito.when(serviceActorProvider.getServiceActor(Mockito.any(), Mockito.anyBoolean())).thenReturn(requestContextExample.getServiceActor());

		// Act
		final UpdateLivingWillRequest request = createUpdateLivingWillRequest(PERSON_IDENTIFIER_3, livingWill);
		UpdateLivingWillResponse response = adminServiceUnderTest.updateLivingWill20180501(request);

		// Assert using business layer
		assertNotNull(response);
		final LivingWillVO livingWillVO = testHelperV2.getLivingWill(businessService, PERSON_IDENTIFIER_3);
		assertNotNull(livingWillVO);
		assertNotNull(livingWillVO.getLivingWillPID());
		assertEquals(2, livingWillVO.getVersion().intValue());
		assertTrue(livingWillVO.getValidFrom().isBefore(livingWillVO.getValidTo()));
		assertEquals(noLifeProlongingIfDyingUpdated, livingWillVO.isNoLifeProlongingIfDying());
		assertEquals(noLifeProlongingIfSeverelyDegradedUpdated, livingWillVO.isNoLifeProlongingIfSeverelyDegraded());
	}

	@Test
	public void deleteLivingWill_testRequestMapping() {
		// Arrange
		RequestContextVO requestContextExample = getSecurityDataForDataOwner(PERSON_IDENTIFIER_3);

		Mockito.when(serviceActorProvider.getServiceActor(Mockito.any(), Mockito.anyBoolean())).thenReturn(requestContextExample.getServiceActor());

		// Act
		final DeleteLivingWillRequest request = createDeleteLivingWillRequest(PERSON_IDENTIFIER_3);
		DeleteLivingWillResponse response = serviceUnderTest.deleteLivingWill20180501(request);

		// Assert using business layer
		// No data should be found
		assertNotNull(response);
		try {
			testHelperV2.getLivingWill(businessService, PERSON_IDENTIFIER_3);
			fail();
		} catch (ServiceException e) {
			assertEquals(FaultCode.LIVING_WILL_NO_DATA_FOUND, e.getFaultCode());
		}
	}

	@Test
	public void adminDeleteLivingWill_testRequestMapping() {
		// Arrange
		RequestContextVO adminUser = getSecurityDataForAdminUser(PERSON_IDENTIFIER_2, ADMIN_PERSON_IDENTIFIER);
		Mockito.when(serviceActorProvider.getServiceActor(Mockito.any(), Mockito.anyBoolean())).thenReturn(adminUser.getServiceActor());

		// Act
		final DeleteLivingWillRequest request = createDeleteLivingWillRequest(PERSON_IDENTIFIER_2);
		DeleteLivingWillResponse response = adminServiceUnderTest.deleteLivingWill20180501(request);

		// Assert using business layer
		// No data should be found
		assertNotNull(response);
		try {
			testHelperV2.getLivingWill(businessService, PERSON_IDENTIFIER_2);
			fail();
		} catch (ServiceException e) {
			assertEquals(FaultCode.LIVING_WILL_NO_DATA_FOUND, e.getFaultCode());
		}
	}

	private UpdateLivingWillRequest createUpdateLivingWillRequest(String personIdentifier, LivingWill livingWill) {
		final UpdateLivingWillRequest request = factory.createUpdateLivingWillRequest();
		request.setId(livingWillHL7CDAMapper.createId(OID.CPR, personIdentifier, TestHelperV2.ASSIGNING_AUTHORITY_NAME_PERSON));
		request.setLivingWill(livingWill);
		return request;
	}

	private DeleteLivingWillRequest createDeleteLivingWillRequest(String personIdentifier) {
		DeleteLivingWillRequest request = factory.createDeleteLivingWillRequest();
		request.setId(livingWillHL7CDAMapper.createId(OID.CPR, personIdentifier, TestHelperV2.ASSIGNING_AUTHORITY_NAME_PERSON));
		return request;
	}

	private GetLivingWillRequest createGetLivingWillRequest(String personIdentifier) {
		GetLivingWillRequest request = factory.createGetLivingWillRequest();
		request.setId(livingWillHL7CDAMapper.createId(OID.CPR, personIdentifier, TestHelperV2.ASSIGNING_AUTHORITY_NAME_PERSON));
		return request;
	}

	private HasLivingWillRequest createHasLivingWillRequest(String personIdentifier) {
		HasLivingWillRequest request = factory.createHasLivingWillRequest();
		request.setId(livingWillHL7CDAMapper.createId(OID.CPR, personIdentifier, TestHelperV2.ASSIGNING_AUTHORITY_NAME_PERSON));
		return request;
	}

	private LivingWill getLivingWillExample() {
		final LivingWill livingWill = LivingWillHL7CDAMapper.oLTR.createLivingWill();
		livingWill.setNoLifeProlongingIfDying(NO_LIFE_PROLONGING_IF_DYING);
		livingWill.setNoLifeProlongingIfSeverelyDegraded(NO_LIFE_PROLONGING_IF_SEVERELY_DEGRADED);
		return livingWill;
	}

	private List<POCDMT000040Entry> getAllEntries(GetLivingWillResponse response) {
		return response.getClinicalDocument().getComponent().getStructuredBody().getComponents().get(0).getSection().getEntries();
	}
}
