/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.webservice;

import dk.sdsd.nsp.slalog.api.SLALogItem;
import dk.sundhedsdatastyrelsen.behandlingstestamente._2018._05._01.*;
import dk.sundhedsdatastyrelsen.livstestamente.auditlog.AuditLogger;
import dk.sundhedsdatastyrelsen.livstestamente.service.business.TreatmentWillService;
import dk.sundhedsdatastyrelsen.livstestamente.domain.RequestContextVO;
import dk.sundhedsdatastyrelsen.livstestamente.domain.TreatmentWillVO;
import dk.sundhedsdatastyrelsen.livstestamente.domain.UserType;
import dk.sundhedsdatastyrelsen.livstestamente.conf.Properties;
import dk.sundhedsdatastyrelsen.livstestamente.util.ServiceException;
import dk.sundhedsdatastyrelsen.livstestamente.service.util.TestHelper;
import dk.sundhedsdatastyrelsen.livstestamente.exceptions.ValidationException;
import dk.sundhedsdatastyrelsen.livstestamente.service.webservice.hl7cda.TreatmentWillHL7CDAMapper;
import dk.sundhedsdatastyrelsen.livstestamente.integrations.personInformation.PersonInformation;
import dk.sundhedsdatastyrelsen.livstestamente.service.context.RequestContextProvider;
import dk.sundhedsdatastyrelsen.livstestamente.context.ServiceActor;
import org.hl7.btr.TreatmentWill;
import org.hl7.btr.TreatmentWillValueType;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.junit.MockitoJUnitRunner;

import javax.servlet.ServletRequest;
import javax.xml.ws.WebServiceContext;
import javax.xml.ws.handler.MessageContext;

import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertThrows;
import static org.mockito.Mockito.*;

@RunWith(MockitoJUnitRunner.class)
public class TreatmentWillWebServiceMockedDependenciesTests {

    @Mock
    public RequestContextProvider requestContextProvider;

    @Mock
    private TreatmentWillService businessService;

    @Mock
    private TreatmentWillHL7CDAMapper treatmentWillHL7CDAMapper;

    @Mock
    private PersonInformation personInformation;

    @Mock
    private Properties properties;

    @Mock
    WebServiceContext context;

    @Mock
    AuditLogger auditLogger;

    @InjectMocks
    TreatmentWillWebService webService;

    private TestHelper testHelper;
    private RequestContextVO requestContext;

    @Before
    public void setup() {
        MessageContext mc = mock(MessageContext.class);
        when(mc.get(MessageContext.SERVLET_REQUEST)).thenReturn(mock(ServletRequest.class));
        when(context.getMessageContext()).thenReturn(mc);
        testHelper = new TestHelper();

        ServiceActor serviceActor = new ServiceActor(UserType.Citizen);
        requestContext = new RequestContextVO(serviceActor, mock(SLALogItem.class));
        requestContext.setPersonIdentifier("1111111111");
        when(requestContextProvider.getRequestContext(any(), anyBoolean(), any()))
                .thenReturn(requestContext);
    }

    @Test
    public void hasTreatmentWill_cprDoesNotExist_falseIsReturned() throws ValidationException {
        doThrow(new ValidationException("Failing validateCpr")).when(personInformation).validateCpr(anyString());
        HasTreatmentWillRequest request = createHasTreatmentWillRequest(requestContext.getPersonIdentifier());

        HasTreatmentWillResponse response = webService.hasTreatmentWill20180501(request);

        assertFalse(response.isWillExists());
    }

    @Test
    public void hasTreatmentWill_ageTooLow_falseIsReturned() throws ValidationException {
        doThrow(new ValidationException("Failing validateAge")).when(personInformation).validateAge(anyString());
        HasTreatmentWillRequest request = createHasTreatmentWillRequest(requestContext.getPersonIdentifier());

        HasTreatmentWillResponse response = webService.hasTreatmentWill20180501(request);

        assertFalse(response.isWillExists());
    }

    @Test
    public void getTreatmentWill_WhenNothingFails_AuditLogs() {
        GetTreatmentWillRequest request = createGetTreatmentWillRequest(requestContext.getPersonIdentifier());
        TreatmentWillVO treatmentWill = testHelper.getTreatmentWillExample();
        when(businessService.getTreatmentWill(eq(requestContext), any())).thenReturn(treatmentWill);

        webService.getTreatmentWill20180501(request);

        verify(auditLogger).logGetTreatmentWill(treatmentWill, requestContext.getServiceActor());
    }

    @Test
    public void getTreatmentWill_WhenOperationFails_ThrowsServiceExceptionAndDoesNotAuditLog() {
        GetTreatmentWillRequest request = createGetTreatmentWillRequest(requestContext.getPersonIdentifier());
        doThrow(RuntimeException.class).when(businessService).getTreatmentWill(any(), any());

        assertThrows(ServiceException.class, () -> webService.getTreatmentWill20180501(request));

        verifyZeroInteractions(auditLogger);
    }

    @Test
    public void hasTreatmentWill_WhenNothingFails_AuditLogs() {
        HasTreatmentWillRequest request = createHasTreatmentWillRequest(requestContext.getPersonIdentifier());

        webService.hasTreatmentWill20180501(request);

        verify(auditLogger).logHasTreatmentWill(requestContext.getPersonIdentifier(), requestContext.getServiceActor());
    }

    @Test
    public void hasTreatmentWill_WhenOperationFails_ThrowsServiceExceptionAndDoesNotAuditLog() {
        HasTreatmentWillRequest request = createHasTreatmentWillRequest(requestContext.getPersonIdentifier());
        doThrow(RuntimeException.class).when(businessService).hasTreatmentWill(any());

        assertThrows(ServiceException.class, () -> webService.hasTreatmentWill20180501(request));

        verifyZeroInteractions(auditLogger);
    }

    @Test
    public void createTreatmentWill_WhenNothingFails_AuditLogs() throws javax.xml.bind.ValidationException {
        CreateTreatmentWillRequest request = createCreateTreatmentWillRequest(requestContext.getPersonIdentifier());
        TreatmentWillVO treatmentWill = testHelper.getTreatmentWillExample();
        when(treatmentWillHL7CDAMapper.fromTreatmentWill(requestContext.getPersonIdentifier(), request.getTreatmentWill())).thenReturn(treatmentWill);

        webService.createTreatmentWill20180501(request);

        verify(auditLogger).logCreateTreatmentWill(treatmentWill, requestContext.getServiceActor());
    }

    @Test
    public void createTreatmentWill_WhenOperationFails_ThrowsServiceExceptionAndDoesNotAuditLog() {
        CreateTreatmentWillRequest request = createCreateTreatmentWillRequest(requestContext.getPersonIdentifier());
        doThrow(RuntimeException.class).when(businessService).createTreatmentWill(any(), any(), anyBoolean(), anyBoolean());

        assertThrows(ServiceException.class, () -> webService.createTreatmentWill20180501(request));

        verifyZeroInteractions(auditLogger);
    }

    @Test
    public void updateTreatmentWill_WhenNothingFails_AuditLogs() throws javax.xml.bind.ValidationException {
        UpdateTreatmentWillRequest request = createUpdateTreatmentWillRequest(requestContext.getPersonIdentifier());
        TreatmentWillVO treatmentWill = testHelper.getTreatmentWillExample();
        when(treatmentWillHL7CDAMapper.fromTreatmentWill(requestContext.getPersonIdentifier(), request.getTreatmentWill())).thenReturn(treatmentWill);

        webService.updateTreatmentWill20180501(request);

        verify(auditLogger).logUpdateTreatmentWill(treatmentWill, requestContext.getServiceActor());
    }

    @Test
    public void updateTreatmentWill_WhenOperationFails_ThrowsServiceExceptionAndDoesNotAuditLog() {
        UpdateTreatmentWillRequest request = createUpdateTreatmentWillRequest(requestContext.getPersonIdentifier());
        doThrow(RuntimeException.class).when(businessService).updateTreatmentWill(any(), any(), anyBoolean());

        assertThrows(ServiceException.class, () -> webService.updateTreatmentWill20180501(request));

        verifyZeroInteractions(auditLogger);
    }

    @Test
    public void deleteTreatmentWill_WhenNothingFails_AuditLogs() {
        DeleteTreatmentWillRequest request = createDeleteTreatmentWillRequest(requestContext.getPersonIdentifier());

        webService.deleteTreatmentWill20180501(request);

        verify(auditLogger).logDeleteTreatmentWill(requestContext.getPersonIdentifier(), requestContext.getServiceActor());
    }

    @Test
    public void deleteTreatmentWill_WhenOperationFails_ThrowsServiceExceptionAndDoesNotAuditLog() {
        DeleteTreatmentWillRequest request = createDeleteTreatmentWillRequest(requestContext.getPersonIdentifier());
        doThrow(RuntimeException.class).when(businessService).deleteTreatmentWill(any());

        assertThrows(ServiceException.class, () -> webService.deleteTreatmentWill20180501(request));

        verifyZeroInteractions(auditLogger);
    }

    private GetTreatmentWillRequest createGetTreatmentWillRequest(String cpr) {
        GetTreatmentWillRequest request = new GetTreatmentWillRequest();
        request.setId(testHelper.getId(cpr));
        return request;
    }

    private HasTreatmentWillRequest createHasTreatmentWillRequest(String cpr) {
        HasTreatmentWillRequest request = new HasTreatmentWillRequest();
        request.setId(testHelper.getId(cpr));
        return request;
    }

    private CreateTreatmentWillRequest createCreateTreatmentWillRequest(String cpr) {
        CreateTreatmentWillRequest request = new CreateTreatmentWillRequest();
        request.setId(testHelper.getId(cpr));
        request.setTreatmentWill(createTreatmentWill());
        return request;
    }

    private UpdateTreatmentWillRequest createUpdateTreatmentWillRequest(String cpr) {
        UpdateTreatmentWillRequest request = new UpdateTreatmentWillRequest();
        request.setId(testHelper.getId(cpr));
        request.setTreatmentWill(createTreatmentWill());
        return request;
    }

    private DeleteTreatmentWillRequest createDeleteTreatmentWillRequest(String cpr) {
        DeleteTreatmentWillRequest request = new DeleteTreatmentWillRequest();
        request.setId(testHelper.getId(cpr));
        return request;
    }

    private static TreatmentWill createTreatmentWill() {
        TreatmentWill treatmentWill = new TreatmentWill();
        treatmentWill.setNoLifeProlongingIfDying(new TreatmentWillValueType());
        treatmentWill.setNoForcedTreatmentIfIncapable(new TreatmentWillValueType());
        treatmentWill.setNoLifeProlongingIfSeverelyDegraded(new TreatmentWillValueType());
        treatmentWill.setNoLifeProlongingIfSeverePain(new TreatmentWillValueType());
        return treatmentWill;
    }
}
