/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.webservice;

import dk.sundhedsdatastyrelsen.behandlingstestamente._2018._05._01.*;
import dk.sundhedsdatastyrelsen.livstestamente.service.business.TreatmentWillServiceImpl;
import dk.sundhedsdatastyrelsen.livstestamente.domain.AcceptanceNeeded;
import dk.sundhedsdatastyrelsen.livstestamente.domain.RequestContextVO;
import dk.sundhedsdatastyrelsen.livstestamente.domain.TreatmentWillVO;
import dk.sundhedsdatastyrelsen.livstestamente.util.FaultCode;
import dk.sundhedsdatastyrelsen.livstestamente.util.ServiceException;
import dk.sundhedsdatastyrelsen.livstestamente.service.webservice.hl7cda.TreatmentWillHL7CDAMapper;
import dk.sundhedsdatastyrelsen.livstestamente.service.webservice.hl7cda.TreatmentWillHL7CDAValidator;
import dk.sundhedsdatastyrelsen.livstestamente.context.ServiceActor;
import dk.sundhedsdatastyrelsen.livstestamente.context.ServiceActorProvider;
import org.flywaydb.core.Flyway;
import org.hl7.btr.TreatmentWill;
import org.hl7.v3.ANY;
import org.hl7.v3.POCDMT000040Entry;
import org.junit.Test;
import org.mockito.Mockito;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.transaction.annotation.Transactional;

import java.util.List;

import static dk.sundhedsdatastyrelsen.livstestamente.service.util.TestHelper.*;
import static org.junit.Assert.*;

@Transactional("btrDataSourceTxManager")
public class TreatmentWillWebServiceTests extends AbstractWebServiceTest {

    @Autowired
    @Qualifier("treatmentWillWebService")
    private TreatmentWillWebService serviceUnderTest;

    @Autowired
    @Qualifier("treatmentWillAdminWebService")
    private TreatmentWillWebService adminServiceUnderTest;

    @Autowired
    private TreatmentWillServiceImpl businessService;

    @Autowired
    private TreatmentWillHL7CDAMapper treatmentWillHL7CDAMapper;

    @Autowired
    private TreatmentWillHL7CDAValidator validator;

    @Autowired
    private ServiceActorProvider serviceActorProvider;

    private static boolean initialized = false;
    private Flyway flyway;
    private String uri;
    private String adminUri;
    private static final String jaxbContextPath = "dk.sundhedsdatastyrelsen.behandlingstestamente._2018._05._01";


    @Test
    public void getTreatmentWill_testResponseStructure() {
        // Arrange
        TreatmentWillVO treatmentWillExample = testHelper.getTreatmentWillExample();
        ServiceActor serviceActorExample = testHelper.getServiceActorExample();
        RequestContextVO requestContextExample = testHelper.getRequestContextExample(serviceActorExample);
        businessService.createTreatmentWill(treatmentWillExample, requestContextExample, false, false);

        Mockito.when(serviceActorProvider.getServiceActor(Mockito.any(), Mockito.anyBoolean())).thenReturn(serviceActorExample);

        // Act
        final GetTreatmentWillRequest request = hl7Helper.createGetTreatmentWillRequest(PERSON_IDENTIFIER);
        final GetTreatmentWillResponse response = serviceUnderTest.getTreatmentWill20180501(request);

        // Assert
        assertNotNull(response);
        final List<POCDMT000040Entry> allEntries = hl7Helper.getAllEntries(response);

        // Should be 1 entry element
        assertEquals(1, allEntries.size());

        // Should be 1 value element
        List<ANY> values = allEntries.get(0).getObservation().getValues();
        assertEquals(1, values.size());

        // Cast the value element to a TreatmentWill-type
        TreatmentWill treatmentWillElement = (TreatmentWill) values.get(0);

        // Test values of TreatmentWill-element
        assertEquals(NO_LIFE_PROLONGING_IF_DYING, treatmentWillElement.getNoLifeProlongingIfDying().isValue());
        assertEquals("relativeAcceptanceRequired", treatmentWillElement.getNoLifeProlongingIfDying().getAcceptanceNeeded());
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERELY_DEGRADED, treatmentWillElement.getNoLifeProlongingIfSeverelyDegraded().isValue());
        assertEquals("relativeAcceptanceRequired", treatmentWillElement.getNoLifeProlongingIfDying().getAcceptanceNeeded());
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERE_PAIN, treatmentWillElement.getNoLifeProlongingIfSeverePain().isValue());
        assertEquals("relativeAcceptanceRequired", treatmentWillElement.getNoLifeProlongingIfDying().getAcceptanceNeeded());
        assertEquals(NO_FORCED_TREATMENT_IF_INCAPABLE, treatmentWillElement.getNoForcedTreatmentIfIncapable().isValue());
        assertEquals("relativeAcceptanceRequired", treatmentWillElement.getNoLifeProlongingIfDying().getAcceptanceNeeded());
    }

    @Test
    public void adminGetTreatmentWill_testResponseStructure() {
        // Arrange
        TreatmentWillVO treatmentWillExample = testHelper.getTreatmentWillExample();
        ServiceActor serviceActorExample = testHelper.getServiceActorAdminExample();
        RequestContextVO requestContextExample = testHelper.getRequestContextExample(serviceActorExample);
        businessService.createTreatmentWill(treatmentWillExample, requestContextExample, false, false);
        Mockito.when(serviceActorProvider.getServiceActor(Mockito.any(), Mockito.anyBoolean())).thenReturn(serviceActorExample);

        // Act
        final GetTreatmentWillRequest request = hl7Helper.createGetTreatmentWillRequest(PERSON_IDENTIFIER);
        final GetTreatmentWillResponse response = adminServiceUnderTest.getTreatmentWill20180501(request);

        // Assert
        assertNotNull(response);
        final List<POCDMT000040Entry> allEntries = hl7Helper.getAllEntries(response);

        // Should be 1 entry element
        assertEquals(1, allEntries.size());

        // Should be 1 value element
        List<ANY> values = allEntries.get(0).getObservation().getValues();
        assertEquals(1, values.size());

        // Cast the value element to a TreatmentWill-type
        TreatmentWill treatmentWillElement = (TreatmentWill) values.get(0);

        // Test values of TreatmentWill-element
        assertEquals(NO_LIFE_PROLONGING_IF_DYING, treatmentWillElement.getNoLifeProlongingIfDying().isValue());
        assertEquals("relativeAcceptanceRequired", treatmentWillElement.getNoLifeProlongingIfDying().getAcceptanceNeeded());
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERELY_DEGRADED, treatmentWillElement.getNoLifeProlongingIfSeverelyDegraded().isValue());
        assertEquals("relativeAcceptanceRequired", treatmentWillElement.getNoLifeProlongingIfDying().getAcceptanceNeeded());
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERE_PAIN, treatmentWillElement.getNoLifeProlongingIfSeverePain().isValue());
        assertEquals("relativeAcceptanceRequired", treatmentWillElement.getNoLifeProlongingIfDying().getAcceptanceNeeded());
        assertEquals(NO_FORCED_TREATMENT_IF_INCAPABLE, treatmentWillElement.getNoForcedTreatmentIfIncapable().isValue());
        assertEquals("relativeAcceptanceRequired", treatmentWillElement.getNoLifeProlongingIfDying().getAcceptanceNeeded());
    }

    @Test
    public void getTreatmentWill_validateResponseStructure() {
        // Arrange
        TreatmentWillVO treatmentWillExample = testHelper.getTreatmentWillExample();
        ServiceActor serviceActorExample = testHelper.getServiceActorExample();
        RequestContextVO requestContextExample = testHelper.getRequestContextExample(serviceActorExample);
        businessService.createTreatmentWill(treatmentWillExample, requestContextExample, false, false);
        Mockito.when(serviceActorProvider.getServiceActor(Mockito.any(), Mockito.anyBoolean())).thenReturn(serviceActorExample);

        // Act
        final GetTreatmentWillRequest request = hl7Helper.createGetTreatmentWillRequest(PERSON_IDENTIFIER);
        final GetTreatmentWillResponse response = serviceUnderTest.getTreatmentWill20180501(request);

        // Assert
        assertNotNull(response);

        validator.validateWithoutThrow(response);

        final int errorCount = validator.getValidationErrors().size();
        assertEquals(errorCount + " validation errors were found: " + validator.getCombinedErrorText(), 0, errorCount);
    }

    @Test
    public void hasTreatmentWill_testResponseStructure() {
        // Arrange
        TreatmentWillVO treatmentWillExample = testHelper.getTreatmentWillExample();
        ServiceActor serviceActorExample = testHelper.getServiceActorExample();
        RequestContextVO requestContextExample = testHelper.getRequestContextExample(serviceActorExample);
        Mockito.when(serviceActorProvider.getServiceActor(Mockito.any(), Mockito.anyBoolean())).thenReturn(serviceActorExample);
        businessService.createTreatmentWill(treatmentWillExample, requestContextExample, false, false);

        // Act
        final HasTreatmentWillRequest request = hl7Helper.createHasTreatmentWillRequest(PERSON_IDENTIFIER);
        final HasTreatmentWillResponse hasTreatmentWillResponse = serviceUnderTest.hasTreatmentWill20180501(request);

        // Assert
        assertNotNull(hasTreatmentWillResponse);
        assertTrue(hasTreatmentWillResponse.isWillExists());
    }

    @Test
    public void createTreatmentWill_testRequestMapping() {
        // Arrange
        final TreatmentWill treatmentWillExample = hl7Helper.getTreatmentWillExample();
        ServiceActor serviceActorExample = testHelper.getServiceActorExample();
        Mockito.when(serviceActorProvider.getServiceActor(Mockito.any(), Mockito.anyBoolean())).thenReturn(serviceActorExample);


        // Act
        final CreateTreatmentWillRequest request = hl7Helper.createCreateTreatmentWillRequest(PERSON_IDENTIFIER, treatmentWillExample);
        final CreateTreatmentWillResponse createTreatmentWillResponse = serviceUnderTest.createTreatmentWill20180501(request);

        // Assert using business layer
        assertNotNull(createTreatmentWillResponse);
        final TreatmentWillVO treatmentWillVO = testHelper.getTreatmentWill(businessService, PERSON_IDENTIFIER);
        assertNotNull(treatmentWillVO);
        assertNotNull(treatmentWillVO.getTreatmentWillPID());
        assertEquals(1, treatmentWillVO.getVersion().intValue());
        assertTrue(treatmentWillVO.getValidFrom().isBefore(treatmentWillVO.getValidTo()));
        assertEquals(NO_LIFE_PROLONGING_IF_DYING, treatmentWillVO.isNoLifeProlongingIfDying());
        assertEquals(null, treatmentWillVO.getNoLifeProlongingIfDyingAcceptanceNeeded());
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERELY_DEGRADED, treatmentWillVO.isNoLifeProlongingIfSeverelyDegraded());
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERELY_DEGRADED_ACCEPTANCE_NEEDED, treatmentWillVO.getNoLifeProlongingIfSeverelyDegradedAcceptanceNeeded());
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERE_PAIN, treatmentWillVO.isNoLifeProlongingIfSeverePain());
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERE_PAIN_ACCEPTANCE_NEEDED, treatmentWillVO.getNoLifeProlongingIfSeverePainAcceptanceNeeded());
        assertEquals(NO_FORCED_TREATMENT_IF_INCAPABLE, treatmentWillVO.isNoForcedTreatmentIfIncapable());
        assertEquals(NO_FORCED_TREATMENT_IF_INCAPABLE_ACCEPTANCE_NEEDED, treatmentWillVO.getNoForcedTreatmentIfIncapableAcceptanceNeeded());
    }

    @Test
    public void adminCreateTreatmentWill_testRequestMapping() {
        // Arrange
        final TreatmentWill treatmentWillExample = hl7Helper.getTreatmentWillExample();
        ServiceActor serviceActorExample = testHelper.getServiceActorAdminExample();
        Mockito.when(serviceActorProvider.getServiceActor(Mockito.any(), Mockito.anyBoolean())).thenReturn(serviceActorExample);

        // Act
        final CreateTreatmentWillRequest request = hl7Helper.createCreateTreatmentWillRequest(PERSON_IDENTIFIER, treatmentWillExample);
        final CreateTreatmentWillResponse createTreatmentWillResponse = adminServiceUnderTest.createTreatmentWill20180501(request);

        // Assert using business layer
        assertNotNull(createTreatmentWillResponse);
        final TreatmentWillVO treatmentWillVO = testHelper.getTreatmentWill(businessService, PERSON_IDENTIFIER);
        assertNotNull(treatmentWillVO);
        assertNotNull(treatmentWillVO.getTreatmentWillPID());
        assertEquals(1, treatmentWillVO.getVersion().intValue());
        assertTrue(treatmentWillVO.getValidFrom().isBefore(treatmentWillVO.getValidTo()));
        assertEquals(NO_LIFE_PROLONGING_IF_DYING, treatmentWillVO.isNoLifeProlongingIfDying());
        assertEquals(null, treatmentWillVO.getNoLifeProlongingIfDyingAcceptanceNeeded());
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERELY_DEGRADED, treatmentWillVO.isNoLifeProlongingIfSeverelyDegraded());
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERELY_DEGRADED_ACCEPTANCE_NEEDED, treatmentWillVO.getNoLifeProlongingIfSeverelyDegradedAcceptanceNeeded());
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERE_PAIN, treatmentWillVO.isNoLifeProlongingIfSeverePain());
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERE_PAIN_ACCEPTANCE_NEEDED, treatmentWillVO.getNoLifeProlongingIfSeverePainAcceptanceNeeded());
        assertEquals(NO_FORCED_TREATMENT_IF_INCAPABLE, treatmentWillVO.isNoForcedTreatmentIfIncapable());
        assertEquals(NO_FORCED_TREATMENT_IF_INCAPABLE_ACCEPTANCE_NEEDED, treatmentWillVO.getNoForcedTreatmentIfIncapableAcceptanceNeeded());
    }

    @Test
    public void createTreatmentWill_testRequestMapping_null() {
        // Arrange
        final TreatmentWill treatmentWillExample = hl7Helper.getTreatmentWillExample();
        ServiceActor serviceActorExample = testHelper.getServiceActorExample();
        Mockito.when(serviceActorProvider.getServiceActor(Mockito.any(), Mockito.anyBoolean())).thenReturn(serviceActorExample);

        treatmentWillExample.getNoLifeProlongingIfDying().setAcceptanceNeeded(null); // <--- is null
        treatmentWillExample.getNoLifeProlongingIfSeverelyDegraded().setAcceptanceNeeded(null); // <--- is null
        treatmentWillExample.getNoLifeProlongingIfSeverePain().setAcceptanceNeeded(null); // <--- is null
        treatmentWillExample.getNoForcedTreatmentIfIncapable().setAcceptanceNeeded(null); // <--- is null

        // Act
        final CreateTreatmentWillRequest request = hl7Helper.createCreateTreatmentWillRequest(PERSON_IDENTIFIER, treatmentWillExample);
        final CreateTreatmentWillResponse createTreatmentWillResponse = serviceUnderTest.createTreatmentWill20180501(request);

        // Assert using business layer
        assertNotNull(createTreatmentWillResponse);
        final TreatmentWillVO treatmentWillVO = testHelper.getTreatmentWill(businessService, PERSON_IDENTIFIER);
        assertNotNull(treatmentWillVO);
        assertNotNull(treatmentWillVO.getTreatmentWillPID());
        assertEquals(1, treatmentWillVO.getVersion().intValue());
        assertNull(treatmentWillVO.getNoLifeProlongingIfDyingAcceptanceNeeded()); // <--- is null
        assertNull(treatmentWillVO.getNoLifeProlongingIfSeverelyDegradedAcceptanceNeeded()); // <--- is null
        assertNull(treatmentWillVO.getNoLifeProlongingIfSeverePainAcceptanceNeeded()); // <--- is null
        assertNull(treatmentWillVO.getNoForcedTreatmentIfIncapableAcceptanceNeeded()); // <--- is null
    }

    @Test
    public void updateTreatmentWill_testRequestMapping() {
        // Note: this also tests null values

        // Arrange
        final boolean noLifeProlongingIfDyingUpdated = false;
        final AcceptanceNeeded noLifeProlongingIfDyingAcceptanceNeededUpdated = null;
        final boolean noLifeProlongingIfSeverelyDegradedUpdated = false;
        final AcceptanceNeeded noLifeProlongingIfSeverelyDegradedAcceptanceNeededUpdated = null;
        final boolean noLifeProlongingIfSeverePainUpdated = false;
        final AcceptanceNeeded noLifeProlongingIfSeverePainAcceptanceNeededUpdated = null;
        final boolean noForcedTreatmentIfIncapableUpdated = false;
        final AcceptanceNeeded noForcedTreatmentIfIncapableAcceptanceNeededUpdated = null;

        TreatmentWill treatmentWill = hl7Helper.getTreatmentWillExample();
        treatmentWill.setNoLifeProlongingIfDying(treatmentWillHL7CDAMapper.toTreatmentWillValueType(noLifeProlongingIfDyingUpdated, noLifeProlongingIfDyingAcceptanceNeededUpdated));
        treatmentWill.setNoLifeProlongingIfSeverelyDegraded(treatmentWillHL7CDAMapper.toTreatmentWillValueType(noLifeProlongingIfSeverelyDegradedUpdated, noLifeProlongingIfSeverelyDegradedAcceptanceNeededUpdated));
        treatmentWill.setNoLifeProlongingIfSeverePain(treatmentWillHL7CDAMapper.toTreatmentWillValueType(noLifeProlongingIfSeverePainUpdated, noLifeProlongingIfSeverePainAcceptanceNeededUpdated));
        treatmentWill.setNoForcedTreatmentIfIncapable(treatmentWillHL7CDAMapper.toTreatmentWillValueType(noForcedTreatmentIfIncapableUpdated, noForcedTreatmentIfIncapableAcceptanceNeededUpdated));

        TreatmentWillVO treatmentWillExample = testHelper.getTreatmentWillExample();
        ServiceActor serviceActorExample = testHelper.getServiceActorExample();
        RequestContextVO requestContextExample = testHelper.getRequestContextExample(serviceActorExample);
        Mockito.when(serviceActorProvider.getServiceActor(Mockito.any(), Mockito.anyBoolean())).thenReturn(serviceActorExample);

        businessService.createTreatmentWill(treatmentWillExample, requestContextExample, false, false);

        // Act
        final UpdateTreatmentWillRequest request = hl7Helper.createUpdateTreatmentWillRequest(PERSON_IDENTIFIER, treatmentWill);
        final UpdateTreatmentWillResponse updateTreatmentWillResponse = serviceUnderTest.updateTreatmentWill20180501(request);

        // Assert using business layer
        assertNotNull(updateTreatmentWillResponse);
        final TreatmentWillVO treatmentWillVO = testHelper.getTreatmentWill(businessService, PERSON_IDENTIFIER);
        assertNotNull(treatmentWillVO);
        assertNotNull(treatmentWillVO.getTreatmentWillPID());
        assertEquals(2, treatmentWillVO.getVersion().intValue());
        assertTrue(treatmentWillVO.getValidFrom().isBefore(treatmentWillVO.getValidTo()));

        assertEquals(noLifeProlongingIfDyingUpdated, treatmentWillVO.isNoLifeProlongingIfDying());
        assertEquals(noLifeProlongingIfDyingAcceptanceNeededUpdated, treatmentWillVO.getNoLifeProlongingIfDyingAcceptanceNeeded());

        assertEquals(noLifeProlongingIfSeverelyDegradedUpdated, treatmentWillVO.isNoLifeProlongingIfSeverelyDegraded());
        assertEquals(noLifeProlongingIfSeverelyDegradedAcceptanceNeededUpdated, treatmentWillVO.getNoLifeProlongingIfSeverelyDegradedAcceptanceNeeded());

        assertEquals(noLifeProlongingIfSeverePainUpdated, treatmentWillVO.isNoLifeProlongingIfSeverePain());
        assertEquals(noLifeProlongingIfSeverePainAcceptanceNeededUpdated, treatmentWillVO.getNoLifeProlongingIfSeverePainAcceptanceNeeded());

        assertEquals(noForcedTreatmentIfIncapableUpdated, treatmentWillVO.isNoForcedTreatmentIfIncapable());
        assertEquals(noForcedTreatmentIfIncapableAcceptanceNeededUpdated, treatmentWillVO.getNoForcedTreatmentIfIncapableAcceptanceNeeded());
    }

    @Test
    public void adminUpdateTreatmentWill_testRequestMapping() {
        // Note: this also tests null values

        // Arrange
        final boolean noLifeProlongingIfDyingUpdated = false;
        final AcceptanceNeeded noLifeProlongingIfDyingAcceptanceNeededUpdated = null;
        final boolean noLifeProlongingIfSeverelyDegradedUpdated = false;
        final AcceptanceNeeded noLifeProlongingIfSeverelyDegradedAcceptanceNeededUpdated = null;
        final boolean noLifeProlongingIfSeverePainUpdated = false;
        final AcceptanceNeeded noLifeProlongingIfSeverePainAcceptanceNeededUpdated = null;
        final boolean noForcedTreatmentIfIncapableUpdated = false;
        final AcceptanceNeeded noForcedTreatmentIfIncapableAcceptanceNeededUpdated = null;

        TreatmentWill treatmentWill = hl7Helper.getTreatmentWillExample();
        treatmentWill.setNoLifeProlongingIfDying(treatmentWillHL7CDAMapper.toTreatmentWillValueType(noLifeProlongingIfDyingUpdated, noLifeProlongingIfDyingAcceptanceNeededUpdated));
        treatmentWill.setNoLifeProlongingIfSeverelyDegraded(treatmentWillHL7CDAMapper.toTreatmentWillValueType(noLifeProlongingIfSeverelyDegradedUpdated, noLifeProlongingIfSeverelyDegradedAcceptanceNeededUpdated));
        treatmentWill.setNoLifeProlongingIfSeverePain(treatmentWillHL7CDAMapper.toTreatmentWillValueType(noLifeProlongingIfSeverePainUpdated, noLifeProlongingIfSeverePainAcceptanceNeededUpdated));
        treatmentWill.setNoForcedTreatmentIfIncapable(treatmentWillHL7CDAMapper.toTreatmentWillValueType(noForcedTreatmentIfIncapableUpdated, noForcedTreatmentIfIncapableAcceptanceNeededUpdated));

        TreatmentWillVO treatmentWillExample = testHelper.getTreatmentWillExample();
        ServiceActor serviceActorExample = testHelper.getServiceActorAdminExample();
        RequestContextVO requestContextExample = testHelper.getRequestContextExample(serviceActorExample);
        Mockito.when(serviceActorProvider.getServiceActor(Mockito.any(), Mockito.anyBoolean())).thenReturn(serviceActorExample);

        businessService.createTreatmentWill(treatmentWillExample, requestContextExample, false, false);

        // Act
        final UpdateTreatmentWillRequest request = hl7Helper.createUpdateTreatmentWillRequest(PERSON_IDENTIFIER, treatmentWill);
        final UpdateTreatmentWillResponse updateTreatmentWillResponse = adminServiceUnderTest.updateTreatmentWill20180501(request);

        // Assert using business layer
        assertNotNull(updateTreatmentWillResponse);
        final TreatmentWillVO treatmentWillVO = testHelper.getTreatmentWill(businessService, PERSON_IDENTIFIER);
        assertNotNull(treatmentWillVO);
        assertNotNull(treatmentWillVO.getTreatmentWillPID());
        assertEquals(2, treatmentWillVO.getVersion().intValue());
        assertTrue(treatmentWillVO.getValidFrom().isBefore(treatmentWillVO.getValidTo()));

        assertEquals(noLifeProlongingIfDyingUpdated, treatmentWillVO.isNoLifeProlongingIfDying());
        assertEquals(noLifeProlongingIfDyingAcceptanceNeededUpdated, treatmentWillVO.getNoLifeProlongingIfDyingAcceptanceNeeded());

        assertEquals(noLifeProlongingIfSeverelyDegradedUpdated, treatmentWillVO.isNoLifeProlongingIfSeverelyDegraded());
        assertEquals(noLifeProlongingIfSeverelyDegradedAcceptanceNeededUpdated, treatmentWillVO.getNoLifeProlongingIfSeverelyDegradedAcceptanceNeeded());

        assertEquals(noLifeProlongingIfSeverePainUpdated, treatmentWillVO.isNoLifeProlongingIfSeverePain());
        assertEquals(noLifeProlongingIfSeverePainAcceptanceNeededUpdated, treatmentWillVO.getNoLifeProlongingIfSeverePainAcceptanceNeeded());

        assertEquals(noForcedTreatmentIfIncapableUpdated, treatmentWillVO.isNoForcedTreatmentIfIncapable());
        assertEquals(noForcedTreatmentIfIncapableAcceptanceNeededUpdated, treatmentWillVO.getNoForcedTreatmentIfIncapableAcceptanceNeeded());
    }

    @Test
    public void deleteTreatmentWill_testRequestMapping() {
        // Arrange
        TreatmentWillVO treatmentWillExample = testHelper.getTreatmentWillExample();
        ServiceActor serviceActorExample = testHelper.getServiceActorExample();
        RequestContextVO requestContextExample = testHelper.getRequestContextExample(serviceActorExample);
        Mockito.when(serviceActorProvider.getServiceActor(Mockito.any(), Mockito.anyBoolean())).thenReturn(serviceActorExample);

        businessService.createTreatmentWill(treatmentWillExample, requestContextExample, false, false);

        // Act
        final DeleteTreatmentWillRequest request = hl7Helper.createDeleteTreatmentWillRequest(PERSON_IDENTIFIER);
        final DeleteTreatmentWillResponse deleteTreatmentWillResponse = serviceUnderTest.deleteTreatmentWill20180501(request);

        // Assert using business layer
        // No data should be found
        assertNotNull(deleteTreatmentWillResponse);
        try {
            testHelper.getTreatmentWill(businessService, PERSON_IDENTIFIER);
            fail();
        } catch (ServiceException e) {
            assertEquals(FaultCode.TREATMENT_WILL_NO_DATA_FOUND, e.getFaultCode());
        }
    }

    @Test
    public void adminDeleteTreatmentWill_testRequestMapping() {
        // Arrange
        TreatmentWillVO treatmentWillExample = testHelper.getTreatmentWillExample();
        ServiceActor serviceActorExample = testHelper.getServiceActorExample();
        RequestContextVO requestContextExample = testHelper.getRequestContextExample(serviceActorExample);
        Mockito.when(serviceActorProvider.getServiceActor(Mockito.any(), Mockito.anyBoolean())).thenReturn(serviceActorExample);

        businessService.createTreatmentWill(treatmentWillExample, requestContextExample, false, false);

        // Act
        final DeleteTreatmentWillRequest request = hl7Helper.createDeleteTreatmentWillRequest(PERSON_IDENTIFIER);
        final DeleteTreatmentWillResponse deleteTreatmentWillResponse = adminServiceUnderTest.deleteTreatmentWill20180501(request);

        // Assert using business layer
        // No data should be found
        assertNotNull(deleteTreatmentWillResponse);
        try {
            testHelper.getTreatmentWill(businessService, PERSON_IDENTIFIER);
            fail();
        } catch (ServiceException e) {
            assertEquals(FaultCode.TREATMENT_WILL_NO_DATA_FOUND, e.getFaultCode());
        }
    }
    
    @Test
    public void createTreatmentWill_testThatAcceptanceNeededNotAllowedForNoLifeProlongingIfDying() {
        // Arrange
        final TreatmentWill treatmentWillExample = hl7Helper.getTreatmentWillExampleWithNoLifeProlongingIfDyinAcceptanceNeeded();
    	String errorExpected = "420: Fejl i request i forbindelse med oprettelse af Behandlingstestamente; 420: Fejl i request i forbindelse med oprettelse af Behandlingstestamente Details: Unsupported value was found: relativeAcceptanceRequired";

        // Act
        final CreateTreatmentWillRequest request = hl7Helper.createCreateTreatmentWillRequest(PERSON_IDENTIFIER, treatmentWillExample);
        CreateTreatmentWillResponse createTreatmentWillResponse = null;
        
        try {
            createTreatmentWillResponse = serviceUnderTest.createTreatmentWill20180501(request);
        } catch (Exception e) {
            // Assert
            assertEquals(errorExpected, e.getMessage());
        }
        assertNull(createTreatmentWillResponse);
    }
    
    @Test
    public void updateTreatmentWill_testThatAcceptanceNeededNotAllowedForNoLifeProlongingIfDying() {
        // Arrange
    	String errorExpected = "430: Fejl i request i forbindelse med opdatering af Behandlingstestamente; 430: Fejl i request i forbindelse med opdatering af Behandlingstestamente Details: Unsupported value was found: relativeAcceptanceRequired";
        TreatmentWill treatmentWill = hl7Helper.getTreatmentWillExampleWithNoLifeProlongingIfDyinAcceptanceNeeded();

        TreatmentWillVO treatmentWillExample = testHelper.getTreatmentWillExample();
        ServiceActor serviceActorExample = testHelper.getServiceActorExample();
        RequestContextVO requestContextExample = testHelper.getRequestContextExample(serviceActorExample);
        Mockito.when(serviceActorProvider.getServiceActor(Mockito.any(), Mockito.anyBoolean())).thenReturn(serviceActorExample);

        businessService.createTreatmentWill(treatmentWillExample, requestContextExample, false, false);

        // Act
        final UpdateTreatmentWillRequest request = hl7Helper.createUpdateTreatmentWillRequest(PERSON_IDENTIFIER, treatmentWill);

        UpdateTreatmentWillResponse updateTreatmentWillResponse = null;
        
        try {
        	updateTreatmentWillResponse = serviceUnderTest.updateTreatmentWill20180501(request);

        } catch (Exception e) {
        	// Assert
        	assertEquals(errorExpected, e.getMessage());
        }
        assertNull(updateTreatmentWillResponse);
    }
    
}
