/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.webservice;

import static dk.sundhedsdatastyrelsen.livstestamente.service.util.TestHelperV2.*;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.util.List;

import dk.sundhedsdatastyrelsen.behandlingstestamente._2020._03._16.*;
import dk.sundhedsdatastyrelsen.livstestamente._2018._05._01.ObjectFactory;
import dk.sundhedsdatastyrelsen.livstestamente.service.webservice.hl7cda.LivingWillHL7CDAMapper;
import org.hl7.btrv2.TreatmentWillV2;
import org.hl7.btrv2.TreatmentWillWithOnlyForcedTreatment;
import org.hl7.ltr.LivingWill;
import org.hl7.v3.ANY;
import org.hl7.v3.POCDMT000040Entry;
import org.junit.Test;
import org.mockito.Mockito;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.transaction.annotation.Transactional;

import dk.sundhedsdatastyrelsen.livstestamente.service.business.TreatmentWillServiceImpl;
import dk.sundhedsdatastyrelsen.livstestamente.domain.AcceptanceNeeded;
import dk.sundhedsdatastyrelsen.livstestamente.domain.RequestContextVO;
import dk.sundhedsdatastyrelsen.livstestamente.domain.TreatmentWillVO;
import dk.sundhedsdatastyrelsen.livstestamente.util.FaultCode;
import dk.sundhedsdatastyrelsen.livstestamente.util.ServiceException;
import dk.sundhedsdatastyrelsen.livstestamente.service.webservice.hl7cda.TreatmentWillHL7CDAMapperV2;
import dk.sundhedsdatastyrelsen.livstestamente.service.webservice.hl7cda.TreatmentWillHL7CDAValidatorV2;
import dk.sundhedsdatastyrelsen.livstestamente.context.ServiceActor;
import dk.sundhedsdatastyrelsen.livstestamente.context.ServiceActorProvider;

@Transactional("btrDataSourceTxManager")
public class TreatmentWillWebServiceV2Tests extends AbstractWebServiceTest {

    @Autowired
    @Qualifier("treatmentWillWebServiceV2")
    private TreatmentWillWebServiceV2 serviceUnderTest;

    @Autowired
    @Qualifier("treatmentWillAdminWebServiceV2")
    private TreatmentWillWebServiceV2 adminServiceUnderTest;

    @Autowired
    @Qualifier("livingWillAdminWebService")
    private LivingWillWebService adminLivingWillServiceUnderTest;

    @Autowired
    private TreatmentWillServiceImpl businessService;

    @Autowired
    private TreatmentWillHL7CDAMapperV2 treatmentWillHL7CDAMapper;

    @Autowired
    private TreatmentWillHL7CDAValidatorV2 validator;

    @Autowired
    private ServiceActorProvider serviceActorProvider;

    @Autowired
    private LivingWillHL7CDAMapper livingWillHL7CDAMapper;

    @Autowired
    private LivingWillWebService livingWillWebService;


    @Test
    public void getTreatmentWill_testResponseStructure() {
        // Arrange
        TreatmentWillVO treatmentWillExample = testHelperV2.getTreatmentWillExample();
        ServiceActor serviceActorExample = testHelperV2.getServiceActorExample();
        RequestContextVO requestContextExample = testHelperV2.getRequestContextExample(serviceActorExample, MESSAGE_ID);
        businessService.createTreatmentWill(treatmentWillExample, requestContextExample, true, false);

        Mockito.when(serviceActorProvider.getServiceActor(Mockito.any(), Mockito.anyBoolean())).thenReturn(serviceActorExample);

        // Act
        final GetTreatmentWillRequest request = hl7HelperV2.createGetTreatmentWillRequest(PERSON_IDENTIFIER);
        GetTreatmentWillResponse response = serviceUnderTest.getTreatmentWill20200316(request);

        // Assert
        assertNotNull(response);
        final List<POCDMT000040Entry> allEntries = hl7HelperV2.getAllEntries(response.getClinicalDocument());

        // Should be 1 entry element
        assertEquals(1, allEntries.size());

        // Should be 1 value element
        List<ANY> values = allEntries.get(0).getObservation().getValues();
        assertEquals(1, values.size());

        // Cast the value element to a TreatmentWill-type
        TreatmentWillV2 treatmentWillElement = (TreatmentWillV2) values.get(0);

        // Test values of TreatmentWill-element
        assertEquals(NO_LIFE_PROLONGING_IF_DYING, treatmentWillElement.getNoLifeProlongingIfDying().isValue());
        
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERELY_DEGRADED, treatmentWillElement.getNoLifeProlongingIfSeverelyDegraded().isValue());
        assertEquals("relativeAcceptanceRequired", treatmentWillElement.getNoLifeProlongingIfSeverelyDegraded().getAcceptanceNeeded());
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERE_PAIN, treatmentWillElement.getNoLifeProlongingIfSeverePain().isValue());
        assertEquals("relativeAcceptanceRequired", treatmentWillElement.getNoLifeProlongingIfSeverePain().getAcceptanceNeeded());
        assertEquals(NO_FORCED_TREATMENT_IF_INCAPABLE, treatmentWillElement.getNoForcedTreatmentIfIncapable().isValue());
        assertEquals("relativeAcceptanceRequired", treatmentWillElement.getNoForcedTreatmentIfIncapable().getAcceptanceNeeded());
    }

    @Test
    public void adminGetTreatmentWill_testResponseStructure() {
        // Arrange
        TreatmentWillVO treatmentWillExample = testHelperV2.getTreatmentWillExample();
        ServiceActor serviceActorExample = testHelperV2.getServiceActorAdminExample();
        RequestContextVO requestContextExample = testHelperV2.getRequestContextExample(serviceActorExample, MESSAGE_ID);
        businessService.createTreatmentWill(treatmentWillExample, requestContextExample, true, false);
        Mockito.when(serviceActorProvider.getServiceActor(Mockito.any(), Mockito.anyBoolean())).thenReturn(serviceActorExample);


        // Act
        final GetTreatmentWillRequest request = hl7HelperV2.createGetTreatmentWillRequest(PERSON_IDENTIFIER);
        GetTreatmentWillResponse response = adminServiceUnderTest.getTreatmentWill20200316(request);

        // Assert
        assertNotNull(response);
        final List<POCDMT000040Entry> allEntries = hl7HelperV2.getAllEntries(response.getClinicalDocument());

        // Should be 1 entry element
        assertEquals(1, allEntries.size());

        // Should be 1 value element
        List<ANY> values = allEntries.get(0).getObservation().getValues();
        assertEquals(1, values.size());

        // Cast the value element to a TreatmentWill-type
        TreatmentWillV2 treatmentWillElement = (TreatmentWillV2) values.get(0);

        // Test values of TreatmentWill-element
        assertEquals(NO_LIFE_PROLONGING_IF_DYING, treatmentWillElement.getNoLifeProlongingIfDying().isValue());

        assertEquals(NO_LIFE_PROLONGING_IF_SEVERELY_DEGRADED, treatmentWillElement.getNoLifeProlongingIfSeverelyDegraded().isValue());
        assertEquals("relativeAcceptanceRequired", treatmentWillElement.getNoLifeProlongingIfSeverelyDegraded().getAcceptanceNeeded());
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERE_PAIN, treatmentWillElement.getNoLifeProlongingIfSeverelyDegraded().isValue());
        assertEquals("relativeAcceptanceRequired", treatmentWillElement.getNoLifeProlongingIfSeverelyDegraded().getAcceptanceNeeded());
        assertEquals(NO_FORCED_TREATMENT_IF_INCAPABLE, treatmentWillElement.getNoForcedTreatmentIfIncapable().isValue());
        assertEquals("relativeAcceptanceRequired", treatmentWillElement.getNoForcedTreatmentIfIncapable().getAcceptanceNeeded());
    }

    @Test
    public void getTreatmentWill_validateResponseStructure() {
        // Arrange
        TreatmentWillVO treatmentWillExample = testHelperV2.getTreatmentWillExample();
        ServiceActor serviceActorExample = testHelperV2.getServiceActorExample();
        RequestContextVO requestContextExample = testHelperV2.getRequestContextExample(serviceActorExample, MESSAGE_ID);
        Mockito.when(serviceActorProvider.getServiceActor(Mockito.any(), Mockito.anyBoolean())).thenReturn(serviceActorExample);

        businessService.createTreatmentWill(treatmentWillExample, requestContextExample, true, false);


        // Act
        final GetTreatmentWillRequest request = hl7HelperV2.createGetTreatmentWillRequest(PERSON_IDENTIFIER);
        GetTreatmentWillResponse response = serviceUnderTest.getTreatmentWill20200316(request);

        // Assert
        assertNotNull(response);

        validator.validateWithoutThrow(response.getClinicalDocument());

        final int errorCount = validator.getValidationErrors().size();
        assertEquals(errorCount + " validation errors were found: " + validator.getCombinedErrorText(), 0, errorCount);
    }

    @Test
    public void getTreatmentWillWithOnlyForcedTreatment_testResponseStructure() {
        // Arrange
        TreatmentWillVO treatmentWillExample = testHelperV2.getTreatmentWillExample();
        ServiceActor serviceActorExample = testHelperV2.getServiceActorExample();
        RequestContextVO requestContextExample = testHelperV2.getRequestContextExample(serviceActorExample, MESSAGE_ID);
        businessService.createTreatmentWill(treatmentWillExample, requestContextExample, true, false);
        Mockito.when(serviceActorProvider.getServiceActor(Mockito.any(), Mockito.anyBoolean())).thenReturn(serviceActorExample);

        // Act
        final GetTreatmentWillWithOnlyForcedTreatmentRequest request = hl7HelperV2.createGetTreatmentWillWithOnlyForcedTreatmentRequest(PERSON_IDENTIFIER);
        GetTreatmentWillWithOnlyForcedTreatmentResponse response = serviceUnderTest.getTreatmentWillWithOnlyForcedTreatment20200316(request);

        // Assert
        assertNotNull(response);
        final List<POCDMT000040Entry> allEntries = hl7HelperV2.getAllEntries(response.getClinicalDocument());

        // Should be 1 entry element
        assertEquals(1, allEntries.size());

        // Should be 1 value element
        List<ANY> values = allEntries.get(0).getObservation().getValues();
        assertEquals(1, values.size());

        TreatmentWillWithOnlyForcedTreatment treatmentWillElement = (TreatmentWillWithOnlyForcedTreatment) values.get(0);

        assertEquals(NO_FORCED_TREATMENT_IF_INCAPABLE, treatmentWillElement.getNoForcedTreatmentIfIncapable().isValue());
        assertEquals("relativeAcceptanceRequired", treatmentWillElement.getNoForcedTreatmentIfIncapable().getAcceptanceNeeded());
    }

    @Test
    public void adminGetTreatmentWillWithOnlyForcedTreatment_testResponseStructure() {
        // Arrange
        TreatmentWillVO treatmentWillExample = testHelperV2.getTreatmentWillExample();
        ServiceActor serviceActorExample = testHelperV2.getServiceActorAdminExample();
        RequestContextVO requestContextExample = testHelperV2.getRequestContextExample(serviceActorExample, MESSAGE_ID);
        businessService.createTreatmentWill(treatmentWillExample, requestContextExample, true, false);
        Mockito.when(serviceActorProvider.getServiceActor(Mockito.any(), Mockito.anyBoolean())).thenReturn(serviceActorExample);


        // Act
        final GetTreatmentWillWithOnlyForcedTreatmentRequest request = hl7HelperV2.createGetTreatmentWillWithOnlyForcedTreatmentRequest(PERSON_IDENTIFIER);
        GetTreatmentWillWithOnlyForcedTreatmentResponse response = adminServiceUnderTest.getTreatmentWillWithOnlyForcedTreatment20200316(request);

        // Assert
        assertNotNull(response);
        final List<POCDMT000040Entry> allEntries = hl7HelperV2.getAllEntries(response.getClinicalDocument());

        // Should be 1 entry element
        assertEquals(1, allEntries.size());

        // Should be 1 value element
        List<ANY> values = allEntries.get(0).getObservation().getValues();
        assertEquals(1, values.size());

        TreatmentWillWithOnlyForcedTreatment treatmentWillElement = (TreatmentWillWithOnlyForcedTreatment) values.get(0);

        assertEquals(NO_FORCED_TREATMENT_IF_INCAPABLE, treatmentWillElement.getNoForcedTreatmentIfIncapable().isValue());
        assertEquals("relativeAcceptanceRequired", treatmentWillElement.getNoForcedTreatmentIfIncapable().getAcceptanceNeeded());
    }

    @Test
    public void getTreatmentWillWithOnlyForcedTreatment_validateResponseStructure() {
        // Arrange
        TreatmentWillVO treatmentWillExample = testHelperV2.getTreatmentWillExample();
        ServiceActor serviceActorExample = testHelperV2.getServiceActorExample();
        RequestContextVO requestContextExample = testHelperV2.getRequestContextExample(serviceActorExample, MESSAGE_ID);
        Mockito.when(serviceActorProvider.getServiceActor(Mockito.any(), Mockito.anyBoolean())).thenReturn(serviceActorExample);

        businessService.createTreatmentWill(treatmentWillExample, requestContextExample, true, false);

        // Act
        final GetTreatmentWillWithOnlyForcedTreatmentRequest request = hl7HelperV2.createGetTreatmentWillWithOnlyForcedTreatmentRequest(PERSON_IDENTIFIER);
        GetTreatmentWillWithOnlyForcedTreatmentResponse response = serviceUnderTest.getTreatmentWillWithOnlyForcedTreatment20200316(request);

        // Assert
        assertNotNull(response);

        validator.validateWithoutThrow(response.getClinicalDocument());

        final int errorCount = validator.getValidationErrors().size();
        assertEquals(errorCount + " validation errors were found: " + validator.getCombinedErrorText(), 0, errorCount);
    }

    @Test
    public void hasTreatmentWill_testResponseStructure() {
        // Arrange
        TreatmentWillVO treatmentWillExample = testHelperV2.getTreatmentWillExample();
        ServiceActor serviceActorExample = testHelperV2.getServiceActorExample();
        RequestContextVO requestContextExample = testHelperV2.getRequestContextExample(serviceActorExample, MESSAGE_ID);
        Mockito.when(serviceActorProvider.getServiceActor(Mockito.any(), Mockito.anyBoolean())).thenReturn(serviceActorExample);
        businessService.createTreatmentWill(treatmentWillExample, requestContextExample, true, false);

        // Act
        final HasTreatmentWillRequest request = hl7HelperV2.createHasTreatmentWillRequest(PERSON_IDENTIFIER);
        HasTreatmentWillResponse response = serviceUnderTest.hasTreatmentWill20200316(request);

        // Assert
        assertNotNull(response);
        assertTrue(response.isWillExists());
    }

    @Test
    public void createTreatmentWill_testRequestMapping() {
        // Arrange
        final TreatmentWillV2 treatmentWillExample = hl7HelperV2.getTreatmentWillExample();
        ServiceActor serviceActorExample = testHelperV2.getServiceActorExample();
        Mockito.when(serviceActorProvider.getServiceActor(Mockito.any(), Mockito.anyBoolean())).thenReturn(serviceActorExample);

        // Act
        final CreateTreatmentWillRequest request = hl7HelperV2.createCreateTreatmentWillRequest(PERSON_IDENTIFIER, treatmentWillExample);
        CreateTreatmentWillResponse response = serviceUnderTest.createTreatmentWill20200316(request);

        // Assert using business layer
        assertNotNull(response);
        final TreatmentWillVO treatmentWillVO = testHelperV2.getTreatmentWill(businessService, PERSON_IDENTIFIER);
        assertNotNull(treatmentWillVO);
        assertNotNull(treatmentWillVO.getTreatmentWillPID());
        assertEquals(1, treatmentWillVO.getVersion().intValue());
        assertTrue(treatmentWillVO.getValidFrom().isBefore(treatmentWillVO.getValidTo()));
        assertEquals(NO_LIFE_PROLONGING_IF_DYING, treatmentWillVO.isNoLifeProlongingIfDying());
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERELY_DEGRADED, treatmentWillVO.isNoLifeProlongingIfSeverelyDegraded());
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERELY_DEGRADED_ACCEPTANCE_NEEDED, treatmentWillVO.getNoLifeProlongingIfSeverelyDegradedAcceptanceNeeded());
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERE_PAIN, treatmentWillVO.isNoLifeProlongingIfSeverePain());
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERE_PAIN_ACCEPTANCE_NEEDED, treatmentWillVO.getNoLifeProlongingIfSeverePainAcceptanceNeeded());
        assertEquals(NO_FORCED_TREATMENT_IF_INCAPABLE, treatmentWillVO.isNoForcedTreatmentIfIncapable());
        assertEquals(NO_FORCED_TREATMENT_IF_INCAPABLE_ACCEPTANCE_NEEDED, treatmentWillVO.getNoForcedTreatmentIfIncapableAcceptanceNeeded());
    }

    @Test
    public void adminCreateTreatmentWill_testRequestMapping() {
        // Arrange
        final TreatmentWillV2 treatmentWillExample = hl7HelperV2.getTreatmentWillExample();
        ServiceActor serviceActorExample = testHelperV2.getServiceActorAdminExample();
        Mockito.when(serviceActorProvider.getServiceActor(Mockito.any(), Mockito.anyBoolean())).thenReturn(serviceActorExample);


        // Act
        final CreateTreatmentWillRequest request = hl7HelperV2.createCreateTreatmentWillRequest(PERSON_IDENTIFIER, treatmentWillExample);
        CreateTreatmentWillResponse response = adminServiceUnderTest.createTreatmentWill20200316(request);

        // Assert using business layer
        assertNotNull(response);
        final TreatmentWillVO treatmentWillVO = testHelperV2.getTreatmentWill(businessService, PERSON_IDENTIFIER);
        assertNotNull(treatmentWillVO);
        assertNotNull(treatmentWillVO.getTreatmentWillPID());
        assertEquals(1, treatmentWillVO.getVersion().intValue());
        assertTrue(treatmentWillVO.getValidFrom().isBefore(treatmentWillVO.getValidTo()));
        assertEquals(NO_LIFE_PROLONGING_IF_DYING, treatmentWillVO.isNoLifeProlongingIfDying());
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERELY_DEGRADED, treatmentWillVO.isNoLifeProlongingIfSeverelyDegraded());
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERELY_DEGRADED_ACCEPTANCE_NEEDED, treatmentWillVO.getNoLifeProlongingIfSeverelyDegradedAcceptanceNeeded());
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERE_PAIN, treatmentWillVO.isNoLifeProlongingIfSeverePain());
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERE_PAIN_ACCEPTANCE_NEEDED, treatmentWillVO.getNoLifeProlongingIfSeverePainAcceptanceNeeded());
        assertEquals(NO_FORCED_TREATMENT_IF_INCAPABLE, treatmentWillVO.isNoForcedTreatmentIfIncapable());
        assertEquals(NO_FORCED_TREATMENT_IF_INCAPABLE_ACCEPTANCE_NEEDED, treatmentWillVO.getNoForcedTreatmentIfIncapableAcceptanceNeeded());
    }


    @Test
    public void adminUpgradeToTreatmentWill_testRequestMapping() {
        // Arrange
        final TreatmentWillV2 treatmentWillExample = hl7HelperV2.getTreatmentWillExample();
        ServiceActor serviceActorExample = testHelperV2.getServiceActorAdminExample();
        Mockito.when(serviceActorProvider.getServiceActor(Mockito.any(), Mockito.anyBoolean())).thenReturn(serviceActorExample);

        // Act
        final UpgradeToTreatmentWillRequest request = hl7HelperV2.createUpgradeToTreatmentWillRequest(PERSON_IDENTIFIER_4, treatmentWillExample);
        UpgradeToTreatmentWillResponse response = adminServiceUnderTest.upgradeToTreatmentWill20200316(request);

        // Assert using business layer
        assertNotNull(response);
        final TreatmentWillVO treatmentWillVO = testHelperV2.getTreatmentWill(businessService, PERSON_IDENTIFIER_4);
        assertNotNull(treatmentWillVO);
        assertNotNull(treatmentWillVO.getTreatmentWillPID());
        assertEquals(1, treatmentWillVO.getVersion().intValue());
        assertTrue(treatmentWillVO.getValidFrom().isBefore(treatmentWillVO.getValidTo()));
        assertEquals(NO_LIFE_PROLONGING_IF_DYING, treatmentWillVO.isNoLifeProlongingIfDying());
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERELY_DEGRADED, treatmentWillVO.isNoLifeProlongingIfSeverelyDegraded());
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERELY_DEGRADED_ACCEPTANCE_NEEDED, treatmentWillVO.getNoLifeProlongingIfSeverelyDegradedAcceptanceNeeded());
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERE_PAIN, treatmentWillVO.isNoLifeProlongingIfSeverePain());
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERE_PAIN_ACCEPTANCE_NEEDED, treatmentWillVO.getNoLifeProlongingIfSeverePainAcceptanceNeeded());
        assertEquals(NO_FORCED_TREATMENT_IF_INCAPABLE, treatmentWillVO.isNoForcedTreatmentIfIncapable());
        assertEquals(NO_FORCED_TREATMENT_IF_INCAPABLE_ACCEPTANCE_NEEDED, treatmentWillVO.getNoForcedTreatmentIfIncapableAcceptanceNeeded());
    }

    @Test
    public void createTreatmentWill_testRequestMapping_null() {
        // Arrange
        final TreatmentWillV2 treatmentWillExample = hl7HelperV2.getTreatmentWillExample();
        ServiceActor serviceActorExample = testHelperV2.getServiceActorExample();
        Mockito.when(serviceActorProvider.getServiceActor(Mockito.any(), Mockito.anyBoolean())).thenReturn(serviceActorExample);
        
        treatmentWillExample.getNoLifeProlongingIfSeverelyDegraded().setAcceptanceNeeded(null); // <--- is null
        treatmentWillExample.getNoLifeProlongingIfSeverePain().setAcceptanceNeeded(null); // <--- is null
        treatmentWillExample.getNoForcedTreatmentIfIncapable().setAcceptanceNeeded(null); // <--- is null

        // Act
        final CreateTreatmentWillRequest request = hl7HelperV2.createCreateTreatmentWillRequest(PERSON_IDENTIFIER, treatmentWillExample);
        CreateTreatmentWillResponse response = serviceUnderTest.createTreatmentWill20200316(request);

        // Assert using business layer
        assertNotNull(response);
        final TreatmentWillVO treatmentWillVO = testHelperV2.getTreatmentWill(businessService, PERSON_IDENTIFIER);
        assertNotNull(treatmentWillVO);
        assertNotNull(treatmentWillVO.getTreatmentWillPID());
        assertEquals(1, treatmentWillVO.getVersion().intValue());
        assertNull(treatmentWillVO.getNoLifeProlongingIfSeverelyDegradedAcceptanceNeeded()); // <--- is null
        assertNull(treatmentWillVO.getNoLifeProlongingIfSeverePainAcceptanceNeeded()); // <--- is null
        assertNull(treatmentWillVO.getNoForcedTreatmentIfIncapableAcceptanceNeeded()); // <--- is null
    }

    @Test
    public void updateTreatmentWill_testRequestMapping() {
        // Note: this also tests null values

        // Arrange
        final boolean noLifeProlongingIfDyingUpdated = false;
        final AcceptanceNeeded noLifeProlongingIfDyingAcceptanceNeededUpdated = null;
        final boolean noLifeProlongingIfSeverelyDegradedUpdated = false;
        final AcceptanceNeeded noLifeProlongingIfSeverelyDegradedAcceptanceNeededUpdated = null;
        final boolean noLifeProlongingIfSeverePainUpdated = false;
        final AcceptanceNeeded noLifeProlongingIfSeverePainAcceptanceNeededUpdated = null;
        final boolean noForcedTreatmentIfIncapableUpdated = false;
        final AcceptanceNeeded noForcedTreatmentIfIncapableAcceptanceNeededUpdated = null;

        TreatmentWillV2 treatmentWill = hl7HelperV2.getTreatmentWillExample();
        treatmentWill.setNoLifeProlongingIfDying(treatmentWillHL7CDAMapper.toTreatmentWillValueTypeWithoutAcceptanceNeeded(noLifeProlongingIfDyingUpdated));
        treatmentWill.setNoLifeProlongingIfSeverelyDegraded(treatmentWillHL7CDAMapper.toTreatmentWillValueType(noLifeProlongingIfSeverelyDegradedUpdated, noLifeProlongingIfSeverelyDegradedAcceptanceNeededUpdated));
        treatmentWill.setNoLifeProlongingIfSeverePain(treatmentWillHL7CDAMapper.toTreatmentWillValueType(noLifeProlongingIfSeverePainUpdated, noLifeProlongingIfSeverePainAcceptanceNeededUpdated));
        treatmentWill.setNoForcedTreatmentIfIncapable(treatmentWillHL7CDAMapper.toTreatmentWillValueType(noForcedTreatmentIfIncapableUpdated, noForcedTreatmentIfIncapableAcceptanceNeededUpdated));

        TreatmentWillVO treatmentWillExample = testHelperV2.getTreatmentWillExample();
        ServiceActor serviceActorExample = testHelperV2.getServiceActorExample();
        RequestContextVO requestContextExample = testHelperV2.getRequestContextExample(serviceActorExample, MESSAGE_ID);
        Mockito.when(serviceActorProvider.getServiceActor(Mockito.any(), Mockito.anyBoolean())).thenReturn(serviceActorExample);

        businessService.createTreatmentWill(treatmentWillExample, requestContextExample, true, false);


        // Act
        final UpdateTreatmentWillRequest request = hl7HelperV2.createUpdateTreatmentWillRequest(PERSON_IDENTIFIER, treatmentWill);
        UpdateTreatmentWillResponse response = serviceUnderTest.updateTreatmentWill20200316(request);

        // Assert using business layer
        assertNotNull(response);
        final TreatmentWillVO treatmentWillVO = testHelperV2.getTreatmentWill(businessService, PERSON_IDENTIFIER);
        assertNotNull(treatmentWillVO);
        assertNotNull(treatmentWillVO.getTreatmentWillPID());
        assertEquals(2, treatmentWillVO.getVersion().intValue());
        assertTrue(treatmentWillVO.getValidFrom().isBefore(treatmentWillVO.getValidTo()));

        assertEquals(noLifeProlongingIfDyingUpdated, treatmentWillVO.isNoLifeProlongingIfDying());

        assertEquals(noLifeProlongingIfSeverelyDegradedUpdated, treatmentWillVO.isNoLifeProlongingIfSeverelyDegraded());
        assertEquals(noLifeProlongingIfSeverelyDegradedAcceptanceNeededUpdated, treatmentWillVO.getNoLifeProlongingIfSeverelyDegradedAcceptanceNeeded());

        assertEquals(noLifeProlongingIfSeverePainUpdated, treatmentWillVO.isNoLifeProlongingIfSeverePain());
        assertEquals(noLifeProlongingIfSeverePainAcceptanceNeededUpdated, treatmentWillVO.getNoLifeProlongingIfSeverePainAcceptanceNeeded());

        assertEquals(noForcedTreatmentIfIncapableUpdated, treatmentWillVO.isNoForcedTreatmentIfIncapable());
        assertEquals(noForcedTreatmentIfIncapableAcceptanceNeededUpdated, treatmentWillVO.getNoForcedTreatmentIfIncapableAcceptanceNeeded());
    }

    @Test
    public void adminUpdateTreatmentWill_testRequestMapping() {
        // Note: this also tests null values

        // Arrange
        final boolean noLifeProlongingIfDyingUpdated = false;
        final AcceptanceNeeded noLifeProlongingIfDyingAcceptanceNeededUpdated = null;
        final boolean noLifeProlongingIfSeverelyDegradedUpdated = false;
        final AcceptanceNeeded noLifeProlongingIfSeverelyDegradedAcceptanceNeededUpdated = null;
        final boolean noLifeProlongingIfSeverePainUpdated = false;
        final AcceptanceNeeded noLifeProlongingIfSeverePainAcceptanceNeededUpdated = null;
        final boolean noForcedTreatmentIfIncapableUpdated = false;
        final AcceptanceNeeded noForcedTreatmentIfIncapableAcceptanceNeededUpdated = null;

        TreatmentWillV2 treatmentWill = hl7HelperV2.getTreatmentWillExample();
        treatmentWill.setNoLifeProlongingIfDying(treatmentWillHL7CDAMapper.toTreatmentWillValueTypeWithoutAcceptanceNeeded(noLifeProlongingIfDyingUpdated));
        treatmentWill.setNoLifeProlongingIfSeverelyDegraded(treatmentWillHL7CDAMapper.toTreatmentWillValueType(noLifeProlongingIfSeverelyDegradedUpdated, noLifeProlongingIfSeverelyDegradedAcceptanceNeededUpdated));
        treatmentWill.setNoLifeProlongingIfSeverePain(treatmentWillHL7CDAMapper.toTreatmentWillValueType(noLifeProlongingIfSeverePainUpdated, noLifeProlongingIfSeverePainAcceptanceNeededUpdated));
        treatmentWill.setNoForcedTreatmentIfIncapable(treatmentWillHL7CDAMapper.toTreatmentWillValueType(noForcedTreatmentIfIncapableUpdated, noForcedTreatmentIfIncapableAcceptanceNeededUpdated));

        TreatmentWillVO treatmentWillExample = testHelperV2.getTreatmentWillExample();
        ServiceActor serviceActorExample = testHelperV2.getServiceActorAdminExample();
        RequestContextVO requestContextExample = testHelperV2.getRequestContextExample(serviceActorExample, MESSAGE_ID);
        Mockito.when(serviceActorProvider.getServiceActor(Mockito.any(), Mockito.anyBoolean())).thenReturn(serviceActorExample);

        businessService.createTreatmentWill(treatmentWillExample, requestContextExample, true, false);


        // Act
        final UpdateTreatmentWillRequest request = hl7HelperV2.createUpdateTreatmentWillRequest(PERSON_IDENTIFIER, treatmentWill);
        UpdateTreatmentWillResponse response = adminServiceUnderTest.updateTreatmentWill20200316(request);

        // Assert using business layer
        assertNotNull(response);
        final TreatmentWillVO treatmentWillVO = testHelperV2.getTreatmentWill(businessService, PERSON_IDENTIFIER);
        assertNotNull(treatmentWillVO);
        assertNotNull(treatmentWillVO.getTreatmentWillPID());
        assertEquals(2, treatmentWillVO.getVersion().intValue());
        assertTrue(treatmentWillVO.getValidFrom().isBefore(treatmentWillVO.getValidTo()));

        assertEquals(noLifeProlongingIfDyingUpdated, treatmentWillVO.isNoLifeProlongingIfDying());

        assertEquals(noLifeProlongingIfSeverelyDegradedUpdated, treatmentWillVO.isNoLifeProlongingIfSeverelyDegraded());
        assertEquals(noLifeProlongingIfSeverelyDegradedAcceptanceNeededUpdated, treatmentWillVO.getNoLifeProlongingIfSeverelyDegradedAcceptanceNeeded());

        assertEquals(noLifeProlongingIfSeverePainUpdated, treatmentWillVO.isNoLifeProlongingIfSeverePain());
        assertEquals(noLifeProlongingIfSeverePainAcceptanceNeededUpdated, treatmentWillVO.getNoLifeProlongingIfSeverePainAcceptanceNeeded());

        assertEquals(noForcedTreatmentIfIncapableUpdated, treatmentWillVO.isNoForcedTreatmentIfIncapable());
        assertEquals(noForcedTreatmentIfIncapableAcceptanceNeededUpdated, treatmentWillVO.getNoForcedTreatmentIfIncapableAcceptanceNeeded());
    }

    @Test
    public void deleteTreatmentWill_testRequestMapping() {
        // Arrange
        TreatmentWillVO treatmentWillExample = testHelperV2.getTreatmentWillExample();
        ServiceActor serviceActorExample = testHelperV2.getServiceActorExample();
        RequestContextVO requestContextExample = testHelperV2.getRequestContextExample(serviceActorExample, MESSAGE_ID);
        Mockito.when(serviceActorProvider.getServiceActor(Mockito.any(), Mockito.anyBoolean())).thenReturn(serviceActorExample);

        businessService.createTreatmentWill(treatmentWillExample, requestContextExample, true, false);


        // Act
        final DeleteTreatmentWillRequest request = hl7HelperV2.createDeleteTreatmentWillRequest(PERSON_IDENTIFIER);
        DeleteTreatmentWillResponse response = serviceUnderTest.deleteTreatmentWill20200316(request);

        // Assert using business layer
        // No data should be found
        assertNotNull(response);
        try {
            testHelperV2.getTreatmentWill(businessService, PERSON_IDENTIFIER);
            fail();
        } catch (ServiceException e) {
            assertEquals(FaultCode.TREATMENT_WILL_NO_DATA_FOUND, e.getFaultCode());
        }
    }

    @Test
    public void adminDeleteTreatmentWill_testRequestMapping() {
        // Arrange
        TreatmentWillVO treatmentWillExample = testHelperV2.getTreatmentWillExample();
        ServiceActor serviceActorExample = testHelperV2.getServiceActorExample();
        RequestContextVO requestContextExample = testHelperV2.getRequestContextExample(serviceActorExample, MESSAGE_ID);
        Mockito.when(serviceActorProvider.getServiceActor(Mockito.any(), Mockito.anyBoolean())).thenReturn(serviceActorExample);

        businessService.createTreatmentWill(treatmentWillExample, requestContextExample, true, false);

        // Act
        final DeleteTreatmentWillRequest request = hl7HelperV2.createDeleteTreatmentWillRequest(PERSON_IDENTIFIER);
        DeleteTreatmentWillResponse response = adminServiceUnderTest.deleteTreatmentWill20200316(request);

        // Assert using business layer
        // No data should be found
        assertNotNull(response);
        try {
            testHelperV2.getTreatmentWill(businessService, PERSON_IDENTIFIER);
            fail();
        } catch (ServiceException e) {
            assertEquals(FaultCode.TREATMENT_WILL_NO_DATA_FOUND, e.getFaultCode());
        }
    }

    private static final dk.sundhedsdatastyrelsen.livstestamente._2018._05._01.ObjectFactory factory = new ObjectFactory();
    private LivingWill getLivingWillExample() {
        final LivingWill livingWill = LivingWillHL7CDAMapper.oLTR.createLivingWill();
        livingWill.setNoLifeProlongingIfDying(NO_LIFE_PROLONGING_IF_DYING);
        livingWill.setNoLifeProlongingIfSeverelyDegraded(NO_LIFE_PROLONGING_IF_SEVERELY_DEGRADED);
        return livingWill;
    }
}
