/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.datacreator.business;

import dk.sundhedsdatastyrelsen.livstestamente.datacreator.DataCreatorProperties;
import dk.sundhedsdatastyrelsen.livstestamente.datacreator.util.LivingWillDataUtil;
import dk.sundhedsdatastyrelsen.livstestamente.datacreator.util.PersonFileUtil;
import dk.sundhedsdatastyrelsen.livstestamente.datacreator.util.RandomDataUtil;
import dk.sundhedsdatastyrelsen.livstestamente.dao.LivingWillDAO;
import dk.sundhedsdatastyrelsen.livstestamente.domain.LivingWillVO;
import dk.sundhedsdatastyrelsen.livstestamente.exceptions.DAOException;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.springframework.stereotype.Component;

import java.util.List;
import java.util.stream.Collectors;

@Component
public class LivingWillDataCreator {

    private static final Logger log = LogManager.getLogger(LivingWillDataCreator.class);

    private final DataCreatorProperties dataCreatorProperties;
    private final LivingWillDAO livingWillDAO;

    private int createRecordCount = 0;
    private int updateRecordCount = 0;

    public LivingWillDataCreator(DataCreatorProperties dataCreatorProperties, LivingWillDAO livingWillDAO) {
        this.dataCreatorProperties = dataCreatorProperties;
        this.livingWillDAO = livingWillDAO;
    }

    public void start() throws DAOException {
        // Get list of person identifiers
        final List<String> personIdentifiers = PersonFileUtil.parsePersonFile(dataCreatorProperties.getPersonFileName());

        // Create data objects for all person identifiers
        final List<LivingWillVO> livingWillData = personIdentifiers.stream()
                .map(personIdentifier ->
                        LivingWillDataUtil.getNewLivingWill(
                                personIdentifier,
                                dataCreatorProperties.getLtr().getDataTimeIntervalStart(),
                                dataCreatorProperties.getLtr().getDataTimeIntervalEnd()))
                .collect(Collectors.toList());

        // Create data in DB
        for (LivingWillVO livingWillVO : livingWillData) {
            final long pid = createLivingWill(livingWillVO);
            livingWillVO.setLivingWillPID(pid);
        }

        // Update data in DB
        for (LivingWillVO livingWillVO : livingWillData) {
            updateLivingWillRecursively(livingWillVO);
        }

        log.info("Number of first records: " + createRecordCount);
        log.info("Number of update records: " + updateRecordCount);
        log.info("Total records: " + (createRecordCount + updateRecordCount));
    }

    private long createLivingWill(LivingWillVO livingWill) throws DAOException {
        final long pid = livingWillDAO.createLivingWill(livingWill);
        createRecordCount++;
        return pid;
    }

    private void updateLivingWillRecursively(LivingWillVO oldLivingWill) throws DAOException {
        // Run if the random decision returns true
        if (RandomDataUtil.getRandomDecision(dataCreatorProperties.getLtr().getChanceOfUpdates())) {
            final LivingWillVO newLivingWill = LivingWillDataUtil.getUpdatedLivingWillData(oldLivingWill);

            livingWillDAO.updateValidToTimestamp(oldLivingWill.getLivingWillPID(), newLivingWill.getValidFrom());
            final long pid = livingWillDAO.createLivingWill(newLivingWill);
            newLivingWill.setLivingWillPID(pid);

            updateRecordCount++;
            updateLivingWillRecursively(newLivingWill);
        }
    }
}
