/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.datacreator.business;

import dk.sundhedsdatastyrelsen.livstestamente.datacreator.DataCreatorProperties;
import dk.sundhedsdatastyrelsen.livstestamente.datacreator.util.PersonFileUtil;
import dk.sundhedsdatastyrelsen.livstestamente.datacreator.util.RandomDataUtil;
import dk.sundhedsdatastyrelsen.livstestamente.datacreator.util.TreatmentWillDataUtil;
import dk.sundhedsdatastyrelsen.livstestamente.dao.TreatmentWillDAO;
import dk.sundhedsdatastyrelsen.livstestamente.domain.TreatmentWillVO;
import dk.sundhedsdatastyrelsen.livstestamente.exceptions.DAOException;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.springframework.stereotype.Component;

import java.util.List;
import java.util.stream.Collectors;

@Component
public class TreatmentWillDataCreator {

    private static final Logger log = LogManager.getLogger(TreatmentWillDataCreator.class);

    private final DataCreatorProperties dataCreatorProperties;
    private final TreatmentWillDAO treatmentWillDAO;

    private int createRecordCount = 0;
    private int updateRecordCount = 0;

    public TreatmentWillDataCreator(DataCreatorProperties dataCreatorProperties, TreatmentWillDAO treatmentWillDAO) {
        this.dataCreatorProperties = dataCreatorProperties;
        this.treatmentWillDAO = treatmentWillDAO;
    }

    public void start() throws DAOException {
        // Get list of person identifiers
        final List<String> personIdentifiers = PersonFileUtil.parsePersonFile(dataCreatorProperties.getPersonFileName());

        // Create data objects for all person identifiers
        final List<TreatmentWillVO> treatmentWillData = personIdentifiers.stream()
                .map(personIdentifier ->
                        TreatmentWillDataUtil.getNewTreatmentWill(
                                personIdentifier,
                                dataCreatorProperties.getBtr().getDataTimeIntervalStart(),
                                dataCreatorProperties.getBtr().getDataTimeIntervalEnd()))
                .collect(Collectors.toList());

        // Create data in DB
        for (TreatmentWillVO treatmentWillVO : treatmentWillData) {
            final long pid = createTreatmentWill(treatmentWillVO);
            treatmentWillVO.setTreatmentWillPID(pid);
        }

        // Update data in DB
        for (TreatmentWillVO treatmentWillVO : treatmentWillData) {
            updateTreatmentWillRecursively(treatmentWillVO);
        }

        log.info("Number of first records: " + createRecordCount);
        log.info("Number of update records: " + updateRecordCount);
        log.info("Total records: " + (createRecordCount + updateRecordCount));
    }

    private long createTreatmentWill(TreatmentWillVO treatmentWill) throws DAOException {
        final long pid = treatmentWillDAO.createTreatmentWill(treatmentWill);
        createRecordCount++;
        return pid;
    }

    private void updateTreatmentWillRecursively(TreatmentWillVO oldTreatmentWill) throws DAOException {
        // Run if the random decision returns true
        if (RandomDataUtil.getRandomDecision(dataCreatorProperties.getBtr().getChanceOfUpdates())) {
            final TreatmentWillVO newTreatmentWill = TreatmentWillDataUtil.getUpdatedTreatmentWillData(oldTreatmentWill);

            treatmentWillDAO.updateValidToTimestamp(oldTreatmentWill.getTreatmentWillPID(), newTreatmentWill.getValidFrom());
            final long pid = treatmentWillDAO.createTreatmentWill(newTreatmentWill);
            newTreatmentWill.setTreatmentWillPID(pid);

            updateRecordCount++;
            updateTreatmentWillRecursively(newTreatmentWill);
        }
    }
}
